/*
 CoNLL2008Format.java

 Created on Mar 14, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: CoNLL2008Format.java,v $
 Revision 1.2  2009/03/06 13:46:39  johansson
 Added removeSUs.

 Revision 1.1  2009/01/16 14:50:55  johansson
 Added to brenta repository.

 Revision 1.6  2008/04/21 09:10:37  richard
 Predicate column optional.

 Revision 1.5  2008/04/10 07:10:40  richard
 Prints lemma.

 Revision 1.4  2008/04/07 07:33:01  richard
 Added mergeSUChains.

 Revision 1.3  2008/04/02 11:28:31  richard
 Complete output.

 Revision 1.2  2008/03/18 15:19:53  richard
 Nodes instead of indices.

 Revision 1.1  2008/03/18 08:38:19  richard
 Added.


 */
package se.lth.cs.nlp.depsrl.format;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.*;

import java.util.*;
import java.io.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class CoNLL2008Format {

	private static final String EMPTY = "_";

	private static final int FORM = 1;
	private static final int LEMMA = 2;
	private static final int PPOS = 4;
	private static final int SPLIT_FORM = 5;
	private static final int SPLIT_LEMMA = 6;
	private static final int PPOSS = 7;
	private static final int HEAD = 8;
	private static final int DEPREL = 9;

	private static final int PRED = 10;
	private static final int ARG_START = 11;

	public static Triple<DepGraph, DepGraph, List<PAStructure>> 
	readNextGraph(BufferedReader br) throws IOException {

		String line = br.readLine();
		if(line == null)
			return null;
		line = line.trim();

		ArrayList<String[]> lines = new ArrayList();        

		while(true) {
			lines.add(line.split("\t"));
			line = br.readLine();
			if(line == null)
				break;
			line = line.trim();
			if(line.equals(""))
				break;
		}

		int len = lines.size();
		DepGraph dg = new DepGraph(len);
		DepGraph dgUnsplit = new DepGraph(len);

		int width = lines.get(0).length;
		for(int i = 1; i < len; i++)
			if(lines.get(i).length != width) {

				for(int j = 0; j < len; j++) {
					if(i == j)
						System.out.print("-->");
					System.out.println(Arrays.asList(lines.get(j)));
				}

				throw new RuntimeException("Illegal format!");
			}

		dg.nodes[0] = new DepNode();
		dgUnsplit.nodes[0] = new DepNode();
		for(int i = 0; i < len; i++) {
			String[] ts = lines.get(i);
			dg.nodes[i+1] = new DepNode(i+1, ts[SPLIT_FORM], ts[PPOSS], ts[SPLIT_LEMMA]);
			dgUnsplit.nodes[i+1] = new DepNode(i+1, ts[FORM], ts[PPOS], ts[LEMMA]);
		}

		for(int i = 0; i < len; i++) {
			String[] ts = lines.get(i);
			int hpos = Integer.parseInt(ts[HEAD]);
			dg.nodes[hpos].addChild(dg.nodes[i+1], ts[DEPREL]);
		}

		ArrayList<PAStructure> pas = new ArrayList();

		if(width > PRED) {
			for(int i = 0; i < len; i++) {
				String[] ts = lines.get(i);
				if(!ts[PRED].equals(EMPTY)) {
					PAStructure pa = new PAStructure();
					pa.pred = dg.nodes[i + 1];
					pa.lemma = ts[PRED];
					pas.add(pa);
				}
			}

			if(lines.get(0).length != ARG_START + pas.size())
				throw new RuntimeException("Illegal number of columns!");

			for(int i = 0; i < len; i++) {
				String[] ts = lines.get(i);
				for(int j = ARG_START; j < ts.length; j++) {
					if(!ts[j].equals(EMPTY)) {
						PAStructure pa = pas.get(j - ARG_START);
						pa.args.add(dg.nodes[i + 1]);
						pa.argLabels.add(ts[j]);
					}
				}
			}
		}

		return new Triple(dg, dgUnsplit, pas);
	}

	public static Triple<DepGraph, DepGraph, List<PAStructure>> readNextGraph(Scanner sc) {
		if(!sc.hasNextLine())
			return null;

		ArrayList<String[]> lines = new ArrayList();        
		String line = sc.nextLine().trim();
		while(!line.equals("")) {
			lines.add(line.split("\t"));            
			line = sc.nextLine().trim(); 
		}

		int len = lines.size();
		DepGraph dg = new DepGraph(len);
		DepGraph dgUnsplit = new DepGraph(len);

		int width = lines.get(0).length;
		for(int i = 1; i < len; i++)
			if(lines.get(i).length != width)
				throw new RuntimeException("Illegal format!");

		dg.nodes[0] = new DepNode();
		dgUnsplit.nodes[0] = new DepNode();
		for(int i = 0; i < len; i++) {
			String[] ts = lines.get(i);
			dg.nodes[i+1] = new DepNode(i+1, ts[SPLIT_FORM], ts[PPOSS], ts[SPLIT_LEMMA]);
			dgUnsplit.nodes[i+1] = new DepNode(i+1, ts[FORM], ts[PPOS], ts[LEMMA]);
		}

		for(int i = 0; i < len; i++) {
			String[] ts = lines.get(i);
			int hpos = Integer.parseInt(ts[HEAD]);
			dg.nodes[hpos].addChild(dg.nodes[i+1], ts[DEPREL]);
		}

		ArrayList<PAStructure> pas = new ArrayList();

		if(width > PRED) {
			for(int i = 0; i < len; i++) {
				String[] ts = lines.get(i);
				if(!ts[PRED].equals(EMPTY)) {
					PAStructure pa = new PAStructure();
					pa.pred = dg.nodes[i + 1];
					pa.lemma = ts[PRED];
					pas.add(pa);
				}
			}

			if(lines.get(0).length != ARG_START + pas.size())
				throw new RuntimeException("Illegal number of columns!");

			for(int i = 0; i < len; i++) {
				String[] ts = lines.get(i);
				for(int j = ARG_START; j < ts.length; j++) {
					if(!ts[j].equals(EMPTY)) {
						PAStructure pa = pas.get(j - ARG_START);
						pa.args.add(dg.nodes[i + 1]);
						pa.argLabels.add(ts[j]);
					}
				}
			}
		}

		return new Triple(dg, dgUnsplit, pas);
	}

	public static void removeSUs(List<PAStructure> pas) {
		for(Iterator<PAStructure> it = pas.iterator(); it.hasNext(); ) {
			PAStructure pa = it.next();
			if(pa.lemma.endsWith(".SU")) {
				it.remove();
				continue;
			}
			for(int i = pa.args.size() - 1; i >= 0; i--) {
				String label = pa.argLabels.get(i);
				if(label.matches("SU(2?)")) {
					pa.args.remove(i);
					pa.argLabels.remove(i);
				}
			}
		}
	}

	public static void mergeSUChains(List<PAStructure> pas) {
		SetMap<DepNode, PAStructure> predsWithSUArg = new SetMap();
		boolean foundSU = false;
		for(PAStructure pa: pas) {
			if(pa.lemma.endsWith("SU")) {
				foundSU = true;
				continue;
			}
			for(int i = 0; i < pa.args.size(); i++) {
				if(pa.argLabels.get(i).equals("SU"))
					predsWithSUArg.put(pa.args.get(i), pa);
			}
		}

		if(!foundSU)
			return;

		boolean changed;
		do {
			changed = false;
			for(Iterator<PAStructure> it = pas.iterator(); it.hasNext(); ) {
				PAStructure pa = it.next();
				if(!pa.lemma.endsWith("SU"))
					continue;
				Set<PAStructure> supportedPreds = predsWithSUArg.get(pa.pred);
				if(supportedPreds == null)
					continue;

				for(PAStructure pa2: supportedPreds)
					for(int i = 0; i < pa.args.size(); i++) {
						pa2.argLabels.add("SU2");
						pa2.args.add(pa.args.get(i));
						predsWithSUArg.put(pa.args.get(i), pa2);
					}
				it.remove();
				changed = true;
			}
		} while(changed);    

		for(Iterator<PAStructure> it = pas.iterator(); it.hasNext(); ) {
			PAStructure pa = it.next();
			if(!pa.lemma.endsWith("SU"))
				continue;
			//System.out.println("Leftover support: " + pa);            
			it.remove();
		}

	}

	private static void checkNBArgs(DepGraph dg, PAStructure pa) {
		/* TODO leta efter redundanta SU. */

		/* TODO coord */
		/* TODO auxiliary+SU, t.ex. has received. 
           I och f�r sig inget problem f�r path. */
		/* TODO relativsatser */

		if(!pa.pred.posFine.startsWith("N"))
			return;
		HashSet<DepNode> sus = new HashSet();
		for(int i = 0; i < pa.args.size(); i++) {
			if(pa.argLabels.get(i).startsWith("SU"))
				sus.add(pa.args.get(i));
		}

		if(pa.pred.relations[0].matches("COORD|CONJ"))
		    return;
		
		//System.out.println("sus = " + sus);

		for(int i = 0; i < pa.args.size(); i++) {
		    DepNode argNode = pa.args.get(i);
			if(sus.contains(argNode))
				continue;
			if(argNode != pa.pred
					&& argNode.parents[0] != pa.pred
					&& pa.pred.parents[0] != argNode
					&& !sus.contains(argNode.parents[0])) {
				System.out.println("Unlinked argument: " + argNode);
				printPAStructure(dg, pa);
				return;
			}
		}
	}

	private static void checkNBArgs2(PAStructure pa, HistogramMap<String, String> m) {

        if(!pa.pred.posFine.startsWith("N"))
            return;
        HashSet<DepNode> sus = new HashSet();
        for(int i = 0; i < pa.args.size(); i++) {
            if(pa.argLabels.get(i).startsWith("SU"))
                sus.add(pa.args.get(i));
        }

        for(int i = 0; i < pa.args.size(); i++) {
            DepNode argNode = pa.args.get(i);
            if(sus.contains(argNode))
                continue;
            if(argNode == pa.pred)
                continue;
            if(argNode.pos.matches("IN|TO|CD"))
                continue;
            String tup = pa.lemma + "/" + argNode.lemma;
            String argLabel = pa.argLabels.get(i);
            m.add(tup, argLabel);
        }
    }
	
	private static void checkNBArgs3(PAStructure pa, DepGraph dg, HistogramMap<String, String> m) {

        if(!pa.pred.posFine.startsWith("N"))
            return;
        HashSet<DepNode> sus = new HashSet();
        for(int i = 0; i < pa.args.size(); i++) {
            if(pa.argLabels.get(i).startsWith("SU"))
                sus.add(pa.args.get(i));
        }

        for(int i = 0; i < pa.args.size(); i++) {
            DepNode argNode = pa.args.get(i);
            if(sus.contains(argNode))
                continue;
            if(argNode == pa.pred)
                continue;
            String tup = pa.lemma + "/" + argNode.lemma;
            
            ArrayList fs = m.getFrequenciesAsSortedList(tup);
            if(fs != null && fs.size() > 1) {
                System.out.println("Inconsistent labeling: " + tup + " / " + fs);
                printPAStructure(dg, pa);
            }
            
        }
    }
	
	private static void checkNB(DepGraph dg, PAStructure pa) {
		if(!pa.pred.posFine.startsWith("N"))
			return;
		for(String lbl: pa.argLabels)
			if(lbl.equals("SU"))
				return;
		for(int i = 0; i < pa.args.size(); i++) {
			DepNode argNode = pa.args.get(i);
			if(argNode != pa.pred
					&& argNode.parents[0] != pa.pred) {
				printPAStructure(dg, pa);
				return;
			}
		}
	}

	private static void checkNBSelfArgs(PAStructure pa, HistogramMap<String, String> hist) {
        if(!pa.pred.posFine.startsWith("N"))
            return;
        for(int i = 0; i < pa.args.size(); i++) {
            DepNode argNode = pa.args.get(i);
            if(argNode != pa.pred)
                continue;
            hist.add(pa.lemma, pa.argLabels.get(i));
        }
    }
	
	private static void checkNBEmpty(PAStructure pa, HistogramMap<String, String> hist) {
        if(!pa.pred.posFine.startsWith("N"))
            return;
        String selfArg = "<NULL>";
        for(int i = 0; i < pa.args.size(); i++) {
            DepNode argNode = pa.args.get(i);
            String argLabel = pa.argLabels.get(i);
            if(argNode == pa.pred) {
                selfArg = argLabel;
                continue;
            }
            if(argLabel.startsWith("SU"))
                continue;
            return;
        }
        hist.add(pa.lemma, selfArg);
    }
	
	private static void checkMultiple(DepGraph dg, PAStructure pa) {
		DepNode predNode = pa.pred;
		if(!predNode.posFine.startsWith("N")) // beh�vs?
			return;
		HashSet<String> s = new HashSet();
		for(String l: pa.argLabels)
			if(s.contains(l)) {
				printPAStructure(dg, pa);
				return;
			} else if(!l.equals("SU"))
				s.add(l);
	}

	private static void checkDeepDescendents(DepGraph dg, PAStructure pa) {
		DepNode predNode = pa.pred;
		for(int i = 0; i < pa.args.size(); i++) {
			DepNode argNode = pa.args.get(i);
			if(argNode != predNode
					&& argNode.parents[0] != predNode
					&& predNode.isSameOrAncestorOf(argNode)) {
				printPAStructure(dg, pa);
				return;
			}
		}
	}

	private static void checkPPs(DepGraph dg, PAStructure pa) {
		DepNode predNode = pa.pred;
		for(int i = 0; i < pa.args.size(); i++) {
			DepNode argNode = pa.args.get(i);
			if(argNode != predNode
					&& argNode.parents[0] != predNode
					&& predNode.isSameOrAncestorOf(argNode)
					&& argNode.parents[0].pos.matches("IN|TO")
					&& argNode.parents[0].parents[0] == predNode) {
				printPAStructure(dg, pa);
				return;
			}
		}
	}

	private static void checkOverlapping(DepGraph dg, PAStructure pa) {

		HashSet<DepNode> seen = new HashSet();
		DepNode predNode = pa.pred;

		for(int i = 0; i < pa.args.size(); i++) {
			DepNode argNode = pa.args.get(i);
			if(argNode != predNode)
				seen.add(argNode);
		}
		for(int i = 0; i < pa.args.size(); i++) {
			DepNode argNode = pa.args.get(i);
			if(argNode != predNode)
				for(DepNode a2: seen)
					if(a2 != argNode
							&& a2.isSameOrAncestorOf(argNode)) {
						printPAStructure(dg, pa);
						return;                        
					}
		}
	}

	private static void printSentence(DepGraph dg) {
		for(int i = 1; i < dg.nodes.length; i++)
			System.out.print(dg.nodes[i].word + " ");
		System.out.println();
	}

	private static void pad(StringBuilder[] sbs) {
		int max = -1;
		for(int i = 0; i < sbs.length; i++)
			if(sbs[i].length() > max)
				max = sbs[i].length();
		for(int i = 0; i < sbs.length; i++) {
			int l = max - sbs[i].length() + 1;
			for(int j = 0; j < l; j++)
				sbs[i].append(' ');
		}        
	}

	public static void printPAStructure(DepGraph dg, PAStructure pa) {
		pa.sortArgs();

		int nextIndex = 0;
		int nextArg = nextIndex == pa.args.size()? -1: pa.args.get(nextIndex).position;
		StringBuilder[] sbs = new StringBuilder[dg.nodes.length-1];
		for(int i = 1; i < dg.nodes.length; i++)
			sbs[i-1] = new StringBuilder();
		for(int i = 1; i < dg.nodes.length; i++)
			sbs[i-1].append(i);
		pad(sbs);
		for(int i = 1; i < dg.nodes.length; i++)
			sbs[i-1].append(dg.nodes[i].word);
		pad(sbs);
		for(int i = 1; i < dg.nodes.length; i++)
			sbs[i-1].append(dg.nodes[i].lemma);
		pad(sbs);
		for(int i = 1; i < dg.nodes.length; i++)
			sbs[i-1].append(dg.nodes[i].pos);
		pad(sbs);
		for(int i = 1; i < dg.nodes.length; i++)
			sbs[i-1].append(dg.nodes[i].parents[0].position);
		pad(sbs);
		for(int i = 1; i < dg.nodes.length; i++)
			sbs[i-1].append(dg.nodes[i].relations[0]);
		pad(sbs);
		for(int i = 1; i < dg.nodes.length; i++)
			if(i == pa.pred.position)
				sbs[i-1].append(pa.lemma);
			else
				sbs[i-1].append(EMPTY);
		pad(sbs);
		for(int i = 1; i < dg.nodes.length; i++)
			if(i == nextArg) {
				sbs[i-1].append(pa.argLabels.get(nextIndex));
				nextIndex++;
				nextArg = nextIndex == pa.args.size()? -1: pa.args.get(nextIndex).position;
			} else
				sbs[i-1].append(EMPTY);
		pad(sbs);
		pad(sbs);
		pad(sbs);

		for(int i = 1; i < dg.nodes.length; i++)
			System.out.println(sbs[i-1]);
		System.out.println();
		System.out.flush();
	}

	private static final Comparator<PAStructure> BY_PRED_POS = new Comparator<PAStructure>() {
		public int compare(PAStructure pa1, PAStructure pa2) {
			return pa1.pred.position - pa2.pred.position;
		}
	};

	public static void printPAStructures(DepGraph dg, DepGraph dgUnsplit, List<PAStructure> pas) {
		Collections.sort(pas, BY_PRED_POS);
		StringBuilder[] lines = new StringBuilder[dg.nodes.length-1];        
		for(int i = 1; i < dg.nodes.length; i++) {
			lines[i-1] = new StringBuilder();
			lines[i-1].append(i);
			lines[i-1].append("\t");
			lines[i-1].append(dgUnsplit.nodes[i].word);
			lines[i-1].append("\t");
			lines[i-1].append(dgUnsplit.nodes[i].lemma);
			lines[i-1].append("\t");
			lines[i-1].append(EMPTY);
			lines[i-1].append("\t");
			lines[i-1].append(dgUnsplit.nodes[i].pos);
			lines[i-1].append("\t");
			lines[i-1].append(dg.nodes[i].word);
			lines[i-1].append("\t");
			lines[i-1].append(dg.nodes[i].lemma);
			lines[i-1].append("\t");
			lines[i-1].append(dg.nodes[i].pos);
			lines[i-1].append("\t");
			lines[i-1].append(dg.nodes[i].parents[0].position);
			lines[i-1].append("\t");
			lines[i-1].append(dg.nodes[i].relations[0]);
			lines[i-1].append("\t");
		}

		Iterator<PAStructure> it = pas.iterator();
		PAStructure nextpa = it.hasNext()? it.next(): null;
		int nextIndex = nextpa == null? -1: nextpa.pred.position;
		for(int i = 1; i < dg.nodes.length; i++)
			if(i == nextIndex) {
				String lemma = nextpa.lemma;
				if(lemma.equals("UNKNOWN.01"))
					lemma = nextpa.pred.lemma + ".01";
				lines[i-1].append(lemma);
				nextpa = it.hasNext()? it.next(): null;
				nextIndex = nextpa == null? -1: nextpa.pred.position;
			} else
				lines[i-1].append(EMPTY);

		if(it.hasNext())
			throw new RuntimeException("!!!");

		for(PAStructure pa: pas) {
			pa.sortArgs();
			int next = 0;
			nextIndex = next == pa.args.size()? -1: pa.args.get(next).position;
			for(int i = 1; i < dg.nodes.length; i++) {
				lines[i-1].append("\t");
				if(i == nextIndex) {
					lines[i-1].append(pa.argLabels.get(next));
					next++;
					nextIndex = next == pa.args.size()? -1: pa.args.get(next).position;                    
				} else
					lines[i-1].append(EMPTY);
			}
			if(next != pa.args.size()) {
				//throw new RuntimeException("!!!!");
				System.out.println("pa: " + pa);
				System.out.println("next = " + next);
				System.out.println("size = " + pa.args.size());
				throw new RuntimeException("!!!!");
			}
		}

		for(int i = 0; i < lines.length; i++)
			System.out.println(lines[i]);
		System.out.println();
		//System.exit(0);
	}

	public static void printPAStructures(PrintWriter pw, DepGraph dg, DepGraph dgUnsplit, List<PAStructure> pas) {
		Collections.sort(pas, BY_PRED_POS);
		StringBuilder[] lines = new StringBuilder[dg.nodes.length-1];        
		for(int i = 1; i < dg.nodes.length; i++) {
			lines[i-1] = new StringBuilder();
			lines[i-1].append(i);
			lines[i-1].append("\t");
			lines[i-1].append(dgUnsplit.nodes[i].word);
			lines[i-1].append("\t");
			lines[i-1].append(dgUnsplit.nodes[i].lemma);
			lines[i-1].append("\t");
			lines[i-1].append(EMPTY);
			lines[i-1].append("\t");
			lines[i-1].append(dgUnsplit.nodes[i].pos);
			lines[i-1].append("\t");
			lines[i-1].append(dg.nodes[i].word);
			lines[i-1].append("\t");
			lines[i-1].append(dg.nodes[i].lemma);
			lines[i-1].append("\t");
			lines[i-1].append(dg.nodes[i].pos);
			lines[i-1].append("\t");
			lines[i-1].append(dg.nodes[i].parents[0].position);
			lines[i-1].append("\t");
			lines[i-1].append(dg.nodes[i].relations[0]);
			lines[i-1].append("\t");
		}

		Iterator<PAStructure> it = pas.iterator();
		PAStructure nextpa = it.hasNext()? it.next(): null;
		int nextIndex = nextpa == null? -1: nextpa.pred.position;
		for(int i = 1; i < dg.nodes.length; i++)
			if(i == nextIndex) {
				String lemma = nextpa.lemma;
				if(lemma.equals("UNKNOWN.01"))
					lemma = nextpa.pred.lemma + ".01";
				lines[i-1].append(lemma);
				nextpa = it.hasNext()? it.next(): null;
				nextIndex = nextpa == null? -1: nextpa.pred.position;
			} else
				lines[i-1].append(EMPTY);

		if(it.hasNext())
			throw new RuntimeException("!!!");

		for(PAStructure pa: pas) {
			pa.sortArgs();
			int next = 0;
			nextIndex = next == pa.args.size()? -1: pa.args.get(next).position;
			for(int i = 1; i < dg.nodes.length; i++) {
				lines[i-1].append("\t");
				if(i == nextIndex) {
					lines[i-1].append(pa.argLabels.get(next));
					next++;
					nextIndex = next == pa.args.size()? -1: pa.args.get(next).position;                    
				} else
					lines[i-1].append(EMPTY);
			}
			if(next != pa.args.size()) {
				//throw new RuntimeException("!!!!");
				System.out.println("pa: " + pa);
				System.out.println("next = " + next);
				System.out.println("size = " + pa.args.size());
				throw new RuntimeException("!!!!");
			}
		}

		for(int i = 0; i < lines.length; i++)
			pw.println(lines[i]);
		pw.println();
	}

	
	public static void mainMain(String[] argv) {
		try {
			Scanner sc = new Scanner(System.in);
			Triple<DepGraph, DepGraph, List<PAStructure>> p = readNextGraph(sc);
			while(p != null) {
				System.out.println("********");                
				mergeSUChains(p.third);
				printSentence(p.first);
				for(PAStructure pa: p.third)
					printPAStructure(p.first, pa);
				//checkNBArgs(p.first, pa);
				p = readNextGraph(sc);
			}

		} catch(Exception e) {
			e.printStackTrace();
		}
	}

	public static void main1(String[] argv) {

	    HistogramMap<String, String> m = new HistogramMap();
	    
	    try {
	        LexicalDB ldb = new LexicalDB("NomBank.ldb");
	        Scanner sc = new Scanner(System.in);
            Triple<DepGraph, DepGraph, List<PAStructure>> p = readNextGraph(sc);
            while(p != null) {
                //removeSUs(p.third);
                mergeSUChains(p.third);
                //printSentence(p.first);
                for(PAStructure pa: p.third)
                    checkNBEmpty(pa, m);

                HashSet<Integer> withPreds = new HashSet();
                for(PAStructure pa: p.third)
                    withPreds.add(pa.pred.position);
                
                DepGraph dg = p.first;
                for(int i = 1; i < dg.nodes.length; i++) {
                    if(withPreds.contains(i))
                        continue;
                    Collection<Roleset> rs = ldb.getRolesets(dg.nodes[i]);
                    if(rs != null) {
                        String id = dg.nodes[i].lemma + ".X";
                        if(rs.size() == 1)
                            id = rs.iterator().next().id;
                        m.add(id, "<EMPTY>");
                    }
                }
                
                p = readNextGraph(sc);
            }

            for(String k: m.keySet()) {
                System.out.println(k + "\t" + m.getFrequenciesAsSortedList(k));
            }
            
            Histogram<String> stats = new Histogram();
            for(String k: m.keySet()) {
                Histogram<String> ks = m.getFrequencies(k);
                for(String k2: ks.keys()) {
                    int f = ks.getFrequency(k2);
                    String nk = k2.startsWith("A")? "SELF": k2;
                    stats.setFrequency(nk, stats.getFrequency(nk) + f);
                }
            }
            
            System.out.println(stats);
            
            
        } catch(Exception e) {
            e.printStackTrace();
        }
    }
	
	   public static void main2(String[] argv) {

	       String filename = "conll2008/conll/conll08st/train/train.closed.bz2";
	       
	        HistogramMap<String, String> m = new HistogramMap();
	        
	        try {
	            LexicalDB ldb = new LexicalDB("NomBank.ldb");
	            BufferedReader br = Util.openFileReader(filename);
	            Triple<DepGraph, DepGraph, List<PAStructure>> p = readNextGraph(br);
	            while(p != null) {
	                //removeSUs(p.third);
	                mergeSUChains(p.third);
	                //printSentence(p.first);
	                for(PAStructure pa: p.third)
	                    checkNBArgs2(pa, m);

	                p = readNextGraph(br);
	            }

	            br = Util.openFileReader(filename);
                p = readNextGraph(br);
                while(p != null) {
                    mergeSUChains(p.third);

                    for(PAStructure pa: p.third) {
                        checkNBArgs3(pa, p.first, m);
                    }

                    p = readNextGraph(br);
                }
	            
	            
	        } catch(Exception e) {
	            e.printStackTrace();
	        }
	    }

	   
	   public static void main(String[] argv) {

           String filename = "conll2008/conll/conll08st/train/train.closed.bz2";
           
            HistogramMap<String, String> m = new HistogramMap();
            
            try {
                LexicalDB ldb = new LexicalDB("NomBank.ldb");
                BufferedReader br = Util.openFileReader(filename);
                Triple<DepGraph, DepGraph, List<PAStructure>> p = readNextGraph(br);
                while(p != null) {
                    mergeSUChains(p.third);
                    //for(PAStructure pa: p.third)
                    //    checkNBArgs(p.first, pa);
                    for(PAStructure pa: p.third)
                        printPAStructure(p.first, pa);
                    p = readNextGraph(br);
                }
                
            } catch(Exception e) {
                e.printStackTrace();
            }
        }

    
	
}
