/*
 Deprojectivization.java
  
 Created on May 7, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: Deprojectivization.java,v $
 Revision 1.1  2009/01/16 14:50:55  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 11:50:20  richard
 Added.

   
 */
package se.lth.cs.nlp.depsrl.format;

import se.lth.cs.nlp.nlputils.depgraph.*;
import it.unitn.disi.depsrl.parser.Parse;

import java.io.*;
import java.util.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class Deprojectivization {

    /* This assumes the Head+Path encoding. */
    
    private static DepNode simpleSearch(DepNode n, String label) {
        for(DepNode c: n.children)
            if(c.relations[0].equals(label + '%'))
                return c;
        
        for(DepNode c: n.children)
            if(c.relations[0].endsWith("%")) {
                DepNode out = simpleSearch(c, label);
                if(out != null)
                    return out;
            }

        for(DepNode c: n.children)
            if(c.relations[0].equals(label))
                return c;

        return null;
    }

    private static DepNode dfs(DepNode lifted, String label, boolean marked,
            boolean vcCheat) {        
        LinkedList<DepNode> queue = new LinkedList();

        for(DepNode c: lifted.parents[0].children)
            if(c != lifted)
                if(!marked || c.relations[0].endsWith("%"))
                    queue.add(c);

        while(!queue.isEmpty()) {
            DepNode n = queue.poll();

            if(n.relations[0].equals(label + '%'))
                return n;
            
            boolean hasVCChild = false;
            if(vcCheat && label.equals("VC"))
                for(DepNode c: n.children)
                    if(c.relations[0].startsWith("VC")) {
                        hasVCChild = true;
                        break;
                    }
                        
            boolean skipVC = hasVCChild && !marked && n.relations[0].startsWith("VC");
                
            if(!skipVC && !marked && n.relations[0].equals(label))
                return n;

            for(DepNode c: n.children)
                if(!marked || c.relations[0].endsWith("%"))
                    queue.add(c);                
        }
        
        return null;
    }
    
    private static DepNode fullSearch(DepNode lifted, String label, boolean vcCheat) {
        DepNode out = dfs(lifted, label, true, vcCheat);
        if(out != null)
            return out;
        return dfs(lifted, label, false, vcCheat);
    }

    private static void checkForCycles(DepNode n, HashSet<DepNode> seen) {
        if(seen.contains(n))
            throw new RuntimeException("Cyclic graph!");
        seen.add(n);
        for(DepNode c: n.children)
            checkForCycles(c, seen);
        seen.remove(n);
    }
    
    public static void deprojectivize(DepGraph dg) {
        for(DepNode n: dg.nodes) {
            if(n.relations.length > 0) {
                int ix = n.relations[0].indexOf('|');
                if(ix != -1) {
                    String label = n.relations[0].substring(0, ix);
                    String headLabel = n.relations[0].substring(ix + 1);

                    //DepNode h = simpleSearch(n.parents[0], headLabel);
                    DepNode h = fullSearch(n, headLabel, true);
                    if(h != null /*&& !label.equals("NMOD")*/) {
                        n.parents[0].removeChild(n);
                        h.addChild(n, label);
                    } else
                        n.relations[0] = label;
                }
            }
        }

        for(DepNode n: dg.nodes)
            if(n.relations.length > 0 && n.relations[0].endsWith("%"))
                n.relations[0] = n.relations[0].substring(0, n.relations[0].length() - 1);

        checkForCycles(dg.nodes[0], new HashSet());
    }
    
    public static void main(String[] argv) {
        try {
            BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
            PrintWriter pw = new PrintWriter(System.out);
            DepGraph dg = CoNLLFormat.readNextGraph(br);
            while(dg != null) {
                deprojectivize(dg);
                CoNLLFormat.printGraph(pw, dg);                
                dg = CoNLLFormat.readNextGraph(br);
            }
            pw.close();
        } catch(Exception e) {
            e.printStackTrace();
        }
    }
    
}
