/*
 Frameset.java
  
 Created on Apr 3, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: Frameset.java,v $
 Revision 1.2  2009/04/16 13:32:13  johansson
 Serializable.

 Revision 1.1  2009/01/16 14:50:55  johansson
 Added to brenta repository.

 Revision 1.5  2008/05/29 11:50:45  richard
 Unknown roleset.

 Revision 1.4  2008/04/21 11:32:31  richard
 Prints warning message if feature can't be set.

 Revision 1.3  2008/04/10 07:11:01  richard
 Added getRoleset.

 Revision 1.2  2008/04/08 11:47:24  richard
 More methods.

 Revision 1.1  2008/04/07 07:32:13  richard
 Added.

   
 */
package se.lth.cs.nlp.depsrl.format;

import java.io.*;
import java.util.*;

import javax.xml.parsers.*;
import org.w3c.dom.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class Frameset implements Serializable {

	private static final long serialVersionUID = 0L;
	
    public ArrayList<Roleset> rolesets = new ArrayList();

    private static HashSet<String> warnings = new HashSet();
    
    Frameset(InputStream is) throws IOException {

        Document doc = null;
        
        try {
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();

            try {
                factory.setFeature("http://apache.org/xml/features/nonvalidating/load-external-dtd",
                        false);
            } catch(Exception e) {
                //System.err.println("Warning: got exception: " + e.getMessage());
                String warning = e.getMessage();
                if(!warnings.contains(warning)) {
                    System.err.println("Warning: got exception " + warning);
                    warnings.add(warning);
                }
            }
            DocumentBuilder builder = factory.newDocumentBuilder();
            doc = builder.parse(is);
        } catch (Exception e) {
            e.printStackTrace();
            //System.err.println("Warning: got exception: " + e.getMessage());
            throw new IOException(e.getMessage());
        }
        
        if(doc == null)
            throw new IOException("Couldn't read frameset.");
        if(!doc.getDocumentElement().getNodeName().equals("frameset")) {
            throw new IOException("This is not a frameset.");
        }
        NodeList l = doc.getDocumentElement().getElementsByTagName("predicate");
        for(int i = 0; i < l.getLength(); i++) {
            Element el = (Element) l.item(i);
            String predicate = el.getAttribute("lemma");
            NodeList l2 = el.getElementsByTagName("roleset");
            for(int j = 0; j < l2.getLength(); j++)
                rolesets.add(new Roleset((Element) l2.item(j), predicate));
        }
    }
    
    Frameset() {
        rolesets.add(new Roleset("UNKNOWN.01"));
    }
    
    String getLemma() {
        Roleset rs = rolesets.get(0);
        String l = rs.id;
        l = l.replaceFirst("\\..*", "");
        return l;
    }
    
    public Collection<String> getSenses() {
        ArrayList<String> out = new ArrayList();
        for(Roleset rs: rolesets)
            out.add(rs.id);
        return out;
    }
    
    public Roleset getRoleset(String id) {
        for(Roleset rs: rolesets)
            if(rs.id.equals(id))
                return rs;
        return null;
    }
    
    public String toString() {
        return "(" + rolesets + ")";
    }
    
}
