/*
 Lemmatizer.java
  
 Created on Apr 9, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: Lemmatizer.java,v $
 Revision 1.1  2009/01/16 14:50:55  johansson
 Added to brenta repository.

 Revision 1.1  2008/04/10 07:08:54  richard
 Added.

   
 */
package se.lth.cs.nlp.depsrl.format;

import java.util.*;
import java.io.*;
import se.lth.cs.nlp.nlputils.core.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class Lemmatizer {

    private HashMap<Pair<String, String>, String> map = new HashMap();
    
    public Lemmatizer(String file) throws IOException {
        Scanner sc = new Scanner(new File(file));
        while(sc.hasNextLine()) {
            String line = sc.nextLine();
            String[] ss = line.split("\t");
            map.put(new Pair(ss[0], ss[1]), ss[2]);
        }
    }

    public String lemmatize(String word, String pos) {
        if(pos.matches("NN|VB|%"))
            return word.toLowerCase();
        return map.get(new Pair(word, pos));
    }
 
    public static void main(String[] argv) {
        try {
            Scanner sc = new Scanner(System.in);            

            HistogramMap<Pair<String, String>, String> hist = new HistogramMap();

            while(sc.hasNextLine()) {
                String line = sc.nextLine().trim();
                if(line.equals(""))
                    continue;
                String[] ss = line.split("\t");
                String pblemma = ss[10];
                if(pblemma.equals("_"))
                    continue;
                pblemma = pblemma.substring(0, pblemma.indexOf('.'));
                String word = ss[5].toLowerCase();
                String pos = ss[7];
                if(pos.matches("(VB|NN).+"))
                    hist.add(new Pair(word, pos), pblemma);
            }

            HashMap m = hist.getMostFrequentAssociations();
            
            for(Pair<String, String> k: hist.keySet()) {
                System.out.println(k.left + "\t" + k.right + "\t" + m.get(k));
            }
            
        } catch(Exception e) {
            e.printStackTrace();
        }
    }
    
}
