/*
 LexicalDB.java
  
 Created on Apr 3, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: LexicalDB.java,v $
 Revision 1.2  2009/04/16 13:32:13  johansson
 Serializable.

 Revision 1.1  2009/01/16 14:50:55  johansson
 Added to brenta repository.

 Revision 1.4  2008/05/29 11:51:07  richard
 Default roleset.

 Revision 1.3  2008/04/10 07:11:17  richard
 Added getRoleset.

 Revision 1.2  2008/04/08 11:47:24  richard
 More methods.

 Revision 1.1  2008/04/07 07:32:13  richard
 Added.

   
 */
package se.lth.cs.nlp.depsrl.format;

import java.util.*;
import java.io.*;
import java.util.zip.GZIPInputStream;
import se.lth.cs.nlp.nlputils.depgraph.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class LexicalDB implements Serializable {

	private static final long serialVersionUID = 0L;
	
    private HashMap<String, Frameset> table = new HashMap();
    private Frameset defaultFrameset;
    
    private static final String SEP = System.getProperty("file.separator");
    
    public LexicalDB(String dirName) throws IOException {
    	if(dirName.endsWith(".ldb")) {
    		ObjectInputStream ois = new ObjectInputStream(new FileInputStream(dirName));
    		try {
    			LexicalDB db = (LexicalDB) ois.readObject();
    			ois.close();
    			this.table = db.table;
    			this.defaultFrameset = db.defaultFrameset;
    		} catch(Exception e) {
    			throw new IOException(e);
    		}
    	} else {
    		File dir = new File(dirName);
    		if(!dir.isDirectory())
    			throw new IllegalArgumentException("Must specify a directory");

    		for(String fileName: dir.list()) {
    			String fullName = dir.getCanonicalPath() + SEP + fileName;
    			InputStream in;
    			try {
    				if(fileName.endsWith(".xml"))
    					in = new FileInputStream(fullName);
    				else if(fileName.endsWith(".xml.gz")) 
    					in = new GZIPInputStream(new FileInputStream(fullName));
    				else
    					continue;
    				Frameset frameset = new Frameset(in);
    				table.put(frameset.getLemma(), frameset);
    			} catch(Exception e) {
    				System.out.println(fileName);
    				e.printStackTrace();
    				System.exit(1);
    			}
    		}
    		defaultFrameset = new Frameset();
    	}
    }

    public Collection<Roleset> getRolesets(DepNode headNode) {
        /* TODO check dependents */
        Frameset fs = table.get(headNode.lemma);
        if(fs == null)
            return null;
        return new ArrayList(fs.rolesets);
    }
    
    public Collection<String> getSenseIDs(DepNode headNode) {
        /* TODO check dependents */
        Frameset fs = table.get(headNode.lemma);
        if(fs == null)
            return null;
        return fs.getSenses();
    }
    
    public Frameset getFrameset(String lemma) {
        return table.get(lemma);
    }       
    
    public Frameset getDefaultFrameset() {
        return defaultFrameset;
    }
    
    public Roleset getRoleset(String senseID) {
        int ix = senseID.indexOf('.');
        if(ix == -1)
            throw new IllegalArgumentException("Must be a sense id");
        Frameset fs = table.get(senseID.substring(0, ix));
        if(fs == null)
            return null;
        return fs.getRoleset(senseID);
    }
    
    public static void main(String[] argv) {
        try {
            LexicalDB db = new LexicalDB(argv[0]);

            //System.out.println(db.getFrameset(argv[1]));
            
            ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(argv[1]));

            oos.writeObject(db);
            oos.close();
            
        } catch(Exception e) {
            e.printStackTrace();
        }
    }
    
}
