/*
 PAStructure.java
  
 Created on Mar 14, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: PAStructure.java,v $
 Revision 1.1  2009/01/16 14:50:55  johansson
 Added to brenta repository.

 Revision 1.5  2008/05/16 11:50:58  richard
 Separate pred pos.

 Revision 1.4  2008/04/02 11:29:01  richard
 Added sortArgs.

 Revision 1.3  2008/03/19 15:52:54  richard
 Added clone.

 Revision 1.2  2008/03/18 15:20:02  richard
 Nodes instead of indices.

 Revision 1.1  2008/03/18 08:38:19  richard
 Added.

   
 */
package se.lth.cs.nlp.depsrl.format;

import se.lth.cs.nlp.nlputils.depgraph.*;
import se.lth.cs.nlp.nlputils.core.*;
import java.util.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class PAStructure implements java.io.Serializable {

	private static final long serialVersionUID = 0L;
	
    public DepNode pred; // such as (interested, JJ)
    public String lemma; // such as interest.01
    public String predPos; // such as VBN

    public ArrayList<DepNode> args = new ArrayList();
    public ArrayList<String> argLabels = new ArrayList();

    public String toString() {
        return "(" + pred + ":" + lemma + ", " + args + ", " + argLabels + ")";
    }
    
    public PAStructure clone() {
        PAStructure out = new PAStructure();
        out.pred = pred;
        out.predPos = predPos;
        out.lemma = lemma;
        out.argLabels = (ArrayList) argLabels.clone();
        out.args = (ArrayList) args.clone();
        return out;
    }
    
    private static final Comparator<Pair> BY_ARG_POS = new Comparator<Pair>() {
        public int compare(Pair p1, Pair p2) {
            DepNode d1 = (DepNode) p1.left;
            DepNode d2 = (DepNode) p2.left;
            return d1.position - d2.position;
        }
    };
    
    void sortArgs() {
        ArrayList<Pair<DepNode, String>> ps = new ArrayList();
        for(int i = 0; i < args.size(); i++)
            ps.add(new Pair(args.get(i), argLabels.get(i)));
        
        Collections.sort(ps, BY_ARG_POS);
        for(int i = 0; i < ps.size(); i++) {
            Pair<DepNode, String> p = ps.get(i);
            args.set(i, p.left);
            argLabels.set(i, p.right);
        }
    }
 
    public ArrayList<DepNode> getArgsByLabel(String argRegex) {
    	ArrayList<DepNode> out = new ArrayList();
    	for(int i = 0; i < args.size(); i++)
    		if(argLabels.get(i).matches(argRegex))
    			out.add(args.get(i));
    	return out;
    }
    
    public DepNode getFirstArgByLabel(String argRegex) {
    	for(int i = 0; i < args.size(); i++)
    		if(argLabels.get(i).matches(argRegex))
    			return args.get(i);
    	return null;
    }
    
}
