/*
 Roleset.java
  
 Created on Apr 3, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: Roleset.java,v $
 Revision 1.2  2009/04/16 13:32:13  johansson
 Serializable.

 Revision 1.1  2009/01/16 14:50:55  johansson
 Added to brenta repository.

 Revision 1.3  2008/05/29 11:51:18  richard
 Default roleset.

 Revision 1.2  2008/04/10 07:11:35  richard
 Added getRoleIDs.

 Revision 1.1  2008/04/07 07:32:13  richard
 Added.

   
 */
package se.lth.cs.nlp.depsrl.format;

import java.io.Serializable;
import java.util.*;
import org.w3c.dom.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class Roleset implements Serializable {
	
	private static final long serialVersionUID = 0L;

    Roleset(Element el, String predicate) {
        lemma = predicate;
        if(lemma == null || lemma.equals(""))
            throw new IllegalArgumentException("no lemma");
        id = el.getAttribute("id");
        if(id == null || id.equals(""))
            throw new IllegalArgumentException("no id");
        name = el.getAttribute("name");
        if(name != null && name.equals(""))
            name = null;
        vncls = el.getAttribute("vncls");
        if(vncls != null && (vncls.equals("") || vncls.equals("-")))
            vncls = null;
        NodeList l = el.getElementsByTagName("roles");
        if(l.getLength() != 1)
            throw new IllegalArgumentException("length != 1");
        Element rl = (Element) l.item(0);
        NodeList l2 = rl.getElementsByTagName("role");
        for(int j = 0; j < l2.getLength(); j++)
            roles.add(new Role((Element) l2.item(j)));
    }

    /* Default roleset for unknown predicates. */
    Roleset(String id) {
        this.id = id;
        for(int i = 0; i <= 5; i++)
            roles.add(new Role(i));
    }
    
    public String lemma;
    public String name;
    public String id;
    public String vncls;

    public ArrayList<Role> roles = new ArrayList();
    
    public Collection<String> getRoleIDs() {
        ArrayList<String> out = new ArrayList();
        for(Role r: roles)
            out.add(r.id);
        return out;
    }
    
    public String toString() {
        String out = "(lemma = " + lemma + ", name = " + name + ", id = " + id;
        if(vncls != null)
            out += ", vncls = " + vncls;
        out += ", roles = " + roles + ")";
        return out;
    }
    
}
