package se.lth.cs.nlp.depsrl.format;

import java.io.*;
import java.util.*;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.*;

public class TestReachability {

	public static void main(String[] argv) {
		try {
			Scanner sc = new Scanner(new FileReader(argv[0]));
			
			int count = 0;
			IntHistogram depthHist = new IntHistogram();
			
			Histogram<String> posHist = new Histogram();
			
			Triple<DepGraph, DepGraph, List<PAStructure>> g = CoNLL2008Format.readNextGraph(sc);
			while(g != null) {
				
				for(PAStructure pa: g.third) {
					if(!pa.pred.pos.startsWith("N"))
						continue;
					if(pa.lemma.endsWith("SU"))
						continue;
					for(int i = 0; i < pa.args.size(); i++) {
						if(!pa.argLabels.get(i).startsWith("SU")) {
							int d = findDepth(pa.pred, pa.args.get(i));
							depthHist.add(d);
							count++;
							if(false && d > 1) {
								System.out.println("depth == " + d);
								CoNLL2008Format.printPAStructure(g.first, pa);
							}
							posHist.add(pa.args.get(i).pos);
						}
					}
				}
				
				g = CoNLL2008Format.readNextGraph(sc);
			}

			for(int i = 0; i < depthHist.size(); i++) {
				int n = depthHist.getFrequency(i);
				System.out.printf("%d: %d / %d = %f\n", i, n, count, (double) n / count);
			}			
			//System.out.println(depthHist);

			System.out.println();
			for(IntObjPair<String> p: posHist.asSortedList()) {
				System.out.println(p);
			}
			
		} catch(Exception e) {
			e.printStackTrace();
		}
	}

	private static int findDepth(DepNode p, DepNode a) {
		HashSet<DepNode> seen = new HashSet();

		while(p != null) {
			seen.add(p);
			if(p.parents.length == 0)
				p = null;
			else
				p = p.parents[0];
		}

		int count = 0;
		while(!seen.contains(a)) {
			a = a.parents[0];
			count++;
		}

		return count;
	}
	
}
