/*
 TestSpans.java
  
 Created on May 22, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: TestSpans.java,v $
 Revision 1.2  2010-06-07 11:43:38  johansson
 Major update.

 Revision 1.1  2009/01/16 14:50:55  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/29 11:51:26  richard
 Added.

   
 */
package se.lth.cs.nlp.depsrl.format;

import java.io.*;
import java.util.*;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.annotations.*;
import se.lth.cs.nlp.nlputils.depgraph.*;

import se.lth.cs.nlp.depsrl.Lemmatizer;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class TestSpans {

    private static ArrayList<String[]> readCoNLL2005Sentence(Scanner sc) {
        if(!sc.hasNextLine())
            return null;
        ArrayList<String[]> out = new ArrayList();
        String line = sc.nextLine().trim();
        while(!line.equals("")) {
            out.add(line.split("\\s+"));
            line = sc.nextLine().trim();
        }
        return out;
    }
    
    private static boolean matches(ArrayList<String[]> tuples2005,
            DepGraph dg) {
        
        StringBuilder sb1 = new StringBuilder();
        for(int i = 1; i < dg.nodes.length; i++) {
            String w = dg.nodes[i].word.replaceAll("\\/", "\\\\/");
            w = w.replaceAll("\\\\\\\\", "\\\\");
            sb1.append(w);
        }

        StringBuilder sb2 = new StringBuilder();
        for(String[] ss: tuples2005) {
            String w = ss[0].replaceAll("\\/", "\\\\/");
            w = w.replaceAll("\\\\\\\\", "\\\\");
            if(!w.equals("_"))
                sb2.append(w);
        }

        //System.out.println(sb1);
        //System.out.println(sb2);
        
        return sb1.toString().equals(sb2.toString());
    }

    private static void printTuples(ArrayList<String[]> tuples) {
        for(String[] ss: tuples)
            System.out.println(Arrays.asList(ss));
        System.out.println();
    }
    
    private static IntPair[] getNewTokenIndices(ArrayList<String[]> tuples,
            DepGraph dg) {
        IntPair[] oldToNewTokens = new IntPair[tuples.size()+1];
        IntPair currentPair = new IntPair(1, -1);
        String current = tuples.get(0)[0];
        int oldIndex = 0;
        for(int i = 1; i < dg.nodes.length; i++) {
            String w = dg.nodes[i].word.replaceAll("\\/", "\\\\/");
            w = w.replaceAll("\\\\\\\\", "\\\\");
            current = current.substring(w.length());
            if(current.equals("")) {
                currentPair.right = i;
                oldToNewTokens[oldIndex+1] = currentPair;
                oldIndex++;
                if(oldIndex < tuples.size()) {
                    current = tuples.get(oldIndex)[0];
                    currentPair = new IntPair(i + 1, -1);
                } else {
                    current = null;
                    currentPair = null;
                }
            }            
        }
        return oldToNewTokens;
    }
    
    private static IntPair[] getDummyTokenIndices(ArrayList<String[]> tuples,
            DepGraph dg) {
        if(dg.nodes.length != tuples.size() + 1)
            throw new RuntimeException("Illegal length!");
        IntPair[] oldToNewTokens = new IntPair[tuples.size()+1];
        for(int i = 0; i < oldToNewTokens.length; i++) {
            oldToNewTokens[i] = new IntPair(i, i);
        }
        return oldToNewTokens;
    }
    
    private static ArrayList<SpanPAStructure> getGoldPAs(ArrayList<String[]> tuples) {

        /*for(String[] xx: tuples) {            
            System.out.println(Arrays.toString(xx));
        }
        System.out.println();
        //System.exit(0);*/
        
        ArrayList<SpanPAStructure> out = new ArrayList();

        int count = 0;
        for(int j = 0; j < tuples.size(); j++) {
            String[] tuple = tuples.get(j);
            if(tuple[1].equals("-"))
                continue;
            SpanPAStructure pa = new SpanPAStructure();
            out.add(pa);

            //pa.predIndex = j;
            pa.predIndex = j+1;
            
            pa.predLemma = tuple[1];
            
            Span current = null;
            for(int i = 0; i < tuples.size(); i++) {
                String s = tuples.get(i)[count + 2];
                if(s.startsWith("(")) {
                    if(current != null)
                        throw new RuntimeException("current != null");
                    int ix = s.indexOf('*');
                    if(ix == -1)
                        throw new RuntimeException("ix == -1");
                    String lbl = s.substring(1, ix);
                    current = new Span(i, -1, lbl);
                }
                if(s.endsWith(")")) {
                    if(current == null)
                        throw new RuntimeException("current == null");                    
                    current.end = i + 1;                    

                    current.start++;
                    current.end++;
                    
                    pa.spans.add(current);
                    current = null;
                }
            }
            
            count++;
        }
        return out;
    }

    private static int getPredIndex(IntPair p, DepGraph dg) {
        if(p.left == p.right)
            return p.left;
        int ix = -1;
        for(int i = p.left; i <= p.right; i++) {
            if(dg.nodes[i].pos.startsWith("V")) {
                if(ix != -1)
                    throw new RuntimeException("ix != -1");
                ix = i;
            }
        }
        if(ix == -1)
            throw new RuntimeException("ix == -1");
        return ix;
    }
    
    private static void expand(ArrayList<SpanPAStructure> pas,
            IntPair[] oldToNewTokens, DepGraph dg) {
        for(SpanPAStructure pa: pas) {
            pa.predIndex = getPredIndex(oldToNewTokens[pa.predIndex], dg);
            for(Span s: pa.spans) {
                s.start = oldToNewTokens[s.start].left;
                s.end = oldToNewTokens[s.end - 1].right + 1;
            }
        }
    }
    
    private static void printPAs(ArrayList<SpanPAStructure> pas, DepGraph dg,
            boolean printWords) {

        //for(SpanPAStructure spa: pas)
        //    System.out.println(spa.predIndex);
        
        StringBuilder[] lines = new StringBuilder[dg.nodes.length-1];
        for(int i = 1; i < dg.nodes.length; i++) {
            lines[i-1] = new StringBuilder();
            if(printWords) {
                lines[i-1].append(dg.nodes[i].word);
                lines[i-1].append(" ");
            }
        }

        Iterator<SpanPAStructure> it = pas.iterator();
        SpanPAStructure nextpa = it.hasNext()? it.next(): null;
        int nextIndex = nextpa == null? -1: nextpa.predIndex;
        for(int i = 1; i < dg.nodes.length; i++)
            if(i == nextIndex) {
                lines[i-1].append(nextpa.predLemma);
                nextpa = it.hasNext()? it.next(): null;
                nextIndex = nextpa == null? -1: nextpa.predIndex;
            } else
                lines[i-1].append("-");
        
        if(it.hasNext()) {
            System.out.println("pas: " + pas);
            throw new RuntimeException("!!!");
        }

        for(SpanPAStructure pa: pas) {
            int next = 0;
            nextIndex = next == pa.spans.size()? -1: pa.spans.get(next).start;
            for(int i = 1; i < dg.nodes.length; i++) {
                lines[i-1].append(" ");
                if(i == nextIndex) {
                    Span s = pa.spans.get(next);
                    lines[i-1].append("(" + s.label);
                    
                    while(i < s.end - 1) {
                        lines[i-1].append("*");
                        lines[i].append(" ");
                        i++;
                    }

                    lines[i-1].append("*)");                    
                    
                    next++;
                    nextIndex = next == pa.spans.size()? -1: pa.spans.get(next).start;                    
                } else
                    lines[i-1].append("*");
            }
            if(next != pa.spans.size()) {
                //throw new RuntimeException("!!!!");
                System.out.println("pa: " + pa);
                System.out.println("next = " + next);
                System.out.println("nextIndex = " + nextIndex);
                System.out.println("size = " + pa.spans.size());
                throw new RuntimeException("!!!!");
            }
        }
        
        for(int i = 0; i < lines.length; i++)
            System.out.println(lines[i]);
        System.out.println();
    }
    
    private static void removeVs(ArrayList<SpanPAStructure> pas) {
        for(SpanPAStructure pa: pas) {
            for(Iterator<Span> it = pa.spans.iterator(); it.hasNext(); ) {
                Span s = it.next();
                if(s.label.matches("V|C-V"))
                    it.remove();
            }
        }
    }
    
    private static void removePuncs(ArrayList<SpanPAStructure> pas, DepGraph dg) {
        for(SpanPAStructure pa: pas)
            for(Span s: pa.spans) {
                while(s.start < dg.nodes.length && dg.nodes[s.start].pos.matches("``|''|\\.|,|:"))
                    s.start++;
                while(s.end > 1 && dg.nodes[s.end-1].pos.matches("``|''|\\.|,|:"))
                    s.end--;
            }
        for(SpanPAStructure pa: pas)
            for(Iterator<Span> it = pa.spans.iterator(); it.hasNext(); ) {
                Span s = it.next();
                if(s.end <= s.start)
                    it.remove();
            }
    }
    
    public static void expandSpans(String[] argv) {
        boolean removePunc = Boolean.parseBoolean(argv[1]);
        String conll2008File = argv[2];
        String conll2005File = argv[3];
        
        try {
            
            Scanner sc1 = new Scanner(new File(conll2008File));
            Scanner sc2 = new Scanner(new File(conll2005File));
            Triple<DepGraph, DepGraph, List<PAStructure>> p = CoNLL2008Format.readNextGraph(sc1);
            if(p != null)
                CoNLL2008Format.mergeSUChains(p.third);
            while(p != null) {
                
                ArrayList<String[]> tuples2005 = readCoNLL2005Sentence(sc2);
                if(tuples2005 == null)
                    throw new RuntimeException("tuples2005 = null");
                
                //if(!matches(tuples2005, p.first))
                //    throw new RuntimeException("Sentences don't match...");
                int count = 0;
                while(!matches(tuples2005, p.first)) {
                    count++;
                    if(count == 10)
                        throw new RuntimeException("Can't match...");
                    tuples2005 = readCoNLL2005Sentence(sc2);
                    if(tuples2005 == null)
                        throw new RuntimeException("tuples2005 = null");
                }
                
                IntPair[] oldToNewTokens = getNewTokenIndices(tuples2005, p.first);
                
                ArrayList<SpanPAStructure> pas = getGoldPAs(tuples2005);
                expand(pas, oldToNewTokens, p.first);
                
                removeVs(pas);
                
                if(removePunc)
                    removePuncs(pas, p.first);
                
                printPAs(pas, p.first, false);
                
                p = CoNLL2008Format.readNextGraph(sc1);
                if(p != null)
                    CoNLL2008Format.mergeSUChains(p.third);
            }
        
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
            
    }

    private static void pruneSubtree(DepNode aboveArgument, DepNode target,
            ArrayList<DepNode> subtree) {

        while(target != aboveArgument) {
            for(Iterator<DepNode> it = subtree.iterator(); it.hasNext(); ) {
                DepNode n = it.next();
                if(target.isSameOrAncestorOf(n)
                        || n != aboveArgument && n.isSameOrAncestorOf(target))
                    it.remove();
            }
            target = target.parents[0];
        }
    }

    
    private static ArrayList<Span> createSpans(DepNode targetNode, DepNode argNode, String label) {
        ArrayList<DepNode> ns = new ArrayList<DepNode>();

        if(argNode != targetNode && !label.equals("AM-MOD")) {
            argNode.collectSubtree(ns);
            if(argNode.isSameOrAncestorOf(targetNode))
                pruneSubtree(argNode, targetNode, ns);
        } else
            ns.add(argNode);                        

        Collections.sort(ns, DepGraph.PRECEDENCE_ORDER);
        
        Span s = null;
        ArrayList<Span> ss = new ArrayList<Span>(1);
        
        for(DepNode n: ns) {
            if(s == null || s.end != n.position) {
                s = new Span();
                ss.add(s);
                s.label = label;
                s.start = n.position;
            }
            s.end = n.position + 1;
        }
        for(int i = 1; i < ss.size(); i++) {
            String l = ss.get(i).label;
            if(!l.startsWith("C-"))
                l = "C-" + l;
            ss.get(i).label = l;
        }
        return ss;
    }

    static int countOverlaps = 0;
    
    private static void removeOverlaps(ArrayList<Span> spans) {
        HashSet<Span> toRemove = new HashSet();        
        for(int i = 0; i < spans.size(); i++) {
            Span s1 = spans.get(i);
            for(int j = i + 1; j < spans.size(); j++) {
                Span s2 = spans.get(j);
                if(s1.overlaps(s2))
                    toRemove.add(s2);             
            }
        }
        countOverlaps += toRemove.size();
        spans.removeAll(toRemove);
    }

    private static void mergeContLabels(ArrayList<Span> spans) {
	Span prev = null;
	for(Iterator<Span> it = spans.iterator(); it.hasNext(); ) {
	    Span s = it.next();
	    if(prev == null || !s.label.startsWith("C-")) {
		prev = s;
		continue;
	    }
	    String l = s.label.substring(2);
	    if(!(prev.label.equals(s.label) || prev.label.equals(l))) {
		prev = s;
		continue;		
	    }
	    if(prev.end != s.start) {
		prev = s;
		continue;		
	    }	
	    prev.end = s.end;
	    it.remove();
	}

	/*for(Span s1: spans)
            for(Span s2: spans) {
                if(s1.start < s2.start
                   && s2.label.equals("C-" + s1.label)) {
                    if(s1.label.startsWith("AM"))
                        s2.label = s1.label;
                }
            }*/
    }
    
    private static void puncHeuristics(DepGraph dg, ArrayList<SpanPAStructure> spas) {
        for(SpanPAStructure spa: spas) {
            for(Span s: spa.spans) {
                DepNode n = dg.nodes[s.end-1];
                if(!n.pos.equals(","))
                    continue;
                boolean found = false;
                for(DepNode c: n.parents[0].children) {
                    if(c == n)
                        continue;
                    if(!c.pos.equals(","))
                        continue;
                    if(c.position >= s.start && c.position < n.position) {
                        found = true;
                        break;
                    }
                }
                if(!found) {
                    s.end--;
                }
            }
            
            for(Span s: spa.spans) {
                DepNode n = dg.nodes[s.end-1];
                if(!n.pos.matches("\\."))
                    continue;
                s.end--;
            }
            
        }
        
        for(SpanPAStructure pa: spas)
            for(Iterator<Span> it = pa.spans.iterator(); it.hasNext(); ) {
                Span s = it.next();
                if(s.end <= s.start)
                    it.remove();
            }
    }
    
    private static SpanPAStructure convertPA(PAStructure pa, SpanPAStructure goldspa) {
    	SpanPAStructure spa = new SpanPAStructure();
    	spa.predIndex = goldspa.predIndex;
    	spa.predLemma = goldspa.predLemma;
    	for(int i = 0; i < pa.args.size(); i++) {
    		String l = pa.argLabels.get(i);
    		if(!l.startsWith("SU"))
    			spa.spans.addAll(createSpans(pa.pred, pa.args.get(i), l));
    	}
    	Collections.sort(spa.spans, Span.NestingOrder.instance());
    	removeOverlaps(spa.spans);
    	mergeContLabels(spa.spans);

    	//puncHeuristics(spa.spans);

    	return spa;
    }
    
    public static SpanPAStructure convertPredictedPA(PAStructure pa) {
    	SpanPAStructure spa = new SpanPAStructure();
    	spa.predIndex = pa.pred.position;
    	spa.predLemma = pa.lemma; 

    	for(int i = 0; i < pa.args.size(); i++) {
    		String l = pa.argLabels.get(i);
    		if(!l.startsWith("SU"))
    			spa.spans.addAll(createSpans(pa.pred, pa.args.get(i), l));
    	}
    	Collections.sort(spa.spans, Span.NestingOrder.instance());
    	removeOverlaps(spa.spans);
    	mergeContLabels(spa.spans);

    	//puncHeuristics(spa.spans);

    	return spa;
    }
    
    private static PAStructure match(Collection<PAStructure> pas, SpanPAStructure goldspa) {
        for(PAStructure pa: pas)
            if(pa.pred.position == goldspa.predIndex)
                return pa;
        return null;
    }
    
    public static void convertDepsToSpans2(String[] argv) {
        boolean removePunc = Boolean.parseBoolean(argv[1]);

        try {

            InputStream is;
            if(argv.length > 2)
            	is = new FileInputStream(argv[2]);
            else
            	is = System.in;        	
            Scanner sc1 = new Scanner(is);

            Triple<DepGraph, DepGraph, List<PAStructure>> p = CoNLL2008Format.readNextGraph(sc1);
            if(p != null)
                CoNLL2008Format.mergeSUChains(p.third);
            while(p != null) {

                ArrayList<SpanPAStructure> predictedSpans = new ArrayList();

                for(PAStructure pa: p.third)
                	predictedSpans.add(convertPredictedPA(pa));

                puncHeuristics(p.first, predictedSpans);
                try {
                    if(removePunc)
                        removePuncs(predictedSpans, p.first);
                } catch(Exception e) {
                    System.out.println(p.first);
                    e.printStackTrace();
                    System.exit(1);
                }
                
                printPAs(predictedSpans, p.first, false);
                
                p = CoNLL2008Format.readNextGraph(sc1);
                if(p != null)
                    CoNLL2008Format.mergeSUChains(p.third);
            }
        
            System.err.println("Number of overlaps: " + countOverlaps);
            
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
    
    public static void convertDepsToSpans(String[] argv) {
        boolean removePunc = Boolean.parseBoolean(argv[1]);
        String conll2008File = argv[2]; // 
        String conll2005File = argv[3]; // gold standard, unexpanded
        
        try {
            
            Scanner sc1 = new Scanner(new File(conll2008File));
            Scanner sc2 = new Scanner(new File(conll2005File));
            Triple<DepGraph, DepGraph, List<PAStructure>> p = CoNLL2008Format.readNextGraph(sc1);
            if(p != null)
                CoNLL2008Format.mergeSUChains(p.third);
            while(p != null) {

                ArrayList<String[]> tuples2005 = readCoNLL2005Sentence(sc2);
                if(tuples2005 == null)
                    throw new RuntimeException("tuples2005 = null");

                int count = 0;
                while(!matches(tuples2005, p.first)) {
                    count++;
                    if(count == 10)
                        throw new RuntimeException("Can't match...");
                    tuples2005 = readCoNLL2005Sentence(sc2);
                    if(tuples2005 == null)
                        throw new RuntimeException("tuples2005 = null");
                }
                
                IntPair[] oldToNewTokens = getNewTokenIndices(tuples2005, p.first);
                
                ArrayList<SpanPAStructure> pas = getGoldPAs(tuples2005);
                expand(pas, oldToNewTokens, p.first);
                
                ArrayList<SpanPAStructure> predictedSpans = new ArrayList();
                for(SpanPAStructure gspa: pas) {
                    PAStructure pa = match(p.third, gspa);
                    if(pa != null)
                        predictedSpans.add(convertPA(pa, gspa));                    
                }

                puncHeuristics(p.first, predictedSpans);
                try {
                    if(removePunc)
                        removePuncs(predictedSpans, p.first);
                } catch(Exception e) {
                    System.out.println(p.first);
                    for(String[] ss: tuples2005)
                        System.out.println(Arrays.toString(ss));
                    e.printStackTrace();
                    System.exit(1);
                }
                
                printPAs(predictedSpans, p.first, false);
                
                p = CoNLL2008Format.readNextGraph(sc1);
                if(p != null)
                    CoNLL2008Format.mergeSUChains(p.third);
            }
        
            System.err.println("Number of overlaps: " + countOverlaps);
            
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
    
    private static boolean checkPA(DepGraph dg, SpanPAStructure pa1, SpanPAStructure pa2) {
        for(Span s1: pa1.spans)
            for(Span s2: pa2.spans) {
                if(s1.start == s2.start && s1.end == s2.end)
                    continue;
                
                int s1s = s1.start;
                int s2s = s2.start;
                int s1e = s1.end;
                int s2e = s2.end;

                while(s1s > 0 && s1s < dg.nodes.length && dg.nodes[s1s].pos.matches("``|''|\\.|,|:"))
                    s1s++;
                while(s2s > 0 && s2s < dg.nodes.length && dg.nodes[s2s].pos.matches("``|''|\\.|,|:"))
                    s2s++;
                while(s1e > 1 && dg.nodes[s1e-1].pos.matches("``|''|\\.|,|:"))
                    s1e--;
                while(s2e > 1 && dg.nodes[s2e-1].pos.matches("``|''|\\.|,|:"))
                    s2e--;

                if(s1s == s2s && s1e == s2e)
                    return true;
            }
        return false;
    }
    
    private static boolean checkPA_A1(DepGraph dg, SpanPAStructure pa1, SpanPAStructure pa2) {
        HashSet<Span> s1s = new HashSet(), s2s = new HashSet();
        
        for(Span s1: pa1.spans)
            for(Span s2: pa2.spans)
                if(s1.start == s2.start && s1.end == s2.end && s1.label.equals(s2.label)) {
                    s1s.add(s1);
                    s2s.add(s2);
                }
        
        for(Span s2: pa2.spans)
            if(!s2s.contains(s2))
                if(s2.label.matches("(C-)?A1"))
                    return true;
        return false;
    }
    
    public static void checkSpans(String[] argv) {
        String conll2005File1 = argv[1];
        String conll2005File2 = argv[2];
        String conll2008File = argv[3];
        
        try {
            
            Scanner sc1 = new Scanner(new File(conll2005File1));
            Scanner sc2 = new Scanner(new File(conll2005File2));
            Scanner sc3 = new Scanner(new File(conll2008File));

            ArrayList<String[]> tuples2005_1 = readCoNLL2005Sentence(sc1);
            
            while(tuples2005_1 != null) {
                ArrayList<String[]> tuples2005_2 = readCoNLL2005Sentence(sc2);
                Triple<DepGraph, DepGraph, List<PAStructure>> p = CoNLL2008Format.readNextGraph(sc3);
                
                ArrayList<SpanPAStructure> pas_1 = getGoldPAs(tuples2005_1);
                ArrayList<SpanPAStructure> pas_2 = getGoldPAs(tuples2005_2);

                //printPAs(pas_1, p.first, true);
                //printPAs(pas_2, p.first, true);

                outer:
                for(SpanPAStructure pa1: pas_1) {
                    for(SpanPAStructure pa2: pas_2)
                        if(pa1.predIndex == pa2.predIndex)
                            if(checkPA_A1(p.first, pa1, pa2)) {
                                printPAs(pas_1, p.first, true);
                                printPAs(pas_2, p.first, true);
                                break outer;
                            }
                }
                    
                tuples2005_1 = readCoNLL2005Sentence(sc1);
            }
        
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
    
    private static IntHistogram hist;

    static {
        if(false)
    		hist = new IntHistogram();
    }
    
    private static int N_TOTAL, N_OK;
    
    private static PAStructure convertSpans(SpanPAStructure spa, DepGraph dg) {
        PAStructure pa = new PAStructure();
        pa.lemma = spa.predLemma + ".99";
        pa.pred = dg.nodes[spa.predIndex];
        for(Span s: spa.spans) {
            ArrayList<DepNode> outgoing = new ArrayList();            
            for(int i = s.start; i < s.end; i++) {
                DepNode n = dg.nodes[i];
                DepNode p = n.parents[0];
                if(p.position < s.start || p.position >= s.end)
                    outgoing.add(n);
            }

            N_TOTAL++;
            //if(outgoing == null)
            //    throw new RuntimeException("!!!");                
            if(hist != null)
                hist.add(outgoing.size());
            
            if(outgoing.size() > 1)
                for(Iterator<DepNode> it = outgoing.iterator(); it.hasNext(); ) {
                    DepNode n = it.next();
                    if(n.relations[0].equals("P"))
                        it.remove();
                }

            if(outgoing.size() == 0)
                continue;
            
            if(outgoing.size() == 1) {
                pa.args.add(outgoing.get(0));
                pa.argLabels.add(s.label);
                N_OK++;
                continue;
            }
            
            if(outgoing.get(1).relations[0].equals("OPRD")) {
                pa.args.add(outgoing.get(0));
                pa.argLabels.add(s.label);
                pa.args.add(outgoing.get(1));
                pa.argLabels.add("C-" + s.label);
                N_OK++;
                continue;
            }
            
            DepNode found = null;
            if(s.label.matches("(C-)?A[0-9]"))
                for(DepNode n: outgoing)
                    if(n.relations[0].matches("SBJ|OBJ|OPRD")) {
                        found = n;
                        break;
                    }
            if(found != null) {
                pa.args.add(found);
                pa.argLabels.add(s.label);
                continue;                
            }

            pa.args.add(outgoing.get(0));
            pa.argLabels.add(s.label);            
        }
        return pa;
    }
    
    public static void convertSpansToDeps(String[] argv) {
        boolean removePunc = Boolean.parseBoolean(argv[1]);
        String conll2008File = argv[2]; // gold standard
        String conll2005File = argv[3]; // *WITH WORDS*
        
        final boolean ALREADY_EXPANDED = false;
        
        try {
            
            Scanner sc1 = new Scanner(new File(conll2008File));
            Scanner sc2 = new Scanner(new File(conll2005File));
            Triple<DepGraph, DepGraph, List<PAStructure>> p = CoNLL2008Format.readNextGraph(sc1);
            if(p != null)
                CoNLL2008Format.mergeSUChains(p.third);
            while(p != null) {
                
                ArrayList<String[]> tuples2005 = readCoNLL2005Sentence(sc2);
                if(tuples2005 == null)
                    throw new RuntimeException("tuples2005 = null");

                /*System.out.println("tuples2005:");
                for(String[] ss: tuples2005)
                    System.out.println(Arrays.toString(ss));
                System.out.println(p.first);
                System.out.println(matches(tuples2005, p.first));
                //System.exit(0);*/
                
                int count = 0;
                while(!matches(tuples2005, p.first)) {
                    count++;
                    if(count == 10)
                        throw new RuntimeException("Can't match...");
                    tuples2005 = readCoNLL2005Sentence(sc2);
                    if(tuples2005 == null)
                        throw new RuntimeException("tuples2005 = null");
                }
                
                IntPair[] oldToNewTokens;
                if(!ALREADY_EXPANDED)
                    oldToNewTokens = getNewTokenIndices(tuples2005, p.first);
                else
                    oldToNewTokens = getDummyTokenIndices(tuples2005, p.first);
                
                ArrayList<SpanPAStructure> pas = getGoldPAs(tuples2005);
                expand(pas, oldToNewTokens, p.first);
                
                removeVs(pas);
                if(removePunc)
                    removePuncs(pas, p.first);
                
                ArrayList<PAStructure> predictedPAs = new ArrayList();
                for(SpanPAStructure spa: pas)
                    predictedPAs.add(convertSpans(spa, p.first));
                
                CoNLL2008Format.printPAStructures(p.first, p.second, predictedPAs);
                System.out.flush();
                
                p = CoNLL2008Format.readNextGraph(sc1);
                if(p != null)
                    CoNLL2008Format.mergeSUChains(p.third);
            }
        
            System.err.println(hist);
            System.err.println(N_OK +"/" + N_TOTAL);
            
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
    
    private static void intersect(ArrayList<SpanPAStructure> predicted,
            ArrayList<SpanPAStructure> gold) {

        for(SpanPAStructure spa: predicted) {            
            for(SpanPAStructure g: gold) {
                if(spa.predIndex != g.predIndex)
                    continue;
                for(Span ps: spa.spans) {
                    for(Span gs: g.spans) {
                        if(!ps.label.equals(gs.label))
                            continue;
                        if(!ps.overlaps(gs))
                            continue;
                        ps.doIntersect(gs);
                    }
                }
            }
        }        
    }
    
    public static void convertSpansToDepsIntersect(String[] argv) {
        boolean removePunc = Boolean.parseBoolean(argv[1]);
        String conll2008File = argv[2]; // gold standard dependencies
        String conll2005FileOutput = argv[3]; // system output segments *WITH WORDS*
        String conll2005FileGold = argv[4]; // gold standard segments *WITH WORDS*
        
        final boolean ALREADY_EXPANDED = false;
        
        try {
            
            Scanner sc1 = new Scanner(new File(conll2008File));
            Scanner sc2 = new Scanner(new File(conll2005FileOutput));
            Scanner sc3 = new Scanner(new File(conll2005FileGold));
            Triple<DepGraph, DepGraph, List<PAStructure>> p = CoNLL2008Format.readNextGraph(sc1);
            if(p != null)
                CoNLL2008Format.mergeSUChains(p.third);
            while(p != null) {
                
                ArrayList<String[]> tuples2005Out = readCoNLL2005Sentence(sc2);
                if(tuples2005Out == null)
                    throw new RuntimeException("tuples2005 = null");

                ArrayList<String[]> tuples2005Gold = readCoNLL2005Sentence(sc3);
                
                /*System.out.println("tuples2005:");
                for(String[] ss: tuples2005)
                    System.out.println(Arrays.toString(ss));
                System.out.println(p.first);
                System.out.println(matches(tuples2005, p.first));
                //System.exit(0);*/
                
                int count = 0;
                while(!matches(tuples2005Out, p.first)) {
                    count++;
                    if(count == 10)
                        throw new RuntimeException("Can't match...");
                    tuples2005Out = readCoNLL2005Sentence(sc2);
                    if(tuples2005Out == null)
                        throw new RuntimeException("tuples2005 = null");
                    tuples2005Gold = readCoNLL2005Sentence(sc3);
                }
                
                IntPair[] oldToNewTokens;
                if(!ALREADY_EXPANDED)
                    oldToNewTokens = getNewTokenIndices(tuples2005Out, p.first);
                else
                    oldToNewTokens = getDummyTokenIndices(tuples2005Out, p.first);
                
                ArrayList<SpanPAStructure> pasOut = getGoldPAs(tuples2005Out);
                ArrayList<SpanPAStructure> pasGold = getGoldPAs(tuples2005Gold);
                
                //System.exit(0);
                
                intersect(pasOut, pasGold);
                
                expand(pasOut, oldToNewTokens, p.first);
                
                removeVs(pasOut);
                if(removePunc)
                    removePuncs(pasOut, p.first);
                
                ArrayList<PAStructure> predictedPAs = new ArrayList();
                for(SpanPAStructure spa: pasOut)
                    predictedPAs.add(convertSpans(spa, p.first));
                
                CoNLL2008Format.printPAStructures(p.first, p.second, predictedPAs);
                System.out.flush();
                
                p = CoNLL2008Format.readNextGraph(sc1);
                if(p != null)
                    CoNLL2008Format.mergeSUChains(p.third);
            }

            System.err.println(hist);
            System.err.println(N_OK +"/" + N_TOTAL);
            
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
    
    private static List<PAStructure> retain2005Data(List<SpanPAStructure> spas,
            List<PAStructure> pas) {
        List<PAStructure> out = new ArrayList();
        for(SpanPAStructure spa: spas) {
            PAStructure pa = match(pas, spa);
            if(pa != null)
                out.add(pa);
            //out.add(match(pas, spa));
        }
        for(PAStructure pa: out)
            pa.lemma = pa.lemma.replaceAll("[0-9][0-9]", "99");
        return out;
    }
    
    public static void simplifyDeps(String[] argv) {
        String conll2008File = argv[1];
        String conll2005File = argv[2]; // gold-standard, unexpanded, with words
        
        try {
            
            Scanner sc1 = new Scanner(new File(conll2008File));
            Scanner sc2 = new Scanner(new File(conll2005File));

            Triple<DepGraph, DepGraph, List<PAStructure>> p = CoNLL2008Format.readNextGraph(sc1);
            if(p != null)
                CoNLL2008Format.mergeSUChains(p.third);
            while(p != null) {

                ArrayList<String[]> tuples2005 = readCoNLL2005Sentence(sc2);
                if(tuples2005 == null)
                    throw new RuntimeException("tuples2005 = null");

                int count = 0;
                while(!matches(tuples2005, p.first)) {
                    count++;
                    if(count == 10)
                        throw new RuntimeException("Can't match...");
                    tuples2005 = readCoNLL2005Sentence(sc2);
                    if(tuples2005 == null)
                        throw new RuntimeException("tuples2005 = null");
                }
                IntPair[] oldToNewTokens = getNewTokenIndices(tuples2005, p.first);
                
                ArrayList<SpanPAStructure> pas = getGoldPAs(tuples2005);
                expand(pas, oldToNewTokens, p.first);
                
                List<PAStructure> remaining = retain2005Data(pas, p.third);
                
                CoNLL2008Format.printPAStructures(p.first, p.second, remaining);
                
                p = CoNLL2008Format.readNextGraph(sc1);
                if(p != null)
                    CoNLL2008Format.mergeSUChains(p.third);
            }
        
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
    
    private static void printMissingSentence(ArrayList<String[]> words,
            ArrayList<String[]> poss, Lemmatizer lemmatizer) {
        for(int i = 0; i < words.size(); i++) {
            String word = words.get(i)[0];
            String pos = poss.get(i)[0];
            String lemma = lemmatizer.lookup(word, pos);
            if(lemma == null)
                lemma = "<NONE>";
            if(false) {
                System.out.print((1 + i) + "\t");
                System.out.print(word + "\t_\t_\t_\t");
                System.out.print(word + "\t" + lemma + "\t" + pos);
                System.out.println("\t0\tUNKN\t_");
            } else {
                String pred = words.get(i)[1];
                System.out.println(pred);
            }
        }
        System.out.println();
    }
            
    private static void printMissingText(String[] argv) {
        String conll2008File = argv[1];
        String conll2005File = argv[2];
        String posFile = argv[3];
        String dictFile = argv[4];
        
        try {
            
            Scanner sc1 = new Scanner(new File(conll2008File));
            Scanner sc2 = new Scanner(new File(conll2005File));
            Scanner sc3 = new Scanner(new File(posFile));
            Lemmatizer lemmatizer = new Lemmatizer(dictFile);
            
            Triple<DepGraph, DepGraph, List<PAStructure>> p = CoNLL2008Format.readNextGraph(sc1);
            if(p != null)
                CoNLL2008Format.mergeSUChains(p.third);
            while(p != null) {
                
                ArrayList<String[]> tuples2005 = readCoNLL2005Sentence(sc2);
                ArrayList<String[]> pos2005 = readCoNLL2005Sentence(sc3);
                if(tuples2005 == null)
                    throw new RuntimeException("tuples2005 = null");
                if(tuples2005.size() != pos2005.size())
                    throw new RuntimeException("tuples2005.size() != pos2005.size()");
                
                int count = 0;
                while(!matches(tuples2005, p.first)) {
                    count++;
                    if(count == 10)
                        throw new RuntimeException("Can't match...");

                    printMissingSentence(tuples2005, pos2005, lemmatizer);
                    
                    tuples2005 = readCoNLL2005Sentence(sc2);
                    pos2005 = readCoNLL2005Sentence(sc3);
                    if(tuples2005 == null)
                        throw new RuntimeException("tuples2005 = null");
                    if(tuples2005.size() != pos2005.size())
                        throw new RuntimeException("tuples2005.size() != pos2005.size()");
                }
                
                p = CoNLL2008Format.readNextGraph(sc1);
                if(p != null)
                    CoNLL2008Format.mergeSUChains(p.third);
            }
        
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }        
    }
    
    private static void removeUnusedPreds(List<PAStructure> pas, List<SpanPAStructure> spas) {
        for(Iterator<PAStructure> it = pas.iterator(); it.hasNext(); ) {
            PAStructure pa = it.next();
            boolean saw = false;
            for(SpanPAStructure spa: spas)
                if(spa.predIndex == pa.pred.position) {
                    saw = true;
                    break;
                }
            if(!saw)
                it.remove();
        }
    }
    
    private static void mergeMissingSentencesParsed(String[] argv) {
        String conll2008File = argv[1];
        String conll2005File = argv[2];
        String missingConll2008File = argv[3];
        
        try {
            
            Scanner sc1 = new Scanner(new File(conll2008File));
            Scanner sc2 = new Scanner(new File(conll2005File));
            Scanner sc3 = new Scanner(new File(missingConll2008File));
            
            Triple<DepGraph, DepGraph, List<PAStructure>> p = CoNLL2008Format.readNextGraph(sc1);
            while(p != null) {
                ArrayList<String[]> tuples2005 = readCoNLL2005Sentence(sc2);
                if(tuples2005 == null)
                    throw new RuntimeException("tuples2005 = null");
                
                int count = 0;
                while(!matches(tuples2005, p.first)) {
                    count++;
                    if(count == 10)
                        throw new RuntimeException("Can't match...");

                    Triple<DepGraph, DepGraph, List<PAStructure>> p2 = CoNLL2008Format.readNextGraph(sc3);
                    CoNLL2008Format.printPAStructures(p2.first, p2.second, p2.third);
                    
                    tuples2005 = readCoNLL2005Sentence(sc2);
                    if(tuples2005 == null)
                        throw new RuntimeException("tuples2005 = null");
                }

                CoNLL2008Format.printPAStructures(p.first, p.second, p.third);
                
                p = CoNLL2008Format.readNextGraph(sc1);                
            }
        
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }        
    }
    
    public static void main(String[] argv) {        
        if(argv[0].equals("-expandSpans"))
            expandSpans(argv);
        else if(argv[0].equals("-convertDepsToSpans"))
            convertDepsToSpans(argv);
        else if(argv[0].equals("-convertDepsToSpans2"))
            convertDepsToSpans2(argv);
        else if(argv[0].equals("-convertSpansToDeps"))
            convertSpansToDeps(argv);
        else if(argv[0].equals("-convertSpansToDepsIntersect"))
            convertSpansToDepsIntersect(argv);
        else if(argv[0].equals("-simplifyDeps"))
            simplifyDeps(argv);
        else if(argv[0].equals("-printMissingText"))
            printMissingText(argv);
        else if(argv[0].equals("-mergeMissingSentencesParsed"))
            mergeMissingSentencesParsed(argv);
        else if(argv[0].equals("-checkSpans"))
            checkSpans(argv);
        else
            throw new RuntimeException("Unknown option: " + argv[0]);
    }
    
}
