package mpqareader;

import java.io.*;
import java.util.*;
import java.util.regex.*;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.ml_long.*;
//import srlpostprocess.TokenSemNode;

import se.lth.cs.nlp.nlputils.depgraph.*;
//import se.lth.cs.nlp.depsrl.format.*;

// TODO idf funkar inte här?
// TODO slå ihop med nya

public class LKSubjSentClassifier {

	static Pattern SEN_ID_PAT = Pattern.compile("id\\s*=\\s*\"(.*?)\"");
	static Pattern SEN_START_PAT = Pattern.compile("start\\s*=\\s*\"#(.*?)\"");
	static Pattern SEN_END_PAT = Pattern.compile("end\\s*=\\s*\"#(.*?)\"");
	static Pattern ON_PAT = Pattern.compile("on\\s*=\\s*\"#(.*?)\"");
	static Pattern SCOPE_PAT = Pattern.compile("scope\\s*=\\s*\"(.*?)\"");
	
	private static int extractIntByPattern(String s, Pattern p) {
		String match = extractStringByPattern(s, p);
		if(match == null)
			throw new IllegalArgumentException("Could not find match");
		return Integer.parseInt(match); //extractStringByPattern(s, p));
	}

	private static String extractStringByPattern(String s, Pattern p) {
		Matcher m = p.matcher(s);
		if(!m.find())
			return null;
		String match = m.group(1);
		return match;
	}
	
	private static Pair<String, ArrayList<int[]>> readSentenceSpans(String prefix, String file) throws IOException {
		BufferedReader br = new BufferedReader(new FileReader(prefix + File.separatorChar + file));
		String line = br.readLine();
		while(line != null && !line.contains("provides=\"SENTENCES\""))
			line = br.readLine();
		if(line == null)
			return null;
		String tokenFile = extractStringByPattern(line, SCOPE_PAT);
		if(tokenFile == null)
			tokenFile = file;
		ArrayList<int[]> out = new ArrayList();
		line = br.readLine();
		while(!line.contains("</annotation>")) {
			line = line.trim();
			if(line.startsWith("<e")) {
				if(!line.endsWith("/>") && !line.endsWith("</e>"))
					throw new RuntimeException("Sorry, we expect sentences line by line...");
				int id = extractIntByPattern(line, SEN_ID_PAT);
				int start = extractIntByPattern(line, SEN_START_PAT);
				int end = extractIntByPattern(line, SEN_END_PAT);
				out.add(new int[] { id, start, end });
			}
			line = br.readLine();
		}
		return new Pair(tokenFile, out);
	}

	private static String extractToken(String line) {
		int s_ix = line.indexOf('>');
		int e_ix = line.indexOf('<', s_ix);
		if(s_ix == -1)
			throw new RuntimeException("Sorry, we expect tokens line by line...");
		if(e_ix == -1)
			throw new RuntimeException("Sorry, we expect tokens line by line...");
		String token = line.substring(s_ix + 1, e_ix);
		return Strings.decodeXML(token);
	}
	
	private static String readTokenPreamble(BufferedReader br) throws IOException {
		String line = br.readLine();
		while(!line.contains("\"base\""))
			line = br.readLine();
		int s_ix = line.indexOf(">");
		int e_ix = line.indexOf('<', s_ix);
		String base = line.substring(s_ix + 1, e_ix);
		while(line != null && !line.contains("provides=\"TOKENS\""))
			line = br.readLine();
		if(line == null)
			throw new RuntimeException("Couldn't find token annotation");
		return base;
	}
	
	private static void readPreamble(BufferedReader br, String prov) throws IOException {
		String line = br.readLine();
		while(line != null && !line.contains("provides=\"" + prov + "\""))
			line = br.readLine();
		if(line == null)
			throw new RuntimeException("Couldn't find annotation layer " + prov);
	}
	
	private static final int LOOKAHEAD = 256;
	
	private static Pair<ArrayList<String>, Triple> readNextSentence(int[] sen, BufferedReader br,
			BufferedReader pbr, BufferedReader lbr) throws IOException {

		if(!lbr.markSupported())
			throw new RuntimeException("!!!");
		
		int senId = sen[0];
		int start = sen[1];
		int end = sen[2];
		ArrayList<String> out = new ArrayList();
		ArrayList<String> tags = new ArrayList();
		ArrayList<String> lemmas = new ArrayList();

		/*String line = br.readLine().trim();
		while(line.equals(""))
			line = br.readLine().trim();
		if(!line.startsWith("<e") || !line.endsWith("</e>")) {
			System.err.println("line = " + line);
			throw new RuntimeException("Sorry, we expect tokens line by line...");
		}
		int id = extractIntByPattern(line, SEN_ID_PAT);
		
		while(id != start) {
			line = br.readLine().trim();
			while(line.equals(""))
				line = br.readLine().trim();
			if(line.contains("</annotation>")) {
				throw new RuntimeException("Problem finding sentence " 
						+ senId + " start=" + start + " end=" + end);				
			}
			if(!line.startsWith("<e") || !line.endsWith("</e>")) {
				System.err.println("line = " + line);
				throw new RuntimeException("Sorry, we expect tokens line by line...");
			}
			id = extractIntByPattern(line, SEN_ID_PAT);
		}
		out.add(extractToken(line));*/

		boolean inside = false;
		
		int id = -1;
		String line;
		
		while(id != end) {
			line = br.readLine().trim();
			while(line.equals(""))
				line = br.readLine().trim();
			if(!line.startsWith("<e") || !line.endsWith("</e>"))
				throw new RuntimeException("Sorry, we expect tokens line by line...");
			id = extractIntByPattern(line, SEN_ID_PAT);

			if(id == start)
				inside = true;

			if(inside) {

				pbr.mark(LOOKAHEAD);
				String posLine = pbr.readLine().trim();
				if(!posLine.startsWith("</annotation")
				    && (!posLine.startsWith("<e") 
						|| !posLine.endsWith("</e>"))) {
					System.err.println("end = " + end);
					System.err.println("line = |" + line + "|");
					System.err.println("posLine = |" + posLine + "|");
					throw new RuntimeException("Sorry, we expect POS tags line by line...");
				}

				if(posLine.startsWith("</annotation"))
					break;
				int on = extractIntByPattern(posLine, ON_PAT);
				if(on != id) {
					//throw new RuntimeException("POS on = " + on + ", token id = " + id);
					System.err.println("*** Warning: POS on = " + on + ", token id = " + id);
					pbr.reset();
					continue;
				}
				
				lbr.mark(LOOKAHEAD);
				String lemmaLine = lbr.readLine().trim();
				if(!lemmaLine.startsWith("</annotation")) {
					if((!lemmaLine.startsWith("<e") 
							|| !lemmaLine.endsWith("</e>"))) {
						System.err.println("line = |" + line + "|");
						System.err.println("posLine = |" + posLine + "|");
						System.err.println("lemmaLine = |" + lemmaLine + "|");
						throw new RuntimeException("Sorry, we expect lemmas line by line...");
					}
				} else {
					lbr.reset();
				}
				
				if(false) {
					System.err.println(line);
					System.err.println(posLine);
					System.err.println(lemmaLine);
				}

				String lemma = null;

				if(!lemmaLine.equals("</annotation>")) {
					on = extractIntByPattern(lemmaLine, ON_PAT);
					if(on != id) {
						lbr.reset();
					} else
						lemma = extractToken(lemmaLine);
				}

				String token = extractToken(line);
				String posTag = extractToken(posLine);

				out.add(token);
				tags.add(posTag);
				lemmas.add(lemma);
			}
		}
		
		DepGraph dg = new DepGraph(out.size());
		for(int i = 1; i < dg.nodes.length; i++) {
			dg.nodes[i] = new DepNode();
			dg.nodes[i].word = out.get(i - 1);
			dg.nodes[i].pos = tags.get(i - 1);
			dg.nodes[i].lemma = lemmas.get(i - 1);
		}

		/*for(int i = 1; i < dg.nodes.length; i++) {
			System.out.println(i + "\t" + dg.nodes[i].word + "\t" + dg.nodes[i].pos + "\t" + dg.nodes[i].lemma);
			System.out.flush();			
		}
		System.out.println();
		System.out.flush();*/
		
		return new Pair(out, new Triple(dg, dg, null));
	}

	public static void processDirectory(SymbolEncoder enc, 
			Classifier<SparseVector, Integer> cl,
			SubjectivityLexicon subjLex, 
			String dirName, String outputDir) throws IOException {

		long t0 = System.currentTimeMillis();
		int countSentences = 0;
		
		File dir = new File(dirName);
		if(!dir.isDirectory())
			throw new IllegalArgumentException(dir + " is not a directory");
		for(File file: dir.listFiles()) {
			if(file.isDirectory())
				continue;
			Pair<String, ArrayList<int[]>> p = readSentenceSpans(dirName, file.getName());
			if(p == null)
				continue;
			String tokenFile = p.left;
			ArrayList<int[]> senSpans = p.right;

			System.out.println("Sentence file " + file.getName() + ", token file " + tokenFile);
			
			BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(dirName + File.separatorChar + tokenFile), "UTF-8"));
			String base = readTokenPreamble(br);

			String outFile;
			if(tokenFile.endsWith(".xml"))
				outFile = tokenFile.replaceFirst("\\.[^\\.]+\\.xml$", ".subjsen.xml");
			else 
				outFile = tokenFile + ".subjsen.xml";
			outFile = outputDir + File.separatorChar + outFile;

			String posFile = null;
			if(tokenFile.endsWith(".xml"))
				posFile = tokenFile.replaceFirst("\\.[^\\.]+\\.xml$", ".pos.xml");
			else 
				posFile = tokenFile + ".pos.xml";

			BufferedReader posReader = null;
			BufferedReader lemmaReader = null;
			try {
				posReader = new BufferedReader(new InputStreamReader(new FileInputStream(dirName + File.separatorChar + posFile), "UTF-8"));
				lemmaReader = new BufferedReader(new InputStreamReader(new FileInputStream(dirName + File.separatorChar + posFile), "UTF-8"));
			} catch(Exception e) {
				System.err.println("Sorry, we expect the POS file name to end with .pos.xml");
				System.exit(1);
			}
			readPreamble(posReader, "POS");
			readPreamble(lemmaReader, "LEMMA");
			
			PrintWriter out = new PrintWriter(new OutputStreamWriter(new FileOutputStream(outFile), "UTF-8"));
			printSubjSenPreamble(base, file.getName(), out);

			int idCounter = 0;
			
			for(int[] senSpan: senSpans) {
				Pair<ArrayList<String>, Triple> pp = readNextSentence(senSpan, br, posReader, lemmaReader);
				ArrayList<String> sen = pp.left;			
				
				if(sen == null)
					throw new RuntimeException("null");
				
				SparseVector sv = BOWSubjSentClassifier.representSentence(sen, pp.right, null, null, null, false, subjLex, enc);
				if(BOWSubjSentClassifier.SET)
					for(int j = 0; j < sv.index; j++)
						sv.values[j] = 1.0;
				if(BOWSubjSentClassifier.NORMALIZE)
					BOWSubjSentClassifier.normalize2(sv);				
				int guess = cl.classify(sv);
				if(guess == BOWSubjSentClassifier.SUBJ)
					printSubjSen(++idCounter, senSpan[0], out);
				countSentences++;
			}
			printSubjSenEnd(out);
			out.close();
			
			posReader.close();
			lemmaReader.close();
			br.close();
		}
		
		long t1 = System.currentTimeMillis();
		double ssPerSec = 1000.0 * countSentences / (t1 - t0);

		System.out.format("Processed %f sentences per second.", ssPerSec);
		
	}
	
	public static void main(String[] argv) {
		String modelFile = argv[0];
		String lkDir = argv[1];
		String outputDir = argv[2];
		
		try {
			ObjectInputStream ois = new ObjectInputStream(new FileInputStream(modelFile));
			SymbolEncoder enc = (SymbolEncoder) ois.readObject();
			Classifier<SparseVector, Integer> cl = (Classifier) ois.readObject();

			SubjectivityLexicon subjLex;
			if(BOWSubjSentClassifier.USE_SUBJLEX) {
				Object o = ois.readObject();
				if(o instanceof HashMap)
					subjLex = new SubjectivityLexicon((HashMap) o);
				else if(o instanceof SubjectivityLexicon)
					subjLex = (SubjectivityLexicon) o;
				else
					throw new RuntimeException("!!!");
			}
			else
				subjLex = null;
			processDirectory(enc, cl, subjLex, lkDir, outputDir);
			
		} catch(Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
		
	}

	private static void printSubjSenPreamble(String base, String sentenceFileName, PrintWriter out) throws IOException {
		out.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
		out.println("<lk-annotation>");
		out.println("<meta-info>");
		out.println(" <tag name=\"base\">" + base + "</tag>");
		out.println(" <tag name=\"annotator\">BOWSubjSentClassifier</tag>");
		out.println("</meta-info>");
		out.println("<annotation scope=\"" + sentenceFileName + "\" provides=\"SUBJECTIVE-SENTENCES\">");
	}
	
	private static void printSubjSen(int nextId, int senId, PrintWriter out) throws IOException {
		out.println(" <e id=\"" + nextId + "\" on=\"#" + senId + "\"><ss/></e>");		
	}

	private static void printSubjSenEnd(PrintWriter out) throws IOException {
		out.println("</annotation>");
		out.println("</lk-annotation>");
	}
	
}
