package mpqa_structlearn.docpolarity;

import gnu.trove.*;

import java.io.*;
import java.util.*;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.*;
import se.lth.cs.nlp.nlputils.ml_long.*;

import mpqa_structlearn.io.*;

public class DocPolarity {

	public static void main(String[] argv) {
		if(argv[0].equals("-train"))
			train(argv);
		else if(argv[0].equals("-test"))
			test(argv);
		else if(argv[0].equals("-train2cv"))
			train2cv(argv);
		else {
			System.err.println("Illegal arguments: " + Arrays.toString(argv));
			System.exit(1);
		}
	}	
	
	public static void train2cv(String[] argv) {
		String ansFile = argv[1]; System.out.println("ansFile = " + ansFile);
		String lkDir = argv[2]; System.out.println("lkDir = " + lkDir);
		String filelist = argv[3]; System.out.println("filelist = " + filelist);
		String subjLexFile = argv[4]; System.out.println("subjLexFile = " + subjLexFile);
		int nfolds = Integer.parseInt(argv[5]); System.out.println("nfolds = " + nfolds);

		boolean useSubjExpr = Boolean.parseBoolean(argv[6]); System.out.println("useSubjExpr = " + useSubjExpr);
		
		String outFile = argv[7]; System.out.println("outFile = " + outFile);
		
		String algName = argv[8]; System.out.println("algName = " + algName);
		String algArgs = argv[9]; System.out.println("algArgs = " + algArgs);
		
		try {
			ArrayList<Pair<SparseVector, Integer>> pos = new ArrayList();
			ArrayList<Pair<SparseVector, Integer>> neg = new ArrayList();

			SymbolEncoder enc = new SymbolEncoder();
			IntHistogram dfs = new IntHistogram();
			
			ArrayList<String> files = new ArrayList();
			Util.readLines(filelist, files);
			
			LKReader r = new LKReader(lkDir, subjLexFile, files);
			BufferedReader br = Util.openFileReader(ansFile);
			
			int count = 0;
			while(r.hasNext()) {
				MPQAText text = r.next();
				String line = br.readLine();
				count++;				

				boolean isPos = isPositive(line);
				boolean isNeg = isNegative(line);

				//if(isPos || isNeg) {
					pos.add(new Pair(represent(text, enc, dfs, useSubjExpr, false), isPos? +1: -1));
					neg.add(new Pair(represent(text, enc, dfs, useSubjExpr, false), isNeg? +1: -1));

					//System.out.println(isPos + "\t" + isNeg + "\t" + line);
					//System.out.println(text.sentences.get(0).dg);
				
			    //}
				if(count % 100 == 0) {
					System.out.print(".");
					System.out.flush();
				}
			}
			System.out.println();
			System.out.println("Read " + count + " examples.");
			
			Collections.shuffle(pos, new Random(0));
			Collections.shuffle(neg, new Random(0));

			if(USE_IDF) {
				double[] idfs = createIDFTable(pos.size(), enc.size(), dfs);
				reweightIDF(pos, idfs);
				reweightIDF(neg, idfs);
			}
			else if(USE_PRESENCE) {
				reweightPresence(pos);
				reweightPresence(neg);
			}
			if(NORMALIZE) {
				normalize2norm(pos);
				normalize2norm(neg);
			}
						
			AlgorithmFactory af = new AlgorithmFactory();
			AlgorithmFactory.setVerbosity(0);
			
			PrintWriter pw = new PrintWriter(new FileWriter(outFile));
			List<Integer> range = Arrays.asList(new Integer[] { -1, 1 });

			double[] scores = new double[2];
			
			int ncorr = 0, ntotal = 0;
			int ncorrp = 0, ngoldp = 0, nguessp = 0;
			int ncorrn = 0, ngoldn = 0, nguessn = 0;
			for(int i = 0; i < nfolds; i++) {
				LearningAlgorithm alg = af.create(algName, algArgs);
				EncodedMulticlassDefinition def = new EncodedMulticlassDefinition();

				ArrayList<Pair<SparseVector, Integer>> trsP = new ArrayList();
				ArrayList<Pair<SparseVector, Integer>> tesP = new ArrayList();			
				CollectionUtils.cvSplit(pos, trsP, tesP, nfolds, i);
				if(trsP.size() + tesP.size() != pos.size())
					throw new RuntimeException("!!!");

				ArrayList<Pair<SparseVector, Integer>> trsN = new ArrayList();
				ArrayList<Pair<SparseVector, Integer>> tesN = new ArrayList();
				CollectionUtils.cvSplit(neg, trsN, tesN, nfolds, i);
				if(trsN.size() + tesN.size() != neg.size())
					throw new RuntimeException("!!!");

				Classifier<SparseVector, Integer> clP = alg.train(def, trsP);
				Classifier<SparseVector, Integer> clN = alg.train(def, trsN);

				for(int j = 0; j < tesN.size(); j++) {
					Pair<SparseVector, Integer> p = tesP.get(j);
					Pair<SparseVector, Integer> n = tesN.get(j);

					clP.computeScoresRestricted(p.left, range, scores);					
					pw.format("%d %f ", p.right, scores[1]);

					//scores[1] = 1;
					boolean posCorr = p.right.equals(1) && scores[1] > 0
					|| p.right.equals(-1) && scores[1] <= 0;
					
					if(scores[1] > 0) {
						if(p.right.equals(1))
							ncorrp++;
						nguessp++;
					}					
					if(p.right.equals(1))
						ngoldp++;
					
					clN.computeScoresRestricted(n.left, range, scores);					
					pw.format("%d %f\n", n.right, scores[1]);

					//scores[1] = -1;
					boolean negCorr = n.right.equals(1) && scores[1] > 0
					|| n.right.equals(-1) && scores[1] <= 0;
					
					if(scores[1] > 0) {
						if(n.right.equals(1))
							ncorrn++;
						nguessn++;
					}					
					if(n.right.equals(1))
						ngoldn++;
					
					ntotal++;
					if(posCorr && negCorr)
						ncorr++;
				}
			}
			pw.close();
			
			double pp = (double) ncorrp / nguessp;
			double pr = (double) ncorrp / ngoldp;
			double pf = 2*pp*pr/(pp+pr);
			double np = (double) ncorrn / nguessn;
			double nr = (double) ncorrn / ngoldn;
			double nf = 2*np*nr/(np+nr);

			System.out.format("Accuracy: %f\n", (double) ncorr / ntotal);
			System.out.format("Positive: %f %f %f\n", pp, pr, pf);
			System.out.format("Negative: %f %f %f\n", np, nr, nf);			
		} catch(Exception e) {
			e.printStackTrace();
		}
		
	}
	

	private static boolean isPositive(String line) {
		int ix = line.indexOf("##");
		if(ix == -1)
			throw new RuntimeException("!!!");
		String prefix = line.substring(0, ix);
		return prefix.contains("[+");
	}
	private static boolean isNegative(String line) {
		int ix = line.indexOf("##");
		if(ix == -1)
			throw new RuntimeException("!!!");
		String prefix = line.substring(0, ix);
		return prefix.contains("[-");
	}

	public static void train(String[] argv) {
		String posDir = argv[1]; System.out.println("posDir = " + posDir);
		String negDir = argv[2]; System.out.println("negDir = " + negDir);
		String subjLexFile = argv[3]; System.out.println("subjLexFile = " + subjLexFile);
		String docListFile = argv[4]; System.out.println("docListFile = " + docListFile);

		boolean useSubjExpr = argv[5].equals("t"); System.out.println("useSubjExpr = " + useSubjExpr);
		boolean useSubjLex = argv[6].equals("t"); System.out.println("useSubjLex = " + useSubjLex);
		
		String algName = argv[7]; System.out.println("algName = " + algName);
		String algArgs = argv[8]; System.out.println("algArgs = " + algArgs);
		
		String outModelFile = argv[9]; System.out.println("outModelFile = " + outModelFile);

		try {
			Random rand = new Random(0);
			ArrayList<Pair<SparseVector, Integer>> ts = new ArrayList();

			SymbolEncoder enc = new SymbolEncoder();
			IntHistogram dfs = new IntHistogram();
			
			ArrayList<String> selectedFiles = new ArrayList();
			Util.readLines(docListFile, selectedFiles);
			ArrayList<String> posFiles = prepareFileList(selectedFiles, true);
			LKReader r = new LKReader(posDir, subjLexFile, posFiles);
			
			System.out.println("Processing positive set...");
			int count = 0;
			while(r.hasNext()) {
				MPQAText text = r.next();
				count++;		
				//System.out.println(text.source);				
				ts.add(new Pair(represent(text, enc, dfs, useSubjExpr, useSubjLex), +1));
			}

			ArrayList<String> negFiles = prepareFileList(selectedFiles, false);
			r = new LKReader(negDir, subjLexFile, negFiles);
			
			System.out.println("Processing negative set...");
			count = 0;
			while(r.hasNext()) {
				MPQAText text = r.next();
				count++;
				//System.out.println(text.source);
				ts.add(new Pair(represent(text, enc, dfs, useSubjExpr, useSubjLex), -1));
			}

			Collections.shuffle(ts, rand);
			double[] idfs = createIDFTable(ts.size(), enc.size(), dfs);
			if(USE_IDF)
				reweightIDF(ts, idfs);			
			else if(USE_PRESENCE)
				reweightPresence(ts);
			if(NORMALIZE)
				normalize2norm(ts);
						
			AlgorithmFactory af = new AlgorithmFactory();
            LearningAlgorithm alg = af.create(algName, algArgs);
            EncodedMulticlassDefinition def = new EncodedMulticlassDefinition();
            Classifier<SparseVector, Integer> cl = alg.train(def, ts);
			
            if(true) {
            	enc.freeze();
            	ObjectOutputStream oos 
            	= new ObjectOutputStream(new FileOutputStream(outModelFile));
            	oos.writeObject(enc);
            	oos.writeObject(cl);
            	oos.writeObject(idfs);
            	oos.writeBoolean(useSubjExpr);
            	oos.writeBoolean(useSubjLex);
            	oos.close();
			}
			
            if(false)
				printWeights(enc, cl, idfs);
            
		} catch(Exception e) {
			e.printStackTrace();
		}
		
	}

	private static void printWeights(SymbolEncoder enc,
			Classifier<SparseVector, Integer> cl,
			double[] idfs) {
		ArrayList<String> encInv = enc.inverse();		
		int max = enc.size();		
		SparseVector sv = new SparseVector();
		ArrayList<Integer> range = new ArrayList();
		range.add(+1);
		range.add(-1);
		double[] scores = new double[2];
		cl.computeScoresRestricted(new SparseVector(), range, scores);
		double bias = scores[0];
		System.out.println("bias = " + bias);		
		ArrayList<DoubleObjPair<String>> l = new ArrayList();		
		sv.index = 1;
		for(int i = 1; i < max; i++) {
			sv.keys[0] = i;
			sv.values[0] = idfs[i];
			cl.computeScoresRestricted(sv, range, scores);
			String k = encInv.get(i);			
			l.add(new DoubleObjPair(scores[0] - bias, k));			
			//System.out.println(i + "\t" + scores[0] + "\t" + scores[1]);
		}
		Collections.sort(l, new NegatedComparator(DoubleObjPair.BY_LEFT));
		for(DoubleObjPair<String> p: l)
			System.out.println(p);		
	}
	
	public static void test(String[] argv) {
		String posDir = argv[1]; System.out.println("posDir = " + posDir);
		String negDir = argv[2]; System.out.println("negDir = " + negDir);
		String subjLexFile = argv[3]; System.out.println("subjLexFile = " + subjLexFile);
		String docListFile = argv[4]; System.out.println("docListFile = " + docListFile);
		String modelFile = argv[5]; System.out.println("modelFile = " + modelFile);
		
		String detailFile = null;
		if(argv.length > 6) {
			detailFile = argv[6]; System.out.println("detailFile = " + detailFile);
		}

		try {
			ObjectInputStream ois = new ObjectInputStream(Util.openFileStream(modelFile));
			SymbolEncoder enc = (SymbolEncoder) ois.readObject();
			Classifier<SparseVector, Integer> cl = (Classifier) ois.readObject();
			double[] idfs = (double[]) ois.readObject();
			boolean useSubjExpr = ois.readBoolean();
			boolean useSubjLex = false; //
			try {
				useSubjLex = ois.readBoolean();
			} catch(Exception e) {
				System.out.println("Warning: setting subjlex to false.");
			}
			ois.close();
			
			ArrayList<Pair<SparseVector, Integer>> ts = new ArrayList();

			ArrayList<String> selectedFiles = new ArrayList();
			Util.readLines(docListFile, selectedFiles);
			ArrayList<String> posFiles = prepareFileList(selectedFiles, true);
			LKReader r = new LKReader(posDir, subjLexFile, posFiles);
			
			System.out.println("Processing positive set...");
			while(r.hasNext()) {
				MPQAText text = r.next();
				ts.add(new Pair(represent(text, enc, null, useSubjExpr, useSubjLex), +1));
			}

			ArrayList<String> negFiles = prepareFileList(selectedFiles, false);
			r = new LKReader(negDir, subjLexFile, negFiles);
			
			System.out.println("Processing negative set...");
			while(r.hasNext()) {
				MPQAText text = r.next();
				ts.add(new Pair(represent(text, enc, null, useSubjExpr, useSubjLex), -1));
			}

			if(USE_IDF)
				reweightIDF(ts, idfs);
			else if(USE_PRESENCE)
				reweightPresence(ts);
			if(NORMALIZE)
				normalize2norm(ts);

			PrintWriter detailPW = null;
			if(detailFile != null)
				detailPW = new PrintWriter(new FileWriter(detailFile));		

			List<Integer> polRange = Arrays.asList(new Integer[] { +1, -1 }); 
			double[] clScores = new double[polRange.size()];
			
			int ncorrect = 0;
			
			for(Pair<SparseVector, Integer> p: ts) {
				cl.computeScoresRestricted(p.left, polRange, clScores);
				if(!Util.equal(clScores[0] + clScores[1], 0))
					throw new RuntimeException("!!!");
				
				int guess = cl.classify(p.left);
				if(p.right.equals(guess))
					ncorrect++;
				if(detailPW != null) {
					detailPW.print(p.right.equals(guess)? "+": "-");
					detailPW.print("\t");
					detailPW.print(p.right);
					detailPW.print("\t");
					detailPW.print(clScores[0]);
					detailPW.println();
				}
			}
			double acc = (double) ncorrect / ts.size();
			System.out.println("Accuracy: " + acc);

			if(detailPW != null)
				detailPW.close();
			
		} catch(Exception e) {
			e.printStackTrace();
		}
		
	}
	
	private static void normalize2norm(ArrayList<Pair<SparseVector, Integer>> ts) {
		for(Pair<SparseVector, Integer> p: ts) {
			SparseVector sv = p.left;
			double sum = 0;
			for(int i = 0; i < sv.index; i++)
				sum += sv.values[i]*sv.values[i];
			double factor = 1.0 / Math.sqrt(sum);
			for(int i = 0; i < sv.index; i++)
				sv.values[i] *= factor;
		}
	}

	private static void reweightIDF(ArrayList<Pair<SparseVector, Integer>> ts,
			double[] idfs) {
		for(Pair<SparseVector, Integer> p: ts) {
			SparseVector sv = p.left;
			for(int i = 0; i < sv.index; i++) {
				double idf = idfs[(int) sv.keys[i]];
				sv.values[i] *= idf;
			}				
		}
	}

	private static void reweightPresence(ArrayList<Pair<SparseVector, Integer>> ts) {
		for(Pair<SparseVector, Integer> p: ts) {
			SparseVector sv = p.left;
			for(int i = 0; i < sv.index; i++) {
				sv.values[i] = 1;
			}				
		}
	}
	
//	private static final boolean USE_IDF = false;
//	private static final boolean USE_PRESENCE = true;
//	
//	private static final boolean NORMALIZE = true;
//	
//	private static final boolean USE_POS = true;
//	private static final boolean USE_POS_BIGRAMS = true;
//	private static final boolean USE_BIGRAMS = false;
//	
//	private static final boolean USE_LEMMAS = false;
//
//	private static final boolean USE_SUBJDICT = false;
	
	private static final boolean USE_IDF = false;
	private static final boolean USE_PRESENCE = false;
	
	private static final boolean NORMALIZE = true;
	
	private static final boolean USE_POS = false;
	private static final boolean USE_POS_BIGRAMS = false;
	private static final boolean USE_BIGRAMS = false;
	
	private static final boolean USE_LEMMAS = false;

	private static SparseVector represent(MPQAText text, SymbolEncoder enc,
			IntHistogram dfs, boolean useSubjExpr, boolean useSubjLex) {
		SparseVector sv = new SparseVector();
		TIntHashSet seen = new TIntHashSet();
		for(MPQASentence sen: text.sentences) {

			for(int i = 1; i < sen.dg.nodes.length; i++) {
				String w;
				if(USE_LEMMAS) {
					w = sen.dg.nodes[i].lemma;
					if(w == null || w.equals("_"))
						w = sen.dg.nodes[i].word;
				} else
					w = sen.dg.nodes[i].word;				
				put(normalize(w), sv, enc, seen);

				if(useSubjLex) {
					String clue = sen.slClues.get(i);
					if(clue != null && !clue.equals("_")) {
						put(clue, sv, enc, seen);
						put(clue + "/" + normalize(w), sv, enc, seen);
					}
				}
			}
			
			if(USE_POS)
				for(int i = 1; i < sen.dg.nodes.length; i++)
					put(sen.dg.nodes[i].pos, sv, enc, seen);

			if(USE_BIGRAMS && sen.dg.nodes.length > 1) {
				String prev = normalize(sen.dg.nodes[1].word);
				for(int i = 2; i < sen.dg.nodes.length; i++) {
					String curr = normalize(sen.dg.nodes[i].word); 
					put(prev + "__" + curr, sv, enc, seen);
					//System.out.println(prev + "__" + curr);
					prev = curr;
				}
			}
			
			if(USE_POS_BIGRAMS && sen.dg.nodes.length > 1) {
				String prev = normalize(sen.dg.nodes[1].word);
				String prevPOS = sen.dg.nodes[1].pos;
				for(int i = 2; i < sen.dg.nodes.length; i++) {
					String currPOS = sen.dg.nodes[i].pos; 
					String curr = normalize(sen.dg.nodes[i].word); 
					
					//put(prev + "__" + curr, sv, enc, seen);
					put(prev + "__" + currPOS, sv, enc, seen);
					put(prevPOS + "__" + curr, sv, enc, seen);
										
					//System.out.println(prevPOS + "__" + curr);
							
					prev = curr;
					prevPOS = currPOS;
				}
			}

			if(useSubjExpr)
				for(MPQAExpression ex: sen.exprs) {
					if(ex.type.equals("os"))
						continue;
					String pol = ex.polarity;
					//String inten = ex.intensity;
					for(DepNode n: ex.span) {
						String w = normalize(n.word);
						put("TYPEw: " + ex.type + "/" + w, sv, enc, seen);
						put("POLw: " + pol + "/" + w, sv, enc, seen);					

						//put("POL/INTw: " + pol + "/" + inten + "/" + w, sv, enc, seen);						
					
					}
					put("POL: " + pol, sv, enc, seen);
					put("POL/INT: " + pol + "/" + ex.intensity, sv, enc, seen);					
				}
		
		}

		sv.sortIndices();
		sv.trim();

		if(dfs != null)
			for(TIntIterator it = seen.iterator(); it.hasNext(); )
				dfs.add(it.next());
		
		return sv;
	}

	private static void put(String feature, SparseVector sv, SymbolEncoder enc, TIntHashSet seen) {
		int k = enc.encode(feature);
		sv.put(k);
		seen.add(k);
	}
	
	private static String normalize(String w) {
		w = w.toLowerCase();
		w = w.replaceAll("[0-9]", "0");
		return w;
	}
	
	private static double[] createIDFTable(int ndocs, int max, IntHistogram dfs) {
		double[] out = new double[max+1];
		for(int i = 0; i < out.length; i++) {
			int df = dfs.getFrequency(i);
			if(df > 0)
				out[i] = Math.log(ndocs / df);
		}
		return out;
	}
	
	private static ArrayList<String> prepareFileList(ArrayList<String> l,
			boolean isPositive) {
		ArrayList<String> out = new ArrayList(l);
		for(ListIterator<String> it = out.listIterator(); it.hasNext(); ) {
			String f = it.next();
			if(isPositive && f.contains("/neg/") || !isPositive && f.contains("/pos/"))
				it.remove();
			else {
				int ix = f.lastIndexOf("/");
				it.set(f.substring(ix + 1));
			}
		}
		return out;
	}
	
}
