package mpqa_structlearn.fullsystem;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.*;

import mpqa_structlearn.io.*;
import mpqa_structlearn.locallink.LocalLinkClassifier;
import mpqa_structlearn.polarity.BaselinePolarityClassifier;
import mpqa_structlearn.sequences.MPQASeqExtractor;
import se.lth.cs.nlp.nlputils.core.*;

import se.lth.cs.nlp.opinions.*;

public class EPEPipeline {

    private static final String DIR_SEP = System.getProperty("file.separator");
    
    private static final int EVAL = Integer.MIN_VALUE;
    
    private static Pair<ArrayList<String>[], ArrayList<String>[]> makeCVSplit(ArrayList<String> files, int nfolds) {
        ArrayList<String>[] folds = new ArrayList[nfolds];
        for(int i = 0; i < nfolds; i++)
            folds[i] = new ArrayList();
        int foldSize = files.size() / nfolds;
        for(int i = 0; i < nfolds; i++) {
            for(int j = i*foldSize; j < (i+1)*foldSize; j++)
                folds[i].add(files.get(j));
        }
        for(int j = foldSize*nfolds; j < files.size(); j++)
            folds[nfolds-1].add(files.get(j));
        
        ArrayList<String>[] trainingSets = new ArrayList[nfolds];
        for(int i = 0; i < nfolds; i++) {
            trainingSets[i] = new ArrayList();
            for(int j = 0; j < nfolds; j++) {
                if(j == i)
                    continue;
                trainingSets[i].addAll(folds[j]);
            }
        }
        
        /*System.out.println(files.size());
        for(int i = 0; i < nfolds; i++) {
            System.out.println(folds[i].size() + " " + trainingSets[i].size() + " " 
                    + (folds[i].size()+trainingSets[i].size()));
        }
        System.exit(0);*/
        
        if(false) {
            System.out.println(files.get(0));
            System.out.println(files.get(files.size()-1));

            ArrayList<String> tmp = new ArrayList(folds[0]);
            tmp.addAll(trainingSets[0]);

            System.out.println(files.equals(tmp));

            Collections.sort(tmp);
            
            for(int i = 1; i < nfolds; i++) {
                ArrayList<String> tmp2 = new ArrayList(folds[i]);
                tmp2.addAll(trainingSets[i]);
                Collections.sort(tmp2);
                System.out.println(tmp.equals(tmp2));
            }
            
            System.exit(0);
        }
        
        
        return new Pair(trainingSets, folds);
    }
    
    private static Iterator<MPQAText> makeTextIterator(String mode, String dir, 
            String subjLexFile, ArrayList<String> files, String epeDir,
            EPEFeatureSpec featureSpec) throws IOException {
        if(mode.equals("EPE"))
            return new MPQAReaderEPE(dir, subjLexFile, files, epeDir, featureSpec);
        else {
            ArrayList<String> l = new ArrayList(files);
            Collections.sort(l);
            for(ListIterator<String> it = l.listIterator(); it.hasNext(); ) {
                String s = it.next();
                s = s.replaceFirst(".*docs/", "");
                s = s.replaceFirst("/", "_");
                it.set(s);
            }
            return new LKReader(dir, subjLexFile, l);
        }
    }
    
    private static void trainSeqLabeler(ArrayList<String> selectedFiles, String prefix,
            String mode, String dir, String subjLexFile, String epeDir, String seqTemplate,
            EPEFeatureSpec featureSpec) throws IOException {

        if(!new File(prefix).exists()) {
            PrintWriter pw = Util.openFileWriter(prefix, "UTF-8");
            for(Iterator<MPQAText> it = makeTextIterator(mode, dir, subjLexFile, selectedFiles, epeDir, featureSpec);
                    it.hasNext(); ) {
                MPQAText text = it.next();
                System.out.println(text.source);
                MPQASeqExtractor.extractSequences(text, pw, false, false, true);
            }
            pw.close();
        }
        
        String modelFile = prefix + ".model";

        if(!new File(modelFile).exists()) {

            String secondOrder = "false";
            String costType = "1"; // for hamming loss

            String methodName = "OnlineAlgorithm";
            String methodArgs = "nRounds=10 printDots=100 useNewLinear=true modelSize=1400000 PAUpdate C=0.1 sorted=true sqrt=true maximizeLoss=true";

            String[] argv = new String[] { null, prefix, seqTemplate, modelFile, 
                    secondOrder, costType, methodName, methodArgs };

            SeqLabeler.train(argv);
        }
        //new File(prefix).delete();        
    }

    private static void generateSeqCandidates(ArrayList<String> selectedFiles, String prefix,
            int fold, int k, String mode, String dir, String subjLexFile, String epeDir,
            EPEFeatureSpec featureSpec) throws IOException {
        
        String foldSeqFile = fold != EVAL? 
                prefix + DIR_SEP + "seq_fold" + fold:
                prefix + DIR_SEP + "seq_eval";
        String candFile = foldSeqFile + ".kbest";
        
        if(!new File(foldSeqFile).exists()) {
            PrintWriter pw = Util.openFileWriter(foldSeqFile, "UTF-8");
            for(Iterator<MPQAText> it = makeTextIterator(mode, dir, subjLexFile, selectedFiles, epeDir, featureSpec); it.hasNext(); ) {
                MPQAText text = it.next();
                System.out.println(text.source);
                MPQASeqExtractor.extractSequences(text, pw, false, false, true);
            }
            pw.close();
        }
        
        if(!new File(candFile).exists()) {
            String modelFile = fold != EVAL?
                    prefix + DIR_SEP + "seq" + fold + ".model":
                    prefix + DIR_SEP + "seq.model";
            SeqLabeler.tagKBest(new String[] { null, modelFile, "" + k, foldSeqFile, candFile });
            // NB: evaluation scores printed after this step are *oracle* scores!
        }
        
    }
    
    private static void mergeSeqCandidates(String prefix, int nfolds) throws IOException {
        String candFileMerged = prefix + DIR_SEP + "seq.kbest";
        if(!new File(candFileMerged).exists()) {
            PrintWriter pw = Util.openFileWriter(candFileMerged, "UTF-8");
            for(int i = 0; i < nfolds; i++) {
                String candFileFold = prefix + DIR_SEP + "seq_fold" + i + ".kbest";            
                for(String line: Util.readLines(candFileFold))
                    pw.println(line);
            }
            pw.close();
        }
    }
    
    private static void trainPolarityClassifier(ArrayList<String> selectedFiles, String prefix,
            String mode, String dir, String subjLexFile, String epeDir, EPEFeatureSpec featureSpec) throws IOException {

        String modelFile = prefix + ".model";

        if(!new File(modelFile).exists()) {
            Iterator<MPQAText> it = makeTextIterator(mode, dir, subjLexFile, selectedFiles, epeDir, featureSpec);
            String methodName = "OnlineAlgorithm";
            String methodArgs = "nRounds=10 printDots=1000 PAUpdate C=1 sorted=true maximizeLoss=true";
            String[] argv = new String[] { null, null, null, null, 
                    modelFile, methodName, methodArgs };
            BaselinePolarityClassifier.train(argv, it);
        }
    }
    
    private static String getModelName(String confFile) throws IOException {
        for(String line: Util.readLines(confFile))
            if(line.contains("<model>"))
                return Strings.extractBetween(line, '>', '<');
        throw new RuntimeException("model file name not found");
    }
    
    private static void trainHolderExtractor(ArrayList<String> selectedFiles, String prefix,
            String mode, String dir, String subjLexFile, String epeDir, String holderConf, EPEFeatureSpec featureSpec) throws IOException {

        String modelFile = prefix + ".model";
        
        if(!new File(modelFile).exists()) {
            Iterator<MPQAText> it = makeTextIterator(mode, dir, subjLexFile, selectedFiles, epeDir, featureSpec);
            String[] argv = new String[] { null, null, null, null, holderConf, modelFile };
            LocalLinkClassifier.train(argv, it);
            
            /*String modelFileTmp = getModelName(holderConf);

            System.out.println("*HOLDER: checking that tmp model file exists.");
            
            System.out.println("*HOLDER: moving '" + modelFileTmp + "' to '" + modelFile);
            
            boolean status = new File(modelFileTmp).renameTo(new File(modelFile));
            if(!status)
                throw new RuntimeException("Could not move holder model file");*/
            
        }
    }
    
    private static void trainReranker(ArrayList<String> selectedFiles, String prefix,
            String mode, String dir, String subjLexFile, String epeDir, EPEFeatureSpec featureSpec) throws IOException {

        /*
        ArrayList<String> tmp = new ArrayList();
        tmp.add(selectedFiles.get(0));
        selectedFiles = tmp;
        */
        
        String modelFile = prefix + DIR_SEP + "reranker.model";
        
        if(!new File(modelFile).exists()) {
            Iterator<MPQAText> r = makeTextIterator(mode, dir, subjLexFile, selectedFiles, epeDir, featureSpec);
            
            String costWeights = "1/0/0/0";
            String cands = prefix + DIR_SEP + "seq.kbest";
            int nseq = 64;
            int nsrcseq = 1;
            int npolseq = 1;
            
            //String polintModel = null;
            String polintModel = prefix + DIR_SEP + "pol.model";
            
            //String linkModel = null;
            String linkModel = prefix + DIR_SEP + "holder.model";
            
            boolean useBigrams = false;
            boolean useSyn = true;
            boolean useSpan = true;
            boolean useSem = false;
            
            String algName = "OnlineAlgorithm";
            String algArgs = "nRounds=8 printDots=100 useNewLinear=true modelSize=1000000 "
                    + "PAUpdate C=0.005 maximizeLoss=true sorted=true sqrt=true";
            
            String[] argv = { null /* -train*/, null /*pre*/, null /* dir */,  
                    subjLexFile, null /*docs*/, null /*synsem*/, 
                    null /* const*/, null /*discconn*/,
                    cands, "" + nseq, "" + nsrcseq, "" + npolseq,
                    linkModel, polintModel, null /*save*/, null /*tk*/,
                    "" + useBigrams, "" + useSyn, "" + useSpan, "" + useSem,
                    costWeights, algName, algArgs, modelFile};
            
            MPQASentenceReranker.train(argv, r);
        
        }
    }
    
    public static void trainPipelinePart1(String[] argv) {
        try {
            String mode = argv[1];
            String graphMode = argv[2];
            
            EPEFeatureSpec epeFeatureSpec = new EPEFeatureSpec(argv[3]);
            
            String dir = argv[4];
            String subjLexFile = argv[5];
            String selectedFiles = argv[6];
            String epeDir = argv[7];
            
            String workDir = argv[8];
            String seqTemplate = argv[9];
            String holderConf = argv[10];
            
            final int NFOLDS = 4;
            final int K = 64;

            if(graphMode.equals("tree"))
                EPEReader.TREE_ONLY = true;
            else if(graphMode.equals("graph")) {
            } else
                throw new RuntimeException("must specify graph or tree");
            
            Random rand = new Random(0);
            ArrayList<String> selected = new ArrayList();
            Util.readLines(selectedFiles, selected);
            Collections.sort(selected);
            Collections.shuffle(selected, rand);
            Util.printToFile(workDir + DIR_SEP + "files", Strings.join(selected, "\n"), "UTF-8");
            
            Pair<ArrayList<String>[], ArrayList<String>[]> cvSplit = makeCVSplit(selected, NFOLDS);
            
            // 1-3 on full training data AND in cross-validation mode:
            
            // 1. seq labeler
            trainSeqLabeler(selected, workDir + DIR_SEP + "seq", mode, dir, subjLexFile, epeDir, seqTemplate, epeFeatureSpec);
            
            for(int i = 0; i < NFOLDS; i++)
                trainSeqLabeler(cvSplit.left[i], workDir + DIR_SEP + "seq" + i, mode, dir, subjLexFile, epeDir, seqTemplate, epeFeatureSpec);
            
            for(int i = 0; i < NFOLDS; i++)
                generateSeqCandidates(cvSplit.right[i], workDir, i, K, mode, dir, subjLexFile, epeDir, epeFeatureSpec);
            mergeSeqCandidates(workDir, NFOLDS);
            
            // 2. polarity classifier
            //trainPolarityClassifier(selected, workDir + DIR_SEP + "pol", mode, dir, subjLexFile, epeDir, epeFeatureSpec);

            /*
            
            //for(int i = 0; i < NFOLDS; i++)
            //    trainPolarityClassifier(cvSplit.left[i], workDir + DIR_SEP + "pol" + i, mode, dir, subjLexFile, epeDir);
            
            // 3. holder extractor
            trainHolderExtractor(selected, workDir + DIR_SEP + "holder", mode, dir, subjLexFile, epeDir, holderConf, epeFeatureSpec);

            //for(int i = 0; i < NFOLDS; i++)
            //    trainHolderExtractor(cvSplit.left[i], workDir + DIR_SEP + "holder" + i, mode, dir, subjLexFile, epeDir, holderConf);
            
            // 4. train reranker
            trainReranker(selected, workDir, mode, dir, subjLexFile, epeDir, epeFeatureSpec);
            
            */
            
        } catch(Exception e) {  
            e.printStackTrace();
            System.exit(1);
        }
    }
    
    public static void trainPipelinePart2(String[] argv) {
        try {
            String mode = argv[1];
            String graphMode = argv[2];
            
            EPEFeatureSpec epeFeatureSpec = new EPEFeatureSpec(argv[3]);
            
            String dir = argv[4];
            String subjLexFile = argv[5];
            String selectedFiles = argv[6];
            String epeDir = argv[7];
            
            String workDir = argv[8];
            String seqTemplate = argv[9];
            String holderConf = argv[10];
            
            final int NFOLDS = 4;
            final int K = 64;

            if(graphMode.equals("tree"))
                EPEReader.TREE_ONLY = true;
            else if(graphMode.equals("graph")) {
            } else
                throw new RuntimeException("must specify graph or tree");
            
            Random rand = new Random(0);
            ArrayList<String> selected = new ArrayList();
            Util.readLines(selectedFiles, selected);
            Collections.sort(selected);
            Collections.shuffle(selected, rand);

            /*
            Util.printToFile(workDir + DIR_SEP + "files", Strings.join(selected, "\n"), "UTF-8");
            
            Pair<ArrayList<String>[], ArrayList<String>[]> cvSplit = makeCVSplit(selected, NFOLDS);
            
            // 1-3 on full training data AND in cross-validation mode:
            
            // 1. seq labeler
            trainSeqLabeler(selected, workDir + DIR_SEP + "seq", mode, dir, subjLexFile, epeDir, seqTemplate, epeFeatureSpec);
            
            for(int i = 0; i < NFOLDS; i++)
                trainSeqLabeler(cvSplit.left[i], workDir + DIR_SEP + "seq" + i, mode, dir, subjLexFile, epeDir, seqTemplate, epeFeatureSpec);
            
            for(int i = 0; i < NFOLDS; i++)
                generateSeqCandidates(cvSplit.right[i], workDir, i, K, mode, dir, subjLexFile, epeDir, epeFeatureSpec);
            mergeSeqCandidates(workDir, NFOLDS);
            */
            
            // 2. polarity classifier
            trainPolarityClassifier(selected, workDir + DIR_SEP + "pol", mode, dir, subjLexFile, epeDir, epeFeatureSpec);

            //for(int i = 0; i < NFOLDS; i++)
            //    trainPolarityClassifier(cvSplit.left[i], workDir + DIR_SEP + "pol" + i, mode, dir, subjLexFile, epeDir);
            
            // 3. holder extractor
            trainHolderExtractor(selected, workDir + DIR_SEP + "holder", mode, dir, subjLexFile, epeDir, holderConf, epeFeatureSpec);

            //for(int i = 0; i < NFOLDS; i++)
            //    trainHolderExtractor(cvSplit.left[i], workDir + DIR_SEP + "holder" + i, mode, dir, subjLexFile, epeDir, holderConf);
            
            // 4. train reranker
            trainReranker(selected, workDir, mode, dir, subjLexFile, epeDir, epeFeatureSpec);
            
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
    
    private static void evalReranker(ArrayList<String> selectedFiles, String prefix,
            String mode, String dir, String subjLexFile, String epeDir, 
            EPEFeatureSpec featureSpec) throws IOException {

        String rerankerFile = prefix + DIR_SEP + "reranker.model";
        
        Iterator<MPQAText> r = makeTextIterator(mode, dir, subjLexFile, selectedFiles, epeDir, featureSpec);

        String expid = "expid";
        boolean detailed = false;
        
        String cands = prefix + DIR_SEP + "seq_eval.kbest";
        int nseq = -1;
        int nsrcseq = -1;
        int npolseq = -1;

        String polintModel = prefix + DIR_SEP + "pol.model";
        String linkModel = prefix + DIR_SEP + "holder.model";
            
        String[] argv = { null /* -eval*/, null /* dir */,  
                subjLexFile, null /*docs*/, null /*synsem*/, 
                null /* const*/, null /*discconn*/,
                cands, "" + nseq, "" + nsrcseq, "" + npolseq,
                linkModel, polintModel, rerankerFile, 
                expid, "" + detailed, null /* classes */ , "false" };

        MPQASentenceReranker.eval(argv, r);

        r = makeTextIterator(mode, dir, subjLexFile, selectedFiles, epeDir, featureSpec);
        
        argv[8] = "1";
        argv[12] = null;
        argv[17] = "true";

        MPQASentenceReranker.eval(argv, r);
        
        new File("expr_" + expid + ".prf").delete();
        new File("link_" + expid + ".prf").delete();
        new File("pol_" + expid + ".prf").delete();        
        
    }
    
    public static void evalPipeline(String[] argv) {
        try {
            String mode = argv[1];
            String graphMode = argv[2];
            EPEFeatureSpec epeFeatureSpec = new EPEFeatureSpec(argv[3]);

            String dir = argv[4];
            String subjLexFile = argv[5];
            String selectedFiles = argv[6];
            String epeDir = argv[7];
            
            String workDir = argv[8];

            if(graphMode.equals("tree"))
                EPEReader.TREE_ONLY = true;
            else if(graphMode.equals("graph")) {
            } else
                throw new RuntimeException("must specify graph or tree");
            
            final int K = 64;
            
            ArrayList<String> selected = new ArrayList();
            Util.readLines(selectedFiles, selected);
            Collections.sort(selected);
            
            generateSeqCandidates(selected, workDir, EVAL, K, mode, dir, subjLexFile, epeDir, epeFeatureSpec);
            
            evalReranker(selected, workDir, mode, dir, subjLexFile, epeDir, epeFeatureSpec);
            
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
    
    public static void main(String[] argv) {
        if(argv[0].equals("-train1"))
            trainPipelinePart1(argv);
        else if(argv[0].equals("-train2"))
            trainPipelinePart2(argv);
        else if(argv[0].equals("-eval"))
            evalPipeline(argv);
        else
            throw new RuntimeException("mode?");
    }
    
}
