package mpqa_structlearn.fullsystem;

import java.io.*;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import se.lth.cs.nlp.depsrl.format.CoNLL2008Format;
import se.lth.cs.nlp.depsrl.format.PAStructure;
import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.*;
import se.lth.cs.nlp.nlputils.ml_long.*;
import se.lth.cs.nlp.opinions.*;

import mpqa_structlearn.io.*;
import mpqa_structlearn.locallink.LocalLinkClassifier;
import mpqa_structlearn.sequences.*;

import mpqareader.SubjectivityLexicon;

public class LKTagger {

	public static void main(String[] argv) {
		if(argv[0].equals("-makePackage")) {
			makePackage(argv);
		} else if(argv[0].equals("-run")) {
			run(argv);
		} else if(argv[0].equals("-run08")) {
			runCoNLL08(argv);
		} else {
			throw new RuntimeException("Illegal options: " + Arrays.asList(argv));
		}
	}
	
	private static void makePackage(String[] argv) {
		try {
			String subjLexFile = argv[1];
			String taggingModelFile = argv[2];
			String polIntModelFile = argv[3];
			String linkModelFile = argv[4];
			String rerankerFile = argv[5];
			
			String outFile = argv[6];
			
			SubjectivityLexicon lex = new SubjectivityLexicon(subjLexFile);			

            System.out.println("Loading sequence model...");
			TaggingModel baseSeqModel = SeqLabeler.loadModel(taggingModelFile);
            
			LocalLinkClassifier llc = LocalLinkClassifier.loadModel(linkModelFile);
			
            System.out.println("Loading polarity and intensity classifiers...");
            ObjectInputStream ois = new ObjectInputStream(Util.openFileStream(polIntModelFile));
			Classifier<SparseVector, Integer> polCl = (Classifier) ois.readObject();
			Classifier<SparseVector, Integer> intCl = (Classifier) ois.readObject();
            ois.close();

			System.out.println("Loading reranking model...");
			ois = new ObjectInputStream(Util.openFileStream(rerankerFile));
            RerankingFE fe = (RerankingFE) ois.readObject();
            Classifier<EncodedCandidateSet, Integer> reranker = (Classifier) ois.readObject();
            
            ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(outFile));

            oos.writeObject(lex);
            oos.writeObject(baseSeqModel);
            oos.writeObject(llc);
            oos.writeObject(polCl);
            oos.writeObject(intCl);

            oos.writeObject(fe);
            oos.writeObject(reranker);

            oos.close();
            
		} catch(Exception e) {
			e.printStackTrace();
		}	
	}
	
	public static void runCoNLL08(String[] argv) {
		try {			

			String packageName = argv[1];			

			boolean extractHolders = Boolean.parseBoolean(argv[2]);
			int k = Integer.parseInt(argv[3]);
			
			ObjectInputStream ois = new ObjectInputStream(Util.openFileStream(packageName));
			
			System.err.println("Loading subjectivity lexicon...");
            SubjectivityLexicon subjLex = (SubjectivityLexicon) ois.readObject();

			System.err.println("Loading sequence model...");
			TaggingModel baseSeqModel = (TaggingModel) ois.readObject();
            
            LocalLinkClassifier llc = null;
            if(extractHolders) {
            	System.err.println("Loading holder link model...");
            	llc = (LocalLinkClassifier) ois.readObject();
            } else
            	ois.readObject();
            
            System.err.println("Loading polarity classifier...");
			Classifier<SparseVector, Integer> polCl = (Classifier) ois.readObject();
            System.err.println("Loading intensity classifier...");
			Classifier<SparseVector, Integer> intCl = (Classifier) ois.readObject();

			Classifier<EncodedCandidateSet, Integer> rcl = null;
			RerankingFE rfe = null;
			if(k > 1) {
				try {
					System.err.println("Loading reranking classifier...");
					rfe = (RerankingFE) ois.readObject();
					rcl  = (Classifier) ois.readObject();
				} catch(Exception e) {
					System.err.println("Found no reranking classifier in this model file.");					
					System.err.println("Setting k to 1.");
				}
			} else {
				System.err.println("Loading no reranking classifier since k = 1.");
			}
			
			ois.close();
            ois = null;

            BufferedReader br = new BufferedReader(new InputStreamReader(System.in, "UTF-8"));
            PrintWriter pw = new PrintWriter(new OutputStreamWriter(System.out, "UTF-8"));
            Triple<DepGraph, DepGraph, List<PAStructure>> tr = CoNLL2008Format.readNextGraph(br);
            while(tr != null) {
            	
                DepGraph dg = tr.first;
    			ArrayList<PAStructure> pas = (ArrayList) tr.third;

    			SentenceData cands = Server.kbest(baseSeqModel, subjLex, dg, pas, k);

				MPQASentence msen;
				if(k == 1) {
					msen = cands.candidates.get(0).mpqaSentence;
					Server.addPolInt(msen, polCl, intCl, subjLex);
				} else {
					final int kpol = 1;
					final int ksrc = 1;
					final SymbolEncoder polintEnc = null; // yes
					// llc not used here				
					Candidate guess = MPQASentenceReranker.rerankSentenceUnscoredNew(cands, null, 
							polCl, intCl, polintEnc, subjLex, ksrc, kpol, rfe, rcl);
					msen = guess.mpqaSentence;
					Server.addPolInt(msen, null, intCl, subjLex); // pol already assigned
				}

				if(llc != null) {
					addHolders(msen, llc);
				}

				MPQAText text = new MPQAText();
				text.sentences = new ArrayList();
				text.sentences.add(msen);
				
				MPQASeqExtractor.extractSequences(text, pw, true, extractHolders, false);
    			
    			tr = CoNLL2008Format.readNextGraph(br);
            }
            
            //processDirectory(inDir, outDir, baseSeqModel, k, subjLex, polCl, intCl, llc, rcl, rfe);            
            pw.close();
            
		} catch(Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
	}
	
	public static void run(String[] argv) {
		try {			
			boolean extractHolders = Boolean.parseBoolean(argv[1]);

			String packageName = argv[2];			
			int k = Integer.parseInt(argv[3]);
			String inDir = argv[4];
			String outDir = argv[5];
			
			ObjectInputStream ois = new ObjectInputStream(Util.openFileStream(packageName));
			
			System.out.println("Loading subjectivity lexicon...");
            SubjectivityLexicon subjLex = (SubjectivityLexicon) ois.readObject();

			System.out.println("Loading sequence model...");
			TaggingModel baseSeqModel = (TaggingModel) ois.readObject();
            
            LocalLinkClassifier llc = null;
            if(extractHolders) {
            	System.out.println("Loading holder link model...");
            	llc = (LocalLinkClassifier) ois.readObject();
            } else
            	ois.readObject();
            
            System.out.println("Loading polarity classifier...");
			Classifier<SparseVector, Integer> polCl = (Classifier) ois.readObject();
            System.out.println("Loading intensity classifier...");
			Classifier<SparseVector, Integer> intCl = (Classifier) ois.readObject();

			Classifier<EncodedCandidateSet, Integer> rcl = null;
			RerankingFE rfe = null;
			if(k > 1) {
				try {
					System.out.println("Loading reranking classifier...");
					rfe = (RerankingFE) ois.readObject();
					rcl  = (Classifier) ois.readObject();
				} catch(Exception e) {
					System.out.println("Found no reranking classifier in this model file.");					
					System.out.println("Setting k to 1.");
				}
			} else {
				System.out.println("Loading no reranking classifier since k = 1.");
			}
			
			ois.close();
            ois = null;
            
            processDirectory(inDir, outDir, baseSeqModel, k, subjLex, polCl, intCl, llc, rcl, rfe);            
            
		} catch(Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
	}
		
	static Pattern SEN_ID_PAT = Pattern.compile("id\\s*=\\s*\"(.*?)\"");
	static Pattern SEN_START_PAT = Pattern.compile("start\\s*=\\s*\"#(.*?)\"");
	static Pattern SEN_END_PAT = Pattern.compile("end\\s*=\\s*\"#(.*?)\"");
	static Pattern ON_PAT = Pattern.compile("on\\s*=\\s*\"#(.*?)\"");
	static Pattern FROM_PAT = Pattern.compile("from\\s*=\\s*\"#(.*?)\"");
	static Pattern TO_PAT = Pattern.compile("to\\s*=\\s*\"#(.*?)\"");
	static Pattern SCOPE_PAT = Pattern.compile("scope\\s*=\\s*\"(.*?)\"");

	static Pattern PAYLOAD_PAT = Pattern.compile(">(.*?)</e>");
	
	private static int extractIntByPattern(String s, Pattern p) {
		String match = extractStringByPattern(s, p);
		if(match == null) {
			throw new IllegalArgumentException("Could not find match: s = |" + s + "|");
		}
		return Integer.parseInt(match); //extractStringByPattern(s, p));
	}

	private static String extractStringByPattern(String s, Pattern p) {
		Matcher m = p.matcher(s);
		if(!m.find())
			return null;
		String match = m.group(1);
		return match;
	}
	
	private static Pair<String, ArrayList<int[]>> readSentenceSpans(String prefix, String file) throws IOException {
		BufferedReader br = new BufferedReader(new FileReader(prefix + File.separatorChar + file));
		String line = br.readLine();
		while(line != null && !line.contains("provides=\"SENTENCES\""))
			line = br.readLine();
		if(line == null) {
			br.close();
			return null;
		}
		String tokenFile = extractStringByPattern(line, SCOPE_PAT);
		if(tokenFile == null)
			tokenFile = file;
		ArrayList<int[]> out = new ArrayList();
		line = br.readLine();
		while(!line.contains("</annotation>")) {
			line = line.trim();
			if(line.startsWith("<e")) {
				if(!line.endsWith("/>") && !line.endsWith("</e>"))
					throw new RuntimeException("Sorry, we expect sentences line by line...");
				int id = extractIntByPattern(line, SEN_ID_PAT);
				int start = extractIntByPattern(line, SEN_START_PAT);
				int end = extractIntByPattern(line, SEN_END_PAT);
				out.add(new int[] { id, start, end });
			}
			line = br.readLine();
		}
		br.close();
		return new Pair(tokenFile, out);
	}

	private static String extractToken(String line) {
		int s_ix = line.indexOf('>');
		int e_ix = line.indexOf('<', s_ix);
		if(s_ix == -1)
			throw new RuntimeException("Sorry, we expect tokens line by line...");
		if(e_ix == -1)
			throw new RuntimeException("Sorry, we expect tokens line by line...");
		String token = line.substring(s_ix + 1, e_ix);
		return Strings.decodeXML(token);
	}
	
	private static String readTokenPreamble(BufferedReader br) throws IOException {
		String line = br.readLine();
		while(!line.contains("\"base\""))
			line = br.readLine();
		int s_ix = line.indexOf(">");
		int e_ix = line.indexOf('<', s_ix);
		String base = line.substring(s_ix + 1, e_ix);
		while(line != null && !line.contains("provides=\"TOKENS\""))
			line = br.readLine();
		if(line == null)
			throw new RuntimeException("Couldn't find token annotation");
		return base;
	}
	
	private static void readPreamble(BufferedReader br, String prov) throws IOException {
		String line = br.readLine();
		while(line != null && !line.contains("provides=\"" + prov + "\""))
			line = br.readLine();
		if(line == null)
			throw new RuntimeException("Couldn't find annotation layer " + prov);
	}
	
	private static final int LOOKAHEAD = 256;	
	
	private static Pair<ArrayList<String>, Triple> readNextSentence(int[] sen, BufferedReader br,
			BufferedReader pbr, BufferedReader lbr,
			BufferedReader depBr, BufferedReader predBr, BufferedReader argBr) throws IOException {

		if(!lbr.markSupported())
			throw new RuntimeException("!!!");

		boolean debug = false;
		
		int senId = sen[0];
		int start = sen[1];
		int end = sen[2];
		ArrayList<String> out = new ArrayList();
		ArrayList<String> tags = new ArrayList();
		ArrayList<String> lemmas = new ArrayList();

		boolean inside = false;
		
		int id = -1;
		String line;
		
		HashMap<Integer, Integer> idMap = new HashMap();
		
		while(id != end) {
			line = br.readLine().trim();
			while(line.equals(""))
				line = br.readLine().trim();
			if(!line.startsWith("<e") || !line.endsWith("</e>"))
				throw new RuntimeException("Sorry, we expect tokens line by line...");
			id = extractIntByPattern(line, SEN_ID_PAT);

			if(id == start)
				inside = true;

			if(inside) {

				pbr.mark(LOOKAHEAD);
				String posLine = pbr.readLine().trim();
				if(!posLine.startsWith("</annotation")
				    && (!posLine.startsWith("<e") 
						|| !posLine.endsWith("</e>"))) {
					System.err.println("end = " + end);
					System.err.println("line = |" + line + "|");
					System.err.println("posLine = |" + posLine + "|");
					throw new RuntimeException("Sorry, we expect POS tags line by line...");
				}

				if(posLine.startsWith("</annotation"))
					break;
				int on = extractIntByPattern(posLine, ON_PAT);
				if(on != id) {
					//throw new RuntimeException("POS on = " + on + ", token id = " + id);
					debug = true;
					System.err.println("*** Warning: POS on = " + on + ", token id = " + id);
					pbr.reset();
					continue;
				}
				
				lbr.mark(LOOKAHEAD);
				String lemmaLine = lbr.readLine().trim();
				if(!lemmaLine.startsWith("</annotation")
				   && (!lemmaLine.startsWith("<e") 
				   	   || !lemmaLine.endsWith("</e>"))) {
					System.err.println("line = |" + line + "|");
					System.err.println("posLine = |" + posLine + "|");
					System.err.println("lemmaLine = |" + lemmaLine + "|");
					throw new RuntimeException("Sorry, we expect lemmas line by line...");
				} else {
					//BUG! fixed 100803 lbr.reset(); // 100727
				}
				
				if(false) {
					System.err.println(line);
					System.err.println(posLine);
					System.err.println(lemmaLine);
				}

				String lemma = null;

				if(!lemmaLine.equals("</annotation>")) {
					on = extractIntByPattern(lemmaLine, ON_PAT);
					if(on != id) {
						lbr.reset();
					} else
						lemma = extractToken(lemmaLine);
				} else {
					lbr.reset(); // 100727 moved here 100803
				}

				String token = extractToken(line);
				String posTag = extractToken(posLine);

				out.add(token);
				tags.add(posTag);
				lemmas.add(lemma);
				
				idMap.put(id, out.size());
			}
		}

		DepGraph dg = new DepGraph(out.size());
		dg.nodes[0] = new DepNode();
		for(int i = 1; i < dg.nodes.length; i++) {
			dg.nodes[i] = new DepNode();
			dg.nodes[i].word = out.get(i - 1);
			dg.nodes[i].pos = tags.get(i - 1);
			dg.nodes[i].lemma = lemmas.get(i - 1);
			dg.nodes[i].position = i;
		}

		if(depBr == null) {
			return new Pair(out, new Triple(dg, dg, null));
		} else {
			for(int i = 1; i < dg.nodes.length; i++) {
				String dline = depBr.readLine().trim();
				while(dline.equals(""))
					dline = depBr.readLine().trim();
				String label = extractStringByPattern(dline, PAYLOAD_PAT);				
				if(dline.contains("on=\""))	{	
					int on = extractIntByPattern(dline, ON_PAT);
					if(!idMap.containsKey(on))
						throw new RuntimeException("!!!");
					DepNode n = dg.nodes[idMap.get(on)];
					//dg.nodes[0].addChild(dg.nodes[on], label);
					n.parents = new DepNode[] { dg.nodes[0] };
					n.relations = new String[] { label };
				} else {
					int from = extractIntByPattern(dline, FROM_PAT);
					int to = extractIntByPattern(dline, TO_PAT);
					//dg.nodes[from].addChild(dg.nodes[to], label);

					if(!idMap.containsKey(from))
						throw new RuntimeException("!!!");
					if(!idMap.containsKey(to))
						throw new RuntimeException("!!!");
					
					DepNode fn = dg.nodes[idMap.get(from)]; 
					DepNode tn = dg.nodes[idMap.get(to)]; 
					
					tn.parents = new DepNode[] { fn };
					tn.relations = new String[] { label };				
				}				
			}

			dg.collectChildren();

			//System.out.println(dg);
			//System.out.println(idMap);
			
			//System.out.println(dg);

			ArrayList<PAStructure> pas = new ArrayList();
			
			while(true) {
				String predLine = nextPredLine(predBr);
				if(predLine == null)
					break;
				int on = extractIntByPattern(predLine, ON_PAT);
				if(on > end) {
					pushBackPredLine(predLine);
					break;
				}

				if(!idMap.containsKey(on))
					throw new RuntimeException("!!!");
				
				PAStructure pa = new PAStructure();
				pa.pred = dg.nodes[idMap.get(on)];
				pa.lemma = extractStringByPattern(predLine, PAYLOAD_PAT);

				pas.add(pa);
				
				//System.out.println(pa.pred);
				//System.out.println(pa.lemma);
				
				if(argBr != null) {
					String argLine = nextArgLine(argBr);
					while(argLine != null) {
						int from = extractIntByPattern(argLine, FROM_PAT);
						if(from != on) {
							pushBackArgLine(argLine);
							break;
						}
						int to = extractIntByPattern(argLine, TO_PAT);

						if(!idMap.containsKey(to))
							throw new RuntimeException("!!!");
						
						DepNode argNode = dg.nodes[idMap.get(to)];
						String argLabel = extractStringByPattern(argLine, PAYLOAD_PAT);

						pa.args.add(argNode);
						pa.argLabels.add(argLabel);

						//System.out.println(argNode);
						//System.out.println(argLabel);						
						argLine = nextArgLine(argBr);					
					}
				}
				
			}

			//if(debug) {
			//	System.out.println(dg);
			//	System.out.println(pas);
			//}
			
			//dg.checkConsistency();
			//System.exit(1);
			
			return new Pair(out, new Triple(dg, dg, pas));
		}				
		
	}
	
	private static final String TERMINATED = "<<<TERMINATED>>>";
	
	private static String cachedPredLine = null;
	private static String nextPredLine(BufferedReader predBr) throws IOException {
		if(cachedPredLine == TERMINATED)
			return null;
		if(cachedPredLine != null) {
			String out = cachedPredLine;
			cachedPredLine = null;
			return out;
		} else {
			String line = predBr.readLine().trim();
			while(line.equals(""))
				line = predBr.readLine().trim();
			if(line.equals("</annotation>")) {
				cachedPredLine = TERMINATED;
				return null;
			} else
				return line;
		}
	}
	private static void pushBackPredLine(String line) {
		if(cachedPredLine != null)
			throw new IllegalStateException("!!!");
		cachedPredLine = line;
	}
	
	private static String cachedArgLine = null;
	private static String nextArgLine(BufferedReader argBr) throws IOException {
		if(cachedArgLine == TERMINATED)
			return null;
		if(cachedArgLine != null) {
			String out = cachedArgLine;
			cachedArgLine = null;
			return out;
		} else {
			String line = argBr.readLine().trim();
			while(line.equals(""))
				line = argBr.readLine().trim();
			if(line.equals("</annotation>")) {
				cachedArgLine = TERMINATED;
				return null;
			} else
				return line;
		}
	}
	private static void pushBackArgLine(String line) {
		if(cachedArgLine != null)
			throw new IllegalStateException("!!!");
		cachedArgLine = line;
	}	
	
	private static void printSubjPreamble(String base, String sentenceFileName, PrintWriter out) throws IOException {
		out.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
		out.println("<lk-annotation>");
		out.println("<meta-info>");
		out.println(" <tag name=\"base\">" + base + "</tag>");
		out.println(" <tag name=\"annotator\">UniTNSubjExprTagger</tag>");
		out.println("</meta-info>");
	}
	
	private static void printSubjEnd(PrintWriter out) throws IOException {
		out.println("</lk-annotation>");
	}
	
	private static final boolean PRINT_COMMENT = false;
	
	private static int printExpressions(int idCounter, int tid, String label, 
			ArrayList<MPQAExpression> spans, StringBuilder out, 
			HashMap<IntPair, Integer> holderIdMap) {
		for(MPQAExpression s: spans)
			if(s.type.equals(label)) {
				idCounter++;
				out.append("  <e id=\"" + idCounter + "\"");
				
				// TODO kolla index
				int start = s.span.get(0).position;
				int end = s.span.get(s.span.size()-1).position;
				
				out.append(" start=\"#" + (tid + start) + "\"");
				out.append(" end=\"#" + (tid + end) + "\"");
				out.append("><" + label);
				
				if(s.polarity != null)
					out.append(" pol=\"" + s.polarity + "\"");
				if(s.intensity != null)
					out.append(" int=\"" + s.intensity + "\"");

				if(s.nestedSource != null) {
					MPQAAgent ag = s.nestedSource.get(0);
					if(ag.isWriter) {
						out.append(" ns=\"#1\"");
					} else if(ag.isImplicit) {
						out.append(" ns=\"#2\"");
					} else {
						MPQAAgentInstance ai = ag.instances.get(0);
						int astart = ai.span.get(0).position;
						int aend = ai.span.get(ai.span.size() - 1).position;
						IntPair sp = new IntPair(astart, aend);
						Integer aid = holderIdMap.get(sp); 
						if(aid == null)
							throw new RuntimeException("!!!");
						out.append(" ns=\"#" + aid + "\"");
					}
				}

				out.append("/></e>");
				
				if(PRINT_COMMENT) {
					StringBuilder comment = new StringBuilder();
					for(DepNode n: s.span)
						comment.append(n.word + " ");
					out.append("<!-- " + comment + "-->");
				}
				
				out.append("\n");
			}		
		return idCounter;
	}
	
	private static Pair<HashMap<IntPair, Integer>, Integer> printAgents(int idCounter, int tid,
			ArrayList<MPQAExpression> spans, StringBuilder out) {

		HashMap<IntPair, Integer> idMap = new HashMap();
		
		for(MPQAExpression s: spans) {
			if(s.nestedSource == null)
				continue;
			MPQAAgent ag = s.nestedSource.get(0);
			if(ag == null || ag.isWriter || ag.isImplicit)
				continue;

			MPQAAgentInstance inst = ag.instances.get(0);
			int start = inst.span.get(0).position;
			int end = inst.span.get(inst.span.size() - 1).position;
			IntPair sp = new IntPair(start, end);
			if(idMap.containsKey(sp))
				continue;
								
			idCounter++;
			idMap.put(sp, idCounter);
			
			out.append("  <e id=\"" + idCounter + "\"");
			out.append(" start=\"#" + (tid + start) + "\"");
			out.append(" end=\"#" + (tid + end) + "\"");
			out.append("><a");
			out.append("/></e>");
				
			if(PRINT_COMMENT) {
				StringBuilder comment = new StringBuilder();
				for(DepNode n: inst.span)
					comment.append(n.word + " ");
				out.append("<!-- " + comment + "-->");
			}
			
			out.append("\n");
		}		
		return new Pair(idMap, idCounter);
	}
	
	private static final int W_ID = 1;
	private static final int IMP_ID = 2;
	
	public static void processDirectory(String dirName, String outputDir,
			TaggingModel model, int k, SubjectivityLexicon sl, 
			Classifier<SparseVector, Integer> polCl, 
			Classifier<SparseVector, Integer> intCl,
			LocalLinkClassifier llc,
			Classifier<EncodedCandidateSet, Integer> rcl,
			RerankingFE rfe) throws IOException {

		long t0 = System.currentTimeMillis();
		
		int countSentences = 0;
		int countWords = 0;
		
		File dir = new File(dirName);
		if(!dir.isDirectory())
			throw new IllegalArgumentException(dir + " is not a directory");
		for(File file: dir.listFiles()) {
			if(file.isDirectory())
				continue;
			Pair<String, ArrayList<int[]>> p = readSentenceSpans(dirName, file.getName());
			if(p == null)
				continue;
			String tokenFile = p.left;
			ArrayList<int[]> senSpans = p.right;

			System.out.println("Sentence file " + file.getName() + ", token file " + tokenFile);
			
			BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(dirName + File.separatorChar + tokenFile), "UTF-8"));
			String base = readTokenPreamble(br);

			String outFile;
			if(tokenFile.endsWith(".xml"))
				outFile = tokenFile.replaceFirst("\\.[^\\.]+\\.xml$", ".subjectivity.xml");
			else 
				outFile = tokenFile + ".subjectivity.xml";
			outFile = outputDir + File.separatorChar + outFile;

			String posFile = null;
			if(tokenFile.endsWith(".xml"))
				posFile = tokenFile.replaceFirst("\\.[^\\.]+\\.xml$", ".pos.xml");
			else 
				posFile = tokenFile + ".pos.xml";

			BufferedReader posReader = null;
			BufferedReader lemmaReader = null;
			try {
				posReader = new BufferedReader(new InputStreamReader(new FileInputStream(dirName + File.separatorChar + posFile), "UTF-8"));
				lemmaReader = new BufferedReader(new InputStreamReader(new FileInputStream(dirName + File.separatorChar + posFile), "UTF-8"));
			} catch(Exception e) {
				System.err.println("Sorry, we expect the POS file name to end with .pos.xml");
				System.exit(1);
			}
			readPreamble(posReader, "POS");
			readPreamble(lemmaReader, "LEMMA");

			BufferedReader depReader = null;
			BufferedReader predReader = null;
			BufferedReader argReader = null;
			if(llc != null || rcl != null) {
				String depFile = null;
				if(tokenFile.endsWith(".xml"))
					depFile = tokenFile.replaceFirst("\\.[^\\.]+\\.xml$", ".depsyntax.xml");
				else 
					depFile = tokenFile + ".depsyntax.xml";
				try {
					depReader = new BufferedReader(new InputStreamReader(new FileInputStream(dirName + File.separatorChar + depFile), "UTF-8"));
				} catch(Exception e) {
					System.err.println("Sorry, we expect the dependency syntax file name to end with .depsyntax.xml");
					System.exit(1);
				}
				
				cachedPredLine = cachedArgLine = null;
				String paFile = null;
				if(tokenFile.endsWith(".xml"))
					paFile = tokenFile.replaceFirst("\\.[^\\.]+\\.xml$", ".predargs.xml");
				else 
					paFile = tokenFile + ".predargs.xml";
				try {
					predReader = new BufferedReader(new InputStreamReader(new FileInputStream(dirName + File.separatorChar + paFile), "UTF-8"));
					argReader = new BufferedReader(new InputStreamReader(new FileInputStream(dirName + File.separatorChar + paFile), "UTF-8"));
				} catch(Exception e) {
					System.err.println("Sorry, we expect the P/A file name to end with .predargs.xml");
					System.exit(1);
				}

				readPreamble(depReader, "DEP-SYNTAX");
				readPreamble(predReader, "SRL-PB-preds");
				readPreamble(argReader, "SRL-PB-args");
			}
			
			PrintWriter out = new PrintWriter(new OutputStreamWriter(new FileOutputStream(outFile), "UTF-8"));
			printSubjPreamble(base, file.getName(), out);

			StringBuilder sbA = new StringBuilder();
			StringBuilder sbE = new StringBuilder();
			StringBuilder sbD = new StringBuilder();
			StringBuilder sbO = new StringBuilder();

			int tokenIdCounter = 0;

			sbA.append("  <e id=\"" + W_ID + "\"><a w=\"true\"/></e>\n");
			sbA.append("  <e id=\"" + IMP_ID + "\"><a imp=\"true\"/></e>\n");

			int idCounter = 2;
			
			for(int[] senSpan: senSpans) {
				Pair<ArrayList<String>, Triple> pp = readNextSentence(senSpan, br, 
						posReader, lemmaReader, depReader, predReader, argReader);
				ArrayList<String> sen = pp.left;
				
				if(sen == null)
					throw new RuntimeException("null");

				DepGraph dg = (DepGraph) pp.right.first;
				ArrayList<PAStructure> pas = (ArrayList) pp.right.third;
				
				//System.out.println("Test: " + dg);
				
				SentenceData cands = Server.kbest(model, sl, dg, pas, k);

				MPQASentence msen;
				if(k == 1) {
					msen = cands.candidates.get(0).mpqaSentence;
					Server.addPolInt(msen, polCl, intCl, sl);
				} else {
					final int kpol = 1;
					final int ksrc = 1;
					final SymbolEncoder polintEnc = null; // yes
					// llc not used here				
					Candidate guess = MPQASentenceReranker.rerankSentenceUnscoredNew(cands, null, 
							polCl, intCl, polintEnc, sl, ksrc, kpol, rfe, rcl);
					msen = guess.mpqaSentence;
					Server.addPolInt(msen, null, intCl, sl); // pol already assigned
				}

				if(llc != null) {
					addHolders(msen, llc);
				}
				
				Pair<HashMap<IntPair, Integer>, Integer> mp = printAgents(idCounter, tokenIdCounter, msen.exprs, sbA);
				idCounter = mp.right;
				idCounter = printExpressions(idCounter, tokenIdCounter, "ds", msen.exprs, sbD, mp.left);
				idCounter = printExpressions(idCounter, tokenIdCounter, "es", msen.exprs, sbE, mp.left);
				idCounter = printExpressions(idCounter, tokenIdCounter, "os", msen.exprs, sbO, mp.left);
				
				//System.out.println(pp.left);
				//System.out.println(spans);
				
				/*

				*/

				tokenIdCounter += sen.size();
				
				countSentences++;
				countWords += dg.nodes.length - 1;
			}

			out.println("<annotation scope=\"" + tokenFile + "\" provides=\"MPQA-agents\">");
			out.print(sbA);
			out.println("</annotation>");
			
			out.println("<annotation scope=\"" + tokenFile + "\" provides=\"MPQA-expressive-subjectivity\">");
			out.print(sbE);
			out.println("</annotation>");
			
			out.println("<annotation scope=\"" + tokenFile + "\" provides=\"MPQA-objective-speech-event\">");
			out.print(sbO);
			out.println("</annotation>");
			
			out.println("<annotation scope=\"" + tokenFile + "\" provides=\"MPQA-direct-subjective\">");
			out.print(sbD);
			out.println("</annotation>");
			
			printSubjEnd(out);
			out.close();
			
			posReader.close();
			lemmaReader.close();
			br.close();
			
			// 110328
			if(depReader != null)
				depReader.close();
			if(predReader != null)
				predReader.close();
			if(argReader != null)
				argReader.close();
		}
		
		long t1 = System.currentTimeMillis();
		double sPerSec = 1000.0 * countSentences / (t1 - t0);
		double wPerSec = 1000.0 * countWords / (t1 - t0);

		System.out.format("Processed %f sentences per second.\n", sPerSec);
		System.out.format("Processed %f words per second.\n", wPerSec);
		
	}	
	
	private static void addHolders(MPQASentence sen, LocalLinkClassifier llm) {
		ArrayList<DepNode> cands = LocalLinkClassifier.findCandidates(sen);
		for(MPQAExpression expr: sen.exprs) {
			MPQAAgent ag = MPQASentenceReranker.bestSource(llm, expr, cands, false);
			if(ag != null) {
				expr.nestedSource = new ArrayList();
				expr.nestedSource.add(ag);
			}
		}
	}
	
	
}
