package mpqa_structlearn.fullsystem;

import java.io.*;
import java.util.*;

//import mpqa_seq_reranker.*;

import mpqa_structlearn.polarity.BaselinePolarityClassifier;
import mpqa_structlearn.sequences.*;
import mpqa_structlearn.io.*;
import mpqa_structlearn.locallink.*;
import mpqareader.SubjectivityLexicon;
import se.lth.cs.nlp.nlputils.annotations.*;
import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.math.BLAS;
import se.lth.cs.nlp.nlputils.math.Resampler;
import se.lth.cs.nlp.nlputils.ml_long.*;
import se.lth.cs.nlp.nlputils.optimization.BestCombinations;
import se.lth.cs.nlp.nlputils.depgraph.*;
import se.lth.cs.nlp.opinions.*;

public class MPQASentenceReranker {

	public static void main(String[] argv) {
		if(argv[0].equals("-train")) {
		    try {
		        String dir = argv[2];
		        String subjLexFile = argv[3];
		        String docListFile = argv[4];
		        ArrayList<String> selectedFiles = new ArrayList();
		        Util.readLines(docListFile, selectedFiles);
		        Iterator<MPQAText> r = new LKReader(dir, subjLexFile, selectedFiles);
		        train(argv, r);
		    } catch(Exception e) {
		        e.printStackTrace();
		        System.exit(1);
		    }
		}
		else if(argv[0].equals("-genTK"))
			genTKData(argv);
		else if(argv[0].equals("-eval")) {
		    try {
                String dir = argv[1];
                String subjLexFile = argv[2];
                String docListFile = argv[3];
                ArrayList<String> selectedFiles = new ArrayList();
                Util.readLines(docListFile, selectedFiles);
                Iterator<MPQAText> r = new LKReader(dir, subjLexFile, selectedFiles);
                eval(argv, r);
            } catch(Exception e) {
                e.printStackTrace();
                System.exit(1);
            }
		} else if(argv[0].equals("-evalTK"))
			evalTK(argv);
		else if(argv[0].equals("-baseline"))
			baseline(argv);
		else if(argv[0].equals("-seqPolBaseline"))
			seqPolBaseline(argv);
		else {
			System.err.println("Illegal arguments: " + Arrays.toString(argv));
			System.exit(1);
		}
	}	
	
	private static final boolean USE_SRC_FEATURES = true;
    private static final double SRC_WEIGHT = 1.0;

	private static final boolean USE_POL_FEATURES = true;
    private static final double POL_WEIGHT = 64.0;

	private static final boolean NORMALIZE_SV = false;
	
	public static void eval(String[] argv, Iterator<MPQAText> r) {

	    System.out.println("*** Reranker evaluation ***");
		for(int i = 1; i < argv.length; i++)
			System.out.println(i + ": " + argv[i]);
		
		//String dir = argv[1];
		String subjLexFile = argv[2];
		//String docListFile = argv[3];

		String synsemFile = argv[4];
		String constFile = argv[5];
		String discConnFile = argv[6];
		
		String seqCandidateFile = argv[7];
		
		int Nseq = Integer.parseInt(argv[8]);
		int Nsrcseq = Integer.parseInt(argv[9]);
		int Npolseq = Integer.parseInt(argv[10]);
		
		String linkModelFile = argv[11];
		String polintModelFile = argv[12];
		String rerankerModelFile = argv[13];

		String expId = argv[14];
		
		boolean printDetailed = Boolean.parseBoolean(argv[15]);

		String classesString = null;
        if(argv.length > 16)
            classesString = argv[16]; 

		boolean goldStandardSpans = false;
        if(argv.length > 17)
            goldStandardSpans = Boolean.parseBoolean(argv[17]); 
		
		try {
			/* Loading trained model. */

			System.out.println("expId = " + expId);
			
			HashMap<String, String> classMap = parseClusters(classesString);
			if(classMap != null)
				System.out.println("Classes: " + classMap);
			
			System.out.println("Loading link model...");
			LocalLinkClassifier llc = LocalLinkClassifier.loadModel(linkModelFile);

			Classifier<SparseVector, Integer> polCl = null;
			Classifier<SparseVector, Integer> intCl = null;
			SymbolEncoder polintEnc = null;
			SubjectivityLexicon sl = null;
			if(!Util.isNullFile(polintModelFile)) {
				System.out.println("Loading polarity and intensity classifiers...");
	            ObjectInputStream ois = new ObjectInputStream(Util.openFileStream(polintModelFile));
	            polCl = (Classifier) ois.readObject();
				intCl = (Classifier) ois.readObject();
				try {
					polintEnc = (SymbolEncoder) ois.readObject();
				} catch(Exception e) {}
				ois.close();
	            sl = new SubjectivityLexicon(subjLexFile, false);
			}
			
			System.out.println("Loading reranker model...");
			//ObjectInputStream ois = new ObjectInputStream(new FileInputStream(rerankerModelFile));
			ObjectInputStream ois = new ObjectInputStream(Util.openFileStream(rerankerModelFile));
            RerankingFE rfe = (RerankingFE) ois.readObject();
            Classifier<EncodedCandidateSet, Integer> rcl = (Classifier) ois.readObject();
            int Nseq_saved = ois.readInt();
            int Nsrcseq_saved = ois.readInt();
            int Npolseq_saved = ois.readInt();
            if(Nseq < 1) {
            	System.out.println("Setting Nseq to saved: " + Nseq_saved);
            	Nseq = Nseq_saved;
            }
            if(Nsrcseq < 1) {
            	System.out.println("Setting Nsrcseq to saved: " + Nsrcseq_saved);
            	Nsrcseq = Nsrcseq_saved;
            }
            if(Npolseq < 1) {
            	System.out.println("Setting Npolseq to saved: " + Npolseq_saved);
            	Npolseq = Npolseq_saved;
            }
            ois.close();

			System.out.println("Loading candidate lists...");
            HashMap<String, SentenceData> candMap;
            ArrayList<SentenceData> seqCands;
            if(!goldStandardSpans)
                seqCands = SeqCandidateReader.readSentences(seqCandidateFile, synsemFile, constFile, discConnFile, Nseq, -1);
            else
                seqCands = new ArrayList();
            candMap = createCandidateMap(seqCands);
            
			PRFStats exprResult = new PRFStats(true);
			PRFStats exprResultSimple = new PRFStats(true);

			PRFStats polResult = new PRFStats(true);
			PRFStats intResult = new PRFStats();
			HashMap<String, PRFStats> polResults = new HashMap();
			polResults.put("positive", new PRFStats(true));
			polResults.put("negative", new PRFStats(true));
			polResults.put("neutral", new PRFStats(true));
			HashMap<String, PRFStats> intResults = new HashMap();
			intResults.put("high", new PRFStats());
			intResults.put("medium", new PRFStats());
			intResults.put("low", new PRFStats());

			PRFStats linkResult = new PRFStats(true);
			HashMap<String, PRFStats> linkLocalResults = new HashMap();
			linkLocalResults.put("all", new PRFStats(true));
			linkLocalResults.put("es", new PRFStats(true));
			linkLocalResults.put("os", new PRFStats(true));
			linkLocalResults.put("ds", new PRFStats(true));
			HashMap<String, PRFStats> linkNonlocalResults = new HashMap();
			linkNonlocalResults.put("all", new PRFStats(true));
			linkNonlocalResults.put("es", new PRFStats(true));
			linkNonlocalResults.put("os", new PRFStats(true));
			linkNonlocalResults.put("ds", new PRFStats(true));
			linkNonlocalResults.put("wr", new PRFStats(true));
			linkNonlocalResults.put("imp", new PRFStats(true));
			
			//ArrayList<String> selectedFiles = new ArrayList();
            //Util.readLines(docListFile, selectedFiles);
            //trimFileNames(selectedFiles);
            //selectedFiles.remove("20020203/20.46.36-9539");
            
			//LKReader r = new LKReader(dir, subjLexFile, selectedFiles);

			System.out.println("Processing test set...");

	        if(false /* Output for Lingjia, 2014-04-09 */) {
	            if(lingjiaOutput == null) {
	                try {
	                    lingjiaOutput = new PrintWriter(new FileWriter("lingjia_pol.txt"));
	                } catch(Exception e) {
	                    e.printStackTrace();
	                    System.exit(1);
	                }
	            }
	        }
			
			//int nExpr = 0, nPos = 0, nNeu = 0, nNeg = 0, nBoth = 0, nDS = 0, nES = 0, nOS = 0;
			
	        int nTexts = 0;
	        
			while(r.hasNext()) {
				MPQAText text = r.next();
				nTexts += 1;
				
				System.out.println(text.source);
				
				if(goldStandardSpans) {
				    candMap.clear();
				    for(MPQASentence sen: text.sentences) {
				        SentenceData sd = new SentenceData();
				        MPQASentence c = sen.clone(); // no holders cloned here
				        sd.goldSpans = sen;
				        sd.text = sen.dg.toTokenString();

				        Candidate cand = new Candidate();
				        cand.stats = sen.compareExpressions(c);
				        cand.mpqaSentence = c;
				        
				        sd.candidates.add(cand);
				        
				        candMap.put(sd.text, sd);
				    }
				}
				
				/*
				for(MPQASentence sen: text.sentences) {
					for(MPQAExpression expr: sen.exprs) {
						if(expr.type.equals("os")) {
							nOS++;
							continue;
						} else if(expr.type.equals("ds"))
							nDS++;
						else if(expr.type.equals("es"))
							nES++;
						nExpr++;
						String pol = expr.polarity;
						if(pol == null)
							pol = "neutral";
						pol = pol.trim();
						if(pol.equals(""))
							pol = "neutral";
						if(pol.startsWith("un"))
							pol = pol.substring(10);
						if(pol.equals("positive"))
							nPos++;
						else if(pol.equals("negative"))
							nNeg++;
						else if(pol.equals("neutral"))
							nNeu++;
						else if(pol.equals("both"))
							nBoth++;
						else throw new RuntimeException("!!!");
					}
				}				
				
				String shortSrc = text.source.substring(23);
				if(!selectedFiles.contains(text.source) && !selectedFiles.contains(shortSrc))
					throw new RuntimeException("source = " + text.source);
                */

				rerankText(text, candMap, llc, polCl, intCl, polintEnc, sl, 
						Nsrcseq, Npolseq, rfe, rcl,
						exprResult, exprResultSimple, linkResult, 
						polResult, intResult, polResults, intResults,
						linkLocalResults, linkNonlocalResults, printDetailed, classMap); 
			}
			
	        if(false /* Output for Lingjia, 2014-04-09 */) {
	            if(lingjiaOutput != null) {
	                try {
	                    lingjiaOutput.close();
	                } catch(Exception e) {
	                    e.printStackTrace();
	                    System.exit(1);
	                }
	            }
	        }
			
			/*
			System.out.println("nExpr = " + nExpr);
			System.out.format("nPos = %d, nNeg = %d, nNeu = %d, nBoth = %d\n",
					nPos, nNeg, nNeu, nBoth);
			System.out.format("nDS = %d, nES = %d, nOS = %d\n",
					nDS, nES, nOS);
			*/
			
	        System.out.println("number of texts: " + nTexts);
	        
			System.out.println("*** Expression result:");
			exprResult.print();
			exprResult.printRaw();
			
			if(true) {
				PrintWriter exprResultSamples = new PrintWriter(new FileWriter("expr_" + expId + ".prf"));
				exprResult.printSamples(exprResultSamples);
				exprResultSamples.close();
			}
			
			System.out.println("*** Expression result (simplified):");
			exprResultSimple.print();
			
			if(linkResult.nOverlap > 0) {
				System.out.println();
				System.out.println("*** Holder result:");
                linkResult.print();
                linkResult.printRaw();
				linkResult.printSamplesToFile("link_" + expId + ".prf");

				if(llc != null && printDetailed) {
					for(String etype: new String[] { "all", "es", "ds", "os" }) {
						PRFStats s = linkLocalResults.get(etype);
						s.printSamplesToFile("link_local_" + etype + "_" + expId + ".prf");
					}
					for(String etype: new String[] { "all", "es", "ds", "os", "wr", "imp" }) {
						PRFStats s = linkNonlocalResults.get(etype);
						s.printSamplesToFile("link_nonlocal_" + etype + "_" + expId + ".prf");
					}
				}
			
			}

			if(polCl != null && printDetailed) {
				System.out.println("*** Polarity results:");
				polResult.print();
				System.out.println();
				System.out.println("Positive:");
				polResults.get("positive").print();
				System.out.println("Neutral:");
				polResults.get("neutral").print();
				System.out.println("Negative:");
				polResults.get("negative").print();

				/*
				System.out.println();
				System.out.println("Intensity results:");
				System.out.println("High:");
				intResults.get("high").print();
				System.out.println("Medium:");
				intResults.get("medium").print();
				System.out.println("Low:");
				intResults.get("low").print();*/

				if(true) {

					/*PrintWriter exsimpResultSamples = new PrintWriter(new FileWriter("exsimp.prf"));
					exprResultSimple.printSamples(exsimpResultSamples);
					exsimpResultSamples.close();*/
					
					PrintWriter polResultSamples = new PrintWriter(new FileWriter("pol.prf"));
					polResult.printSamples(polResultSamples);
					polResultSamples.close();
					
					PrintWriter polResultSamplesPos = new PrintWriter(new FileWriter("pol_pos.prf"));
					polResults.get("positive").printSamples(polResultSamplesPos);
					polResultSamplesPos.close();
					
					PrintWriter polResultSamplesNeu = new PrintWriter(new FileWriter("pol_neu.prf"));
					polResults.get("neutral").printSamples(polResultSamplesNeu);
					polResultSamplesNeu.close();
					
					PrintWriter polResultSamplesNeg = new PrintWriter(new FileWriter("pol_neg.prf"));
					polResults.get("negative").printSamples(polResultSamplesNeg);
					polResultSamplesNeg.close();
					
					/*int n = polResult.size();
					System.out.println(Resampler.resamplingTest(polResult.getPartialF1Evaluator(), 
										n, n, new Random(0), 0.1, 100000, 100000));*/
				
				}
				
			} else {

				if(true) {

					if(polCl != null) {
						//System.out.println();
						System.out.println("*** Polarity results:");
						polResult.print();

						PrintWriter polResultSamples = new PrintWriter(new FileWriter("pol_" + expId + ".prf"));
						polResult.printSamples(polResultSamples);
						polResultSamples.close();
						/*int n = polResult.size();
					System.out.println(Resampler.resamplingTest(polResult.getPartialF1Evaluator(), 
										n, n, new Random(0), 0.1, 100000, 100000));*/

					}

				
				
				}
				
				
				
				/*if(intResult.nOverlap > 0) {
					System.out.println();
					System.out.println("Intensity results:");
					intResult.print();
				}*/
			}
			
		} catch(Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
		
	}
	
	private static void trimFileNames(List<String> list) {
		for(ListIterator<String> it = list.listIterator(); it.hasNext(); ) {
			String s = it.next();
			if(s.startsWith("database.mpqa.2.0/docs/")) {
				s = s.substring(23);
				it.set(s);
			}
		}
	}

	/* Output for Lingjia, 2014-04-09 */
	private static PrintWriter lingjiaOutput = null;
	
	private static void rerankText(MPQAText text, HashMap<String, SentenceData> candMap,
			LocalLinkClassifier llc, 
			Classifier<SparseVector, Integer> polCl, Classifier<SparseVector, Integer> intCl,
    		SymbolEncoder polintEnc, SubjectivityLexicon sl,
			int kbestSrc, int kbestPol, RerankingFE rfe, 
			Classifier<EncodedCandidateSet, Integer> rcl, 
			PRFStats exprResult, PRFStats exprResultSimple, PRFStats linkResult,
			PRFStats polResult, PRFStats intResult, HashMap<String, PRFStats> polResults, 
			HashMap<String, PRFStats> intResults, HashMap<String, PRFStats> linkLocalResults, 
			HashMap<String, PRFStats> linkNonlocalResults,
			boolean printDetailed, HashMap<String, String> classMap) {

        HashMap<MPQAAgent, String> agentToNbr = null;
        HashMap<MPQASentence, String> senToNbr = null;
	    if(lingjiaOutput != null) {
	        lingjiaOutput.println("*** TEXT: " + text.source + " ***");
	        
	        senToNbr = new HashMap();
	        for(MPQASentence s: text.sentences)
	            senToNbr.put(s, "" + (senToNbr.size()+1));

            lingjiaOutput.println("** Gold-standard agents:");
	        
            agentToNbr = new HashMap();
            for(MPQAAgent a: text.agents) {
	            String id = "a" + (agentToNbr.size() + 1);
	            agentToNbr.put(a, id);
	            
	            //lingjiaOutput.print("Gold-standard agent " + id + ": ");
	            lingjiaOutput.print(id + ": ");
	            lingjiaOutput.println(a.toString2(senToNbr));
	        }
	    }

        long t0 = System.currentTimeMillis();
	    
    	for(int j = 0; j < text.sentences.size(); j++) {
    	    System.out.print(".");
    	    System.out.flush();
    	    
    	    MPQASentence goldSen = text.sentences.get(j);

            //System.out.println(goldSen.dg.nodes.length);
    	    
    		SentenceData sd = candMap.get(goldSen.dg.toTokenString());
    		//System.out.println(sd.text);
    		
    		if(sd == null) {
    			System.out.println("*** WARNING: skipped");
    			return;
    		}
    		
    		//System.out.println(goldSen.dg);
    		
    		Candidate guessCand = null;
    		
    		/* For EPE: */
    		if(goldSen.dg.nodes[0] instanceof EPEDepNode) {
    		    if(sd.candidates.get(0).mpqaSentence.dg != goldSen.dg)
    		        //System.out.println("isEmpty");
    		        for(Candidate c: sd.candidates)
    		            c.mpqaSentence.replaceDepTree(goldSen.dg);
    		    sd.goldSpans.replaceDepTree(goldSen.dg);
    		    //System.out.println(sd.candidates.get(0).mpqaSentence.dg);

    		    //for(MPQAExpression expr: goldSen.exprs)
    		    //    expr.setCharSpan();

    		    EPEPaths.prepare(goldSen.dg);

    		} else /* for old code */
    		    goldSen.replaceDepTree(sd.candidates.get(0).mpqaSentence.dg);

    		if(goldSen.dg.nodes.length < 256) {
    		    
    		    for(Candidate c: sd.candidates) {
    		        //System.out.println(c.mpqaSentence.exprs);
    		        for(MPQAExpression expr: c.mpqaSentence.exprs) {
    		            if(!expr.span.isEmpty()) {
    		                //System.out.println(expr);
    		                //System.out.println(expr.start + "," + expr.end);
    		                //System.exit(0);
    		                if(expr.start == Integer.MIN_VALUE || expr.end == Integer.MIN_VALUE)
    		                    throw new RuntimeException("!");
    		                //System.out.println(expr + " / " + expr.span + " / " + expr.start + "," + expr.end);
    		            }
    		        }     
    		    }

    		    ArrayList<Candidate> newCands;

    		    if(polCl != null) {
    		        addSubjClues(sl, sd.candidates);
    		        newCands = expandCandidateSetWithPolarity(polCl, polintEnc, sd.candidates, goldSen, kbestPol);
    		    } else
    		        newCands = sd.candidates;
    		    if(llc != null)
    		        newCands = expandCandidateSetWithSources(llc, newCands, kbestSrc, goldSen, true);

    		    double[] lls = new double[newCands.size()];

    		    for(int i = 0; i < newCands.size(); i++)
    		        lls[i] = newCands.get(i).baseScore; // hmm vad händer med baseSrcScore?

    		    Reranker.normalize(lls, newCands.size());

    		    EncodedCandidateSet nb = new EncodedCandidateSet();
    		    nb.reps = new SparseVector[newCands.size()];
    		    for(int i = 0; i < newCands.size(); i++) {
    		        Candidate c = newCands.get(i);
    		        SparseVector sv = new SparseVector();
    		        rfe.extractFeatures(c, lls[i], sv);

    		        if(USE_SRC_FEATURES && llc != null)
    		            rfe.extractSourceFeatures(c, sv);

    		        if(USE_POL_FEATURES && (polCl != null))
    		            rfe.extractPolarityFeatures(c, sv);

    		        sv.sortIndices();

    		        nb.reps[i] = sv;
    		    }

    		    if(true) {
    		        double[] scores = new double[nb.reps.length];
    		        ArrayList<Integer> range = new ArrayList();
    		        for(int i = 0; i < nb.reps.length; i++)
    		            range.add(i);
    		        rcl.computeScoresRestricted(nb, range, scores);
    		        //if(biases.length != scores.length)
    		        //	throw new RuntimeException("!!!");
    		        //System.out.println(Arrays.toString(scores));
    		        //System.out.println(Arrays.toString(biases));
    		        //int guess0 = CollectionUtils.maxIndex(scores);            	
    		        //BLAS.daxpy(scores.length, 1.0, biases, 1, scores, 1, 0, 0);
    		        //System.out.println(Arrays.toString(scores));
    		        int guess = CollectionUtils.maxIndex(scores);         		
    		        //System.out.println("guess0 = " + guess0 + ", guess = " + guess);            	
    		        //System.exit(0);
    		        //int guess = rcl.classify(nb);
    		        guessCand = newCands.get(guess);
    		    } else {
    		        double maxF = Double.NEGATIVE_INFINITY;
    		        for(Candidate c: newCands) {
    		            PRFStats res = goldSen.comparePolInt(c.mpqaSentence, true);
    		            if(res.getPartialF1() > maxF) {
    		                maxF = res.getPartialF1();
    		                guessCand = c;
    		            }            			
    		        }
    		    }

    		    if(lingjiaOutput != null /* Output for Lingjia, 2014-04-09 */) {
    		        String sid = "" + (j+1);

    		        lingjiaOutput.print("** Sentence " + sid + " tokens: ");
    		        lingjiaOutput.println(goldSen.dg.toTokenString());
    		        Collections.sort(goldSen.exprs, MPQAExpression.EXPR_POS_COMP);
    		        Collections.sort(guessCand.mpqaSentence.exprs, MPQAExpression.EXPR_POS_COMP);
    		        lingjiaOutput.println("** Sentence " + sid + " gold expressions: ");
    		        //lingjiaOutput.println(goldSen.exprs);

    		        for(MPQAExpression e: goldSen.exprs) {
    		            lingjiaOutput.print(e + ": ");
    		            List<MPQAAgent> ns = e.nestedSource;
    		            if(ns == null || ns.size() == 0)
    		                lingjiaOutput.println("[]");
    		            else {
    		                MPQAAgent a = ns.get(ns.size()-1);
    		                //lingjiaOutput.println(a.toString2(senToNbr));
    		                lingjiaOutput.println("[" + agentToNbr.get(a) + "]");
    		            }
    		        }

    		        lingjiaOutput.println("** Sentence " + sid + " predicted expressions: ");
    		        //lingjiaOutput.println(guessCand.mpqaSentence.exprs);

    		        for(MPQAExpression e: guessCand.mpqaSentence.exprs) {
    		            lingjiaOutput.print(e + ": ");
    		            List<MPQAAgent> ns = e.nestedSource;
    		            if(ns == null || ns.size() == 0)
    		                lingjiaOutput.println("[]");
    		            else {
    		                if(ns.size() > 1)
    		                    throw new RuntimeException("???");
    		                MPQAAgent a = ns.get(0);
    		                if(a.instances == null || a.instances.size() == 0)
    		                    lingjiaOutput.println(a);
    		                else {
    		                    MPQAAgentInstance i = a.instances.get(0);
    		                    lingjiaOutput.println("[" + i.toString2(sid) + "]");
    		                }
    		            }
    		        }
    		    }

    		} else {
    		    guessCand = sd.candidates.get(0);
    		}
            
            //System.out.println("guessCand:\n" + guessCand.mpqaSentence.exprs);
            //System.exit(0);
            
            //MPQASentence guessedSen = toMPQASentence(guessCand, goldSen);

            // TODO här
            if(classMap == null)
            	exprResult.add(goldSen.compareExpressions(guessCand.mpqaSentence));
            else
            	exprResult.add(goldSen.compareExpressionsClustered(guessCand.mpqaSentence, classMap));
            
            polResult.add(goldSen.comparePolInt(guessCand.mpqaSentence, true));
            intResult.add(goldSen.comparePolInt(guessCand.mpqaSentence, false));

            exprResultSimple.add(goldSen.compareExpressionsSimple(guessCand.mpqaSentence));
            
            for(String pol: polResults.keySet()) {
            	PRFStats s = polResults.get(pol);
            	s.add(goldSen.comparePolInt(guessCand.mpqaSentence, true, pol));
            }
            for(String in: intResults.keySet()) {
            	PRFStats s = intResults.get(in);
            	s.add(goldSen.comparePolInt(guessCand.mpqaSentence, false, in));
            }

            if(true) {
            	linkResult.add(goldSen.compareLinks(guessCand.mpqaSentence));

            	if(llc != null && printDetailed) {
            		{
            			PRFStats s = linkLocalResults.get("all");
            			s.add(goldSen.compareLinks(guessCand.mpqaSentence, null, false, null, null));
            		}
            		for(String etype: new String[] { "es", "ds", "os" }) {
            			PRFStats s = linkLocalResults.get(etype);
            			s.add(goldSen.compareLinks(guessCand.mpqaSentence, etype, false, null, null));
            		}
            		{
            			PRFStats s = linkNonlocalResults.get("all");
            			s.add(goldSen.compareLinks(guessCand.mpqaSentence, null, true, null, null));
            		}
            		for(String etype: new String[] { "es", "ds", "os" }) {
            			PRFStats s = linkNonlocalResults.get(etype);
            			s.add(goldSen.compareLinks(guessCand.mpqaSentence, etype, true, null, null));
            		}
            		{
            			PRFStats s = linkNonlocalResults.get("wr");
            			s.add(goldSen.compareLinks(guessCand.mpqaSentence, null, true, true, null));
            		}
            		{
            			PRFStats s = linkNonlocalResults.get("imp");
            			s.add(goldSen.compareLinks(guessCand.mpqaSentence, null, true, null, true));
            		}
            	}
            }
            
    	}

    	System.out.println();
    	
    	/*long t1 = System.currentTimeMillis();
    	System.out.println(t1 - t0);
    	System.exit(0);*/

	}
	
	static HashMap<String, String> parseClusters(String clString) {
		if(Util.isNullFile(clString))
			return null;
		HashMap<String, String> out = new HashMap();
		String[] cls = clString.split("/");
		for(String cl: cls) {
			cl = cl.trim();
			String[] types = cl.split("[&\\+]");
			for(String t: types)
				out.put(t.trim(), cl);
		}
		return out;
	}
	
	static Candidate rerankSentenceUnscoredNew(SentenceData sd,
			LocalLinkClassifier llc, 
			Classifier<SparseVector, Integer> polCl, Classifier<SparseVector, Integer> intCl,
			SymbolEncoder polintEnc, SubjectivityLexicon sl,
			int kbestSrc, int kbestPol, RerankingFE rfe, 
			Classifier<EncodedCandidateSet, Integer> rcl) {

		/*System.out.println("syntax:");
		System.out.println(sd.candidates.get(0).mpqaSentence.dg);
		System.out.println("semantics:");
		System.out.println(sd.candidates.get(0).mpqaSentence.pas);*/
		
		ArrayList<Candidate> newCands = sd.candidates;
		if(polCl != null) {
			addSubjClues(sl, newCands);
			newCands = expandCandidateSetWithPolarity(polCl, polintEnc, newCands, null, kbestPol);
		}
		if(llc != null)
			newCands = expandCandidateSetWithSources(llc, newCands, kbestSrc, null, true);

		double[] lls = new double[newCands.size()];

		for(int i = 0; i < newCands.size(); i++) 
			lls[i] = newCands.get(i).baseScore;

		Reranker.normalize(lls, newCands.size());

		EncodedCandidateSet nb = new EncodedCandidateSet();
		nb.reps = new SparseVector[newCands.size()];
		for(int i = 0; i < newCands.size(); i++) {
			Candidate c = newCands.get(i);
			SparseVector sv = new SparseVector();
			rfe.extractFeatures(c, lls[i], sv);

			if(USE_SRC_FEATURES && llc != null)
				rfe.extractSourceFeatures(c, sv);

			if(USE_POL_FEATURES && (polCl != null))
				rfe.extractPolarityFeatures(c, sv);

			sv.sortIndices();

			nb.reps[i] = sv;
		}

		return newCands.get(rcl.classify(nb));
	}
	
	static Candidate rerankSentenceUnscored(SentenceData sd,
			LocalLinkClassifier llc, int kbestSrc, RerankingFE rfe, 
			Classifier<EncodedCandidateSet, Integer> rcl) {
    		
		ArrayList<Candidate> newCands = expandCandidateSetWithSources(llc, 
				sd.candidates, kbestSrc, null, true);

		double[] lls = new double[newCands.size()];

		for(int i = 0; i < newCands.size(); i++)
			lls[i] = newCands.get(i).baseScore; // hmm vad händer med baseSrcScore?

		Reranker.normalize(lls, newCands.size());

		EncodedCandidateSet nb = new EncodedCandidateSet();
		nb.reps = new SparseVector[newCands.size()];
		for(int i = 0; i < newCands.size(); i++) {
			Candidate c = newCands.get(i);
			SparseVector sv = new SparseVector();
			rfe.extractFeatures(c, lls[i], sv);
            	
			if(USE_SRC_FEATURES)
				rfe.extractSourceFeatures(c, sv);
            	
			nb.reps[i] = sv;
		}
		int guess = rcl.classify(nb);    		
		return newCands.get(guess);            
	}
	
	private static void rerankTextTK(MPQAText text, List<SentenceData> sds,			
			LocalLinkClassifier llc, int kbestSrc, 
			BufferedReader tkGuessReader,
			PRFStats exprResult, PRFStats linkResult) throws IOException {

    	for(int j = 0; j < text.sentences.size(); j++) {
    		MPQASentence goldSen = text.sentences.get(j);
    		SentenceData sd = sds.get(j);
    		goldSen.replaceDepTree(sd.goldSpans.dg); // behövs??
    		
    		ArrayList<Candidate> cands;
    		if(llc != null)
    			cands = expandCandidateSetWithSources(llc, 
    					sd.candidates, kbestSrc, goldSen, true);
    		else
    			cands = sd.candidates;

    		//double[] lls = new double[cands.size()];

    		//for(int i = 0; i < cands.size(); i++)
            //    lls[i] = cands.get(i).baseScore;                

    		//Reranker.normalize(lls, cands.size());

            int maxI = -1;
            double maxTKVal = Double.NEGATIVE_INFINITY;
            
            for(int i = 0; i < cands.size(); i++) {
            	String guessLine = tkGuessReader.readLine();
            	//System.out.println(guessLine);
            	//System.out.flush();
            	double tkVal = Double.parseDouble(guessLine);
            	if(tkVal > maxTKVal) {
            		maxI = i;
            		maxTKVal = tkVal;
            	}
            }

            Candidate guessCand = cands.get(maxI);            

            //MPQASentence guessedSen = toMPQASentence(guessCand, goldSen);

            exprResult.add(goldSen.compareExpressions(guessCand.mpqaSentence));
            linkResult.add(goldSen.compareLinks(guessCand.mpqaSentence));
    	}
		
	}
	
	private static double[] parseCostWeights(String ws) {
		String[] ts = ws.split("/");
		return CollectionUtils.toDoubleArray(ts);		
	}
	
	public static void train(String[] argv, Iterator<MPQAText> r) {
		int ap = 0;
		String precomputed = argv[++ap]; System.out.println("precomputed = " + precomputed);

		String dir = null;
		String subjLexFile = null;
		String docListFile = null;
		String synsemFile = null;
		String constFile = null;
		String discConnFile = null;
		String candidateFile = null;
		int Nseq = -1;
		int Nsrcseq = -1;
		int Npolseq = -1;
		String linkModelFile = null;
		String polintModelFile = null;
		String saveCandidates = null;
		
		boolean extractSrcFeatures, extractPolFeatures;
		String generateTKData = null; // = null, train or test
		
        System.out.println("*** Reranker training ***");
		
		if(Util.isNullFile(precomputed)) {
			precomputed = null;
			dir = argv[++ap]; System.out.println("dir = " + dir);
			subjLexFile = argv[++ap]; System.out.println("subjLexFile = " + subjLexFile);
			docListFile = argv[++ap]; System.out.println("docListFile = " + docListFile);
			synsemFile = argv[++ap]; System.out.println("synsemFile = " + synsemFile);
			constFile = argv[++ap]; System.out.println("constFile = " + constFile);
			discConnFile = argv[++ap]; System.out.println("discConnFile = " + discConnFile);
			candidateFile = argv[++ap];  System.out.println("candidateFile = " + candidateFile);
			Nseq = Integer.parseInt(argv[++ap]); System.out.println("Nseq = " + Nseq);
			Nsrcseq = Integer.parseInt(argv[++ap]); System.out.println("Nsrcseq = " + Nsrcseq);
			Npolseq = Integer.parseInt(argv[++ap]); System.out.println("Npolseq = " + Npolseq);
			linkModelFile = argv[++ap]; System.out.println("linkModelFile = " + linkModelFile);
			polintModelFile = argv[++ap]; System.out.println("polintModelFile = " + polintModelFile);
			saveCandidates = argv[++ap]; System.out.println("saveCandidates = " + saveCandidates);
			generateTKData = argv[++ap];  System.out.println("generateTKData = " + generateTKData);
		} else {
			extractSrcFeatures = Boolean.parseBoolean(argv[++ap]); System.out.println("extractSrcFeatures = " + extractSrcFeatures);
			extractPolFeatures = Boolean.parseBoolean(argv[++ap]); System.out.println("extractPolFeatures = " + extractPolFeatures);
		}
		
		/* Seq reranker feature params */
		boolean useBigrams = Boolean.parseBoolean(argv[++ap]); System.out.println("useBigrams = " + useBigrams);
		boolean useSyntax = Boolean.parseBoolean(argv[++ap]);  System.out.println("useSyntax = " + useSyntax);
		boolean useSpans = Boolean.parseBoolean(argv[++ap]);     System.out.println("useSpans = " + useSpans);
		boolean useSemantics = Boolean.parseBoolean(argv[++ap]); System.out.println("useSemantics = " + useSemantics);

		double[] costWeights = parseCostWeights(argv[++ap]); System.out.println("costWeights = " + Arrays.toString(costWeights));
		
		String algName = argv[++ap]; System.out.println("algName = " + algName);
		String algArgs = argv[++ap]; System.out.println("algArgs = ...");
		
		String outModelFile = argv[++ap]; System.out.println("outModelFile = " + outModelFile);

		try {

			/* Feature extractor for seq reranker. */
			RerankingFE rfe = new RerankingFE(useBigrams, useSyntax, useSpans, useSemantics);

			ArrayList<SentenceData> processedCandidates = null;
			ArrayList<Pair<EncodedCandidateSet, Integer>> trainingSet = new ArrayList();

			/* Loading trained model. */
			LocalLinkClassifier llc = null;
			if(!Util.isNullFile(linkModelFile)) {
				System.out.println("Loading link model...");
				llc = LocalLinkClassifier.loadModel(linkModelFile);
			}

			Classifier<SparseVector, Integer> polCl = null;
			Classifier<SparseVector, Integer> intCl = null;
			SymbolEncoder polintEnc = null;
			SubjectivityLexicon sl = null;
			if(!Util.isNullFile(polintModelFile)) {
				System.out.println("Loading polarity and intensity classifiers...");
				ObjectInputStream ois = new ObjectInputStream(Util.openFileStream(polintModelFile));
				polCl = (Classifier) ois.readObject();
				intCl = (Classifier) ois.readObject();				
				try {
					polintEnc = (SymbolEncoder) ois.readObject();
				} catch(Exception e) {}
				ois.close();
				sl = new SubjectivityLexicon(subjLexFile, false);
			}
			
			extractSrcFeatures = USE_SRC_FEATURES && llc != null;
			extractPolFeatures = USE_POL_FEATURES && polCl != null;
			
			if(extractSrcFeatures)
				System.out.println("Source link classifier specified, will extract source features.");
			else
				System.out.println("Will NOT extract source features.");
			if(extractPolFeatures)
				System.out.println("Polarity classifier specified, will extract polarity features.");
			else
				System.out.println("Will NOT extract polarity features.");
			
			if(Util.isNullFile(precomputed)) {
				long t0 = System.currentTimeMillis();
				
				System.out.println("Loading candidate lists...");
				ArrayList<SentenceData> seqCands = SeqCandidateReader.readSentences(candidateFile,
						synsemFile, constFile, discConnFile, Nseq, -1);
				HashMap<String, SentenceData> candMap = createCandidateMap(seqCands);

				//ArrayList<String> selectedFiles = new ArrayList();
				//Util.readLines(docListFile, selectedFiles);
				//trimFileNames(selectedFiles);
				//selectedFiles.remove("20020203/20.46.36-9539");

				//LKReader r = new LKReader(dir, subjLexFile, selectedFiles);
				//int candSenIndex = 0;

				System.out.println("Processing training set...");

				processedCandidates = new ArrayList();

				int nExpr = 0, nPos = 0, nNeu = 0, nNeg = 0, nBoth = 0, nDS = 0, nES = 0, nOS = 0;

				int count = 0;
				while(r.hasNext()) {
					MPQAText text = r.next();
					
					System.out.println(text.source);
					
					/*count++;
					if(count % 10 == 0) {
					    System.out.print(".");
					    System.out.flush();
					}*/
					
					/*
					for(MPQASentence sen: text.sentences) {
						for(MPQAExpression expr: sen.exprs) {
							if(expr.type.equals("os")) {
								nOS++;
								continue;
							} else if(expr.type.equals("ds"))
								nDS++;
							else if(expr.type.equals("es"))
								nES++;
							nExpr++;
							String pol = expr.polarity;
							if(pol == null)
								pol = "neutral";
							pol = pol.trim();
							if(pol.equals(""))
								pol = "neutral";
							if(pol.startsWith("un"))
								pol = pol.substring(10);
							if(pol.equals("positive"))
								nPos++;
							else if(pol.equals("negative"))
								nNeg++;
							else if(pol.equals("neutral"))
								nNeu++;
							else if(pol.equals("both"))
								nBoth++;
							else throw new RuntimeException("!!!");
						}
					}		

					String shortSrc = text.source.substring(23);
					if(!selectedFiles.contains(text.source) && !selectedFiles.contains(shortSrc))
						throw new RuntimeException("source = " + text.source);
                    */

					//System.out.println(text.source);
					//int nextSenIndex = candSenIndex + text.sentences.size();
					//List<SentenceData> sds = seqCands.subList(candSenIndex, nextSenIndex);
					//candSenIndex = nextSenIndex;

					try {
						//generateRerankerTrainingData(llc, polCl, intCl, polintEnc, sl,
						//		text, candMap, Nsrcseq, Npolseq, costWeights, rfe, trainingSet);

						generateRerankerCandidates(llc, polCl, intCl, polintEnc, sl,
								text, candMap, Nsrcseq, Npolseq, processedCandidates);

					} catch(Exception e) {

						System.err.println("Error when processing " + text.source);

						e.printStackTrace();
						System.exit(1);
					}

				}

				System.out.println();

				/*
				System.out.println("nExpr = " + nExpr);
				System.out.format("nPos = %d, nNeg = %d, nNeu = %d, nBoth = %d\n",
						nPos, nNeg, nNeu, nBoth);
				System.out.format("nDS = %d, nES = %d, nOS = %d\n",
						nDS, nES, nOS);*/

				seqCands = null;
				candMap = null;

				long t1 = System.currentTimeMillis();
				System.out.println("Loading time: " + Util.toTimeString(t1 - t0, Util.SECONDS));
				
				if(!Util.isNullFile(saveCandidates)) {					
					System.out.println("Writing candidates to " + saveCandidates);
					long t2 = System.currentTimeMillis();
					ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(saveCandidates));
					oos.writeObject(processedCandidates);
					oos.close();
					long t3 = System.currentTimeMillis();
					System.out.println("Writing time: " + Util.toTimeString(t3 - t2, Util.SECONDS));
				}
			} else {
				long t0 = System.currentTimeMillis();
				System.out.println("Reading candidates from " + precomputed);
				ObjectInputStream ois = new ObjectInputStream(Util.openFileStream(precomputed));
				processedCandidates = (ArrayList) ois.readObject();
				ois.close();
				ois = null;
				long t1 = System.currentTimeMillis();
				System.out.println("Loading time: " + Util.toTimeString(t1 - t0, Util.SECONDS));
			}

			if(Util.isNullFile(generateTKData)) {

				long t0 = System.currentTimeMillis();
				System.out.println("Encoding candidates...");
				encodeCandidates(processedCandidates, extractSrcFeatures, extractPolFeatures,
						costWeights, rfe, trainingSet);
				long t1 = System.currentTimeMillis();
				System.out.println("Encoding time: " + Util.toTimeString(t1 - t0, Util.SECONDS));

				//int n = seqCands.size();
				//System.out.format("Average number of candidates: %d / %d = %f\n",
				//		nCandidates, n, (double) nCandidates / n);
				//System.out.println();

				AlgorithmFactory af = new AlgorithmFactory();
				LearningAlgorithm alg = af.create(algName, algArgs);
				RerankingDefinition def = new RerankingDefinition();
				Classifier<EncodedCandidateSet, Integer> cl = alg.train(def, trainingSet);

				try {
					ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(outModelFile));
					oos.writeObject(rfe);
					oos.writeObject(cl);
					oos.writeInt(Nseq);
					oos.writeInt(Nsrcseq);
					oos.writeInt(Npolseq);
					oos.close();
				} catch(Exception e) {
					e.printStackTrace();
					System.exit(1);
				}

				System.out.println("Done.");

				if(Reranker.PRINT_TOP_FEATURES) {
					EncodedCandidateSet ecs = new EncodedCandidateSet();
					ecs.reps = new SparseVector[1];
					ecs.reps[0] = new SparseVector();
					ecs.reps[0].put(1, 1);
					ArrayList<Integer> range = new ArrayList();
					range.add(0);
					double[] scores = new double[1];
					ArrayList<DoubleIntPair> weights = new ArrayList();
					long max = RerankingFE.getMaxFeatureIndex();
					System.out.println("max = " + max);
					for(int i = 1; i < max; i++) {
						ecs.reps[0].keys[0] = i;
						cl.computeScoresRestricted(ecs, range, scores);
						if(scores[0] > 1e-6 || scores[0] < -1e-6)	
							weights.add(new DoubleIntPair(scores[0], i));
					}
					Collections.sort(weights, new Comparator() {
						public int compare(Object o1, Object o2) {
							DoubleIntPair d1 = (DoubleIntPair) o1;
							DoubleIntPair d2 = (DoubleIntPair) o2;
							return -Double.compare(Math.abs(d1.left), Math.abs(d2.left));
						}					
					});
					int n = Math.min(weights.size(), 250);
					RerankingFE.printProminentFeatures(weights.subList(0, n));
				}
			} else {
				System.out.println("Writing TK data...");
				String fn = generateTKData;
				PrintWriter treePW = new PrintWriter(new OutputStreamWriter(new FileOutputStream(fn + ".conllx"), "UTF-8"));
				PrintWriter bracketPW = new PrintWriter(new OutputStreamWriter(new FileOutputStream(fn + ".brackets"), "UTF-8"));
				PrintWriter tktemplatePW = new PrintWriter(new OutputStreamWriter(new FileOutputStream(fn + ".template"), "UTF-8"));
				PrintWriter scorePW = null;
				if(generateTKData.equals("test"))
					scorePW = new PrintWriter(new FileWriter("test.scores"));

				//PrintWriter debugPW = new PrintWriter("debug2.xxx");				
				int sid = 0;
				for(SentenceData sd: processedCandidates) {
					//debugPW.println(sd.candidates.get(0).stats.toRawString());
					printTKData(sd, treePW, bracketPW, tktemplatePW, scorePW, generateTKData.equals("train"), ++sid);
				}
				//debugPW.close();
				
				treePW.close();
				bracketPW.close();
				tktemplatePW.close();
				if(scorePW != null)
					scorePW.close();
				
				System.out.println("Done.");
			}
			
		} catch(Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
	}
			
	private static void printTKData(SentenceData sd, PrintWriter treePW,
			PrintWriter bracketPW, PrintWriter tktemplatePW, 
			PrintWriter scorePW,
			boolean train,
			int id) throws IOException {

//		boolean debug = sd.text.equals("Now Mr. Ruckauf is considered a major presidential contender .");
//		if(debug) {
//			System.out.println(sd.candidates.get(0).stats.toRawString());
//		}
		
		// Print depgraph to file
		
		treePW.println("#" + id);
		CoNLLFormat.printGraph(treePW, sd.goldSpans.dg);

		double[] lls = new double[sd.candidates.size()];
		
		double maxF1 = Double.NEGATIVE_INFINITY;
		Candidate selected = null; // the first one with the highest
		int i = 0;
		int selectedIndex = 0;

		for(Candidate cand: sd.candidates) {

			//For the NEW prf evaluation
			//PRFStats prf = sd.goldSpans.compareExpressions(cand.mpqaSentence);
			
			//For the OLD prf evaluation
			PRFStats prf = cand.stats;

			//For the NEW UNIQUE prf evaluation
			//PRFStats prf = sd.goldSpans.compareExpressionsUnique(cand.mpqaSentence);

			//if(debug && i == 0)
			//	System.out.println(prf.toRawString());
			
			lls[i] = cand.baseScore;
			if(prf.getPartialF1() > maxF1) {
				maxF1 = prf.getPartialF1();
				selected = cand;
				selectedIndex = i;
			}
			i++;
		}

		Reranker.normalize(lls, sd.candidates.size());
		
		tktemplatePW.println("#" + id);
		bracketPW.println("#" + id);
		if(scorePW != null)
			scorePW.println("#" + id);
		
		i = 0;
		for(Candidate cand: sd.candidates) {

			//For the NEW prf evaluation
			//PRFStats prf = sd.goldSpans.compareExpressions(cand.mpqaSentence);

			//For the OLD prf evaluation
			PRFStats prf = cand.stats;

			//For the NEW UNIQUE prf evaluation
			//PRFStats prf = sd.goldSpans.compareExpressionsUnique(cand.mpqaSentence);
			
			if(prf.getPartialF1() < maxF1) {
				if(train) {
					tktemplatePW.print("+1 |BT| (???) |BT| (???) |ET| 1:");
					tktemplatePW.print(lls[selectedIndex]);
					tktemplatePW.print(" |BV| 1:");
					tktemplatePW.print(lls[i]);
					tktemplatePW.println(" |EV|");				

					bracketPW.println(selected.mpqaSentence.toBrackets() 
							+ "\t" + cand.mpqaSentence.toBrackets());
					
					tktemplatePW.print("-1 |BT| (???) |BT| (???) |ET| 1:");
					tktemplatePW.print(lls[i]);
					tktemplatePW.print(" |BV| 1:");
					tktemplatePW.print(lls[selectedIndex]);
					tktemplatePW.println(" |EV|");

					bracketPW.println(cand.mpqaSentence.toBrackets() 
							+ "\t" + selected.mpqaSentence.toBrackets());
				} else {
					tktemplatePW.print("-1 |BT| (???) |BT| |ET| 1:");
					tktemplatePW.print(lls[i]);
					tktemplatePW.println(" |BV| |EV|");
					bracketPW.println(cand.mpqaSentence.toBrackets());
					scorePW.println(prf.toRawString());
				}								
			} else if(!train) {
				tktemplatePW.print("+1 |BT| (???) |BT| |ET| 1:");
				tktemplatePW.print(lls[i]);
				tktemplatePW.println(" |BV| |EV|");
				bracketPW.println(cand.mpqaSentence.toBrackets());
				scorePW.println(prf.toRawString());
			}
			i++;
		}
		
	}

	private static String[][] readNextSentence(BufferedReader br) throws IOException {
		String line = br.readLine();
		if(line == null)
			return null;
		line = line.trim();
		ArrayList<String[]> ls = new ArrayList();
		while(line != null && !line.equals("")) {
			String[] ts = line.split("\\s+");
			ls.add(ts);
			line = br.readLine();
			if(line != null)
				line = line.trim();			
		}
		return (String[][]) ls.toArray(new String[0][]);
	}
	
	private static void seqPolBaseline(String[] argv) {

		String seqLabelerFile = argv[1];
		String polintModelFile = argv[2];
		int Nseq = Integer.parseInt(argv[3]);
		int Npol = Integer.parseInt(argv[4]);		
		String seqFile = argv[5];
		String outFile = argv[6];
		
		try {

			Classifier<SparseVector, Integer> polCl = null;
			SymbolEncoder polintEnc = null;
			SubjectivityLexicon sl = null;
			if(!polintModelFile.matches("NULL|NONE|EMPTY")) {
				System.out.println("Loading polarity and intensity classifiers...");
	            ObjectInputStream ois = new ObjectInputStream(Util.openFileStream(polintModelFile));
	            polCl = (Classifier) ois.readObject();
	            ois.readObject();				
				try {
					polintEnc = (SymbolEncoder) ois.readObject();
				} catch(Exception e) {}
				ois.close();
			}
			
			TaggingModel seqModel = null;
			ObjectInputStream ois = new ObjectInputStream(Util.openFileStream(seqLabelerFile));
			seqModel = (TaggingModel) ois.readObject();
			ois.close();

			BufferedReader br = Util.openFileReader(seqFile);

			PrintWriter outPW = new PrintWriter(new FileWriter(outFile));

			System.err.println("Generating candidates...");
			
			String[][] sen = readNextSentence(br);
			while(sen != null) {
				outPW.print(sen[0][0]);
				for(int i = 1; i < sen.length; i++)
					outPW.print(" " + sen[i][0]);
				outPW.println();					
				ArrayList<Candidate> cands = Candidate.convertList(sen,
						SeqLabeler.tagSentenceKBest(seqModel, sen, Nseq));
				ArrayList<Candidate> candsP = expandCandidateSetWithPolarity(polCl,
						polintEnc, cands, null, Npol);
				for(Candidate cand: candsP) {
					outPW.println(cand.baseScore + " " + cand.basePolScore);
					boolean first = true;
					for(MPQAExpression expr: cand.mpqaSentence.exprs) {
						if(!first)
							outPW.print(" ");
						else
							first = false;
						outPW.print(expr.type + "," + (expr.span.get(0).position-1) + ","
								+ (expr.span.get(expr.span.size()-1).position));
						if(expr.polarity != null) {
							outPW.print(",");
							if(expr.polarity.startsWith("pos"))
								outPW.print("+");
							else if(expr.polarity.startsWith("neg"))
								outPW.print("-");
							else if(expr.polarity.startsWith("neu"))
								outPW.print("0");
							else
								throw new RuntimeException("!!!");
						}
					}
					outPW.println();
				}
				outPW.println("---");
				sen = readNextSentence(br);
			}
			
			outPW.close();
			
		} catch(Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
	}

	
	private static HashMap<String, SentenceData> createCandidateMap(ArrayList<SentenceData> sds) {
		HashMap<String, SentenceData> out = new HashMap();
		for(SentenceData sd: sds) {
		    out.put(sd.text, sd);
		}
		return out;
	}
	
	private static void evalTK(String[] argv) {
		String dir = argv[1];
		String subjLexFile = argv[2];
		String docListFile = argv[3];
		
		String seqCandidateFile = argv[4];
		int Nseq = Integer.parseInt(argv[5]);
		int Nsrcseq = Integer.parseInt(argv[6]);
		
		String linkModelFile = argv[7];
		String tkPredictions = argv[8];
		
		try {
			/* Loading trained model. */
			LocalLinkClassifier llc = null;
			if(!linkModelFile.matches("NULL|NONE|EMPTY"))
				llc = LocalLinkClassifier.loadModel(linkModelFile);

			BufferedReader tkPredReader = new BufferedReader(new FileReader(tkPredictions));
			
			System.out.println("Loading candidate lists...");
            ArrayList<SentenceData> seqCands = SeqCandidateReader.readSentences(seqCandidateFile, null, null, null, Nseq, -1);

			PRFStats exprResult = new PRFStats();
			PRFStats linkResult = new PRFStats();
            
            ArrayList<String> selectedFiles = new ArrayList();
			Util.readLines(docListFile, selectedFiles);
            
			LKReader r = new LKReader(dir, subjLexFile, selectedFiles);
			int candSenIndex = 0;

			System.out.println("Processing test set...");
			
			while(r.hasNext()) {
				MPQAText text = r.next();
				String src = text.source;
				src = src.replaceAll("database.mpqa.2.0/docs/", "");
				src = src.replaceAll("/", "_");
				src = src + ".lktext.xml";
				if(!selectedFiles.contains(src)) {
					System.out.println("src = " + src);
					throw new RuntimeException("shouldn't get this text");
				}
				//System.out.println(text.source);
				int nextSenIndex = candSenIndex + text.sentences.size();					
				List<SentenceData> sds = seqCands.subList(candSenIndex, nextSenIndex);
				candSenIndex = nextSenIndex;					
				
				rerankTextTK(text, sds, llc, Nsrcseq, tkPredReader, exprResult, linkResult);
			}
			
			if(tkPredReader.readLine() != null)
				throw new RuntimeException("not all lines read");
			
			System.out.println("Expression result:");
			exprResult.print();

			if(llc != null) {
				System.out.println();
				System.out.println("Link result:");
				linkResult.print();
			}
			
		} catch(Exception e) {
			e.printStackTrace();
		}
		
	}
	
	private static void genTKData(String[] argv) {
		String dir = argv[1];
		String subjLexFile = argv[2];
		String docListFile = argv[3];

		String synsemFile = argv[4]; 
		String pstreeFile = argv[5]; 
		
		String seqCandidateFile = argv[6];
		int Nseq = Integer.parseInt(argv[7]);
		int Nsrcseq = Integer.parseInt(argv[8]);

		String linkModelFile = argv[9];
		boolean train = Boolean.parseBoolean(argv[10]);		
		String tkDataFile = argv[11];

		try {

			/* Loading trained model. */
			LocalLinkClassifier llc = null;
			if(!linkModelFile.matches("NULL|NONE|EMPTY"))
				llc = LocalLinkClassifier.loadModel(linkModelFile);

			PrintWriter pw = new PrintWriter(new FileWriter(tkDataFile));

			ArrayList<SentenceData> seqCands = SeqCandidateReader.readSentences(seqCandidateFile, synsemFile, pstreeFile, null, Nseq, -1);
			
            ArrayList<String> selectedFiles = new ArrayList();
			Util.readLines(docListFile, selectedFiles);

			LKReader r = new LKReader(dir, subjLexFile);
			int candSenIndex = 0;
			int count = 0;
			while(r.hasNext()) {
				MPQAText text = r.next();
				count++;
				if(count % 10 == 0) {
					System.out.print(".");
					System.out.flush();
				}
					
				if(selectedFiles.contains(text.source)) {
					System.out.println(text.source);
					int nextSenIndex = candSenIndex + text.sentences.size();
					List<SentenceData> sds = seqCands.subList(candSenIndex, nextSenIndex);
					candSenIndex = nextSenIndex;					

					generateRerankerTKData(llc, text, sds, Nsrcseq, train, 1.0, pw);
					
				}
			}			
			pw.close();			
			System.out.println("Done.");
			
		} catch(Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
	}	
	
	private static void generateRerankerTKData(LocalLinkClassifier llm,
			MPQAText text, List<SentenceData> sds, int k, boolean train,
			double fAlpha, PrintWriter pw) {

		throw new RuntimeException("deprecated");

		/*
		for(int j = 0; j < text.sentences.size(); j++) {
    		MPQASentence sen = text.sentences.get(j);
    		SentenceData sd = sds.get(j);
    		sen.replaceDepTree(sd.goldSpans.dg);

    		ArrayList<Candidate> cands;
    		if(llm != null)
    			cands = expandCandidateSetWithSources(llm, sd.candidates, k, sen, true);
    		else
    			cands = sd.candidates;
    		
    		int n = cands.size();
    		
    		double[] costs = new double[n];
            double[] lls = new double[n];
            double minCost = Double.POSITIVE_INFINITY;
            double maxCost = Double.NEGATIVE_INFINITY;

            for(int i = 0; i < n; i++) {
                lls[i] = cands.get(i).baseScore;                
                costs[i] = 1.0 - cands.get(i).stats.getPartialF(fAlpha);
                
                if(costs[i] < minCost)
                    minCost = costs[i];
                if(costs[i] > maxCost)
                    maxCost = costs[i];
            }
            
            for(int i = 0; i < n; i++)
                costs[i] -= minCost;
            Reranker.normalize(lls, cands.size());

            int o = CollectionUtils.minIndex(costs, costs.length);

            if(train && minCost == maxCost)
                continue;

            //String[] depEncs = new String[n];
            String[] depEncsU = new String[n];
            String[] psEncs = new String[n];
            String[] seqEncs = new String[n];
            String[] pasEncs = new String[n];
            
            for(int i = 0; i < cands.size(); i++) {
            	//depEncs[i] = TKEncodings.markNodesDG(cands.get(i), true);
            	depEncsU[i] = TKEncodings.markNodesDG(cands.get(i), false);
            	psEncs[i] = TKEncodings.markNodesPS2(cands.get(i));
            	seqEncs[i] = TKEncodings.markSeqAroundWords(cands.get(i));
            	pasEncs[i] = TKEncodings.markNodesPA(cands.get(i));
            }
            
            if(train) {
                for(int i = 0; i < n; i++) {
                	if(Util.equal(costs[i], 0))
                		continue;
                	TKEncodings.printCand(pw, true, 
                			//depEncs[o], depEncs[i], 
                			depEncsU[o], depEncsU[i], 
                			psEncs[o], psEncs[i], 
                			seqEncs[o], seqEncs[i], 
                			pasEncs[o], pasEncs[i],
                			lls[o], lls[i]);
                	TKEncodings.printCand(pw, false, 
                			//depEncs[i], depEncs[o], 
                			depEncsU[i], depEncsU[o], 
                			psEncs[i], psEncs[o], 
                			seqEncs[i], seqEncs[o], 
                			pasEncs[i], pasEncs[o],
                			lls[i], lls[o]);
                }
            } else {
                for(int i = 0; i < n; i++) {
                	TKEncodings.printCand(pw, Util.equal(costs[i], 0),
                			//depEncs[i], null,
                			depEncsU[i], null,
                			psEncs[i], null,
                			seqEncs[i], null,
                			pasEncs[i], null,
                			lls[i], 0);
                }
            }
    	}	
		*/
	}
	

	private static void baseline(String[] argv) {
		String dir = argv[1];
		String subjLexFile = argv[2];
		String docListFile = argv[3];
		String seqCandidateFile = argv[4];
		int N = Integer.parseInt(argv[5]);
		
		String linkModelFile = argv[6];
		String seqRerankerFile = argv[7];
		
		try {

			/* Loading trained model. */
			LocalLinkClassifier llc = LocalLinkClassifier.loadModel(linkModelFile);
			Pair<RerankingFE, Classifier<EncodedCandidateSet, Integer>> p = loadReranker(seqRerankerFile);
			
			ArrayList<SentenceData> seqCands = SeqCandidateReader.readSentences(seqCandidateFile, null, null, null, N, -1);
			
			/* Creation of example feature vectors. */
            ArrayList<String> selectedFiles = new ArrayList();
			Util.readLines(docListFile, selectedFiles);

			LKReader r = new LKReader(dir, subjLexFile);
			int candSenIndex = 0;

			PRFStats exprResult = new PRFStats();
			PRFStats linkResult = new PRFStats();
			
			while(r.hasNext()) {
				MPQAText text = r.next();
				if(selectedFiles.contains(text.source)) {
					//System.out.println(text.source);
					int nextSenIndex = candSenIndex + text.sentences.size();					
					List<SentenceData> cands = seqCands.subList(candSenIndex, nextSenIndex);
					candSenIndex = nextSenIndex;					
					testTextBaseline(text, cands, llc, p.left, p.right,
							exprResult, linkResult);
				}
			}
			
			System.out.println("Expression result:");
			exprResult.print();
			
			System.out.println();
			System.out.println("Link result:");
			linkResult.print();
			
		} catch(Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
	}

	private static Pair<RerankingFE, Classifier<EncodedCandidateSet, Integer>> loadReranker(
			String file) throws IOException {
		ObjectInputStream ois = new ObjectInputStream(new FileInputStream(file));
		try {
			RerankingFE fe = (RerankingFE) ois.readObject();
			Classifier<EncodedCandidateSet, Integer> cl = (Classifier) ois.readObject();
			ois.close();
			return new Pair(fe, cl);
		} catch(Exception e) {
			throw new IOException(e);
		}
	}

	private static void testTextBaseline(MPQAText text, List<SentenceData> candList,
			LocalLinkClassifier llc, RerankingFE rfe, 
			Classifier<EncodedCandidateSet, Integer> rcl, 
			PRFStats exprResult, PRFStats linkResult) {
		
		if(true)
			throw new RuntimeException("unimplemented");
		
		for(int i = 0; i < text.sentences.size(); i++) {
			MPQASentence goldSen = text.sentences.get(i);

			//goldSen.pruneIdentical();
			//goldSen.pruneOverlapping();
			
			MPQASentence guessSen;
			
			Candidate cand = null; // getCandidate(candList.get(i), goldSen, rfe, rcl);			
			//double origF1 = cand.stats.getSoftF1();
			guessSen = null; //toMPQASentence(cand, goldSen);

			//guessSen = toMPQASentence(candList.get(i).goldSpans, goldSen);

			addSources(llc, guessSen);
			//kbestSourcesOracle(llc, guessSen, 64, goldSen);

			if(false && i == 1) {
				PRFStats er = goldSen.compareExpressions(guessSen);
				PRFStats lr = goldSen.compareLinks(guessSen);
				
				System.out.println(goldSen);
				System.out.println(guessSen);

				er.print();
				lr.print();
				
				System.exit(0);
			}

			PRFStats er = goldSen.compareExpressions(guessSen);			

			if(false && er.getRecall() < 1.0) {
				System.out.println(goldSen.dg);
				System.out.println(goldSen.exprs);
				System.out.println(guessSen.exprs);
				System.exit(0);
			}
			
			/*if(Math.abs(origF1 - newF1) > 1e-3) {
				System.out.println("Old f1: " + origF1);
				System.out.println("New f1: " + newF1);
				System.exit(0);
			}*/
			
			exprResult.add(goldSen.compareExpressions(guessSen));
			linkResult.add(goldSen.compareLinks(guessSen));
		}		
	}

	private static boolean FIRST = false;
	private static boolean ORACLE = false;
	
//	private static Candidate getCandidate(SentenceData candList, 
//			MPQASentence goldSen, 
//			RerankingFE rfe, Classifier<EncodedCandidateSet, Integer> rcl) {
//		if(true)
//			throw new RuntimeException("unimplemented!");
//		
//		if(FIRST)
//			return candList.candidates.get(0);
//		if(!ORACLE) {
//			double[] scores = new double[candList.candidates.size()];
//			Reranker.rerankSentence(rfe, rcl, candList.candidates, scores);
//			int ix = CollectionUtils.maxIndex(scores, scores.length);
//			return candList.candidates.get(ix);
//		} else {
//			Candidate bestCand = null;
//			double bestScore = Double.NEGATIVE_INFINITY;
//			for(Candidate cand: candList.candidates) {
//				double score = cand.stats.getPartialF1();
//				if(score > bestScore) {
//					bestScore = score;
//					bestCand = cand;
//				}				
//			}
//			return bestCand;
//		}
//	}
//
//	private static MPQASentence toMPQASentence(Candidate cd, MPQASentence s0) {
//		if(true)
//			throw new RuntimeException("unimplemented!");
//		
//		MPQASentence out = new MPQASentence();
//		out.dg = s0.dg;
//		out.pas = s0.pas;
//		out.slClues = s0.slClues;
//		out.exprs = new ArrayList();
//		for(Span sp: cd.spans)
//			out.exprs.add(toMPQAExpression(sp, out));
//
//		if(cd.srcIndex != null) {
//			for(int i = 0; i < cd.spans.size(); i++) {
//				MPQAAgent src = new MPQAAgent();
//				if(cd.isWriter[i])
//					src.isWriter = true;
//				else if(cd.isImplicit[i])
//					src.isImplicit = true;
//				else if(cd.srcIndex[i] > 0) {
//					MPQAAgentInstance inst = new MPQAAgentInstance();
//					DepNode srcHead = out.dg.nodes[cd.srcIndex[i]];
//					inst.span = createSourceSpan(srcHead, out.exprs.get(i).span, false);
//					inst.agent = src;
//					inst.containingSentence = out;
//					src.instances.add(inst);
//				} else
//					continue; // skip
//				ArrayList<MPQAAgent> ns = new ArrayList();
//				ns.add(src);
//				out.exprs.get(i).nestedSource = ns;
//			}
//		}
//		
//		return out;
//	}

	private static MPQAExpression toMPQAExpression(Span s, MPQASentence sen) {
		MPQAExpression out = new MPQAExpression();
		out.type = s.label;
		out.span = new ArrayList();
		for(int i = s.tokenStart; i < s.tokenEnd; i++)
			out.span.add(sen.dg.nodes[i+1]);		
		out.containingSentence = sen;
		return out;
	}
	
	private static void addSources(LocalLinkClassifier llm, MPQASentence sen) {
		if(sen.exprs.isEmpty())
			return;
		ArrayList<DepNode> cands = LocalLinkClassifier.findCandidates(sen);
		for(MPQAExpression expr: sen.exprs) {
			MPQAAgent agent = findSource(llm, expr, cands);
			expr.nestedSource = new ArrayList();
			expr.nestedSource.add(agent);
		}
	}
	
	public static ArrayList<DepNode> createSourceSpan(DepNode headNode, 
			ArrayList<DepNode> exprSpan, boolean onlyNode) {
		if(onlyNode) {
			ArrayList<DepNode> out = new ArrayList();
			out.add(headNode);
			return out;
		} else {
			HashSet<DepNode> out = new HashSet();
			for(DepNode n: headNode.children) {
				boolean dominatesExpr = false;
				for(DepNode sn: exprSpan)
					if(n.isSameOrAncestorOf(sn)) {
						dominatesExpr = true;
						break;
					}
				if(!dominatesExpr)
					//n.collectSubtree(out);
				    collectSubtreeNew(n, out);
			}		
			out.add(headNode);
			ArrayList<DepNode> out2 = new ArrayList(out);
			Collections.sort(out2);
			
			pruneDiscontinuous(out2, headNode);

			if(out2.isEmpty())
			    out2.add(headNode);
			
			return out2;
		}
	}
	
	private static void collectSubtreeNew(DepNode h, HashSet<DepNode> seen) {
	    if(seen.contains(h))
	        return;
	    seen.add(h);
	    for(DepNode c: h.children)
	        collectSubtreeNew(c, seen);
	}
	
	private static void pruneDiscontinuous(ArrayList<DepNode> span, DepNode head) {
	    HashSet<DepNode> toRetain = new HashSet();
	    toRetain.add(head);
	    int ix = span.indexOf(head);

	    int position = head.position;
	    for(int i = ix - 1; i >= 0; i--) {
	        DepNode n = span.get(i);
	        if(n.position == position - 1) {
	            position = n.position;
	            toRetain.add(n);
	        } else
	            break;
	    }
	    position = head.position;
	    for(int i = ix + 1; i < span.size(); i++) {
	        DepNode n = span.get(i);
	        if(n.position == position + 1) {
	            position = n.position;
	            toRetain.add(n);
	        } else
	            break;
	    }
	    span.retainAll(toRetain);	    
    }
	
	private static final List<String> LOCAL_CLASSES = Arrays.asList(new String[] { "T", "F" }); 
	private static final List<String> NONLOCAL_CLASSES = Arrays.asList(new String[] { "0", "W", "I", "E" }); 
	
	private static MPQAAgent findSource(LocalLinkClassifier llm, MPQAExpression expr, 
			ArrayList<DepNode> cands) {
		
		FeatureList localFeatures = new FeatureList(llm.localDecl);

		Pair<MPQAExpression, MPQASentence> p = new Pair(expr, expr.containingSentence);
		llm.localExprExtractor.extractFeatures(p, localFeatures);
		
		DepNode bestNode = null;
		double bestScore = Double.NEGATIVE_INFINITY;
		
		Triple<MPQAExpression, DepNode, MPQASentence> tr = new Triple(expr, null, expr.containingSentence);

		double[] localScores = new double[2];
		for(DepNode cand: cands) {
			tr.second = cand;
			llm.localNodeExtractor.extractFeatures(tr, localFeatures);
			llm.localCl.computeScoresRestricted(localFeatures, LOCAL_CLASSES, localScores);
			if(localScores[0] > bestScore) {
				bestNode = cand;
				bestScore = localScores[0];
			}
		}
		
		FeatureList nonlocalFeatures = new FeatureList(llm.nonlocalDecl);
		llm.nonlocalExtractor.extractFeatures(p, nonlocalFeatures);
		double[] nonlocalScores = new double[4];		
		llm.nonlocalCl.computeScoresRestricted(nonlocalFeatures, NONLOCAL_CLASSES, nonlocalScores);
		
		// No coreference handling: we create one agent per expression
		MPQAAgent out = new MPQAAgent();

		int nlIx = -1;

		for(int i = 1; i < 4; i++) {
			if(nonlocalScores[i] > bestScore) {
				bestScore = nonlocalScores[i];
				nlIx = i;
			}	
		}
		
		switch(nlIx) {
		case -1: { // node
			MPQAAgentInstance inst = new MPQAAgentInstance();
			inst.containingSentence = expr.containingSentence;
			inst.agent = out;			
			inst.agent.instances.add(inst);
			inst.span = createSourceSpan(bestNode, expr.span, false);

			//System.out.println("Added instance: " + inst.span);
		}
		break;
		case 1: { // writer
			//System.out.println("Added writer");
			out.isWriter = true;
		}
		break;
		case 2: { // implicit
			//System.out.println("Added implicit");
			out.isImplicit = true;			
		}
		break;
		case 3: { // external
			// don't handle
			out = null;
		}
		break;
		default: throw new RuntimeException("!!!");
		}
		
		return out;
	}
	
	private static final Comparator<DoubleObjPair> NEG_LEFT = new NegatedComparator(DoubleObjPair.BY_LEFT);
	
    private static final double SCALE = 1.0;
    private static final double OFFSET = 0.0;    

    static void normalize(ArrayList<DoubleObjPair<MPQAAgent>> l) {
    	double m = Double.MIN_VALUE;
    	for(DoubleObjPair p: l)
    		if(p.left > m)
    			m = p.left;
    	
        m = SCALE*m + OFFSET;
        double expSum = 0;

    	for(DoubleObjPair p: l) {
    		p.left = SCALE*p.left + OFFSET;
    		expSum += Math.exp(p.left - m);
    	}

        double logExpSum = m + Math.log(expSum);
    	for(DoubleObjPair p: l)
    		p.left -= logExpSum;
    }
    
    static void normalizeSrc(ArrayList<Candidate> l) {
    	double m = Double.MIN_VALUE;
    	for(Candidate p: l)
    		if(p.baseSrcScore > m)
    			m = p.baseSrcScore;
    	
        m = SCALE*m + OFFSET;
        double expSum = 0;

    	for(Candidate p: l) {
    		p.baseSrcScore = SCALE*p.baseSrcScore + OFFSET;
    		expSum += Math.exp(p.baseSrcScore - m);
    	}

        double logExpSum = m + Math.log(expSum);
    	for(Candidate p: l)
    		p.baseSrcScore -= logExpSum;
    }

    static void normalizeSV(SparseVector sv) {
    	/* We assume sorted. */
    	double sum = 0;
    	for(int i = 0; i < sv.index; i++)
    		sum += sv.values[i]*sv.values[i];
    	if(sum < 1e-9)
    		return;
    	double factor = 1.0 / Math.sqrt(sum);
    	for(int i = 0; i < sv.index; i++)
    		sv.values[i] *= factor;
    }
    
    private static int nCandidates;
    
    private static int getNSubj(MPQASentence sen) {
    	int out = 0;
    	for(MPQAExpression expr: sen.exprs)
    		if(!expr.type.equals("os"))
    			out++;
    	return out;
    }
    
    private static void generateRerankerCandidates(LocalLinkClassifier llm,
    		Classifier<SparseVector, Integer> polCl, Classifier<SparseVector, Integer> intCl,
    		SymbolEncoder polintEnc, SubjectivityLexicon sl,
    		MPQAText text, HashMap<String, SentenceData> candMap, int k, int kpol,
    		ArrayList<SentenceData> trainingSet) {

    	if(llm == null && k > 1)
    		throw new IllegalArgumentException("!!!");
    	
    	for(int j = 0; j < text.sentences.size(); j++) {
    		MPQASentence goldSen = text.sentences.get(j);    		
    		if(goldSen.dg.nodes.length > 256)
    		    continue;
    		
    		SentenceData sd = candMap.get(goldSen.dg.toTokenString());
    		if(sd == null) {
                System.err.println("Warning: no candidate list found: |" + goldSen.dg.toTokenString() + "|");
                continue;
    			//throw new RuntimeException("candidate list not found");
    		}

    		if(sd.candidates.get(0).mpqaSentence.dg.nodes.length
    		   != goldSen.dg.nodes.length) {
                System.err.println("Warning: incompatible parse trees!");
                continue;    		    
    		}
    		
    		//System.out.println(goldSen.dg);
            //System.out.println(sd.candidates.get(0).mpqaSentence.dg);
    		
    		/* For EPE: */
    		if(goldSen.dg.nodes[0] instanceof EPEDepNode) {
    		    //System.out.println("isEmpty");
    		    for(Candidate c: sd.candidates)
    		        c.mpqaSentence.replaceDepTree(goldSen.dg);
                //System.out.println(sd.candidates.get(0).mpqaSentence.dg);
                sd.goldSpans.replaceDepTree(goldSen.dg);
    		    
    		    EPEPaths.prepare(goldSen.dg);
    		    
    		} else /* for old code */
    		    goldSen.replaceDepTree(sd.candidates.get(0).mpqaSentence.dg);

    		//System.exit(0);
    		
    		ArrayList<Candidate> newCands;

    		if(polCl != null) {
    			addSubjClues(sl, sd.candidates);
    			newCands = expandCandidateSetWithPolarity(polCl, polintEnc, sd.candidates, goldSen, kpol);
    		} else
    			newCands = sd.candidates;
    		
    		if(llm != null)
    			newCands = expandCandidateSetWithSources(llm, newCands, k, goldSen, false);

    		sd.candidates = newCands;
    		trainingSet.add(sd);
    	}
    }
    
    private static void encodeCandidates(ArrayList<SentenceData> allSentences,
    		boolean extractSourceFeatures, boolean extractPolarityFeatures,
    		double[] costWeights,
    		RerankingFE rfe, 
    		ArrayList<Pair<EncodedCandidateSet, Integer>> trainingSet) {

    	for(SentenceData sd: allSentences) {
    	    
    	    EPEPaths.prepare(sd.goldSpans.dg);
    	    
    		ArrayList<Candidate> newCands = sd.candidates;
    		double[] costs = new double[newCands.size()];
            double[] lls = new double[newCands.size()];
            double minCost = Double.POSITIVE_INFINITY;
            double maxCost = Double.NEGATIVE_INFINITY;

            for(int i = 0; i < newCands.size(); i++) {
                lls[i] = newCands.get(i).baseScore; // TODO + något annat?

                double pf1 = newCands.get(i).stats.getPartialF1();
                double pr = newCands.get(i).stats.getPartialRecall();
                double of1 = newCands.get(i).stats.getOverlapF1();
                double op = newCands.get(i).stats.getOverlapPrecision();

                costs[i] = -costWeights[0]*pf1 - costWeights[1]*of1
                	- costWeights[2]*op - costWeights[3]*pr;
                
                if(costs[i] < minCost)
                    minCost = costs[i];
                if(costs[i] > maxCost)
                    maxCost = costs[i];
            }
            
            for(int i = 0; i < newCands.size(); i++)
                costs[i] -= minCost;

            // TODO borde vi inte normalisera innan expansion?
            Reranker.normalize(lls, newCands.size());

            int oraclePrediction = CollectionUtils.minIndex(costs, costs.length);

            if(minCost == maxCost)
                continue;
            
            EncodedCandidateSet nb = new EncodedCandidateSet();
            nb.costs = costs;
            nb.reps = new SparseVector[newCands.size()];

            for(int i = 0; i < newCands.size(); i++) {
            	Candidate c = newCands.get(i);
            	SparseVector sv = new SparseVector();
            	rfe.extractFeatures(c, lls[i], sv);
            	
            	if(extractSourceFeatures)
            		rfe.extractSourceFeatures(c, sv);

            	if(extractPolarityFeatures)
            		rfe.extractPolarityFeatures(c, sv);
            	
            	sv.sortIndices();
            	
            	if(NORMALIZE_SV)
            		normalizeSV(sv);
            	
            	nb.reps[i] = sv;
            }            
            
            //System.exit(0);
            
            if(RerankingFE.debug) {
            	System.out.println("newCands.size = " + newCands.size());
            	System.out.println("text: |" + sd.text + "|");
            	System.exit(0);
            }
            
            SparseVector op = nb.reps[oraclePrediction];
            for(int i = 0; i < newCands.size(); i++) {
            	SparseVector sv = new SparseVector();
            	SparseVector.sortedSubtract(sv, nb.reps[i], op);
            	sv.trim();
            	nb.reps[i] = sv;
            }

            /*System.out.println(sd.text);
			nb.print();
            System.exit(0);*/
            
            Pair p = new Pair(nb, oraclePrediction);
            trainingSet.add(p);

    		nCandidates += newCands.size();
    	}
    }
    
    private static void generateRerankerTrainingData(LocalLinkClassifier llm,
    		Classifier<SparseVector, Integer> polCl, Classifier<SparseVector, Integer> intCl,
    		SymbolEncoder polintEnc, SubjectivityLexicon sl,
    		MPQAText text, HashMap<String, SentenceData> candMap, int k, int kpol,
    		double[] costWeights,
    		RerankingFE rfe, 
    		ArrayList<Pair<EncodedCandidateSet, Integer>> trainingSet) {

    	if(llm == null && k > 1)
    		throw new IllegalArgumentException("!!!");
    	
    	for(int j = 0; j < text.sentences.size(); j++) {
    		MPQASentence goldSen = text.sentences.get(j);    		
    		
    		SentenceData sd = candMap.get(goldSen.dg.toTokenString());
    		if(sd == null) {
    			System.err.println("Warning: no candidate list found: |" + goldSen.dg.toTokenString() + "|");
    			continue;
    			//throw new RuntimeException("candidate list not found");
    		}
    		
    		goldSen.replaceDepTree(sd.candidates.get(0).mpqaSentence.dg);

    		ArrayList<Candidate> newCands;

    		if(polCl != null) {
    			addSubjClues(sl, sd.candidates);
    			newCands = expandCandidateSetWithPolarity(polCl, polintEnc, sd.candidates, goldSen, kpol);
    		} else
    			newCands = sd.candidates;
    		
    		if(llm != null)
    			newCands = expandCandidateSetWithSources(llm, newCands, k, goldSen, false);
    		
    		double[] costs = new double[newCands.size()];
            double[] lls = new double[newCands.size()];
            double minCost = Double.POSITIVE_INFINITY;
            double maxCost = Double.NEGATIVE_INFINITY;

            for(int i = 0; i < newCands.size(); i++) {
                lls[i] = newCands.get(i).baseScore; // TODO + något annat?

                double pf1 = newCands.get(i).stats.getPartialF1();
                double pr = newCands.get(i).stats.getPartialRecall();
                double of1 = newCands.get(i).stats.getOverlapF1();
                double op = newCands.get(i).stats.getOverlapPrecision();

                costs[i] = 1.0 - pf1;
                
                if(costs[i] < minCost)
                    minCost = costs[i];
                if(costs[i] > maxCost)
                    maxCost = costs[i];
            }
            
            for(int i = 0; i < newCands.size(); i++)
                costs[i] -= minCost;

            // TODO borde vi inte normalisera innan expansion?
            Reranker.normalize(lls, newCands.size());

            int oraclePrediction = CollectionUtils.minIndex(costs, costs.length);

            if(minCost == maxCost)
                continue;
            
            EncodedCandidateSet nb = new EncodedCandidateSet();
            nb.costs = costs;
            nb.reps = new SparseVector[newCands.size()];

            for(int i = 0; i < newCands.size(); i++) {
            	Candidate c = newCands.get(i);
            	SparseVector sv = new SparseVector();
            	rfe.extractFeatures(c, lls[i], sv);
            	
            	if(llm != null && USE_SRC_FEATURES)
            		rfe.extractSourceFeatures(c, sv);

            	if(USE_POL_FEATURES && (polCl != null))
            		rfe.extractPolarityFeatures(c, sv);
            	
            	sv.sortIndices();
            	
            	if(NORMALIZE_SV)
            		normalizeSV(sv);
            	
            	nb.reps[i] = sv;
            }            
            
            if(RerankingFE.debug) {
            	System.out.println("newCands.size = " + newCands.size());
            	System.out.println("text: |" + sd.text + "|");
            	System.exit(0);
            }
            
            SparseVector op = nb.reps[oraclePrediction];
            for(int i = 0; i < newCands.size(); i++) {
            	SparseVector sv = new SparseVector();
            	SparseVector.sortedSubtract(sv, nb.reps[i], op);
            	sv.trim();
            	nb.reps[i] = sv;
            }

            /*System.out.println(sd.text);
			nb.print();
            System.exit(0);*/
            
            Pair p = new Pair(nb, oraclePrediction);
            trainingSet.add(p);

    		nCandidates += newCands.size();
    	}
    }
    
    private static void addSubjClues(SubjectivityLexicon sl, ArrayList<Candidate> candidates) {
    	MPQASentence s0 = candidates.get(0).mpqaSentence;
    	if(s0.slClues != null)
    		return;
    	ArrayList<String> slClues = new ArrayList();
    	slClues.add(null);
    	for(int i = 1; i < s0.dg.nodes.length; i++) {
			DepNode n = s0.dg.nodes[i];
			slClues.add(sl.lookup(n.word, n.pos, n.lemma));
		}
    	for(Candidate cand: candidates)
    		cand.mpqaSentence.slClues = slClues;
    }
    
	private static void addPolInt(Classifier<SparseVector, Integer> polCl,
			Classifier<SparseVector, Integer> intCl,
			ArrayList<Candidate> candidates) {

		boolean warned = false;
		
		for(Candidate cand: candidates) {
			MPQASentence sen = cand.mpqaSentence;
			for(MPQAExpression expr: sen.exprs) {
				if(expr.type.equals("os"))
					continue;
				if(expr.polarity != null) {
					//throw new IllegalArgumentException("already added");
					if(!warned) {
						System.err.println("Warning: already added polarity: |" + sen.dg.toTokenString() + "|");
						warned = true;
					}
				}
				SparseVector sv = BaselinePolarityClassifier.represent(expr, null);

				int polId = polCl.classify(sv);				
				expr.polarity = BaselinePolarityClassifier.decodePol(polId);
				if(expr.polarity == null)
					throw new RuntimeException("No polarity assigned! polId = " + polId);

				int intId = intCl.classify(sv);
				expr.intensity = BaselinePolarityClassifier.decodeInt(intId);
				if(expr.intensity == null)
					throw new RuntimeException("No intensity assigned! intId = " + intId);
			}
		}
		
	}

	/*
	
		private static ArrayList<Candidate> expandCandidateSetWithPolarity(Classifier<SparseVector, Integer> polCl,
			ArrayList<Candidate> originalSet,
			MPQASentence goldSen) {
		if(polCl == null)
			throw new IllegalArgumentException("polCl shouldn't be null here");

		ArrayList<Candidate> out = new ArrayList();
		
		for(Candidate cand: originalSet) {
			if(cand.mpqaSentence.exprs.isEmpty()) {				
				Candidate clone = cand.clone();
				if(goldSen != null)
					evalCandPol(clone, goldSen, POL_WEIGHT);
				out.add(clone);
				continue;
			}

			// Förenklat -- enbart ett förslag till att börja med
			Candidate clone = cand.clone();

			boolean warned = false;
			MPQASentence sen = clone.mpqaSentence;
			for(MPQAExpression expr: sen.exprs) {
				if(expr.type.equals("os"))
					continue;
				if(expr.polarity != null) {
					if(!warned) {
						System.err.println("Warning: already added polarity: |" + sen.dg.toTokenString() + "|");
						warned = true;
					}
				}
				SparseVector sv = BaselinePolarityClassifier.represent(expr, null);

				int polId = polCl.classify(sv);				
				expr.polarity = BaselinePolarityClassifier.decodePol(polId);
				if(expr.polarity == null)
					throw new RuntimeException("No polarity assigned! polId = " + polId);
			}

			if(goldSen != null)
				evalCandPol(clone, goldSen, POL_WEIGHT);

			out.add(clone);
		}
		
		return out;
	}
	
	*/
	
	private static ArrayList<Candidate> expandCandidateSetWithPolarity(Classifier<SparseVector, Integer> polCl,
			SymbolEncoder polintEnc, ArrayList<Candidate> originalSet,
			MPQASentence goldSen, int kpol) {
		if(polCl == null)
			throw new IllegalArgumentException("polCl shouldn't be null here");

		ArrayList<Candidate> out = new ArrayList();
		
		ArrayList<Integer> polRange = new ArrayList(); 
		polRange.add(BaselinePolarityClassifier.POSITIVE);
		polRange.add(BaselinePolarityClassifier.NEUTRAL);
		polRange.add(BaselinePolarityClassifier.NEGATIVE);
		double[] polScores = new double[polRange.size()];

		boolean debug = false;
		
		if(false) {
			int nop1 = 0;
			for(MPQAExpression e: goldSen.exprs)
				if(!e.type.equals("os") && e.polarity != null && e.polarity.equals("negative"))
					nop1++;
			int nop2 = 0;
			Candidate c0 = originalSet.get(0);
			for(MPQAExpression e: c0.mpqaSentence.exprs)
				if(!e.type.equals("os"))
					nop2++;
			if(nop1 > 1 && nop2 > 1)
				debug = true;
		}

		if(debug) {
			System.out.println("Gold labeling: " + goldSen.exprs);
		}
		
		for(Candidate cand: originalSet) {
			if(cand.mpqaSentence.exprs.isEmpty()) {
				Candidate clone = cand.clone();
				if(goldSen != null)
					evalCandPol(clone, goldSen, POL_WEIGHT);
				out.add(clone);
				continue;
			}
			
			MPQASentence sen = cand.mpqaSentence;
			ArrayList<ArrayList<DoubleObjPair<String>>> table = new ArrayList();
			
			for(MPQAExpression expr: sen.exprs) {
				if(expr.type.equals("os"))
					continue;

				SparseVector sv = BaselinePolarityClassifier.represent(expr, polintEnc);

				polCl.computeScoresRestricted(sv, polRange, polScores);
				CollectionUtils.normalizeToLogProbs(polScores);
				
				ArrayList<DoubleObjPair<String>> polCandList = new ArrayList();
				for(int i = 0; i < polScores.length; i++)
					polCandList.add(new DoubleObjPair(polScores[i], 
							BaselinePolarityClassifier.decodePol(polRange.get(i))));
				Collections.sort(polCandList, NEG_LEFT);

				if(debug)
					System.out.println(polCandList);
				
				table.add(polCandList);
			}

			ArrayList<ArrayList<DoubleObjPair<String>>> kbest
				= BestCombinations.findBestCombinations(table, kpol);

			int dbgix = 0;
			for(ArrayList<DoubleObjPair<String>> labeling: kbest) {
				
				if(debug) {
					dbgix++;
					double sum = 0;
					for(DoubleObjPair<String> p: labeling)
						sum += p.left;
					//System.out.println(dbgix + ": (" + sum + ") " + labeling);
				}
				
				Candidate clone = cand.clone();
				
				int index = 0;
				for(MPQAExpression expr: clone.mpqaSentence.exprs) {
					if(expr.type.equals("os"))
						continue;
					DoubleObjPair<String> p = labeling.get(index);
					clone.basePolScore += p.left;
					expr.polarity = p.right;
					index++;
				}
				
				if(goldSen != null)
					evalCandPol(clone, goldSen, POL_WEIGHT);				

				if(debug) {
					PRFStats pp = goldSen.comparePolInt(clone.mpqaSentence, true);
					System.out.println(clone.mpqaSentence.exprs + ": bs=" + clone.basePolScore + ", f1=" + pp.getPartialF1());
				}
				
				out.add(clone);
			}

			/* Normalize once again... */
			
			/*double[] totalProbs = new double[out.size()];
			for(int i = 0; i < out.size(); i++)
				totalProbs[i] = out.get(i).basePolScore;
			CollectionUtils.normalizeToLogProbs(totalProbs);
			for(int i = 0; i < out.size(); i++)
				out.get(i).basePolScore = totalProbs[i];*/
			/*double p0 = out.get(0).basePolScore;
			for(int i = 0; i < out.size(); i++)
				out.get(i).basePolScore -= p0;*/
			
		}

		if(debug)
			System.exit(0);
		
		//if(true)
		//	Collections.shuffle(out, debugTestRand);
		
		return out;
	}
	
	private static ArrayList<Candidate> expandCandidateSetWithSources(LocalLinkClassifier llm,
			ArrayList<Candidate> originalSet,
			int k,
			MPQASentence goldSen,
			boolean expandSrcSpans) {

		if(llm == null)
			throw new IllegalArgumentException("llm shouldn't be null here");
		
        //System.out.println("sources");
		
		ArrayList<Candidate> out = new ArrayList();
		for(Candidate cand: originalSet) {

			if(cand.mpqaSentence.exprs.isEmpty()) {				
				Candidate clone = cand.clone();
				if(goldSen != null)
					evalCandSrc(clone, goldSen, SRC_WEIGHT);
				out.add(clone);
				continue;
			}

			ArrayList<Candidate> tmp = new ArrayList();
			
			//MPQASentence sen = toMPQASentence(cand, goldSen);
			
			//ArrayList<DepNode> cands = LocalLinkClassifier.findCandidates(sen);

            //System.out.println("1");
			
			ArrayList<DepNode> cands = LocalLinkClassifier.findCandidates(cand.mpqaSentence);

			//System.out.println("2");
			
			ArrayList<ArrayList<DoubleObjPair<MPQAAgent>>> table = new ArrayList();
			for(MPQAExpression expr: cand.mpqaSentence.exprs) {
	            long l0 = System.currentTimeMillis();
				ArrayList<DoubleObjPair<MPQAAgent>> l = allSources(llm, expr, cands, true);
	            long l1 = System.currentTimeMillis();
	            STEP2_TIME += l1 - l0;
	            normalize(l);
				table.add(l);
			}
			
            //System.out.println(STEP2_TIME + " / " + FEATURE_TIME + " / " + llm.localNodeExtractor.PATH_TIME);
			
			//System.out.println("3");
			
			ArrayList<ArrayList<DoubleObjPair<MPQAAgent>>> kbest
			= BestCombinations.findBestCombinations(table, k);

			//System.out.println("4");
			
			int nExpr = cand.mpqaSentence.exprs.size();
			for(ArrayList<DoubleObjPair<MPQAAgent>> srcSequence: kbest) {

				Candidate clone = cand.clone();
				clone.isImplicit = new boolean[nExpr];
				clone.isWriter = new boolean[nExpr];
				clone.srcIndex = new int[nExpr];
				
				for(int i = 0; i < nExpr; i++) {
					clone.baseSrcScore += srcSequence.get(i).left;
					MPQAAgent src = srcSequence.get(i).right;
					if(src == null)
						continue;
					if(src.isWriter)
						clone.isWriter[i] = true;
					else if(src.isImplicit)
						clone.isImplicit[i] = true;
					else if(src != null) {
						DepNode srcHead = src.instances.get(0).span.get(0);
						clone.srcIndex[i] = srcHead.position;
						
						if(expandSrcSpans) {
							MPQAExpression e = clone.mpqaSentence.exprs.get(i);
							setAgentInstanceSpan(srcHead, e, src.instances.get(0));
						}
					}
					
					ArrayList<MPQAAgent> ns = new ArrayList();
					ns.add(src);
					clone.mpqaSentence.exprs.get(i).nestedSource = ns;
				}
				
				//addCandAgentsToMPQASentence(clone);
				if(goldSen != null)
					evalCandSrc(clone, goldSen, SRC_WEIGHT);
				tmp.add(clone);				
			}

			//System.out.println("5");
			
			normalizeSrc(tmp);
			
			//System.out.println("6");
			
			out.addAll(tmp);			
		}
		
        //System.out.println("end sources")    ;
		
		return out;		
	}

	private static void setAgentInstanceSpan(DepNode srcHead, MPQAExpression expr,
            MPQAAgentInstance inst) {
	    inst.span = createSourceSpan(srcHead, expr.span, false);
	    if(inst.span.get(0) instanceof EPEDepNode) {
            inst.start = ((EPEDepNode) inst.span.get(0)).start;
            inst.end = ((EPEDepNode) inst.span.get(inst.span.size()-1)).end;
	    }
	}

    private static void evalCandSrc(Candidate cd, MPQASentence goldSen,
			double srcCostWeight) {
		//MPQASentence ms = toMPQASentence(cd, goldSen);	
		PRFStats res = goldSen.compareLinks(cd.mpqaSentence);

//		cd.stats.nCorrect += srcCostWeight * res.nCorrect;
//		cd.stats.nGuesses += srcCostWeight * res.nGuesses;
//		cd.stats.nInGold += srcCostWeight * res.nInGold;
//		cd.stats.propFound += srcCostWeight * res.propFound;
//		cd.stats.propGuessed += srcCostWeight * res.propGuessed;
//		cd.stats.nOverlap += srcCostWeight * res.nOverlap;

		cd.stats.nCorrect += srcCostWeight * res.nOverlap; // pga att vi bara har ett ord
		cd.stats.nGuesses += srcCostWeight * res.nGuesses;
		cd.stats.nInGold += srcCostWeight * res.nInGold;
		cd.stats.propFound += srcCostWeight * res.nOverlap;
		cd.stats.propGuessed += srcCostWeight * res.nOverlap;
		cd.stats.nOverlap += srcCostWeight * res.nOverlap;
	
	}
	
	private static final boolean EVAL_POL_OVERLAP = false;
	
	private static void evalCandPol(Candidate cd, MPQASentence goldSen,
			double polCostWeight) {
		PRFStats res = goldSen.comparePolInt(cd.mpqaSentence, true);

		if(EVAL_POL_OVERLAP) {		
			cd.stats.nCorrect += polCostWeight * res.nOverlap; // res.nCorrect 
			cd.stats.nGuesses += polCostWeight * res.nGuesses;
			cd.stats.nInGold += polCostWeight * res.nInGold;
			cd.stats.propFound += polCostWeight * res.nOverlap; // res.propFound
			cd.stats.propGuessed += polCostWeight * res.nOverlap; // res.propGuessed;
			cd.stats.nOverlap += polCostWeight * res.nOverlap;
		} else {
			cd.stats.nCorrect += polCostWeight * res.nCorrect;
			cd.stats.nGuesses += polCostWeight * res.nGuesses;
			cd.stats.nInGold += polCostWeight * res.nInGold;
			cd.stats.propFound += polCostWeight * res.propFound;
			cd.stats.propGuessed += polCostWeight * res.propGuessed;
			cd.stats.nOverlap += polCostWeight * res.nOverlap;			
		}
	
	}
	
	private static void kbestSourcesOracle(LocalLinkClassifier llm,
			MPQASentence sen,
			int k,
			MPQASentence goldSen) {
		if(sen.exprs.isEmpty())
			return;

		ArrayList<DepNode> cands = LocalLinkClassifier.findCandidates(sen);
		
		ArrayList<ArrayList<DoubleObjPair<MPQAAgent>>> table = new ArrayList();
		for(MPQAExpression expr: sen.exprs) {
			ArrayList<DoubleObjPair<MPQAAgent>> cand = allSources(llm, expr, cands, false);
			normalize(cand);
			table.add(cand);
		}
		ArrayList<ArrayList<DoubleObjPair<MPQAAgent>>> kbest
		= BestCombinations.findBestCombinations(table, k);

		double bestResult = Double.NEGATIVE_INFINITY;
		ArrayList<DoubleObjPair<MPQAAgent>> bestCand = null;
		
		for(ArrayList<DoubleObjPair<MPQAAgent>> candidate: kbest) {
			for(int i = 0; i < sen.exprs.size(); i++) {
				MPQAExpression expr = sen.exprs.get(i);
				expr.nestedSource = new ArrayList();
				expr.nestedSource.add(candidate.get(i).right);
			}
			PRFStats result = goldSen.compareLinks(sen);
			double pf = result.getPartialF1();
			if(pf > bestResult) {
				bestResult = pf;
				bestCand = candidate;
			}
		}

		for(int i = 0; i < sen.exprs.size(); i++) {
			MPQAExpression expr = sen.exprs.get(i);
			expr.nestedSource = new ArrayList();
			expr.nestedSource.add(bestCand.get(i).right);
		}
				
	}
	
    private static long FEATURE_TIME;
    private static long STEP2_TIME;
	
	private static ArrayList<DoubleObjPair<MPQAAgent>> allSources(LocalLinkClassifier llm, 
			MPQAExpression expr, 
			ArrayList<DepNode> cands,
			boolean onlySrcHead) {

		ArrayList<DoubleObjPair<MPQAAgent>> out = new ArrayList();		
		
		FeatureList localFeatures = new FeatureList(llm.localDecl);
		Pair<MPQAExpression, MPQASentence> p = new Pair(expr, expr.containingSentence);
		
        long l0 = System.currentTimeMillis();
		llm.localExprExtractor.extractFeatures(p, localFeatures);	
		long l1 = System.currentTimeMillis();
        
		Triple<MPQAExpression, DepNode, MPQASentence> tr = new Triple(expr, null, expr.containingSentence);
		double[] localScores = new double[2];
		for(DepNode cand: cands) {
			tr.second = cand;

			l0 = System.currentTimeMillis();
			llm.localNodeExtractor.extractFeatures(tr, localFeatures);
			l1 = System.currentTimeMillis();
	        FEATURE_TIME += l1-l0;
			
			llm.localCl.computeScoresRestricted(localFeatures, LOCAL_CLASSES, localScores);

			MPQAAgent ag = new MPQAAgent();
			MPQAAgentInstance inst = new MPQAAgentInstance();
			inst.containingSentence = expr.containingSentence;
			inst.agent = ag;
			inst.agent.instances.add(inst);
			
			inst.span = createSourceSpan(cand, expr.span, onlySrcHead);			
			if(inst.span.get(0) instanceof EPEDepNode) {
                inst.start = ((EPEDepNode) inst.span.get(0)).start;
                inst.end = ((EPEDepNode) inst.span.get(inst.span.size()-1)).end;
			}
			
			out.add(new DoubleObjPair(localScores[0], ag));
		}
		
		FeatureList nonlocalFeatures = new FeatureList(llm.nonlocalDecl);
		llm.nonlocalExtractor.extractFeatures(p, nonlocalFeatures);
		double[] nonlocalScores = new double[4];		
		llm.nonlocalCl.computeScoresRestricted(nonlocalFeatures, NONLOCAL_CLASSES, nonlocalScores);

		// writer
		MPQAAgent agW = new MPQAAgent();
		agW.isWriter = true;
		out.add(new DoubleObjPair(nonlocalScores[1], agW));

		// implicit
		MPQAAgent agI = new MPQAAgent();
		agI.isImplicit = true;
		out.add(new DoubleObjPair(nonlocalScores[2], agI));

		// external -- just add null
		out.add(new DoubleObjPair(nonlocalScores[3], null));

		Collections.sort(out, NEG_LEFT);		
		return out;
	}
	
	public static MPQAAgent bestSource(LocalLinkClassifier llm, 
			MPQAExpression expr,
			ArrayList<DepNode> cands,
			boolean onlySrcHead) {

		FeatureList localFeatures = new FeatureList(llm.localDecl);
		Pair<MPQAExpression, MPQASentence> p = new Pair(expr, expr.containingSentence);
		llm.localExprExtractor.extractFeatures(p, localFeatures);
		Triple<MPQAExpression, DepNode, MPQASentence> tr = new Triple(expr, null, expr.containingSentence);

		MPQAAgent bestAgent = null;
		double bestScore = Double.NEGATIVE_INFINITY;
		
		double[] localScores = new double[2];
		for(DepNode cand: cands) {
			tr.second = cand;
			llm.localNodeExtractor.extractFeatures(tr, localFeatures);
			llm.localCl.computeScoresRestricted(localFeatures, LOCAL_CLASSES, localScores);

			if(localScores[0] > bestScore) {
				bestScore = localScores[0];

				MPQAAgent ag = new MPQAAgent();
				MPQAAgentInstance inst = new MPQAAgentInstance();
				inst.containingSentence = expr.containingSentence;
				inst.agent = ag;
				inst.agent.instances.add(inst);				
				inst.span = createSourceSpan(cand, expr.span, onlySrcHead);
				
				bestAgent = ag;
			}
			
		}
		
		FeatureList nonlocalFeatures = new FeatureList(llm.nonlocalDecl);
		llm.nonlocalExtractor.extractFeatures(p, nonlocalFeatures);
		double[] nonlocalScores = new double[4];		
		llm.nonlocalCl.computeScoresRestricted(nonlocalFeatures, NONLOCAL_CLASSES, nonlocalScores);

		if(nonlocalScores[1] > bestScore) {
			bestScore = nonlocalScores[1];		
			// writer
			MPQAAgent agW = new MPQAAgent();
			agW.isWriter = true;
			bestAgent = agW;
		}
		
		if(nonlocalScores[2] > bestScore) {
			bestScore = nonlocalScores[2];		
			// implicit
			MPQAAgent agI = new MPQAAgent();
			agI.isImplicit = true;
			bestAgent = agI;
		}
		
		if(nonlocalScores[3] > bestScore) {
			bestScore = nonlocalScores[3];		
			// external -- just return null
			bestAgent = null;
		}

		return bestAgent;
	}
	
	
	
}
