package mpqa_structlearn.fullsystem;

import java.io.*;
import java.net.*;
import java.util.*;

import mpqa_structlearn.io.*;
import mpqa_structlearn.locallink.*;
import mpqa_structlearn.polarity.BaselinePolarityClassifier;
import mpqa_structlearn.sequences.*;
import mpqareader.BOWSubjSentClassifier;
import mpqareader.SubjectivityLexicon;

import se.lth.cs.nlp.nlputils.annotations.*;
import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.*;

import se.lth.cs.nlp.nlputils.ml_long.*;
import se.lth.cs.nlp.opinions.*;

import se.lth.cs.nlp.depsrl.format.*;

public class Server {

	public static void main(String[] argv) {
		server(argv);
	}
	
	private static void server(String[] argv) {		
		String subjLexFile = argv[1];	
		int Nseq_arg = Integer.parseInt(argv[2]);
		int Nsrcseq_arg = Integer.parseInt(argv[3]);
		String taggingModelFile = argv[4];
		String linkModelFile = argv[5];
		String rerankerModelFile = argv[6];

		String subjSenModelFile = argv[7];
		String polIntModelFile = argv[8];
		
		String parserCmd = argv[9];

		String forwardHost = argv[10];
		int port = Integer.parseInt(argv[11]);
		
		try {
			/* Loading trained models. */

			System.out.println("Loading sequence model...");
			final TaggingModel baseSeqModel = SeqLabeler.loadModel(taggingModelFile);
			
			System.out.println("Loading link model...");
			final LocalLinkClassifier llc = LocalLinkClassifier.loadModel(linkModelFile);

			System.out.println("Loading reranker model...");
            ObjectInputStream ois = new ObjectInputStream(Util.openFileStream(rerankerModelFile));
            final RerankingFE rfe = (RerankingFE) ois.readObject();
            final Classifier<EncodedCandidateSet, Integer> rcl = (Classifier) ois.readObject();
            
            int Nseq_saved = ois.readInt();
            int Nsrcseq_saved = ois.readInt();
            final int Nseq, Nsrcseq;
            if(Nseq_arg < 1) {
            	System.out.println("Setting Nseq to saved: " + Nseq_saved);
            	Nseq = Nseq_saved;
            } else
            	Nseq = Nseq_arg;
            if(Nsrcseq_arg < 1) {
            	System.out.println("Setting Nsrcseq to saved: " + Nsrcseq_saved);
            	Nsrcseq = Nsrcseq_saved;
            } else
            	Nsrcseq = Nsrcseq_arg;
            ois.close();

        	System.out.println("Loading subjectivity lexicon...");
        	final SubjectivityLexicon subjLex = new SubjectivityLexicon(subjLexFile);
            
            System.out.println("Loading sentence classifier...");
            ois = new ObjectInputStream(Util.openFileStream(subjSenModelFile));
			final SymbolEncoder senEnc = (SymbolEncoder) ois.readObject();
			final Classifier<SparseVector, Integer> senCl = (Classifier) ois.readObject();
            ois.close();
			
            System.out.println("Loading polarity and intensity classifiers...");
            ois = new ObjectInputStream(Util.openFileStream(polIntModelFile));
			final Classifier<SparseVector, Integer> polCl = (Classifier) ois.readObject();
			final Classifier<SparseVector, Integer> intCl = (Classifier) ois.readObject();
            ois.close();
            
            System.out.println("Starting parser...");
            Process process = Runtime.getRuntime().exec(parserCmd);
            final PrintWriter toParser = new PrintWriter(process.getOutputStream());
            final Scanner fromParser = new Scanner(process.getInputStream());

            toParser.println("I am testing .");
            toParser.flush();
            Triple<DepGraph, DepGraph, List<PAStructure>> parse = CoNLL2008Format.readNextGraph(fromParser);
            if(parse == null)
            	throw new RuntimeException("could not read parse!");
            
            System.out.println("Done.");
            
            if(!forwardHost.equals("LOCAL")) {
            	Socket socket = new Socket(forwardHost, port);
            	System.out.println("Forwarding to " + forwardHost + ":" + port + ".");
                handleSocket(socket, fromParser, toParser, subjLex,
                		senCl, senEnc, baseSeqModel, Nseq, polCl, intCl,
                		llc, Nsrcseq, rfe, rcl);
            } else {           
            	ServerSocket ss = new ServerSocket(port);
            	System.out.println("Listening on port " + port + ".");
                while(true) {
                	final Socket socket = ss.accept();
                    System.out.println("Incoming connection from " + socket.getInetAddress());                    
                    new Thread() {
                    	public void run() {
                    		try {
                    			handleSocket(socket, fromParser, toParser, subjLex,
                    					senCl, senEnc, baseSeqModel, Nseq, polCl, intCl,
                    					llc, Nsrcseq, rfe, rcl);
                    		} catch(IOException e) {
                    			e.printStackTrace();
                    		}
                    	}
                    }.start();
                }
            }
            
		} catch(Exception e) {
			e.printStackTrace();
		}
	}

	private static void handleSocket(Socket socket,
			Scanner fromParser, PrintWriter toParser,
			SubjectivityLexicon subjLex, 
			Classifier<SparseVector, Integer> senCl, SymbolEncoder senEnc,
			TaggingModel baseSeqModel, int Nseq,
			Classifier<SparseVector, Integer> polCl, 
			Classifier<SparseVector, Integer> intCl,
			LocalLinkClassifier llc, int Nsrcseq,
			RerankingFE rfe, Classifier<EncodedCandidateSet, Integer> rcl
			) throws IOException {
		BufferedReader fromSocket = new BufferedReader(new InputStreamReader(socket.getInputStream()));
        PrintWriter toSocket = new PrintWriter(new OutputStreamWriter(socket.getOutputStream()));
        
        String tokenizedLine = fromSocket.readLine();
        while(tokenizedLine != null) {
        	System.out.println("Read line: |" + tokenizedLine + "|");
        	if(tokenizedLine.isEmpty()) {
        		toSocket.println();
        		toSocket.flush();
        	} else {
        		Candidate guess;
        		boolean isSubj;
        		synchronized(Server.class) {
        			toParser.println(tokenizedLine);                
        			toParser.flush();
        			Triple<DepGraph, DepGraph, List<PAStructure>> parse = CoNLL2008Format.readNextGraph(fromParser);
        			if(parse == null)
        				throw new RuntimeException("parse == null");        		
        			isSubj = BOWSubjSentClassifier.classifySentence(subjLex, parse, senCl, senEnc);

        			SentenceData sd = kbest(baseSeqModel, subjLex, parse.first, (ArrayList) parse.third, Nseq);
        			guess 
        			= MPQASentenceReranker.rerankSentenceUnscored(sd, llc, 
        					Nsrcseq, rfe, rcl);

        			addPolInt(guess.mpqaSentence, polCl, intCl, subjLex);
        		}
        		toSocket.println((isSubj? "T ": "F ") + senToString(guess.mpqaSentence));
        		toSocket.flush();
        	}
        	tokenizedLine = fromSocket.readLine();
        }   
        try {
        	fromSocket.close();
        } catch(Exception e) {}
        try {
        	toSocket.close();
        } catch(Exception e) {}
        System.out.println("Closed connection.");
	}
	
	public static SentenceData kbest(TaggingModel model, 
			SubjectivityLexicon sl, DepGraph dg, ArrayList<PAStructure> pas, int k) {
		int n = dg.nodes.length - 1;
		int m = 6;		
		String[][] tokens = new String[n][m];
		for(int i = 0; i < n; i++) {
			DepNode dn = dg.nodes[i+1];
			tokens[i][0] = dn.word;
			tokens[i][1] = dn.pos;
			tokens[i][2] = dn.lemma;
			String ss = sl.lookup(dn.word, dn.pos, dn.lemma);
			if(ss == null)
				tokens[i][3] = "_";
			else
				tokens[i][3] = ss;			
			tokens[i][4] = "_";
			tokens[i][5] = "_";
		}		
		/*for(int i = 0; i < n; i++) {
			for(int j = 0; j < tokens[i].length; j++) {
				if(j > 0)
					System.out.print("\t");
				System.out.print(tokens[i][j]);
			}
			System.out.println();
		}
		System.out.println();*/	
		ArrayList<DoubleObjPair<AnnotationLayer>> out = 
			SeqLabeler.tagSentenceKBest(model, tokens, k);		
		//System.out.println(out);
		return convertSentence(out, dg, pas);
	}
	
	private static SentenceData convertSentence(ArrayList<DoubleObjPair<AnnotationLayer>> cands,
			DepGraph dg, ArrayList<PAStructure> pas) {
		SentenceData out = new SentenceData();		
		for(DoubleObjPair<AnnotationLayer> p: cands) {
			Candidate cand = new Candidate();
			cand.baseScore = p.left;
			MPQASentence sen = new MPQASentence();
			sen.dg = dg;
			sen.pas = pas;
			for(Span s: (AnnotationLayer<Span>) p.right) {
				MPQAExpression expr = new MPQAExpression();
				expr.type = s.label;
				expr.span = new ArrayList();
				for(int i = s.tokenStart; i < s.tokenEnd; i++) 
					expr.span.add(dg.nodes[i+1]);
				expr.containingSentence = sen;
				sen.exprs.add(expr);
			}
			cand.mpqaSentence = sen;
			out.candidates.add(cand);
		}
		return out;
	}

	public static void addPolInt(MPQASentence sen,
			Classifier<SparseVector, Integer> polCl,
			Classifier<SparseVector, Integer> intCl,
			SubjectivityLexicon sl) {
		if(sen.slClues == null)
			addSlClues(sen, sl);
		for(MPQAExpression expr: sen.exprs) {
			if(expr.type.equals("os"))
				continue;
			SparseVector sv = BaselinePolarityClassifier.represent(expr, null);
			if(polCl != null)
				expr.polarity = BaselinePolarityClassifier.decodePol(polCl.classify(sv));
			if(intCl != null)
				expr.intensity = BaselinePolarityClassifier.decodeInt(intCl.classify(sv));
		}
	}

	private static void addSlClues(MPQASentence sen, SubjectivityLexicon lex) {
		if(sen.slClues != null)
			throw new IllegalArgumentException("already added");
		if(sen.dg == null)
			throw new IllegalArgumentException("dg == null");
		sen.slClues = new ArrayList();
		sen.slClues.add(null);
		for(int i = 1; i < sen.dg.nodes.length; i++) {
			DepNode n = sen.dg.nodes[i];
			sen.slClues.add(lex.lookup(n.word, n.pos, n.lemma));
		}
	}
	
	private static String senToString(MPQASentence sen) {
		StringBuilder sb = new StringBuilder();
		boolean first = true;
		for(MPQAExpression expr: sen.exprs) {
			if(!first)
				sb.append(", ");
			else
				first = false;
			sb.append(exprToString(expr));
		}		
		return sb.toString();
	}
	
	private static String exprToString(MPQAExpression expr) {
		StringBuilder sb = new StringBuilder();
		
		sb.append("[" + expr.type + " ");	
		sb.append(expr.span.get(0).position + "," + (expr.span.get(expr.span.size()-1).position));
		sb.append(" ");

		if(expr.nestedSource != null) {
			MPQAAgent ag;
			ag = expr.nestedSource.get(expr.nestedSource.size() - 1);
			if(ag == null)
				sb.append("n");
			else if(ag.isImplicit)
				sb.append("i");
			else if(ag.isWriter)
				sb.append("w");
			else {
				if(ag.instances.size() != 1)
					throw new IllegalArgumentException("size != 1");					
				MPQAAgentInstance ai = ag.instances.get(0);
				sb.append(ai.span.get(0).position + "," + (ai.span.get(ai.span.size()-1).position));
			}			
		} else {
			sb.append("n");
		}
		
		sb.append(" " + expr.polarity + " " + expr.intensity + "]");
		
		return sb.toString();
	}
	
}
