package mpqa_structlearn.fullsystem;

import se.lth.cs.nlp.nlputils.core.*;
import java.io.*;
import java.util.ArrayList;

public class TKEvaluator {

	public static void main(String[] argv) {
		try {
			if(argv.length == 0) {
				System.err.println("Allowed modes: -eval -bl -oracle -sig1 -sig2");
				System.exit(1);
			}
				
			if(argv[0].equals("-eval"))
				eval(argv);
			else if(argv[0].equals("-sig1"))
				sig1(argv);
			else if(argv[0].equals("-sig2"))
				sig2(argv);
			else if(argv[0].equals("-bl"))
				bl(argv);
			else if(argv[0].equals("-oracle"))
				oracle(argv);
			else {
				System.err.println("Allowed modes: -eval -bl -oracle -sig1 -sig2");
				System.exit(1);
			}
					
		} catch(Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
	}

	private static void sig1(String[] argv) {
		if(argv.length != 4) {
			System.err.println("three arguments necessary: svmlight output file, score file, attribute (ep|er|ef|pp|pr|pf|op|or|of)");
			System.exit(1);	
		}
		throw new UnsupportedOperationException("unimplemented");
	}
	
	private static void sig2(String[] argv) {
		if(argv.length != 5) {
			System.err.println("four arguments necessary: svmlight output file, score file, attribute (ep|er|ef|pp|pr|pf|op|or|of), and test type (perm|res)");
			System.exit(1);
		}
		throw new UnsupportedOperationException("unimplemented");
	}

	private static void eval(String[] argv) throws IOException {
		if(argv.length != 3) {
			System.err.println("two arguments necessary: svmlight output file and score file");
			System.exit(1);	
		}
		BufferedReader tkOutputBR = Util.openFileReader(argv[1]);
		BufferedReader scoreBR = Util.openFileReader(argv[2]);

		PRFStats total = new PRFStats();
		
		String sline = scoreBR.readLine();
		if(!sline.startsWith("#"))
			throw new RuntimeException("Illegal format of the score file: |" + sline + "|");
		while(sline != null) {
			ArrayList<PRFStats> prfs = new ArrayList();
			sline = scoreBR.readLine();
			while(sline != null && !sline.startsWith("#")) {
				prfs.add(new PRFStats(sline.split(" ")));
				sline = scoreBR.readLine();
			}
			
			PRFStats guess = null;
			double max = Double.NEGATIVE_INFINITY;			
			for(int i = 0; i < prfs.size(); i++) {
				double tscore = Double.parseDouble(tkOutputBR.readLine());
				if(tscore > max) {
					max = tscore;
					guess = prfs.get(i);
				}
			}
			total.add(guess);
		}

		System.out.println("Evaluation scores:");
		total.print();		
	}
	
	private static void bl(String[] argv) throws IOException {
		if(argv.length != 2) {
			System.err.println("one argument necessary: score file");
			System.exit(1);	
		}
		BufferedReader scoreBR = Util.openFileReader(argv[1]);

		PRFStats total = new PRFStats();
		
		String sline = scoreBR.readLine();
		if(!sline.startsWith("#"))
			throw new RuntimeException("Illegal format of the score file: |" + sline + "|");
		while(sline != null) {
			ArrayList<PRFStats> prfs = new ArrayList();
			sline = scoreBR.readLine();
			while(sline != null && !sline.startsWith("#")) {
				prfs.add(new PRFStats(sline.split(" ")));
				sline = scoreBR.readLine();
			}
			
			PRFStats guess = prfs.get(0);
			total.add(guess);			
		}

		System.out.println("Baseline evaluation scores:");
		total.print();		
		
		//total.printRaw();
		
	}
	
	private static void oracle(String[] argv) throws IOException {
		if(argv.length != 3) {
			System.err.println("two arguments necessary: score file and evalutor type (e|p|o)");
			System.exit(1);	
		}
		BufferedReader scoreBR = Util.openFileReader(argv[1]);
		String maxBy = argv[2];
		if(!maxBy.matches("e|p|o")) {
			System.err.println("second argument must be e, p, or o");
			System.exit(1);
		}			
		char mb = maxBy.charAt(0);
		
		PRFStats total = new PRFStats();
		
		String sline = scoreBR.readLine();
		if(!sline.startsWith("#"))
			throw new RuntimeException("Illegal format of the score file: |" + sline + "|");
		while(sline != null) {
			ArrayList<PRFStats> prfs = new ArrayList();
			sline = scoreBR.readLine();
			while(sline != null && !sline.startsWith("#")) {
				prfs.add(new PRFStats(sline.split(" ")));
				sline = scoreBR.readLine();
			}
			
			PRFStats guess = null;
			double max = Double.NEGATIVE_INFINITY;			
			for(int i = 0; i < prfs.size(); i++) {
				double tscore = 0;
				switch(mb) {
				case 'e':
					tscore = prfs.get(i).getF1();
					break;
				case 'p':
					tscore = prfs.get(i).getPartialF1();
					break;
				case 'o':
					tscore = prfs.get(i).getOverlapF1();
					break;
				}
				if(tscore > max) {
					max = tscore;
					guess = prfs.get(i);
				}
			}			
			total.add(guess);
		}

		System.out.println("Oracle evaluation scores:");
		total.print();		
	}
	
}
