package mpqa_structlearn.io;

import java.util.PriorityQueue;

import mpqa_structlearn.io.EPEPaths.Edge;
import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.DepNode;

public class DepPath implements Comparable<DepPath> {

    public double cost;        
    public DepNode n;
    
    public DepPath prev;

    public String prevRel;   // relation on the edge, regardless of direction
    public boolean isParent; // this is a parent of the previous node
    
    SimpleBitSet seen;
    
    public int compareTo(DepPath o) {
        return Double.compare(cost, o.cost);
    }
    
    DepPath(DepNode init) {
        this(init, true);
    }
    
    DepPath(DepNode init, boolean noRepeat) {
        cost = 0;
        n = init;
        prev = null;
        if(noRepeat)
            seen = new SimpleBitSet().add(n.position);
    }
    
    DepPath(DepPath prev, double incr, String prevRel, DepNode n) {
        this.n = n;
        this.prev = prev;
        cost = prev.cost + incr;
        this.prevRel = prevRel;
        if(prev.seen != null)
            seen = prev.seen.add(n.position);
    }
    
    void advance(PriorityQueue<DepPath> q, ListMap<DepNode, Edge> graph) {
        for(Edge e: graph.get(n)) {
            //if(prev != null && (e.from == prev.n || e.to == prev.n))
            //    continue;
            DepPath next;
            if(e.from == n) {
                if(seen != null && seen.contains(e.to.position))
                    continue;
                next = new DepPath(this, 1, e.label, e.to);
                next.isParent = false;
            } else if(e.to == n) {
                if(seen != null && seen.contains(e.from.position))
                    continue;
                next = new DepPath(this, 1, e.label, e.from);
                next.isParent = true;
            } else
                throw new RuntimeException("?");
            q.add(next);
        }
    }
    
    void stringRepr(StringBuilder sb, boolean includePosition) {
        if(includePosition)
            sb.append(n.position);
        if(prev != null) {
            if(isParent)
                sb.append("<-[" + prevRel + "]-");
            else
                sb.append("-[" + prevRel + "]->");
            prev.stringRepr(sb, includePosition);
        }
    }
    
    public String toString() {
        StringBuilder sb = new StringBuilder();
        stringRepr(sb, true);
        return sb.toString();
    }
    
    public String relString() {
        StringBuilder sb = new StringBuilder();
        stringRepr(sb, false);
        return sb.toString();
    }
    
    void stringReprPoS(StringBuilder sb) {
        sb.append(n.pos);
        if(prev != null) {
            if(isParent)
                sb.append("<-");
            else
                sb.append("->");
            prev.stringReprPoS(sb);
        }
    }
    
    public String posString() {
        StringBuilder sb = new StringBuilder();
        stringReprPoS(sb);
        return sb.toString();
    }
    
}