package mpqa_structlearn.io;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashSet;

import se.lth.cs.nlp.nlputils.core.CollectionUtils;
import se.lth.cs.nlp.nlputils.depgraph.DepNode;

public class MPQAAgentInstance implements java.io.Serializable {

	private static final long serialVersionUID = 0L;
	
	public MPQAAgent agent;
	
	public String mpqaNbr;
	public ArrayList<DepNode> span;

	/* for EPE; in case of gold-standard, this comes directly from MPQA */
    public int start = Integer.MIN_VALUE, end = Integer.MIN_VALUE;
	
	public ArrayList<MPQAAgent> nestedSource;
	
	public MPQASentence containingSentence;
	
	public boolean overlaps(MPQAAgentInstance other) {
		HashSet<DepNode> s = new HashSet(span);
		return CollectionUtils.containsSome(s, other.span);
	}
	
	/* Returns how well other covers this. */ 
	public double getSpanCoverage(MPQAAgentInstance other) {

        if(false) { /* old code */
            HashSet<DepNode> s = new HashSet(span);
            int count = 0;
            for(DepNode n: other.span)
                if(s.contains(n))
                    count++;
            return (double) count / span.size();
        }
	    
	    int start, end, ostart, oend;
        if(MPQAExpression.CHAR_EVAL) {
            start = this.start;
            end = this.end;
            ostart = other.start;
            oend = other.end;
            if(start == Integer.MIN_VALUE || end == Integer.MIN_VALUE) {
                System.out.println(this.toStringDetailed() + " / " + span);

                System.out.println(other.toStringDetailed() + " / " + other.span);
                
                throw new RuntimeException("inst start/end not set");
            }
            if(ostart == Integer.MIN_VALUE || oend == Integer.MIN_VALUE) {
                System.out.println(other);
                System.out.println(other.span);
                
                System.out.println(this);
                System.out.println(this.span);
                
                throw new RuntimeException("oinst start/end not set");
            }
        } else {
            //if(true)
            //    throw new RuntimeException("?");
            start = span.get(0).position;
            end = span.get(span.size()-1).position;
            ostart = other.span.get(0).position;
            oend = other.span.get(other.span.size()-1).position;
        }
        
        int maxstart = Math.max(start, ostart);
        int minend = Math.min(end, oend);       

        if(maxstart > minend)
            return 0;
        
        int intersectionLength = minend - maxstart + 1;
        int length = end - start + 1;
        
        return (double) intersectionLength / length;
	}

    public String toString() {
        if(span == null)
            return "null";
        int start = span.get(0).position;
        int end = span.get(span.size() - 1).position + 1;
        StringBuilder out = new StringBuilder("(" +  start + "," + end + ")");
        return out.toString();
    }
    
    public String toString2(String sid) {
        if(span == null)
            return "null";
        int start = span.get(0).position;
        int end = span.get(span.size() - 1).position + 1;
        StringBuilder out = new StringBuilder("(" + sid + ": "+  start + "," + end + ")");
        return out.toString();
    }
    
    public String toWordString() {
        StringBuilder out = new StringBuilder();
        if(span.size() > 0) {
            for(int i = 0; i < span.size(); i++) {
                if(i > 0)
                    out.append(" ");
                out.append(span.get(i).word);
            }
            return out.toString();
        } else
            return "(empty)";
    }

    public String toStringDetailed() {
        int sstart = span.get(0).position;
        int send = span.get(span.size() - 1).position + 1;
        StringBuilder out = new StringBuilder("(" + sstart + "," + send + " [" + start + "," + end + "]");
        out.append(" [" + toWordString() + "]");
        out.append(")");
        return out.toString();
    }

}

