package mpqa_structlearn.io;

import java.util.*;

import se.lth.cs.nlp.nlputils.depgraph.DepNode;

public class MPQAExpression implements java.io.Serializable {

	private static final long serialVersionUID = 0L;
	
	public String type;
	public String polarity;

	public String intensity;
	public String exprIntensity;
	
	public ArrayList<MPQAAgent> nestedSource;

	public ArrayList<DepNode> span;

	public MPQASentence containingSentence;

	/* for EPE; in case of gold-standard, this comes directly from MPQA */
	public int start = Integer.MIN_VALUE, end = Integer.MIN_VALUE;
	
	static final boolean CHAR_EVAL = true;
	
	public boolean overlaps(MPQAExpression other) {
		int start = span.get(0).position;
		int end = span.get(span.size()-1).position;
		
		int ostart = other.span.get(0).position;
		int oend = other.span.get(other.span.size()-1).position;
		
        if(start >= ostart && start <= oend)
            return true;
        if(ostart >= start && ostart <= end)
            return true;
        return false;
	}

	public void setCharSpan() {
	    if(span.isEmpty())
	        return;
	    int sstart = ((EPEDepNode) span.get(0)).start;
	    int send = ((EPEDepNode) span.get(span.size()-1)).end;
	    if(start != Integer.MIN_VALUE && start != sstart
	       || end != Integer.MIN_VALUE && end != send)
	        System.out.println(this + " / " + span + " / " + start + "," + end);
	    start = sstart;
	    end = send;
	}
	
	/* Returns how well other covers this. */ 
	public double getSpanCoverage(MPQAExpression other) {

	    int start, end, ostart, oend;
	    if(CHAR_EVAL) {
	        start = this.start;
	        end = this.end;
	        ostart = other.start;
	        oend = other.end;
            if(start == Integer.MIN_VALUE || end == Integer.MIN_VALUE) {
                System.out.println(this);
                System.out.println(span);

                throw new RuntimeException("expr start/end not set");
            }
            if(ostart == Integer.MIN_VALUE || oend == Integer.MIN_VALUE) {
                System.out.println(other);
                System.out.println(other.span);
                
                System.out.println(this);
                System.out.println(this.span);
                
                throw new RuntimeException("oexpr start/end not set");
            }
	        
	    } else {
	        //if(true)
	        //    throw new RuntimeException("?");
	        start = span.get(0).position;
	        end = span.get(span.size()-1).position;
	        ostart = other.span.get(0).position;
	        oend = other.span.get(other.span.size()-1).position;
	    }
	    
        int maxstart = Math.max(start, ostart);
        int minend = Math.min(end, oend);		

        if(maxstart > minend)
        	return 0;
		
        int intersectionLength = minend - maxstart + 1;
        int length = end - start + 1;
        
        /*HashSet<DepNode> s = new HashSet(span);
		int count = 0;
		for(DepNode n: other.span)
			if(s.contains(n))
				count++;*/

        return (double) intersectionLength / length;
	}

	public String toString() {
		int start = span.get(0).position;
		int end = span.get(span.size() - 1).position + 1;
		StringBuilder out = new StringBuilder("(" + type + " " +  start + "," + end);
		if(polarity != null)
			out.append(" p=" + polarity);
		if(intensity != null)
			out.append(" i=" + intensity);
		out.append(")");
		return out.toString();
	}
	
	public String toWordString() {
	    StringBuilder out = new StringBuilder();
        if(span.size() > 0) {
            for(int i = 0; i < span.size(); i++) {
                if(i > 0)
                    out.append(" ");
                out.append(span.get(i).word);
            }
            return out.toString();
        } else
            return "(empty)";
	}
	
	public String toStringDetailed() {
        int sstart = span.get(0).position;
        int send = span.get(span.size() - 1).position + 1;
        StringBuilder out = new StringBuilder("(" + type + " " +  sstart + "," + send
                + " [" + start + "," + end + "]");
        out.append(" [" + toWordString() + "],");
        if(polarity != null)
            out.append(" p=" + polarity);
        if(intensity != null)
            out.append(" i=" + intensity);

        if(nestedSource != null && nestedSource.size() > 0) {
            MPQAAgent ai = nestedSource.get(nestedSource.size()-1);
            out.append(" [a=" + ai.mpqaID + "]");
        }
            
        out.append(")");
        return out.toString();
    }

	public boolean equalsAnnotation(MPQAExpression e2) {
		if(!type.equals(e2.type))
			return false;
		if(span == null)
			return e2.span == null;
		else if(e2.span == null)
			return false;
		if(span.get(0).position != e2.span.get(0).position)
			return false;
		int l = span.size();
		if(e2.span.size() != l)
			return false;
		if(span.get(l-1).position != e2.span.get(l-1).position)
			return false;
		return true;
	}
	
	public MPQAExpression clone() {
		MPQAExpression out = new MPQAExpression();
		out.type = type;
		out.exprIntensity = exprIntensity;
		out.intensity = intensity;
		out.polarity = polarity;
		out.span = (ArrayList) span.clone();
		out.start = start;
		out.end = end;
		
		//if(nestedSource != null) 
		//throw new RuntimeException("cannot clone expressions with sources");
		out.nestedSource = null;
		
		return out;
	}
	
	public static final Comparator<MPQAExpression> EXPR_POS_COMP = new Comparator<MPQAExpression>() {
        public int compare(MPQAExpression o1, MPQAExpression o2) {
            if(o1.span.isEmpty())
                if(o2.span.isEmpty())
                    return o1.type.compareTo(o2.type);
                else
                    return -1;
            if(o2.span.isEmpty())
                return 1;
            
            int start1 = o1.span.get(0).position;
            int end1 = o1.span.get(o1.span.size() - 1).position + 1;
            int start2 = o2.span.get(0).position;
            int end2 = o2.span.get(o2.span.size() - 1).position + 1;

            if(start1 < start2)
                return -1;
            if(start2 < start1)
                return 1;
            
            if(end1 > end2)
                return -1;
            if(end2 > end1)
                return 1;
            
            return o1.type.compareTo(o2.type);
        }
	    
	};
	
}
