package mpqa_structlearn.io;

import java.util.*;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.*;

public class Test {

	private static String getSentenceDistance(MPQAExpression expr, 
			HashMap<DepNode, Integer> nodeIndexMap) {

		int eix = nodeIndexMap.get(expr.span.get(0));

		if(expr.nestedSource == null)
			return "ns_null";
		
		MPQAAgent src = expr.nestedSource.get(expr.nestedSource.size() - 1);
		if(src == null)
			return "src_null";

		for(MPQAAgentInstance inst: src.instances) {
			if(inst == null || inst.span == null)
				continue; // TODO detta är en bug
			int aix2 = nodeIndexMap.get(inst.span.get(0));
			if(aix2 == eix)
				return "local";
		}

		if(src.isWriter)
			return "w";
		if(src.isImplicit)
			return "imp";

		if(src.instances.isEmpty())
			return "agent_empty";
		
		return "nonlocal";		
	}
	
	private static void testAgents(MPQAText text, HashMap<String, Histogram<String>> hist) {
		HashMap<DepNode, Integer> m = new HashMap();
		for(int i = 0; i < text.sentences.size(); i++)
			for(DepNode n: text.sentences.get(i).dg.nodes)
				m.put(n, i);
		for(MPQASentence sen: text.sentences)
			for(MPQAExpression expr: sen.exprs) {
				hist.get(expr.type).add(getSentenceDistance(expr, m));
			}				
	}
	
	public static void mainOld(String[] argv) {
		String dir = argv[0];
		String subjLexFile = argv[1];
		
		try {
			LKReader r = new LKReader(dir, subjLexFile);

			HashMap<String, Histogram<String>> agentDistHist = new HashMap();
			agentDistHist.put("os", new Histogram());
			agentDistHist.put("es", new Histogram());
			agentDistHist.put("ds", new Histogram());
			
			int i = 0;
			while(r.hasNext()) {
				i++;
				System.out.println(i);
				MPQAText text = r.next();
				System.out.println(text.source);
				testAgents(text, agentDistHist);
			}
			
			for(String k: agentDistHist.keySet()) {
				Histogram<String> hist = agentDistHist.get(k);
				System.out.println(k + ": " + hist.asSortedList());
			}
		
		} catch(Exception e) {
			e.printStackTrace();
		}
	}
	
	public static void main(String[] argv) {
	    String mode = argv[0];
        String graphMode = argv[1];
        EPEFeatureSpec epeFeatureSpec = new EPEFeatureSpec(argv[2]);
	    String dir = argv[3];
        String subjLexFile = argv[4];
        String selectedFiles = argv[5];
        String epeDir = argv[6];

        if(graphMode.equals("tree"))
            EPEReader.TREE_ONLY = true;
        else if(graphMode.equals("graph")) {
        } else
            throw new RuntimeException("must specify graph or tree");
        
        try {
            ArrayList<String> selected = new ArrayList();
            Util.readLines(selectedFiles, selected);

            Iterator<MPQAText> r;
            
            if(mode.equals("EPE"))
                r = new MPQAReaderEPE(dir, subjLexFile, selected, epeDir, epeFeatureSpec);
            else if(mode.equals("LK")) {
                ArrayList<String> l = new ArrayList(selected);
                Collections.sort(l);
                for(ListIterator<String> it = l.listIterator(); it.hasNext(); ) {
                    String s = it.next();
                    s = s.replaceFirst(".*docs/", "");
                    s = s.replaceFirst("/", "_");
                    //System.out.println(s);
                    //System.exit(0);
                    it.set(s);
                }
                r = new LKReader(dir, subjLexFile, l);
            } else
                throw new RuntimeException("mode!");
            
            int maxLen = 0;
            
            while(r.hasNext()) {
                MPQAText text = r.next();
                                
                System.out.println(text.source);

                if(text.sentences != null)
                for(MPQASentence sen: text.sentences) {
                    if(sen.dg.nodes.length > maxLen) {
                        maxLen = sen.dg.nodes.length;
                    }
                    /*if(sen.dg.nodes.length > 200) {
                        System.out.println(sen.dg);
                    }*/
                }
                
                if(false) {
                    for(MPQASentence sen: text.sentences)
                        System.out.println(sen.dg);
                    System.exit(0);
                }
                
                
                /*
                for(MPQAAgent a: text.agents)
                    System.out.println(a.toStringDetailed());
                System.out.println("---");
                
                for(MPQASentence sen: text.sentences) {
                    for(MPQAExpression expr: sen.exprs)
                        System.out.println(expr.toStringDetailed());
                }              */      
                
                //System.out.println(text.original.text);

                
            }

            System.out.println("Maximal length: " + maxLen);
            
        } catch(Exception e) {
            e.printStackTrace();
        }
    }


	
}
