/*
 EnglishFeatures.java
  
 Created on Nov 15, 2006 by Richard Johansson (richard@cs.lth.se).

 $Log: EnglishFeatures.java,v $
 Revision 1.1  2010-06-07 11:33:30  johansson
 Added.

 Revision 1.1  2009/01/16 14:50:55  johansson
 Added to brenta repository.

 Revision 1.8  2008/05/16 11:48:21  richard
 POS paths, siblings.

 Revision 1.7  2008/04/23 09:06:07  richard
 Handles multiple roots.

 Revision 1.6  2008/04/17 14:51:11  richard
 Replaces numbers.

 Revision 1.5  2008/04/10 07:09:31  richard
 New features.

 Revision 1.4  2008/04/07 07:34:47  richard
 Support features.

 Revision 1.3  2008/04/02 11:30:21  richard
 Added getRelativePosition.

 Revision 1.2  2008/03/19 15:54:12  richard
 Serializable.

 Revision 1.1  2008/03/18 15:18:48  richard
 Imported from fnlabeler.

 Revision 1.7  2007/10/12 06:34:44  richard
 Some more functions.

 Revision 1.6  2007/10/08 12:33:02  richard
 Refactored again to remove atoms.

 Revision 1.5  2007/10/04 12:28:02  richard
 Refactored to atom-based.

 Revision 1.4  2007/08/21 10:19:00  richard
 Changed collapsing in ps and dep paths.

 Revision 1.3  2007/08/20 15:03:41  richard
 Added phrases.

 Revision 1.2  2007/08/17 10:54:22  richard
 Cleaned up. Loads feature declarations.

 Revision 1.1  2007/08/17 10:15:10  richard
 Refactored.

 Revision 1.1  2007/04/05 13:26:42  richard
 Version used in SemEval.

 Revision 1.2  2006/11/20 16:19:13  richard
 Labels a treebank.

 Revision 1.1  2006/11/15 14:21:46  richard
 Added the file.

   
 */
package mpqa_structlearn.locallink;

import java.util.*;

import mpqa_structlearn.io.EPEPaths;
import mpqa_structlearn.io.MPQAAgent;
import mpqa_structlearn.io.MPQAAgentInstance;
import mpqa_structlearn.io.MPQASentence;
import mpqa_structlearn.io.DepPath;

import java.io.Serializable;
import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.*;
import se.lth.cs.nlp.depsrl.format.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class EnglishFeaturesEPE implements Serializable {

    private static final long serialVersionUID = 0L;
    
    /* Whether the preposition should be attached to PP for prepositional
     * phrases, i.e. a PP headed by "at" gets the phrase type PP-at.
     */
    private static final boolean PP_ATTACH_PREP = false; //true;

    private static final String IDENTICAL = "<IDENTICAL>";
    private static final String EMPTY = "<EMPTY>";
    private static final String NONE = "<NONE>";
    private static final String ACTIVE = "active";
    private static final String PASSIVE = "passive";
    
    private static final String LEFT = "left";
    private static final String RIGHT = "right";
    private static final String ON = "on";

    /*
    public Pair<String, DepNode> getRelAndHead(DepNode node, DepNode target) {
        if(node == target)
            return new Pair(IDENTICAL, node);
        if(node.isSameOrAncestorOf(target))
            //for(DepNode c: node.children)
            //if(target == c)
            return new Pair("UP->" + target.relations[0], node);

        return new Pair(node.relations[0], node);
    }
    */
    
    public void findDeps(DepNode target, HashSet<String> deps) {
        deps.clear();
        for(DepNode c: target.children)
            deps.addAll(getChildLabels(target, c));
    }
    
    private ArrayList<String> getChildLabels(DepNode p, DepNode c) {
        ArrayList<String> out = new ArrayList();
        for(int i = 0; i < c.parents.length; i++) {
            if(c.parents[i] == p)
                out.add(c.relations[i]);
        }
        Collections.sort(out);
        return out;
    }
    
    public String findDepSubCat(DepNode target) {
        if(!target.pos.startsWith("V"))
            return EMPTY;
        List<String> deps = new ArrayList<String>();
        for(DepNode c: target.children)
            deps.addAll(getChildLabels(target, c));
        if(deps.size() == 0)
            return EMPTY;
        return Strings.join(deps, "_");
    }

    public DepNode getLeftmostChild(DepNode node) {
        ArrayList<DepNode> l = new ArrayList<DepNode>();
        for(DepNode c: node.children)
            if(c.position < node.position)
                l.add(c);
        int min = Integer.MAX_VALUE;
        DepNode minNode = null;
        for(DepNode d: l)
            if(d.position < min) {
                min = d.position;
                minNode = d;
            }
        return minNode;
    }
    
    public DepNode getRightmostChild(DepNode node) {
        ArrayList<DepNode> l = new ArrayList<DepNode>();
        for(DepNode c: node.children)
            if(c.position > node.position)
                l.add(c);
        int max = -1;
        DepNode maxNode = null;
        for(DepNode d: l)
            if(d.position > max) {
                max = d.position;
                maxNode = d;
            }
        return maxNode;
    }
    
    public String getRelativePosition(DepNode node, DepNode target) {
        if(node.position < target.position)
            return LEFT;
        if(node.position == target.position)
            return ON;
        return RIGHT;
    }
    
    public DepNode getLeftSibling(DepNode node) {
        if(node.parents.length == 0)
            return null;
        DepNode p = node.parents[0];
        int max = -1;
        DepNode maxn = null;
        for(DepNode n: p.children)
            if(n.position < node.position && n.position > max) {
                max = n.position;
                maxn = n;
            }
        return maxn;
    }
    
    public DepNode getRightSibling(DepNode node) {
        if(node.parents.length == 0)
            return null;
        DepNode p = node.parents[0];
        int min = Integer.MAX_VALUE;
        DepNode minn = null;
        for(DepNode n: p.children)
            if(n.position > node.position && n.position < min) {
                min = n.position;
                minn = n;
            }
        return minn;
    }
 
    /*

private static ArrayList<DepNode> findOutgoingLinksInSentence(MPQAAgent agent, MPQASentence sen) {
        ArrayList<DepNode> out = new ArrayList();
        for(MPQAAgentInstance inst: agent.instances) {
            if(inst.containingSentence != sen)
                continue;

            int min = inst.span.get(0).position;
            int max = inst.span.get(inst.span.size() - 1).position;
            for(DepNode n: inst.span) {
                for(DepNode p: n.parents) {
                    int pp = p.position;
                    if(pp < min || pp > max) {
                        out.add(n);
                        break;
                    }
                }
            }
        }
        return out;
    }
    
    */
    
    public DepNode getHighest(ArrayList<DepNode> span) {
    	int size = span.size();
    	if(size == 0)
    		throw new IllegalArgumentException("span is empty");
    	if(size == 1)
    		return span.get(0);

    	ArrayList<DepNode> out = new ArrayList();
    	
    	int min = span.get(0).position;
    	int max = span.get(span.size()-1).position;

    	for(DepNode n: span) {
    	    for(DepNode p: n.parents) {
    	        int pp = p.position;
                if(pp < min || pp > max) {
                    out.add(n);
                    break;
                }
    	    }
    	}
    	
    	if(out.size() == 0) {
    	    //for(DepNode n: span)
    	    //    System.out.println(n);
    	    //throw new RuntimeException("no nodes!");
    	    // no incoming edges... just return the first one
    	    return span.get(0);
    	}
    	
    	if(out.size() == 1)
    	    return out.get(0);
    	
    	HashSet<DepNode> toRemove = new HashSet();
        for(DepNode n1: out)
            for(DepNode n2: out)
                if(n1 != n2 && n1.isSameOrAncestorOf(n2))
                    toRemove.add(n2);
    	
        if(toRemove.size() < out.size()) {
            out.removeAll(toRemove);
            return out.get(0);
        }

    	toRemove.clear();
        for(DepNode n1: out)
            for(DepNode n2: out)
                if(n1 != n2 && CollectionUtils.contains(n2.parents, n1))
                    toRemove.add(n2);

        if(toRemove.size() < out.size()) {
            out.removeAll(toRemove);
            return out.get(0);
        }

        return out.get(0);
    }
    
}
