package mpqa_structlearn.locallink;

import java.io.*;
import java.util.*;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.*;
import se.lth.cs.nlp.nlputils.ml_long.*;

import mpqa_structlearn.io.*;

public class LocalLinkClassifier implements Serializable {

	private static final long serialVersionUID = 0L;
	
	   static final FeatureDeclaration FEATURE_DECL 
	    = new FeatureDeclaration(new String[] {

	            "P_EXPR_HEAD",
	            "P_EXPR_HEAD_LEMMA",
	            
	            //"P_VOICE",
                "P_EXPR_HEAD_POS",
                "P_EXPR_HEAD_FEATS",

	            "P_DEPS",
	            "P_DEP_SUBCAT",

	            "P_PARENT_WORD",
	            "P_PARENT_POS",
	            "P_REL_TO_PARENT",

	            "P_EXPRTYPE",
	            "P_DOMINATING_EXPR",	            
	            
	            //"P_EXPR_WORDS", // compatibility
	            
	            "N_POSITION",

	            "N_NEAREST_EXPR_WORD",
	            
	            "N_HOLDER_HEAD",
	            "N_HOLDER_HEAD_POS",

	            "N_PATH",
	            "N_LEFT_WORD",
	            "N_LEFT_POS",
	            "N_RIGHT_WORD",
	            "N_RIGHT_POS",

	            "N_DEPREL", // lowest

	            "N_LEFTSIBLING_WORD",
	            "N_LEFTSIBLING_POS",

	            "N_RIGHTSIBLING_WORD",
	            "N_RIGHTSIBLING_POS",
	            
	            "N_POS_PATH",
	            "N_PB_REL"
	    });
	    
	private static final boolean TRAIN = true;
	   
	public static void train(String[] argv, Iterator<MPQAText> r) {
	    //String dir = argv[1];
		//String subjLexFile = argv[2];
		//String docListFile = argv[3];
		String confFile = argv[4];
		
		try {
			Configuration config = new Configuration(new FileInputStream(confFile));
			
			String modelFileName;
			if(argv.length > 5)
			    modelFileName = argv[5];
			else
			    modelFileName = config.getModelName();
			
			FeatureDeclaration localDecl = config.getClassifierFeatures(Configuration.ClassifierID.LOCAL);
			FeatureDeclaration nonlocalDecl = config.getClassifierFeatures(Configuration.ClassifierID.NONLOCAL);

			EnglishFeaturesEPE lang = new EnglishFeaturesEPE();
			NodeFeatureExtractorEPE localNodeExtractor = new NodeFeatureExtractorEPE(lang, localDecl);
			ExprFeatureExtractorEPE localExprExtractor = new ExprFeatureExtractorEPE(lang, localDecl);
			ExprFeatureExtractorEPE nonlocalExtractor = new ExprFeatureExtractorEPE(lang, nonlocalDecl);
			
			ArrayList<Pair<FeatureList, String>> localExamples = new ArrayList();
			ArrayList<Pair<FeatureList, String>> nonlocalExamples = new ArrayList();

			/*HashSet<String> selectedFiles = new HashSet();
	            Util.readLines(docListFile, selectedFiles);
			LKReader r = new LKReader(dir, subjLexFile);
			while(r.hasNext()) {
				MPQAText text = r.next();
				if(selectedFiles.contains(text.source)) {
					System.out.println(text.source);
					extractLinks(text, localDecl, nonlocalDecl,
							localNodeExtractor, localExprExtractor, nonlocalExtractor,
							localExamples, nonlocalExamples);
				}
			}
			*/
			while(r.hasNext()) {
                MPQAText text = r.next();
                System.out.println(text.source);
                //System.out.print(".");
                //System.out.flush();

                extractLinks(text, localDecl, nonlocalDecl,
                        localNodeExtractor, localExprExtractor, nonlocalExtractor,
                        localExamples, nonlocalExamples);
            }
			System.out.println();

			double lcov = (double) nLocalLinksCovered / nLocalLinks;
			System.out.format("Local link coverage: %d / %d = %f\n", 
					nLocalLinksCovered, nLocalLinks, lcov);
			double lfrac = (double) nPositiveInstances / nInstances;
			System.out.format("Local link fraction positive: %d / %d = %f\n", 
					nPositiveInstances, nInstances, lfrac);
			
			if(TRAIN) {
				ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(modelFileName));
				oos.writeObject(new java.util.Date().toString());
				oos.writeObject(config.toXMLString());

				{					
					System.out.println("Training local classifier... (" + localExamples.size() + " instances)");
					Configuration.ClassifierID cid = Configuration.ClassifierID.LOCAL;
					LearningAlgorithm alg = config.getAlgorithm(cid);
					HashSet<Pair<String, String>> pairs = config.getClassifierPairs(cid);

					//System.out.println("pairs1 = " + pairs);
					
					FeatureListEncoder fle = new PairingFeatureListEncoder(localDecl, pairs);
					MulticlassDefinition<FeatureList, String> def
					= new MulticlassDefinition(localExamples, localDecl, 
							MulticlassDefinition.ID, fle);
					Classifier localCl = alg.train(def, localExamples);
					System.out.println("Done.");
					oos.writeObject(localCl);
					oos.flush();
					oos.reset();
				}

				{					
					System.out.println("Training nonlocal classifier... (" + nonlocalExamples.size() + " instances)");
					Configuration.ClassifierID cid = Configuration.ClassifierID.NONLOCAL;
					LearningAlgorithm alg = config.getAlgorithm(cid);
					HashSet<Pair<String, String>> pairs = config.getClassifierPairs(cid);

					System.out.println("pairs2 = " + pairs);
					
					FeatureListEncoder fle = new PairingFeatureListEncoder(nonlocalDecl, pairs);
					MulticlassDefinition<FeatureList, String> def
					= new MulticlassDefinition(nonlocalExamples, nonlocalDecl, 
							MulticlassDefinition.ID, fle);
					Classifier nonlocalCl = alg.train(def, nonlocalExamples);
					System.out.println("Done.");
					oos.writeObject(nonlocalCl);
					oos.flush();
					oos.reset();
				}		

				oos.writeObject(localNodeExtractor);
				oos.writeObject(localExprExtractor);
				oos.writeObject(nonlocalExtractor);

				oos.writeObject(localDecl);
				oos.writeObject(nonlocalDecl);
				
				oos.close();
				
			} else {
				printDataset("local.txt", localExamples, localDecl);
				printDataset("nonlocal.txt", nonlocalExamples, nonlocalDecl);
			}
			
		} catch(Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
	}

	private static void test(String[] argv) {
		try {
			String dir = argv[1];
			String subjLexFile = argv[2];
			String docListFile = argv[3];
			String modelFile = argv[4];

			/* Loading trained model. */
			LocalLinkClassifier llc = loadModel(modelFile);

			/* Creation of example feature vectors. */
			HashSet<String> selectedFiles = new HashSet();
			Util.readLines(docListFile, selectedFiles);
						
			LKReader r = new LKReader(dir, subjLexFile);
			ArrayList<Pair<FeatureList, String>> localExamples = new ArrayList();
			ArrayList<Pair<FeatureList, String>> nonlocalExamples = new ArrayList();
			while(r.hasNext()) {
				MPQAText text = r.next();
				if(selectedFiles.contains(text.source)) {
					//System.out.println(text.source);
					extractLinks(text, llc.localDecl, llc.nonlocalDecl,
							llc.localNodeExtractor, llc.localExprExtractor, llc.nonlocalExtractor,
							localExamples, nonlocalExamples);
				}
			}
			
			/* Evaluation. */
			System.out.println("*** Local links: ***");
			evalClassifier(localExamples, llc.localCl, "F");
			
			System.out.println("*** Nonlocal links: ***");
			evalClassifier(nonlocalExamples, llc.nonlocalCl, "0");
			
		} catch(Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
	}
	
	private static void evalClassifier(ArrayList<Pair<FeatureList, String>> l2,
			Classifier<FeatureList, String> cl, String NULL) {
        int n = 0, ncorrect = 0;            
        int nguesses = 0, ningold = 0, ncorrectGuesses = 0;

        HashMap<Object, int[]> perClassResults = new HashMap();
        HashMap<Object, int[]> perExprResults = new HashMap();
        
        for(Pair<FeatureList, String> p: l2) {
            n++;
            if(n % 100 == 0) {
                System.out.print(".");
                System.out.flush();
            }
            String answer = cl.classify(p.left);
            if(answer.equals(p.right))
                ncorrect++;
            int[] c1 = null;

            if(!answer.equals(NULL)) {
                c1 = perClassResults.get(answer);
                if(c1 == null) {
                    c1 = new int[3];
                    perClassResults.put(answer, c1);
                }
                c1[0]++;
                if(answer.equals(p.right))
                	c1[2]++;
            }
            if(!p.right.equals(NULL)) {
                int[] c2 = perClassResults.get(p.right);
                if(c2 == null) {
                    c2 = new int[3];
                    perClassResults.put(p.right, c2);
                }
                c2[1]++;
            }

            /* Temporary for holder test. */
            String f2 = (String) p.left.getFeature("P_EXPRTYPE");
            if(!answer.equals(NULL)) {
            	c1 = perExprResults.get(f2);
                if(c1 == null) {
                    c1 = new int[3];
                    perExprResults.put(f2, c1);
                }
                c1[0]++;
                if(answer.equals(p.right))
                	c1[2]++;
            }
            if(!p.right.equals(NULL)) {
                int[] c2 = perExprResults.get(f2);
                if(c2 == null) {
                    c2 = new int[3];
                    perExprResults.put(f2, c2);
                }
                c2[1]++;
            }

        }

        double acc = (double) ncorrect / n;
        double p = (double) ncorrectGuesses / nguesses;
        double r = (double) ncorrectGuesses / ningold;
        double f1 = 2*p*r/(p+r);

        System.out.println();           
        System.out.println(ncorrect + "/" + n + " = " + acc);

        if(true) {
            System.out.println();
            System.out.println("p = " + ncorrectGuesses + "/" + nguesses + " = " + p);
            System.out.println("r = " + ncorrectGuesses + "/" + ningold + " = " + r);
            System.out.println("f1 = " + f1);
        }

        if(true)
            for(Object c: perClassResults.keySet()) {
                int[] res = perClassResults.get(c);
                double pc = (double) res[2] / res[0];
                double rc = (double) res[2] / res[1];
                double fc = 2*pc*rc/(pc + rc);
                pc = Math.round(10000.0*pc)/10000.0;
                rc = Math.round(10000.0*rc)/10000.0;
                fc = Math.round(10000.0*fc)/10000.0;
                System.out.println(c + ": p = " + pc + ", r = " + rc + ", f1 = " + fc);
            }

        if(true)
            for(Object c: perExprResults.keySet()) {
                int[] res = perExprResults.get(c);
                double pc = (double) res[2] / res[0];
                double rc = (double) res[2] / res[1];
                double fc = 2*pc*rc/(pc + rc);
                pc = Math.round(10000.0*pc)/10000.0;
                rc = Math.round(10000.0*rc)/10000.0;
                fc = Math.round(10000.0*fc)/10000.0;
                System.out.println(c + ": p = " + pc + ", r = " + rc + ", f1 = " + fc);
            }
	}
	
	private static void printDataset(String filename, 
			ArrayList<Pair<FeatureList, String>> dataset,
			FeatureDeclaration decl) throws IOException {
		OutputStream os = new FileOutputStream(filename);
		PrintWriter pw = new PrintWriter(new OutputStreamWriter(os, "UTF-8"));
		if(decl != null) {
			pw.println("decl=" + decl);
		}
		for(Pair<FeatureList, String> p: dataset) {
			if(p.left.decl != decl)
				throw new RuntimeException("illegal declaration");
			pw.print(p.right);

			for(int i = 0; i < decl.size(); i++) {
				pw.print("|");				
				Object o = p.left.getFeature(i);
				if(o instanceof String)
					pw.print(o);
				else {
					HashSet<String> s = (HashSet) o;
					if(s.isEmpty())
						pw.print("<E>");
					else {
						boolean first = true;
						for(String f: s) {
							if(first)
								first = false;
							else
								pw.print("#");
							pw.print(f);
						}
					}						
				}
			}			
			pw.println();
		}
		pw.close();
	}
	
	private static final boolean isNominal(DepNode n) {
		String pos = n.pos;
		if(pos.startsWith("NN"))
			return true;
		if(pos.equals("PRP"))
			return true;
		// "this" or "that" as pronoun shouldn't be a holder in any case
		//if(pos.equals("DT"))
		//	return true;
		//if(pos.equals("DT") && n.word.toLowerCase().matches("this|that"))
		//	return true;
		
		if(pos.equals("PRP$")) 
			return true;
		
		// TODO PRP$ också
		
		return false;
	}
	
	private static ArrayList<DepNode> findNominals(DepGraph dg) {
		ArrayList<DepNode> out = new ArrayList();
		for(int i = 1; i < dg.nodes.length; i++) {
			DepNode n = dg.nodes[i];
			if(isNominal(dg.nodes[i]))
				out.add(n);
		}
		return out;
	}
	
	public static ArrayList<DepNode> findCandidatesOld(MPQASentence sen) {

        // TODO
	    if(true)
	        throw new RuntimeException("TODO");
	    
	    DepGraph dg = sen.dg;
		HashSet<DepNode> out = new HashSet();
		for(int i = 1; i < dg.nodes.length; i++) {
			DepNode n = dg.nodes[i];
			if(isNominal(dg.nodes[i]))
				out.add(n);
		}
		for(MPQAExpression expr: sen.exprs)
			out.removeAll(expr.span);
		HashSet<DepNode> toRemove = new HashSet();
		for(DepNode n1: out)
			for(DepNode n2: out)
				if(n1 != n2 && n2.parents[0] == n1)
					toRemove.add(n2);
		out.removeAll(toRemove);
		ArrayList<DepNode> out2 = new ArrayList(out);
		Collections.sort(out2);
		return out2;
	}
	
	public static ArrayList<DepNode> findCandidates(MPQASentence sen) {
        DepGraph dg = sen.dg;
        HashSet<DepNode> out = new HashSet();
        for(int i = 1; i < dg.nodes.length; i++)
            out.add(dg.nodes[i]);
        for(MPQAExpression expr: sen.exprs)
            out.removeAll(expr.span);
        ArrayList<DepNode> out2 = new ArrayList(out);
        Collections.sort(out2);
        return out2;
    }

	private static ArrayList<DepNode> findOutgoingLinksInSentence(MPQAAgent agent, MPQASentence sen) {
	    ArrayList<DepNode> out = new ArrayList();
		for(MPQAAgentInstance inst: agent.instances) {
			if(inst.containingSentence != sen)
				continue;
			/* We assume that they form a contiguous span. */
			int min = inst.span.get(0).position;
			int max = inst.span.get(inst.span.size() - 1).position;
			for(DepNode n: inst.span) {
			    for(DepNode p: n.parents) {
			        int pp = p.position;
			        if(pp < min || pp > max) {
			            out.add(n);
			            break;
			        }
			    }
			}
		}
		return out;
	}
	
	/* Denna lösning är egentligen inte riktigt som den ska!
	 * Detta är ett strukturprediktionsproblem i sig.
	 * Med PA skulle det kunna göras annorlunda.
	 */
	
	private static int nLocalLinks;
	private static int nLocalLinksCovered;

	private static int nInstances, nPositiveInstances;
	
	private static boolean hasInstanceInSentence(MPQAAgent ag, MPQASentence sen) {
		for(MPQAAgentInstance inst: ag.instances)
			if(inst.containingSentence == sen)
				return true;
		return false;
	}
	
	private static void extractLinks(MPQAText text,
			FeatureDeclaration localDecl, FeatureDeclaration nonlocalDecl,
			NodeFeatureExtractorEPE localNodeExtractor,
			ExprFeatureExtractorEPE localExprExtractor,
			ExprFeatureExtractorEPE nonlocalExtractor,
			ArrayList<Pair<FeatureList, String>> localExamples,
			ArrayList<Pair<FeatureList, String>> nonlocalExamples) {

		/*for(MPQASentence sen: text.sentences)
			System.out.println(sen.dg);
		System.out.println("---");*/
		
		for(MPQASentence sen: text.sentences) {
		    if(sen.dg.nodes.length > 256)
		        continue;		        
		    
		    EPEPaths.prepare(sen.dg);
		    
		    // all potential dep nodes we might link to in this sentence
			ArrayList<DepNode> candidates = findCandidates(sen);
			
			for(MPQAExpression expr: sen.exprs) {
				if(expr.nestedSource == null)
					continue;
				MPQAAgent holder = expr.nestedSource.get(expr.nestedSource.size() - 1);
				if(holder == null)
					continue;

				if(hasInstanceInSentence(holder, sen))
					nLocalLinks++;
				
				// "head" nodes of instances of this holder in this sentence
				ArrayList<DepNode> instanceNodes = findOutgoingLinksInSentence(holder, sen);
				
				Pair<MPQAExpression, MPQASentence> p = new Pair(expr, sen);

				boolean coveredLink = false;
				
				/* Extract features for potential local links. */
				FeatureList linkFeatures = new FeatureList(localDecl);
				localExprExtractor.extractFeatures(p, linkFeatures);
				Triple<MPQAExpression, DepNode, MPQASentence> tr = new Triple(expr, null, sen);
				for(DepNode cand: candidates) {
					boolean answer = instanceNodes.contains(cand); //holderContainsNode(cand, holder);
					tr.second = cand;
					localNodeExtractor.extractFeatures(tr, linkFeatures);
					localExamples.add(new Pair(linkFeatures.clone(), answer? "T": "F"));
					if(answer) {
						coveredLink = true;
						nPositiveInstances++;
					}
					nInstances++;
				}
				if(coveredLink)
					nLocalLinksCovered++;
				
				/* Extract features for potential nonlocal links. */
				FeatureList nonlocalFeatures = new FeatureList(nonlocalDecl);
				nonlocalExtractor.extractFeatures(p, nonlocalFeatures);
				nonlocalExamples.add(new Pair(nonlocalFeatures, getNonlocalType(sen, holder)));
			}
		}
	}

	private static boolean holderContainsNode(DepNode cand, MPQAAgent holder) {
		if(holder.instances == null)
			return false;
		for(MPQAAgentInstance inst: holder.instances)
			if(inst.span != null && inst.span.contains(cand))
				return true;
		return false;
	}

	private static String getNonlocalType(MPQASentence sen, MPQAAgent holder) {
		if(holder.isWriter)
			return "W";
		if(isExternal(sen, holder))
			return "E";
		if(holder.isImplicit)
			return "I";
		return "0";
	}
	
	private static boolean isExternal(MPQASentence sen, MPQAAgent holder) {
		boolean seenConcrete = false;
		for(MPQAAgentInstance inst: holder.instances) {
			if(inst.containingSentence == sen)
				return false;
			if(inst.containingSentence != null)
				seenConcrete = true;
		}
		return seenConcrete;
	}

	/* TODO implementera dessa i extractor. */
	
	private static boolean containsQuote(MPQASentence sen) {
		for(DepNode n: sen.dg.nodes)
			if(n.pos != null && (n.pos.equals("``") || n.pos.equals("''")))
				return true;
		return false;
	}

	private static boolean containsExprType(MPQASentence sen, String type) {
		for(MPQAExpression expr: sen.exprs)
			if(expr.type.equals(type))
				return true;
		return false;		
	}
	
	private static boolean exprSynDominatedByType(MPQAExpression expr, String type) {
		return false;
	}
	
	public static void main(String[] argv) {
		if(argv[0].equals("-train")) {
		    try {
		        String dir = argv[1];
		        String subjLexFile = argv[2];
		        String docListFile = argv[3];
		        ArrayList<String> selectedFiles = new ArrayList();
		        Util.readLines(docListFile, selectedFiles);
		        LKReader r = new LKReader(dir, subjLexFile);
		        train(argv, r);
		    } catch(Exception e) {
		        e.printStackTrace();
		    }
		}
		else if(argv[0].equals("-test"))
			test(argv);
		else {
			System.err.println("Illegal arguments: " + Arrays.toString(argv));
			System.exit(1);
		}
	}
	
	/* Private attributes */
	
	public Classifier<FeatureList, String> localCl, nonlocalCl;

	public NodeFeatureExtractorEPE localNodeExtractor;
	public ExprFeatureExtractorEPE localExprExtractor;
	public ExprFeatureExtractorEPE nonlocalExtractor;

	public FeatureDeclaration localDecl;
	public FeatureDeclaration nonlocalDecl;

	public static LocalLinkClassifier loadModel(String modelFile) throws IOException {
		if(modelFile == null || modelFile.matches("null|none|empty|NULL|NONE|EMPTY"))
			return null;

        System.out.println("Loading link model...");
		
		ObjectInputStream ois = new ObjectInputStream(Util.openFileStream(modelFile));
		LocalLinkClassifier out = new LocalLinkClassifier();
		try {
			ois.readObject(); // date
			ois.readObject(); // conf xml string
			out.localCl = (Classifier) ois.readObject();
			out.nonlocalCl = (Classifier) ois.readObject();
			out.localNodeExtractor = (NodeFeatureExtractorEPE) ois.readObject();
			out.localExprExtractor = (ExprFeatureExtractorEPE) ois.readObject();
			out.nonlocalExtractor = (ExprFeatureExtractorEPE) ois.readObject();
			out.localDecl = (FeatureDeclaration) ois.readObject();
			out.nonlocalDecl = (FeatureDeclaration) ois.readObject();
		} catch(Exception e) {
			throw new IOException(e);
		}
		ois.close();
		return out;
	}
	
}
