package mpqa_structlearn.polarity;

import java.io.*;
import java.util.*;

import mpqa_structlearn.io.*;
import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.*;
import se.lth.cs.nlp.nlputils.ml_long.*;

public class BaselinePolarityClassifier {

	public static void main(String[] argv) {
		if(argv[0].equals("-train")) {
		    try {
		        String dir = argv[1];
		        String subjLexFile = argv[2];
		        String docListFile = argv[3];
		        ArrayList<String> selectedFiles = new ArrayList();
		        Util.readLines(docListFile, selectedFiles);
		        Iterator<MPQAText> r = new LKReader(dir, subjLexFile, selectedFiles);
		        train(argv, r);
		    } catch(Exception e) {
		        e.printStackTrace();
		        System.exit(1);
		    }
		} else if(argv[0].equals("-test"))
			test(argv);
		else
			throw new RuntimeException("!!!");
	}
	
	public static void train(String[] argv, Iterator<MPQAText> r) {
		//String dir = argv[1];
		//String subjLexFile = argv[2];
	    //String docListFile = argv[3];
		String outModelFile = argv[4];

		String algName = argv[5];
		String algArgs = argv[6];
		
		ArrayList<Pair<SparseVector, Integer>> polTS = new ArrayList();
		ArrayList<Pair<SparseVector, Integer>> intTS = new ArrayList();
		
		SymbolEncoder enc = new SymbolEncoder();
		
		try {
			/*HashSet<String> selectedFiles = new HashSet();
			Util.readLines(docListFile, selectedFiles);
			//LKReader r = new LKReader(dir, subjLexFile);
			while(r.hasNext()) {
				MPQAText text = r.next();
				String shortSource = text.source.substring(23);				
				if(selectedFiles.contains(text.source) || selectedFiles.contains(shortSource)) {
					System.out.print(".");
					System.out.flush();
					extractTrainingData(text, enc, polTS, intTS);
				}
			}*/
		    while(r.hasNext()) {
                MPQAText text = r.next();
                System.out.print(".");
                extractTrainingData(text, enc, polTS, intTS);
		    }
		    
			System.out.println();
			
			//for(Pair<SparseVector, Integer> p: polTS)
			//	System.out.println(p.right + " " + p.left);
			
			AlgorithmFactory af = new AlgorithmFactory();

			LearningAlgorithm alg = af.create(algName, algArgs);
			int[] polRange = new int[] { NEUTRAL, POSITIVE, NEGATIVE, BOTH};
			EncodedMulticlassDefinition def = new EncodedMulticlassDefinition(polRange);
            Classifier<SparseVector, Integer> polCl = alg.train(def, polTS);
            polTS = null;
            
            alg = af.create(algName, algArgs);
            int[] intRange = new int[] { LOW, MEDIUM, HIGH, EXTREME };
            def = new EncodedMulticlassDefinition(intRange);
            Classifier<SparseVector, Integer> intCl = alg.train(def, intTS);

            ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(outModelFile));
            oos.writeObject(polCl);
			oos.writeObject(intCl);
			if(enc != null) {
				enc.freeze();
	            oos.writeObject(enc);
			}
			oos.close();

			if(false) {
				ArrayList<String> einv = enc.inverse();
				SparseVector sv = new SparseVector();
				sv.put(1, 1);
				ArrayList<Integer> range = new ArrayList();
				for(int p: polRange)
					range.add(p);
				// polRange and intRange identical
				double[] scores = new double[polRange.length];

				ArrayList<DoubleObjPair<String>> ptable = new ArrayList();
				ArrayList<DoubleObjPair<String>> itable = new ArrayList();
				
				for(int i = 1; i < einv.size(); i++) {
					sv.keys[0] = i;
					
					polCl.computeScoresRestricted(sv, range, scores);					
					for(int j = 0; j < polRange.length; j++) {
						String s = einv.get(i) + "/" + polRange[j];
						ptable.add(new DoubleObjPair(scores[j], s));
					}
					
					intCl.computeScoresRestricted(sv, range, scores);					
					for(int j = 0; j < intRange.length; j++) {
						String s = einv.get(i) + "/" + intRange[j];
						itable.add(new DoubleObjPair(scores[j], s));
					}					
				}
				
				Comparator<DoubleObjPair> comp = new Comparator<DoubleObjPair>() {
					public int compare(DoubleObjPair o1, DoubleObjPair o2) {
						return -Double.compare(Math.abs(o1.left), Math.abs(o2.left));
					}
				};
			
				Collections.sort(ptable, comp);
				Collections.sort(itable, comp);				

				System.out.println("*** Polarity:");
				for(int i = 0; i < 100; i++) {
					System.out.println(ptable.get(i));
				}

				System.out.println("*** Intensity:");
				for(int i = 0; i < 100; i++) {
					System.out.println(itable.get(i));
				}
				
			}
			
		} catch(Exception e) {
			e.printStackTrace();
		}
	}
	
	public static final int NEUTRAL = 1;
	public static final int POSITIVE = 2;
	public static final int NEGATIVE = 3;
	public static final int BOTH = 4;
	
	private static final int LOW = 1;
	private static final int MEDIUM = 2;
	private static final int HIGH = 3;
	private static final int EXTREME = 4;
	
	private static void extractTrainingData(MPQAText text,
			SymbolEncoder enc,
			ArrayList<Pair<SparseVector, Integer>> polTS,
			ArrayList<Pair<SparseVector, Integer>> intTS) throws IOException {		
		
		for(MPQASentence sen: text.sentences) {
			for(MPQAExpression expr: sen.exprs) {

				if(expr.type.equals("os"))
					continue;
				
				int polCl = getPolCode(expr.polarity);
				int intCl = getIntCode(expr.intensity);
				
				SparseVector sv = represent(expr, enc);
				SparseVector sv1 = sv.copy();

				//if(expr.polarity != null && !expr.polarity.startsWith("unc"))				
				polTS.add(new Pair(sv, polCl));

				intTS.add(new Pair(sv1, intCl));
			}
		}
	}

	private static int LEFT_CONTEXT = 1;
	private static int RIGHT_CONTEXT = 1;

	private static final boolean USE_BIGRAMS = true;
	
	private static final boolean NORMALIZE_BY_NORM = false;
	
	private static final boolean SIMPLIFIED_SET = false; //true;
	
	public static SparseVector represent(MPQAExpression expr, SymbolEncoder enc) {
		SparseVector sv = new SparseVector();		

		ArrayList<String> slClues = expr.containingSentence.slClues;
		DepGraph dg = expr.containingSentence.dg;
		if(slClues.size() != dg.nodes.length ) {
			System.out.println("slClues.size = " + slClues.size());
			System.out.println("dg.nodes.length = " + dg.nodes.length);
			throw new RuntimeException("!!!!!");
		}

		String prevWord = "<SW>";
		//String prevPos = "<SP>";
		
		for(DepNode n: expr.span) {
			String scl = slClues.get(n.position);
			if(scl == null)
				scl = "0";
			putFeature(sv, enc, "scl:" + scl);
			
			String w = n.word.toLowerCase();

			if(!SIMPLIFIED_SET)
				putFeature(sv, enc, "w:" + w);

			if(!SIMPLIFIED_SET)
				putFeature(sv, enc, "p:" + n.pos);
			//putFeature(sv, enc, "l:" + n.lemma);

			if(USE_BIGRAMS) {

				if(!SIMPLIFIED_SET)
					putFeature(sv, enc, "bi:" + prevWord + "_:_" + w);
				//putFeature(sv, enc, "bi:" + prevPos + "_:_" + n.pos);
				//putFeature(sv, enc, "bi:" + prevPos + "_:_" + w);
				//putFeature(sv, enc, "bi:" + prevWord + "_:_" + n.pos);
			}

			prevWord = w;
			//prevPos = n.pos;
		}
		
		if(!SIMPLIFIED_SET)
			putFeature(sv, enc, "bi:" + prevWord + "_:_<EW>");
		//putFeature(sv, enc, "bi:" + prevPos + "_:_<EP>");
		
		sv.sortIndices();
		normalize(sv);
		
		if(!SIMPLIFIED_SET) {
			int start = expr.span.get(0).position;
			for(int i = 0; i < LEFT_CONTEXT; i++) {
				int position = start - i - 1;
				if(position <= 0) {
					putFeature(sv, enc, "Lscl:BEF");
					putFeature(sv, enc, "Lw:BEF");
					putFeature(sv, enc, "Lp:BEF");
					//putFeature(sv, enc, "Ll:BEF");
				} else {
					String scl = slClues.get(position);
					if(scl == null)
						scl = "0";
					DepNode n = dg.nodes[position];
					putFeature(sv, enc, "Lscl:" + scl);
					putFeature(sv, enc, "Lw:" + n.word.toLowerCase());

					putFeature(sv, enc, "Lp:" + n.pos);
					//putFeature(sv, enc, "Ll:" + n.lemma);
				}			
			}

			int end = expr.span.get(expr.span.size() - 1).position;
			for(int i = 0; i < RIGHT_CONTEXT; i++) {
				int position = end + i + 1;
				if(position >= dg.nodes.length) {
					putFeature(sv, enc, "Rscl:AFT");
					putFeature(sv, enc, "Rw:AFT");
					putFeature(sv, enc, "Rp:AFT");
					//putFeature(sv, enc, "Rl:AFT");
				} else {
					String scl = slClues.get(position);
					if(scl == null)
						scl = "0";
					DepNode n = dg.nodes[position];
					putFeature(sv, enc, "Rscl:" + scl);
					putFeature(sv, enc, "Rw:" + n.word.toLowerCase());

					putFeature(sv, enc, "Rp:" + n.pos);
					//putFeature(sv, enc, "Rl:" + n.lemma);
				}
			}
		}
		
		sv.sortIndices();
		normalize(sv);
		
		return sv;
	}

	private static int getPolCode(String pol) {
		pol = simplifyPol(pol);
		int polCl = -1;
		if(pol.equals("neutral"))
			polCl = NEUTRAL;
		else if(pol.equals("positive"))
			polCl = POSITIVE;
		else if(pol.equals("negative"))
			polCl = NEGATIVE;
		else if(pol.equals("both"))
			polCl = BOTH;
		return polCl;
	}
	
	private static int getIntCode(String intensity) {
		intensity = simplifyInt(intensity);
		int intCl = -1;
		if(intensity.equals("low"))
			intCl = LOW;
		else if(intensity.equals("medium"))
			intCl = MEDIUM;
		else if(intensity.equals("high"))
			intCl = HIGH;
		else if(intensity.equals("extreme"))
			intCl = EXTREME;
		return intCl;
	}	

	public static String decodePol(int c) {
		switch(c) {
		case NEUTRAL:
			return "neutral";
		case POSITIVE:
			return "positive";
		case NEGATIVE:
			return "negative";
		case BOTH:
			return "both";
		default:
			throw new IllegalArgumentException("unknown code: " + c);
		}
	}
	
	public static String decodeInt(int c) {
		switch(c) {
		case LOW:
			return "low";
		case MEDIUM:
			return "medium";
		case HIGH:
			return "high";
		case EXTREME:
			return "extreme";
		default:
			throw new IllegalArgumentException("unknown code: " + c);
		}
	}
	
	private static String simplifyPol(String polarity) {
		if(polarity == null)
			polarity = "neutral";
		polarity = polarity.trim();
		if(polarity.startsWith("uncertain-"))
			polarity = polarity.substring("uncertain-".length());					
		if(polarity.equals(""))
			polarity = "neutral";	
		return polarity;
	}
	
	private static String simplifyInt(String intensity) {
		if(intensity == null)
			intensity = "medium";
		intensity = intensity.trim();				
		if(intensity.equals("") || intensity.equals("neutral"))
			intensity = "medium";
		return intensity;
	}
	
	private static void putFeature(SparseVector sv, SymbolEncoder enc, String f) {
		//sv.put(enc.encode(f), 1.0);
		sv.put(hash(f), 1.0);
	}
	
	private static void normalize(SparseVector sv) {
		if(NORMALIZE_BY_NORM) {
			double norm = Math.sqrt(SparseVector.sortedSqrLength(sv));
			if(norm > 0) {
				double inorm = 1.0 / norm;
				for(int i = 0; i < sv.index; i++)
					sv.values[i] *= inorm;
			}
		} else {			
			for(int i = 0; i < sv.index; i++)
				sv.values[i] = 1;
		}
	}
	
	private static long hash(String s) {
		long hash = 0;
		int l = s.length();
		for(int i = 0; i < l; i++) {			
			hash += s.charAt(i);
			hash += (hash << 10L);
			hash ^= (hash >> 6L);
		}
		hash += (hash << 3L);
		hash ^= (hash >> 11L);
		hash += (hash << 15L);
		return hash;
	}

	private static int nExprs, nCorrPol, nCorrInt, nCorrPN;
	
	public static void test(String[] argv) {
		String dir = argv[1];
		String subjLexFile = argv[2];
		String docListFile = argv[3];
		String modelFile = argv[4];

		try {

			ObjectInputStream ois = new ObjectInputStream(Util.openFileStream(modelFile));
			Classifier<SparseVector, Integer> polCl = (Classifier) ois.readObject();
			Classifier<SparseVector, Integer> intCl = (Classifier) ois.readObject();
			SymbolEncoder enc = (SymbolEncoder) ois.readObject();
			ois.close();
			ois = null;
			
			ConfusionMatrix polCm = new ConfusionMatrix();
			ConfusionMatrix intCm = new ConfusionMatrix();
			
			HashSet<String> selectedFiles = new HashSet();
			Util.readLines(docListFile, selectedFiles);
			LKReader r = new LKReader(dir, subjLexFile);
			while(r.hasNext()) {
				MPQAText text = r.next();
				if(selectedFiles.contains(text.source)) {
					System.out.print(".");
					System.out.flush();
					testText(text, enc, polCl, intCl, polCm, intCm);					
				}
			}
			System.out.println();
			
			System.out.println("Polarity:");
			System.out.format("Acc: %g\n", (double) nCorrPol / nExprs);
			System.out.format("Acc Pol/Nonpol: %g\n", (double) nCorrPN / nExprs);
			//System.out.println(polCm);
			
			System.out.println("\nIntensity:");
			System.out.format("Acc: %g\n", (double) nCorrInt / nExprs);
			//System.out.println(intCm);
			
		} catch(Exception e) {
			e.printStackTrace();
		}
	}

	private static void testText(MPQAText text, SymbolEncoder enc,
			Classifier<SparseVector, Integer> polCl,
			Classifier<SparseVector, Integer> intCl, 
			ConfusionMatrix polCm,
			ConfusionMatrix intCm) {

		for(MPQASentence sen: text.sentences) {
			for(MPQAExpression expr: sen.exprs) {

				if(expr.type.equals("os"))
					continue;
				
				SparseVector sv = represent(expr, enc);
				
				int polGuess = polCl.classify(sv);
				int intGuess = intCl.classify(sv);
	
				int polGold = getPolCode(expr.polarity);
				int intGold = getIntCode(expr.intensity);

				nExprs++;
				if(polGuess == polGold)
					nCorrPol++;
				if(intGuess == intGold)
					nCorrInt++;

				if(polGold == NEUTRAL && polGuess == NEUTRAL)
					nCorrPN++;
				if(polGold != NEUTRAL && polGuess != NEUTRAL)
					nCorrPN++;
				
				polCm.add(polGold, polGuess);
				intCm.add(intGold, intGuess);				
			}
		}
		
	}
	
	
	
	
	
	
	
	
	
}
