package mpqa_structlearn.sequences;

import se.lth.cs.nlp.depsrl.format.PAStructure;
import se.lth.cs.nlp.nlputils.ml_long.*;
import se.lth.cs.nlp.nlputils.annotations.Span;
import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.*;
import java.io.Serializable;
import java.util.*;

import mpqa_structlearn.io.*;

public class RerankingFE implements Serializable {

	private static final long serialVersionUID = 0;

	//SparseVector tmp = new SparseVector();

	private boolean useBigrams;
	private boolean useSyntax;
	private boolean useSpans;
	private boolean useSemantics;
	
	private static final boolean COUNT_INSIDE = true;

	private static final boolean USE_DISC_CONN = false; //true;
	
	//private boolean scoreSources;

	public RerankingFE(boolean useBigrams, boolean useSyntax, boolean useSpans, boolean useSemantics) {
		this.useBigrams = useBigrams;
		this.useSyntax = useSyntax;
		this.useSpans = useSpans;
		this.useSemantics = useSemantics;
	}

	public void extractFeatures(Candidate candidate, double baseScore, SparseVector sv) {

		sv.clear();
		sv.put(1, baseScore);		

		if(useBigrams)
			extractBigrams(candidate, sv);

		//extractUnigrams(candidate, sv);

		if(useSyntax)
			extractPaths(candidate, sv);

		if(useSpans)
			extractSpanFeatures(candidate, sv);

		if(useSemantics)
			extractSemPaths(candidate, sv);

		if(COUNT_INSIDE)
			countInsideTokens(candidate, sv);

		if(USE_DISC_CONN)
			extractDiscConnFeatures(candidate, sv);
		
		//sv.sortIndices();
	}
	
	private void countInsideTokens(Candidate c, SparseVector sv) {
		int nInside = 0;
		for(MPQAExpression ex: c.mpqaSentence.exprs)
			if(ex.type.matches("ds|es"))
				nInside += ex.span.size();			
		putFeature("N_INSIDE_TOKENS", sv, nInside);
	}
	
	private void countInsideTokensPol(Candidate c, SparseVector sv) {
		int nInsidePos = 0, nInsideNeg = 0, nInsideNeu = 0;
		for(MPQAExpression ex: c.mpqaSentence.exprs)
			if(ex.polarity != null) {
				if(ex.polarity.startsWith("p"))
					nInsidePos += ex.span.size();
				else if(ex.polarity.startsWith("neg"))
					nInsideNeg += ex.span.size();
				else if(ex.polarity.startsWith("neu"))
					nInsideNeu += ex.span.size();				
			}
		if(nInsidePos > 0)
			putFeature("N_INSIDE_POS", sv, nInsidePos);
		if(nInsideNeg > 0)
			putFeature("N_INSIDE_NEG", sv, nInsideNeg);
		if(nInsideNeu > 0)
			putFeature("N_INSIDE_NEU", sv, nInsideNeu);
	}

	public void extractSourceFeatures(Candidate c, SparseVector sv) {
		
		ArrayList<MPQAExpression> el = c.mpqaSentence.exprs;
		DepGraph dg = c.mpqaSentence.dg;
		putFeature("srcBase", sv, c.baseSrcScore);
		for(int i = 0; i < el.size(); i++) {
			MPQAExpression s1 = el.get(i);
			for(int j = i+1; j < el.size(); j++) {
				if(i == j)
					continue;
				MPQAExpression s2 = el.get(j);
				
				//DepNode[] ns = findClosestNodes(s1, s2, dg);
				//Object[] ns = findClosestNodes(s1, s2, dg);

				DepNode n1 = null, n2 = null, a = null;
				
				if(false) { 
					//if(true)
					//	throw new RuntimeException("unimplemented");
				    
					String depPath = depPath(n1, n2, a);
					String w1 = c.isWriter[i]? "t": "f";
					String w2 = c.isWriter[j]? "t": "f";
					String i1 = c.isImplicit[i]? "t": "f";
					String i2 = c.isImplicit[j]? "t": "f";
					String nn1 = c.srcIndex[i] > 0? "t": "f";
					String nn2 = c.srcIndex[j] > 0? "t": "f";

					String f2 = "src:" + s1.type + ":" + depPath + ":" + s2.type;
					f2 += w1 + w2 + i1 + i2 + nn1 + nn2;
					putFeature(f2, sv);

					String path1 = "<E>";
					if(c.srcIndex[i] > 0) {
						//Span s_src = null; //new Span();
						//s_src.tokenStart = c.srcIndex[i] - 1;
						//s_src.tokenEnd = c.srcIndex[i];
						MPQAExpression s_src = null;
						DepNode[] ns_src = findClosestNodesOld(s1, s_src, dg);
						path1 = depPath(ns_src[0], ns_src[1], ns_src[2]);
					}

					String f3 = "srcX:" + s1.type + ":" + depPath + ":" + path1 + ":" + s2.type;
					putFeature(f3, sv);
				}

				if(true) {
					if(true || a == n1) {

                        String shFeature = s1.type + /*":" + depPath +*/ ":" + s2.type; 
                        shFeature = "shs:" + sharedSource(c, i, j) + ":" + shFeature;

                        // Shared holders
                        if(true)
                            putFeature(shFeature, sv);

					    ArrayList<Object[]> paths = shortestPaths(s1, s2, dg);
					    
					    for(Object[] t: paths) {

					        //String depPath = depPath(n1, n2, a);
					        String depPathStr = ((DepPath) t[3]).relString();
					        
					        String w1 = c.isWriter[i]? "t": "f";
					        String w2 = c.isWriter[j]? "t": "f";
					        String i1 = c.isImplicit[i]? "t": "f";
					        String i2 = c.isImplicit[j]? "t": "f";
					        String nn1 = c.srcIndex[i] > 0? "t": "f";
					        String nn2 = c.srcIndex[j] > 0? "t": "f";

					        String xxxFeature = "src:" + s1.type + ":" + depPathStr + ":" + s2.type;
					        xxxFeature += w1 + w2 + i1 + i2 + nn1 + nn2;

					        // ?
					        if(true)
					            putFeature(xxxFeature, sv);

					        String htpFeature = "src2:" + s1.type + ":" + depPathStr + ":" + s2.type;
					        htpFeature += getSourceString(c, i, dg) + ":" + getSourceString(c, j, dg);

					        // Holder type + path?
					        if(true)
					            putFeature(htpFeature, sv);

					        if(false) {
					            if(a == n1) {
					                String domFeature = "dom" + s1.type + "->" + s2.type;
					                domFeature += w1 + w2 + i1 + i2 + n1 + n2;
					                putFeature(domFeature, sv);
					            } else if(a == n2) {
					                String domFeature = "dom" + s2.type + "->" + s1.type;
					                domFeature += w2 + w1 + i2 + i1 + n2 + n1;							
					                putFeature(domFeature, sv);
					            }
					        }
					    }
						
					} else if(a == n2) {
						// OBS DETTA UTFÖRS INTE!						
						
						String depPath = depPath(n2, n1, a);
						String shFeature = s2.type + /*":" + depPath +*/ ":" + s1.type; 
						shFeature = "shs:" + sharedSource(c, i, j) + ":" + shFeature;
						putFeature(shFeature, sv);
						
						String w2 = c.isWriter[i]? "t": "f";
						String w1 = c.isWriter[j]? "t": "f";
						String i2 = c.isImplicit[i]? "t": "f";
						String i1 = c.isImplicit[j]? "t": "f";
						String nn2 = c.srcIndex[i] > 0? "t": "f";
						String nn1 = c.srcIndex[j] > 0? "t": "f";

						String xxxFeature = "src:" + s2.type + ":" + depPath + ":" + s1.type;
						xxxFeature += w1 + w2 + i1 + i2 + nn1 + nn2;

						putFeature(xxxFeature, sv);

						String htpFeature = "src2:" + s2.type + ":" + depPath + ":" + s1.type;
						htpFeature += getSourceString(c, j, dg) + ":" + getSourceString(c, i, dg);
						putFeature(htpFeature, sv);

					} else
						continue;
				}


			}
		}

		if(true) { /* "semantic path": exprs are args of the same pred */
			HashMap<DepNode, MPQAExpression> nodeToExpr = new HashMap();
			for(MPQAExpression ex: c.mpqaSentence.exprs)
				for(DepNode n: ex.span)
					nodeToExpr.put(n, ex);
			
			if(false)
				for(PAStructure pa: c.mpqaSentence.pas) {

					for(int i = 0; i < pa.args.size(); i++) {
						DepNode n1 = pa.args.get(i);
						MPQAExpression e1 = nodeToExpr.get(n1);
						if(e1 == null)
							for(DepNode n1b: n1.children) {
								if(pa.args.contains(n1b))
									continue;
								e1 = nodeToExpr.get(n1);
								if(e1 != null)
									break;
							}
						if(e1 == null)
							continue;
						String l1 = pa.argLabels.get(i);

						int ix1 = c.mpqaSentence.exprs.indexOf(e1);

						for(int j = i + 1; j < pa.args.size(); j++) {
							DepNode n2 = pa.args.get(j);
							MPQAExpression e2 = nodeToExpr.get(n2);
							if(e2 == null)
								for(DepNode n2b: n2.children) {
									if(pa.args.contains(n2b))
										continue;
									e2 = nodeToExpr.get(n2);
									if(e2 != null)
										break;
								}
							if(e2 == null)
								continue;
							String l2 = pa.argLabels.get(j);

							int ix2 = c.mpqaSentence.exprs.indexOf(e2);						

							putFeature("shs_sempath:" + l1 + ":" + l2 + ":" + sharedSource(c, ix1, ix2), sv);
						}
					}

				}
			if(false)
				for(PAStructure pa: c.mpqaSentence.pas) {
					MPQAExpression e1 = nodeToExpr.get(pa.pred);
					if(e1 == null)
						continue;
					int ix1 = c.mpqaSentence.exprs.indexOf(e1);
					for(int i = 0; i < pa.args.size(); i++) {
						DepNode n2 = pa.args.get(i);
						MPQAExpression e2 = nodeToExpr.get(n2);
						if(e2 == null)
							for(DepNode n2b: n2.children) {
								if(pa.args.contains(n2b))
									continue;
								e2 = nodeToExpr.get(n2);
								if(e2 != null)
									break;
							}
						if(e2 == null)
							continue;
						String l2 = pa.argLabels.get(i);
						int ix2 = c.mpqaSentence.exprs.indexOf(e2);
						putFeature("shs_consem:" + e1.type + ":" + e2.type + ":" + l2 + ":" + sharedSource(c, ix1, ix2), sv);
					}

				}
		}
		
		//sv.sortIndices();
	}

	public static boolean debug = false;
	
	public void extractPolarityFeatures(Candidate c, SparseVector sv) {
		ArrayList<MPQAExpression> el = c.mpqaSentence.exprs;

		DepGraph dg = c.mpqaSentence.dg;
		
		if(true)
			putFeature("polBase", sv, c.basePolScore);
		
		//if(COUNT_INSIDE)
		//	countInsideTokensPol(c, sv);
		
		int nop = 0;
		for(MPQAExpression e: el)
			if(!e.type.equals("os"))
				nop++;

		if(debug) {
			System.out.println("Expressions:");
			System.out.println(el);
		}

		//if(true)
		//	return;
		
		HashMap<DepNode, MPQAExpression> inverse = new HashMap();
		for(MPQAExpression expr: c.mpqaSentence.exprs)
			for(DepNode n: expr.span)
				inverse.put(n, expr);
		
		for(int i = 0; i < el.size(); i++) {
			MPQAExpression s1 = el.get(i);
			String pol1 = s1.polarity;
			if(pol1 == null && !s1.type.equals("os"))
				throw new RuntimeException("no polarity assigned");
			
			for(int j = 0 /*i+1*/; j < el.size(); j++) {
				if(i == j)
					continue;
				MPQAExpression s2 = el.get(j);

				// 101102
				//if(j > i + 1)
				//	break;
				// end 101102
				
				String pol2 = s2.polarity;
				if(pol2 == null && !s2.type.equals("os"))
					throw new RuntimeException("no polarity assigned");

				if(pol1 != null && pol2 != null) {

					// 1. par av polariteter
					if(true) { // default TRUE
						String pairF = "polp:" + pol1 + ":" + pol2;
						putFeature(pairF, sv);
					}

					//DepNode[] ns = findClosestNodes(s1, s2, dg);
					ArrayList<Object[]> paths = shortestPaths(s1, s2, dg);
					for(Object[] t: paths) {
					    //String depPath = depPath(ns[0], ns[1], ns[2]);
					    String depPathStr = ((DepPath) t[3]).relString();
					    
					    // 2. par + path
					    if(true) { // default TRUE
					        String polPathF = "polpath:" + pol1 + ":" + depPathStr + ":" + pol2;
					        putFeature(polPathF, sv);
					    }
					}

                    DepNode n1 = (DepNode) paths.get(0)[0];
                    DepNode n2 = (DepNode) paths.get(0)[1];
                    DepNode a = (DepNode) paths.get(0)[2];
					
					// 3. par + dom
					if(true) { // default TRUE

						if(n1 == a) {// ns[0] dominates ns[1]
							boolean foundIntervening = false;
		
							/*
							DepNode n = ns[1].parents[0];
							while(n != ns[0]) {
								MPQAExpression intExpr = inverse.get(n);
								if(intExpr != s1 && intExpr != s2) {
									foundIntervening = true;
									break;
								}
								n = n.parents[0];
							}*/
							
							foundIntervening = hasInterveningExpression(inverse,
							        s1, s2, n2, n1, null);
							
							if(!foundIntervening) {
								putFeature("pdom:" + pol1 + "->" + pol2, sv);
								putFeature("pdom:" + pol1 + "/" + n1.word.toLowerCase() 
										+ "->" + pol2, sv);
								putFeature("pdom:" + pol1 + "->" + pol2 
										+ "/" + n2.word.toLowerCase(), sv);
								putFeature("pdom:" + pol1 + "/" + n1.word.toLowerCase()  
										+ "->" + pol2 + "/" + n2.word.toLowerCase(), sv);
							}
								
						} else if(n2 == a) { // ns[1] dominates ns[0]
							boolean foundIntervening = false;

							foundIntervening = hasInterveningExpression(inverse,
                                    s1, s2, n1, n2, null);
							
							/*DepNode n = ns[0].parents[0];
							while(n != ns[1]) {
								MPQAExpression intExpr = inverse.get(n);
								if(intExpr != s1 && intExpr != s2) {
									foundIntervening = true;
									break;
								}
								n = n.parents[0];								
							}*/
							
							if(!foundIntervening) {
								putFeature("pdom:" + pol2 + "->" + pol1, sv);
								putFeature("pdom:" + pol2 + "/" + n2.word.toLowerCase() 
										+ "->" + pol1, sv);
								putFeature("pdom:" + pol2 + "->" + pol1 
										+ "/" + n1.word.toLowerCase(), sv);
								putFeature("pdom:" + pol2 + "/" + n2.word.toLowerCase()  
										+ "->" + pol1 + "/" + n1.word.toLowerCase(), sv);
							}
						}
							
						/*if(depPath.startsWith("|"))
							putFeature("pdom:" + pol1 + "->" + pol2, sv);
						else if(depPath.endsWith("|"))
							putFeature("pdom:" + pol2 + "->" + pol1, sv);*/
					}

					// 4. par + typer
					if(false) { // default FALSE
						String polTypesF = "poltypes:" + pol1 + "/" + s1.type + ":" + pol2 + "/" + s2.type;
						putFeature(polTypesF, sv);
					}

					// 5. par + typer + path
					if(false) { // default FALSE
					    String depPathStr = null;
					    if(true)
					        throw new RuntimeException("unimplemented");
						String polTypesPathF = "poltypespath:" + pol1 + "/" + s1.type + ":" + pol2 + "/" + s2.type + ":" + depPathStr;
						putFeature(polTypesPathF, sv);
					}
					
					// 6. par + lex
					if(true) { // default TRUE
						String polLex = "pollex:" + pol1 + "/" + n1.word.toLowerCase()
						+ ":" + pol2 + "/" + n2.word.toLowerCase();
						putFeature(polLex, sv);
					}

					/// NYA FEATURES 101102
					if(true) {

						/* ord mellan exprs */
						if(false) { // default FALSE
							if(j == i + 1) {
								DepNode r1 = s1.span.get(s1.span.size()-1);
								DepNode l2 = s2.span.get(0);
								if(l2.position <= r1.position)
									throw new RuntimeException("!!!");
								for(int k = r1.position + 1; k < l2.position; k++) {
									String polWordBetween = "pol_betw:" + pol1 + "/"
									+ dg.nodes[k].word.toLowerCase() + "/" + pol2;
									putFeature(polWordBetween, sv);
								}
							}
						}
						
						/* ord längs path */
						if(false) { // default FALSE

						    /*
						    if(j == i + 1) {
								DepNode n = ns[0];
								if(n != ns[2]) {
									n = n.parents[0];
									while(n != ns[2]) {
										String pwInPath = "pol_inpath:" + pol1 + "/"
										+ n.word.toLowerCase() + "/" + pol2;
										putFeature(pwInPath, sv);
										n = n.parents[0];
									}
									String pwInPath = "pol_inpath:" + pol1 + "/"
									+ n.word.toLowerCase() + "/" + pol2;
									putFeature(pwInPath, sv);
								}
								n = ns[1];
								if(n != ns[2]) {
									n = n.parents[0];
									while(n != ns[2]) {
										String pwInPath = "pol_inpath:" + pol1 + "/"
										+ n.word.toLowerCase() + "/" + pol2;
										putFeature(pwInPath, sv);
										n = n.parents[0];
									}
								}
							}
							*/
						}
						
					}					
					
				}
				
			}
		}
		
		if(false) { // default FALSE
			for(PAStructure pa: c.mpqaSentence.pas) {
				for(MPQAExpression s: c.mpqaSentence.exprs) {
					if(s.type.equals("os"))
						continue;
					if(s.span.contains(pa.pred)) {
						for(int i = 0; i < pa.argLabels.size(); i++) {
							String al = pa.argLabels.get(i);
							DepNode arg = pa.args.get(i);
							for(MPQAExpression s2: c.mpqaSentence.exprs) {
								if(s == s2)
									continue;
								if(s2.type.equals("os"))
									continue;
								if(s2.span.contains(arg)) {
									// CONNECTING ARGLABEL
									putFeature("prd_al_l_POL:" + s.polarity + /*":" + pa.lemma +*/ ":" + al + ":" + s2.polarity, sv);
									break;
								}
							}
						}

						break;
					}
				}
			}
		}
		
		

	}
	
	private String getSourceString(Candidate c, int i, DepGraph dg) {
		if(c.isWriter[i])
			return "W";
		if(c.isImplicit[i])
			return "I";
		if(c.srcIndex[i] > 0)
			return dg.nodes[c.srcIndex[i]].word.toLowerCase();
		return "E";
	}

	private boolean sharedSource(Candidate c, int i, int j) {
		if(c.isWriter[i])
			return c.isWriter[j];
		if(c.isImplicit[i])
			return c.isImplicit[j];
		if(c.srcIndex[i] > 0)
			return c.srcIndex[i] == c.srcIndex[j];
		return false;
	}

	private static void extractBigrams(Candidate c, SparseVector sv) {
		ArrayList<MPQAExpression> el = c.mpqaSentence.exprs;
		if(false) {
			String prev = "<START>";
			for(MPQAExpression s: el) {
				String bigram = "BG:" + prev + "|" + s.type;
				sv.put(hash(bigram), 1.0);
				prev = s.type;
			}
			String bigram = "BG:" + prev + "|<END>";
			sv.put(hash(bigram), 1.0);
		} else {
			for(int i = 0; i < el.size(); i++) {
				for(int j = 0; j < el.size(); j++) {
					if(i == j)
						continue;
					MPQAExpression s1 = el.get(i);
					MPQAExpression s2 = el.get(j);
					String bigram;
					if(s1.type.compareTo(s2.type) < 0)
						bigram = s1.type + "_" + s2.type;
					else
						bigram = s2.type + "_" + s2.type;
					sv.put(hash("BG2:" + bigram), 1.0);
				}
			}
		}
	}

	private static void extractUnigrams(Candidate c, SparseVector sv) {
		if(false) {
			for(MPQAExpression s: c.mpqaSentence.exprs) {
				String unigram = "UG:" + s.type;
				sv.put(hash(unigram), 1.0);
			}
		} else {
			for(MPQAExpression s: c.mpqaSentence.exprs) {
				String unigram = "UG2";
				sv.put(hash(unigram), 1.0);
			}
		}
	}

	private static IntObjPair<DepNode> findCommonAncestor(DepNode n1, DepNode n2) {
		HashMap<DepNode, Integer> m = new HashMap();
		int steps = 0;
		while(true) {
			m.put(n1, steps);
			if(n1.parents.length == 0)
				break;
			else
				n1 = n1.parents[0];
			steps++;
		}

		steps = 0;
		while(true) {
			Integer otherSteps = m.get(n2);
			if(otherSteps != null)
				return new IntObjPair(otherSteps + steps, n2);
			if(n2.parents.length == 0)
				break;
			else
				n2 = n2.parents[0];			
			steps++;
		}
		
		/* EPE: */
		System.out.println(n1);
		System.out.println(n2);
		
		throw new RuntimeException("coulnd't find ancestor");
	}

	/*private static IntObjPair<DepNode> findCommonAncestor(DepNode n1, DepNode n2) {
	    DepPath p = EPEPaths.treePath(n1, n2)[0];
	    
	    throw new RuntimeException("coulnd't find ancestor");
    }*/
	
	private static DepNode[] findClosestNodesOld(MPQAExpression s1, MPQAExpression s2, 
			DepGraph dg) {
		int minLength = Integer.MAX_VALUE;
		DepNode bestN1 = null, bestN2 = null, bestAnc = null;
		for(DepNode n1: s1.span) {
			for(DepNode n2: s2.span) {
				IntObjPair<DepNode> p = findCommonAncestor(n1, n2); 
				if(p.left < minLength) {
					minLength = p.left;
					bestAnc = p.right;
					bestN1 = n1;
					bestN2 = n2;
				}				
			}
		}
		return new DepNode[] { bestN1, bestN2, bestAnc };				
	}

    private static final int MAX_PATHS = 4;
    private static final int MAX_PATH_LEN = 8;
	
	private static ArrayList<Object[]> shortestPaths(MPQAExpression s1, MPQAExpression s2, 
            DepGraph dg) {  
        double minLength = Double.MAX_VALUE;

        ArrayList<DepPath> bestPaths = new ArrayList();
        
        for(DepNode n1: s1.span) {
            for(DepNode n2: s2.span) {
                
                double upperBound = Double.min(MAX_PATH_LEN, minLength);
                
                //DepPath[] ps = EPEPaths.treePath(n1, n2);
                DepPath[] ps = EPEPaths.findPaths(n1, n2, MAX_PATHS, upperBound);
                
                if(Util.equal(minLength, ps[0].cost)) {
                    for(DepPath p: ps)
                        if(p != null && bestPaths.size() < MAX_PATHS && Util.equal(minLength, p.cost))
                            bestPaths.add(p);
                        else
                            break;
                } else if(ps[0].cost < minLength) {
                    minLength = ps[0].cost;
                    bestPaths.clear();
                    for(DepPath p: ps)
                        if(p != null && bestPaths.size() < MAX_PATHS && Util.equal(minLength, p.cost))
                            bestPaths.add(p);
                        else
                            break;
                }
            }
        }
        
        /* TODO what if multiple paths? */
        if(bestPaths.isEmpty())
            throw new RuntimeException("?");
        
        /*
        if(bestPaths.size() != 1) {
            System.out.println(dg);
            System.out.println(s1.toStringDetailed() + " -- " + s2.toStringDetailed());
            for(DepPath p: bestPaths) {
                System.out.println(p);
            }
            System.exit(0);
        }
        */
        /*
        for(DepPath p: bestPaths) {
            if(p.prevRel.equals("ROOT")) {
                System.out.println(dg);
                System.out.println(s1.toStringDetailed() + " -- " + s2.toStringDetailed());
                for(DepPath pp: bestPaths) {
                    System.out.println(pp);
                }
                System.exit(0);
            }
        }
        */
        
        ArrayList<Object[]> out = new ArrayList();

        for(DepPath p: bestPaths) {
            DepPath pp = p;
            DepNode n2 = p.n;
            while(p.prev != null && !p.isParent)
                p = p.prev;
            DepNode a = p.n;
            while(p.prev != null)
                p = p.prev;
            DepNode n1 = p.n;
            out.add(new Object[] { n1, n2, a, pp });
        }
        return out;
    }
	
	private static String depPath(DepNode n1, DepNode n2, DepNode ancestor) {		
	    StringBuilder sb = new StringBuilder();		
		while(n1 != ancestor) {
			sb.append(n1.relations[0] + "_");
			n1 = n1.parents[0];
		}		
		sb.append("|");		
		while(n2 != ancestor) {
			sb.append(n2.relations[0] + "_");
			n2 = n2.parents[0];
		}
		return sb.toString();		
	}

    // ns[1] == from
    // ns[0] == to
	private static boolean hasInterveningExpression(HashMap<DepNode, MPQAExpression> inverse,
            MPQAExpression s1, MPQAExpression s2, DepNode from, DepNode to, HashSet<DepNode> seen) {
	    if(seen == null)
	        seen = new HashSet();
	    if(seen.contains(from))
	        return false;
	    else
	        seen.add(from);
	    if(from == to)
	        return false;
	    if(CollectionUtils.contains(from.parents, to))
	        return false;
	    for(DepNode n: from.parents) {
	        MPQAExpression intExpr = inverse.get(n);
	        if(intExpr != s1 && intExpr != s2)
	            return true;
	        if(hasInterveningExpression(inverse, s1, s2, n, to, seen))
	            return true;
	    }
	    return false;
    }

	private static boolean dominatesDirectly(MPQAExpression e1, MPQAExpression e2) {
	    HashSet<DepNode> e1ns = new HashSet(e1.span);
	    for(DepNode n: e2.span)
	        for(DepNode p: n.parents)
	            if(e1ns.contains(p))
	                return true;
	    return false;
	}
	
	private static void extractPaths(Candidate c, SparseVector sv) {
		ArrayList<MPQAExpression> el = c.mpqaSentence.exprs;
		DepGraph dg = c.mpqaSentence.dg;
		for(int i = 0; i < el.size(); i++) {
			MPQAExpression s1 = el.get(i);
			for(int j = i + 1; j < el.size(); j++) {
				MPQAExpression s2 = el.get(j);

				//DepNode[] ns = findClosestNodes(s1, s2, dg);
				//Object[] ns = findClosestNodes(s1, s2, dg);
				//DepNode n1 = (DepNode) ns[0];
				//DepNode n2 = (DepNode) ns[1];
				//String depPath = depPath(ns[0], ns[1], ns[2]);
				
				//DepPath path = (DepPath) ns[3];

				ArrayList<Object[]> paths = shortestPaths(s1, s2, dg);
				for(Object[] t: paths) {
				    String depPathStr = ((DepPath) t[3]).relString();
				    DepNode n1 = (DepNode) t[0]; DepNode n2 = (DepNode) t[1];
				    String depPathFeature = "p:" + s1.type + "-" + depPathStr + "-" + s2.type;
				    putFeature(depPathFeature, sv);	

				    //System.out.println(depPathFeature);
				    
				    /*if(depPathStr.equals("-[OBJ]->")) { 
				        System.out.println(dg);
				        System.out.println(s1.toStringDetailed() + " / " + s2.toStringDetailed());
				        System.exit(0);
				    }*/
				    
				    String depPathWordFeature = "pw:" + s1.type + "|" + n1.word.toLowerCase() + "|" + depPathStr + "|" + n2.word.toLowerCase() + "|" + s2.type;
				    putFeature(depPathWordFeature, sv);
				}

				DepNode n1 = (DepNode) paths.get(0)[0]; DepNode n2 = (DepNode) paths.get(0)[0];
				
				String dominanceFeature = null;
				//String dominanceFeatureUnlex = null;
				// ancestor contained in s1 ==> s1 dominates s2
				if(/*ns[2] == ns[0]*/ dominatesDirectly(s1, s2)) {
					dominanceFeature = "dom_w:" + s1.type + "/" + n1.word.toLowerCase() + "->" + s2.type + "/" + n2.word.toLowerCase();
					//dominanceFeatureUnlex = "dom_u:" + s1.type + "->" + s2.type;

                    /*System.out.println(dg);
                    System.out.println(s1.toStringDetailed() + " / " + s2.toStringDetailed());
                    System.exit(0);*/                   
				}
				else if(/*ns[2] == ns[1]*/ dominatesDirectly(s2, s1)) {
					dominanceFeature = "dom_w:" + s2.type + "/" + n2.word.toLowerCase() + "->" + s1.type + "/" + n1.word.toLowerCase();
					//dominanceFeatureUnlex = "dom_u:" + s2.type + "->" + s1.type;
					/*System.out.println(dg);
                    System.out.println(s1.toStringDetailed() + " / " + s2.toStringDetailed());
                    System.exit(0);*/                   
				}
				if(dominanceFeature != null) {
					putFeature(dominanceFeature, sv);
					//putFeature(dominanceFeatureUnlex, sv);
				}

				/* ord mellan exprs */
				if(false) {
					if(j == i + 1) {
						DepNode r1 = s1.span.get(s1.span.size()-1);
						DepNode l2 = s2.span.get(0);
						if(l2.position <= r1.position)
							throw new RuntimeException("!!!");
						for(int k = r1.position + 1; k < l2.position; k++) {
							String polWordBetween = "w_betw:" + s1.type + "/" 
							+ dg.nodes[k].word.toLowerCase() + s2.type;
							putFeature(polWordBetween, sv);
						}
					}
				}
				
				/* ord längs path */
				if(false) {

				    
				    /*
				    if(j == i + 1) {
						DepNode n = ns[0];
						if(n != ns[2]) {
							n = n.parents[0];
							while(n != ns[2]) {
								String pwInPath = "w_inpath:" + s1.type + "/"
								+ n.word.toLowerCase() + "/" + s2.type;
								putFeature(pwInPath, sv);
								n = n.parents[0];
							}
							String pwInPath = "w_inpath:" + s1.type + "/"
							+ n.word.toLowerCase() + "/" + s2.type;
							putFeature(pwInPath, sv);
						}
						n = ns[1];
						if(n != ns[2]) {
							n = n.parents[0];
							while(n != ns[2]) {
								String pwInPath = "w_inpath:" + s1.type + "/"
								+ n.word.toLowerCase() + "/" + s2.type;
								putFeature(pwInPath, sv);
								n = n.parents[0];
							}
						}
					}
					*/
				}
				
				
			}
		}
	}

	private static void extractSpanFeatures(Candidate c, SparseVector sv) {
		for(MPQAExpression ex: c.mpqaSentence.exprs) {
			ArrayList<DepNode> ns = ex.span;
			String startWord = ns.get(0).word.toLowerCase();
			String endWord = ns.get(ns.size() - 1).word.toLowerCase();
			putFeature("st:" + ex.type + ":" + startWord, sv);
			putFeature("en:" + ex.type + ":" + endWord, sv);
			putFeature("st_en:" + ex.type + ":" + startWord + ":" + endWord, sv);
		}
	}

	private static void extractSemPaths(Candidate c, SparseVector sv) {

		HashMap<DepNode, MPQAExpression> nodeToExpr = new HashMap();
		for(MPQAExpression ex: c.mpqaSentence.exprs)
			for(DepNode n: ex.span)
				nodeToExpr.put(n, ex);

		for(PAStructure pa: c.mpqaSentence.pas) {
			int predPosition = pa.pred.position - 1;
			for(MPQAExpression s: c.mpqaSentence.exprs) 				
				if(s.span.contains(pa.pred)) {

					// PREDICATE
					putFeature("prd:" + s.type + ":" + pa.lemma, sv);

					for(int i = 0; i < pa.argLabels.size(); i++) {
						String al = pa.argLabels.get(i);

						// PREDICATE + ARGLABEL
						putFeature("prd_al:" + s.type + ":" + pa.lemma + ":" + al, sv);

						DepNode arg = pa.args.get(i);
						//int argPosition = pa.args.get(i).position - 1;
						for(MPQAExpression s2: c.mpqaSentence.exprs) // todo kolla också att det är ett annat arg?
							if(s2.span.contains(arg)) {
								//putFeature("prd_al_l:" + s.label + ":" + pa.lemma + ":" + al + ":" + s2.label, sv);

								// CONNECTING ARGLABEL
								putFeature("prd_al_l:" + s.type + /*":" + pa.lemma +*/ ":" + al + ":" + s2.type, sv);

								break;
							}
					}

					if(true)
						for(MPQAExpression s2: c.mpqaSentence.exprs) {
							if(s == s2)
								continue;

							// check whether s2 is dominated by some argument

							DepNode[] ns = findClosestNodesOld(s, s2, c.mpqaSentence.dg);
							//return new DepNode[] { bestN1, bestN2, bestAnc };
							DepNode n1= ns[0];
							DepNode n2 = ns[1];
							DepNode anc = ns[2];

							//boolean found = pa.args.contains(n2);
							int ix = pa.args.indexOf(n2);
							while(ix == -1) {
								if(n2 != anc) {
									n2 = n2.parents[0];
									MPQAExpression s_n2 = nodeToExpr.get(n2);
									if(s_n2 != null && s_n2 != s && s_n2 != s2)
										break;
									ix = pa.args.indexOf(n2);
								} else
									break;
							}

							while(n1 != anc) {
								n1 = n1.parents[0];
								MPQAExpression s_n1 = nodeToExpr.get(n1);
								if(s_n1 != null && s_n1 != s && s_n1 != s2) {
									ix = -1;
									break;
								}
							}

							if(ix == -1)
								continue;

							String al = pa.argLabels.get(ix);

							// CONNECTING ARGLABEL VARIANT 2
							putFeature("prd_al2_l:" + s.type + /*":" + pa.lemma +*/ ":" + al + ":" + s2.type, sv);

						}					

					break;

				}
		}
		
		if(false) /* "semantic path" */
			for(PAStructure pa: c.mpqaSentence.pas) {

				for(int i = 0; i < pa.args.size(); i++) {
					DepNode n1 = pa.args.get(i);
					MPQAExpression e1 = nodeToExpr.get(n1);
					if(e1 == null)
						for(DepNode n1b: n1.children) {
							if(pa.args.contains(n1b))
								continue;
							e1 = nodeToExpr.get(n1);
							if(e1 != null)
								break;
						}
					if(e1 == null)
						continue;
					String l1 = pa.argLabels.get(i);

					for(int j = i + 1; j < pa.args.size(); j++) {
						DepNode n2 = pa.args.get(j);
						MPQAExpression e2 = nodeToExpr.get(n2);
						if(e2 == null)
							for(DepNode n2b: n2.children) {
								if(pa.args.contains(n2b))
									continue;
								e2 = nodeToExpr.get(n2);
								if(e2 != null)
									break;
							}
						if(e2 == null)
							continue;
						String l2 = pa.argLabels.get(j);

						putFeature("sempath:" + e1.type + ":" + l1 + ":" + l2 + ":" + e2.type, sv);

					}
				}

			}
		
	}
	
	private void extractDiscConnFeatures(Candidate c, SparseVector sv) {

		if(c.mpqaSentence.discConns == null)
			throw new IllegalArgumentException("No connectives!");			
		
		HashMap<DepNode, String> nodeToSense = new HashMap();
		for(Span sp: c.mpqaSentence.discConns)
			for(int i = sp.tokenStart; i < sp.tokenEnd; i++)
				nodeToSense.put(c.mpqaSentence.dg.nodes[i+1], sp.label);

		ArrayList<MPQAExpression> el = c.mpqaSentence.exprs;

		DepGraph dg = c.mpqaSentence.dg;

		boolean found = false;

		//debug = true;
		
		for(int i = 0; i < el.size(); i++) {
			MPQAExpression s1 = el.get(i);

			String pol1 = s1.polarity;
			if(pol1 == null && !s1.type.equals("os"))
				throw new RuntimeException("no polarity assigned");

			for(int j = 0 /*i+1*/; j < el.size(); j++) {
				if(i == j)
					continue;
				MPQAExpression s2 = el.get(j);

				// 101102
				//if(j > i + 1)
				//	break;
				// end 101102

				String pol2 = s2.polarity;
				if(pol2 == null && !s2.type.equals("os"))
					throw new RuntimeException("no polarity assigned");

				DepNode[] ns = findClosestNodesOld(s1, s2, dg);
				
				String lm0 = ns[0].lemma.toLowerCase();
				String lm1 = ns[1].lemma.toLowerCase();
				
				//String depPath = depPath(ns[0], ns[1], ns[2]);
				
				if(true) { //pol1 != null && pol2 != null) {

					/* discourse connectives mellan exprs */
					if(true) {
						if(j == i + 1) {
							DepNode r1 = s1.span.get(s1.span.size()-1);
							DepNode l2 = s2.span.get(0);
							if(l2.position <= r1.position)
								throw new RuntimeException("!!!");
							for(int k = r1.position + 1; k < l2.position; k++) {

								DepNode n = dg.nodes[k];
								String sense = nodeToSense.get(n);
								if(sense != null) {
									putFeature("dc_sense_betw:" + s1.type + "/" + sense + "/" + s2.type, sv);
									String polSenseBetween = "dc_pol_sense_betw:" + pol1 + "/"
									+ sense + "/" + pol2;
									putFeature(polSenseBetween, sv);
									String polTypeSenseBetween = "dc_pol_type_sense_betw:" + pol1 + "/"
									+ sense + "/" + pol2 + "/" + s1.type + "/" + s2.type;
									putFeature(polTypeSenseBetween, sv);										

									found = false; //true;
								}

							}
						}
					}

					/* längs path */
					if(true) {
						if(true || j == i + 1) {
							DepNode n = ns[0];
							if(n != ns[2]) {
								n = n.parents[0];
								while(n != ns[2]) {
									String sense = nodeToSense.get(n);
									if(sense != null) {										
										putFeature("edct:" + s1.type + "/" + s2.type + "/" + sense, sv);
										putFeature("pdcp:" + pol1 + "/" + pol2 + "/" + sense, sv);
										putFeature("edctw:" + s1.type + "/" + s2.type + "/" + sense
												+ "/" + n.word.toLowerCase(), sv);
										putFeature("pdcpw:" + pol1 + "/" + pol2 + "/" + sense
												+ "/" + n.word.toLowerCase(), sv);
										putFeature("epdcpw:" + s1.type + "/" + pol1 
												+ "/" + s2.type + "/" + pol2 
												+ "/" + sense + "/" + n.word.toLowerCase(), sv);

										putFeature("pdcplex:" + pol1 + "/" + pol2 
												+ "/" + sense
												+ "/" + lm0 + "/" + lm1, sv);
										putFeature("pdcpwlex:" + pol1 + "/" + pol2 
												+ "/" + sense + "/" + n.word.toLowerCase()
												+ "/" + lm0 + "/" + lm1, sv);

//										putFeature("edct_p:" + s1.type + "/" + s2.type + "/" + sense + "/" + depPath, sv);
//										putFeature("pdcp_p:" + pol1 + "/" + pol2 + "/" + sense + "/" + depPath, sv);										
//										putFeature("edctw_p:" + s1.type + "/" + s2.type + "/" + sense
//												+ "/" + n.word.toLowerCase() + "/" + depPath, sv);
//										putFeature("pdcpw_p:" + pol1 + "/" + pol2 + "/" + sense
//												+ "/" + n.word.toLowerCase() + "/" + depPath, sv);
										
										found = true;
									}
									n = n.parents[0];
								}
								String sense = nodeToSense.get(n);
								if(sense != null) {
									putFeature("edct:" + s1.type + "/" + s2.type + "/" + sense, sv);
									putFeature("pdcp:" + pol1 + "/" + pol2 + "/" + sense, sv);
									putFeature("edctw:" + s1.type + "/" + s2.type + "/" + sense
											+ "/" + n.word.toLowerCase(), sv);
									putFeature("pdcpw:" + pol1 + "/" + pol2 + "/" + sense
											+ "/" + n.word.toLowerCase(), sv);
									putFeature("epdcpw:" + s1.type + "/" + pol1 
											+ "/" + s2.type + "/" + pol2 
											+ "/" + sense + "/" + n.word.toLowerCase(), sv);
									putFeature("pdcplex:" + pol1 + "/" + pol2 
											+ "/" + sense
											+ "/" + lm0 + "/" + lm1, sv);
									putFeature("pdcpwlex:" + pol1 + "/" + pol2 
											+ "/" + sense + n.word.toLowerCase()
											+ "/" + lm0 + "/" + lm1, sv);

//									putFeature("edct_p:" + s1.type + "/" + s2.type + "/" + sense + "/" + depPath, sv);
//									putFeature("pdcp_p:" + pol1 + "/" + pol2 + "/" + sense + "/" + depPath, sv);										
//									putFeature("edctw_p:" + s1.type + "/" + s2.type + "/" + sense
//											+ "/" + n.word.toLowerCase() + "/" + depPath, sv);
//									putFeature("pdcpw_p:" + pol1 + "/" + pol2 + "/" + sense
//											+ "/" + n.word.toLowerCase() + "/" + depPath, sv);
									
									found = true;
								}

							}
							n = ns[1];
							if(n != ns[2]) {
								n = n.parents[0];
								while(n != ns[2]) {
									String sense = nodeToSense.get(n);
									if(sense != null) {
										putFeature("edct:" + s1.type + "/" + s2.type + "/" + sense, sv);
										putFeature("pdcp:" + pol1 + "/" + pol2 + "/" + sense, sv);
										putFeature("edctw:" + s1.type + "/" + s2.type + "/" + sense
												+ "/" + n.word.toLowerCase(), sv);
										putFeature("pdcpw:" + pol1 + "/" + pol2 + "/" + sense
												+ "/" + n.word.toLowerCase(), sv);
										putFeature("epdcpw:" + s1.type + "/" + pol1 
												+ "/" + s2.type + "/" + pol2 
												+ "/" + sense + "/" + n.word.toLowerCase(), sv);
										putFeature("pdcplex:" + pol1 + "/" + pol2 
												+ "/" + sense
												+ "/" + lm0 + "/" + lm1, sv);
										putFeature("pdcpwlex:" + pol1 + "/" + pol2 
												+ "/" + sense + n.word.toLowerCase()
												+ "/" + lm0 + "/" + lm1, sv);

//										putFeature("edct_p:" + s1.type + "/" + s2.type + "/" + sense + "/" + depPath, sv);
//										putFeature("pdcp_p:" + pol1 + "/" + pol2 + "/" + sense + "/" + depPath, sv);										
//										putFeature("edctw_p:" + s1.type + "/" + s2.type + "/" + sense
//												+ "/" + n.word.toLowerCase() + "/" + depPath, sv);
//										putFeature("pdcpw_p:" + pol1 + "/" + pol2 + "/" + sense
//												+ "/" + n.word.toLowerCase() + "/" + depPath, sv);
										
										found = true;
									}
									n = n.parents[0];
								}
							}
						}
					}
				}
			}
		}
		
		if(false && found) {
			System.out.println(dg);
//			System.out.println(c.mpqaSentence.discConns);
			System.out.println(nodeToSense);
			System.out.println(el);
			System.out.println();
		}
		
		debug = false;
		
	}

	private static SymbolEncoder enc = new SymbolEncoder();
	static {
		if(enc.encode("(base model score)") != 1)
			throw new RuntimeException("!!!");
	}

	private static void putFeature(String f, SparseVector sv) {
		if(debug)
			System.out.println(f + ": 1.0");
		if(Reranker.PRINT_TOP_FEATURES) {
			int encoded = enc.encode(f);
			sv.put(encoded, 1.0);
		} else
			sv.put(hash(f), 1.0);	
	}
	
	private static void putFeature(String f, SparseVector sv, double val) {
		if(debug)
			System.out.println(f + ": " + val);
		if(Reranker.PRINT_TOP_FEATURES) {
			int encoded = enc.encode(f);
			sv.put(encoded, val);
		} else
			sv.put(hash(f), val);	
	}

	public static long getMaxFeatureIndex() {
		if(!Reranker.PRINT_TOP_FEATURES)
			throw new IllegalStateException("illegal");
		return enc.size() + 1;
	}
	
	public static void printProminentFeatures(List<DoubleIntPair> l) {
		int index = 0;
		ArrayList<String> encInv = enc.inverse();
		for(DoubleIntPair p: l) {
			index++;
			System.out.println(index + ":\t" + p.left + "\t" + encInv.get(p.right));
		}
	}

	private static long hash(String s) {
		long hash = 0;
		int l = s.length();
		for(int i = 0; i < l; i++) {			
			hash += s.charAt(i);
			hash += (hash << 10L);
			hash ^= (hash >> 6L);
		}
		hash += (hash << 3L);
		hash ^= (hash >> 11L);
		hash += (hash << 15L);
		return hash;
	}

	public String toString() {
		return "FE:" + useBigrams + "/" + useSyntax + "/" + useSpans + "/" + useSemantics;
	}

}
