package mpqa_structlearn.sequences;

import java.util.*;
import java.io.*;

import mpqa_structlearn.io.*;

import se.lth.cs.nlp.nlputils.annotations.Span;
import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.*;
import se.lth.cs.nlp.nlputils.pstree.*;
import se.lth.cs.nlp.nlputils.pstree.english.ModifiedYamadaHeadFinder;
import se.lth.cs.nlp.depsrl.format.*;

public class SeqCandidateReader {

	private static SentenceData readNextSentence(BufferedReader br,
			HashMap<String, Parses> sm,
			int N) throws IOException {
		String text = br.readLine();
		if(text == null)
			return null;

		SentenceData out = new SentenceData();
		
		Parses parses;
		if(sm == null) {
			parses = new Parses();
			parses.dg = new DepGraph(text.split(" "));
		} else {
			parses = sm.get(text);
			if(parses == null)
				throw new RuntimeException("No parse found for this sentence: |" + text + "|");			
		}
		
		/* Read gold standard annotation. */
		String line = br.readLine().trim();
		out.goldSpans = createSentenceAnnotation(parses, line);
		
		int i = 0;
		line = br.readLine().trim();
		while(!line.equals("---")) {
			if(Integer.parseInt(line) != i)
				throw new RuntimeException("wrong index, should be " + i);
			
			Candidate c = new Candidate();
			c.baseScore = Double.parseDouble(br.readLine());

			String[] ss = br.readLine().trim().split(" ");
			c.stats = new PRFStats(ss);

			line = br.readLine().trim();
			c.mpqaSentence = createSentenceAnnotation(parses, line);

			if(i < N)
				out.candidates.add(c);
			
			i++;
			line = br.readLine().trim();
		}
		
		out.text = text;
		
		//System.out-
		
		for(i = 0; i < out.candidates.size(); i++) {			
			Candidate c1 = out.candidates.get(i);
			Iterator<Candidate> it = out.candidates.listIterator(i + 1);
			while(it.hasNext()) {
				Candidate c2 = it.next();
				if(c1.mpqaSentence.equalsAnnotation(c2.mpqaSentence))
					it.remove();
			}
		}

		return out;
	}
	
	private static SentenceData readNextSentencePol(BufferedReader br,
		HashMap<String, Parses> sm, int N) throws IOException {

		String text = br.readLine();
		if(text == null)
			return null;

		SentenceData out = new SentenceData();
		
		Parses parses = sm.get(text);
		if(parses == null)
			throw new RuntimeException("No parse found for this sentence: |" + text + "|");			
		
		/* Read gold standard annotation. */
		//String line = br.readLine().trim();
		out.goldSpans = null; //
		
		int i = 0;
		String line = br.readLine().trim();
		while(!line.equals("---")) {
			String[] pss = line.split(" ");
			if(pss.length != 2)
				throw new RuntimeException("should see 2 scores, line = |" + line + "|");
			
			Candidate c = new Candidate();
			c.baseScore = Double.parseDouble(pss[0]);
			c.basePolScore = Double.parseDouble(pss[1]);
			line = br.readLine().trim();
			c.mpqaSentence = createSentenceAnnotation(parses, line);
			out.candidates.add(c);			
			i++;
			line = br.readLine().trim();
		}		
		out.text = text;		
		
		// TODO		
		/*for(i = 0; i < out.candidates.size(); i++) {			
			Candidate c1 = out.candidates.get(i);
			Iterator<Candidate> it = out.candidates.listIterator(i + 1);
			while(it.hasNext()) {
				Candidate c2 = it.next();
				if(c1.mpqaSentence.equalsAnnotation(c2.mpqaSentence))
					it.remove();
			}
		}*/

		/*System.out.println("Sentence read: " + out.text);
		for(Candidate c: out.candidates) {
			System.out.println(c.mpqaSentence.exprs);
		}*/

		return out;
	}
	
	private static MPQASentence createSentenceAnnotation(Parses parses, String line) {
		MPQASentence out = new MPQASentence();
		out.dg = parses.dg;
		out.pas = parses.pas;
		out.pstree = parses.pstree;
		out.discConns = parses.discConns;
		if(!line.equals("")) {
			String[] ss = line.split(" ");
			for(String s: ss) {
				String[] ss2 = s.split(",");
				MPQAExpression expr = new MPQAExpression();
				expr.containingSentence = out;
				expr.type = ss2[0];
				expr.span = new ArrayList();
				int start = Integer.parseInt(ss2[1]);
				int end = Integer.parseInt(ss2[2]);
				for(int i = start; i < end; i++)
					expr.span.add(out.dg.nodes[i+1]);
				if(ss2.length == 4) {
					if(ss2[3].equals("+"))
						expr.polarity = "positive";
					else if(ss2[3].equals("-"))
						expr.polarity = "negative";
					else if(ss2[3].equals("0"))
						expr.polarity = "neutral";
					else
						throw new RuntimeException("!!!");
				}
				out.exprs.add(expr);
			}			
		}
		return out;
	}
	
	/*private static void toSubjectivityLayer(AnnotationLayer l) {
		for(Iterator<Span> it = l.iterator(); it.hasNext(); ) {
			Span s = it.next();
			if(s.label.equals("os"))
				it.remove();
			else
				s.label = "ss";				
		}
	}*/
	
	private static class Parses {
		DepGraph dg;
		PhraseStructureTree pstree;
		ArrayList<PAStructure> pas;
		ArrayList<Span> discConns;
	}
	
	public static ArrayList<SentenceData> readSentences() {
	    return null;
	}
	
	public static ArrayList<SentenceData> readSentences(String candFile,
			String synsemFile, String pstreeFile, String discConnFile,
			int N, int maxNSentences) throws IOException {	
		
		HashMap<String, Parses> senMap;

		TreeParser tp = null;
		if(pstreeFile != null && !pstreeFile.matches("NULL|NONE")) {
			BufferedReader psR = Util.openFileReader(pstreeFile, "UTF-8");
			ModifiedYamadaHeadFinder hf = ModifiedYamadaHeadFinder.instance();
			hf.setConjAsHead(true);
			hf.setIMAsHead(true);
			hf.setPosAsHead(true);
			hf.setPrepAsHead(true);
			hf.setSubAsHead(true);
			hf.setWhAsHead(true);
			tp = new PennTreeParser(psR, hf);
		}

		BufferedReader dcr = Util.isNullFile(discConnFile)? null: Util.openFileReader(discConnFile);		
		
		if(synsemFile == null || synsemFile.matches("NULL|NONE"))
			senMap = null;
		else {
			senMap = new HashMap();
			Scanner ssc = new Scanner(Util.openFileReader(synsemFile, "UTF-8"));

			Triple<DepGraph, DepGraph, List<PAStructure>> g = CoNLL2008Format.readNextGraph(ssc);
			while(g != null) {
				Parses ssp = new Parses();
				ssp.dg = g.first;
				ssp.pas = (ArrayList) g.third;
				if(tp != null)
					ssp.pstree = tp.parseTree();
				ssp.discConns = readDiscConns(dcr);
				//for(Span dc: ssp.discConns) {
				//	System.out.println(dc.label + ssp.dg.nodes[dc.tokenStart+1]);
				//}
				
				StringBuilder sb = new StringBuilder();
				for(int i = 1; i < g.first.nodes.length; i++) {
					if(i > 1)
						sb.append(" ");
					sb.append(g.first.nodes[i].word);
				}
				senMap.put(sb.toString(), ssp);				
				g = CoNLL2008Format.readNextGraph(ssc);
			}
		}
		
		ArrayList<SentenceData> out = new ArrayList();		
		BufferedReader candReader = Util.openFileReader(candFile, "UTF-8");
		
		SentenceData sd = readNextSentence(candReader, senMap, N);
		while(sd != null && out.size() != maxNSentences) {
			out.add(sd);
			sd = readNextSentence(candReader, senMap, N);				
		}

		return out;
	}
	
	private static ArrayList<Span> readDiscConns(BufferedReader br) throws IOException {
		if(br == null)
			return null;
		String line = br.readLine();
		ArrayList<Span> out = new ArrayList();
		if(!line.equals(""))
			for(String s: line.split(", ")) {
				String[] ts = s.substring(1, s.length()-1).split("/");
				Span sp = new Span();
				sp.label = ts[0];
				sp.tokenStart = Integer.parseInt(ts[1]);
				sp.tokenEnd = Integer.parseInt(ts[2]);
				out.add(sp);
			}
		return out;
	}
	
	public static ArrayList<SentenceData> readSentencesPol(String candFile,
			String synsemFile, int N) throws IOException {

		HashMap<String, Parses> senMap;		

		if(synsemFile == null || synsemFile.matches("NULL|NONE"))
			throw new RuntimeException("synsem file mandatory in this case");
		else {
			senMap = new HashMap();
			Scanner ssc = new Scanner(Util.openFileReader(synsemFile, "UTF-8"));

			Triple<DepGraph, DepGraph, List<PAStructure>> g = CoNLL2008Format.readNextGraph(ssc);
			while(g != null) {
				Parses ssp = new Parses();
				ssp.dg = g.first;
				ssp.pas = (ArrayList) g.third;
					
				StringBuilder sb = new StringBuilder();
				for(int i = 1; i < g.first.nodes.length; i++) {
					if(i > 1)
						sb.append(" ");
					sb.append(g.first.nodes[i].word);
				}
				senMap.put(sb.toString(), ssp);				
				g = CoNLL2008Format.readNextGraph(ssc);
			}
		}
		
		ArrayList<SentenceData> out = new ArrayList();		
		BufferedReader candReader = Util.openFileReader(candFile, "UTF-8");
		
		SentenceData sd = readNextSentencePol(candReader, senMap, N);
		while(sd != null) {
			out.add(sd);
			if(out.size() % 100 == 0)
				System.out.println(out.size());
			sd = readNextSentencePol(candReader, senMap, N);				
		}

		return out;
	}
	

}
