package mpqa_structlearn.synsem;

import java.io.*;
import java.util.*;

import mpqa_structlearn.io.*;
import mpqa_structlearn.sequences.*;
import se.lth.cs.nlp.depsrl.format.PAStructure;
import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.DepGraph;
import se.lth.cs.nlp.nlputils.ml_long.*;

public class Baseline {

	private static void baseline(String[] argv) {
		String dir = argv[1];
		String subjLexFile = argv[2];
		String docListFile = argv[3];

		String synsemScoreFile = argv[4];
		
		String seqCandidateFile = argv[5];
		int Nseq = Integer.parseInt(argv[6]);

		String rerankerModelFile = argv[7];
		
		double synWeight = Double.parseDouble(argv[8]);
		double semWeight = Double.parseDouble(argv[9]);
		
		try {
			/* Loading trained model. */

			System.out.println("Loading reranker model...");
            ObjectInputStream ois = new ObjectInputStream(new FileInputStream(rerankerModelFile));
            RerankingFE rfe = (RerankingFE) ois.readObject();
            Classifier<EncodedCandidateSet, Integer> rcl = (Classifier) ois.readObject();
            int Nseq_saved = ois.readInt();
            if(Nseq < 1) {
            	System.out.println("Setting Nseq to saved: " + Nseq_saved);
            	Nseq = Nseq_saved;
            }
            ois.close();
            
			System.out.println("Loading sequence candidate lists...");
            ArrayList<SentenceData> seqCands = SeqCandidateReader.readSentences(seqCandidateFile, null, null, null, Nseq, -1);

			System.out.println("Loading parse tree candidate lists...");
            Scanner sssc = new Scanner(Util.openFileReader(synsemScoreFile, "UTF-8"));
            HashMap<String, NBestSynsem> nbestParseTable = NBestSynsem.readAll(sssc);
            
			PRFStats exprResult = new PRFStats();
			PRFStats linkResult = new PRFStats();
            
            ArrayList<String> selectedFiles = new ArrayList();
			Util.readLines(docListFile, selectedFiles);
            
			LKReader r = new LKReader(dir, subjLexFile, selectedFiles);
			int candSenIndex = 0;			
			
			System.out.println("Processing test set...");
			
			while(r.hasNext()) {
				MPQAText text = r.next();
				System.out.println(text.source);
				int nextSenIndex = candSenIndex + text.sentences.size();					
				List<SentenceData> sds = seqCands.subList(candSenIndex, nextSenIndex);
				candSenIndex = nextSenIndex;
				rerankTextBaseline(text, sds, rfe, rcl, nbestParseTable,
						synWeight, semWeight,
						exprResult, linkResult);
			}
			
			System.out.println("Expression result:");
			exprResult.print();

			if(false) {
				System.out.println();
				System.out.println("Link result:");
				linkResult.print();
			}
			
		} catch(Exception e) {
			e.printStackTrace();
		}
		
	}
	
	private static void rerankTextBaseline(MPQAText text,
			List<SentenceData> sds,			
			RerankingFE rfe, 
			Classifier<EncodedCandidateSet, Integer> rcl,
			HashMap<String, NBestSynsem> parseTable,
			double synWeight, double semWeight,
			PRFStats exprResult, PRFStats linkResult) {

    	for(int j = 0; j < text.sentences.size(); j++) {

    		MPQASentence goldSen = text.sentences.get(j);
    		SentenceData sd = sds.get(j);
    		
			ArrayList<Candidate> cands = sd.candidates;
			int n = cands.size();
    		
			double[] lls = new double[n];

			for(int i = 0; i < n; i++)
				lls[i] = cands.get(i).baseScore; // hmm vad händer med baseSrcScore?

			Reranker.normalize(lls, n);

    		NBestSynsem parses = parseTable.get(goldSen.dg.toTokenString());
			
    		HashMap<Candidate, Double> candScoreMap = new HashMap();
    		
    		for(int k = 0; k < parses.dgs.length; k++) {
    			
    			if(k == 1)
    				break;
    			
    			EncodedCandidateSet nb = new EncodedCandidateSet();
    			nb.reps = new SparseVector[n];
    			for(int i = 0; i < n; i++) {
    				Candidate c = cands.get(i);

    				c.mpqaSentence.replaceDepTree(parses.dgs[k]);
    				c.mpqaSentence.pas = (ArrayList) parses.pas[k];
    				
    				SparseVector sv = new SparseVector();
    				rfe.extractFeatures(c, lls[i], sv);
    				nb.reps[i] = sv;
    			}

    			ArrayList<Integer> range = new ArrayList();
    			for(int i = 0; i < n; i++)
    				range.add(i);

    			double[] scores = new double[n];
    			rcl.computeScoresRestricted(nb, range, scores);

    			CollectionUtils.normalizeToLogProbs(scores);
    			
    			for(int i = 0; i < n; i++) {
    				Candidate c = cands.get(i);
        			Double prevScore = candScoreMap.get(c);    				

        			double fullScore = Math.exp(scores[i]);
        			fullScore *= Math.exp(synWeight * parses.synScores[k]);
        			
        			/*if(!Double.isInfinite(parses.synScores[k]))
        				fullScore += synWeight * parses.synScores[k];*/
        			if(!Double.isInfinite(parses.semScores[k]))
        				fullScore *= Math.exp(semWeight * parses.semScores[k]);
        			
        			if(prevScore == null)
        				candScoreMap.put(c, fullScore);
        			else
        				candScoreMap.put(c, prevScore + fullScore);
    			}    			
    		}

    		double maxScore = Double.NEGATIVE_INFINITY;
    		Candidate maxCand = null;    		
    		for(Candidate c: candScoreMap.keySet()) {
    			double score = candScoreMap.get(c);
    			if(score > maxScore) {
    				maxScore = score;
    				maxCand = c;
    			}
    		}
    		
    		if(maxCand == null) {
    			System.err.println("Warning: maxCand = null");
    			for(Candidate c: candScoreMap.keySet()) {
    				System.out.println(candScoreMap.get(c));
    			}
    			maxCand = cands.get(0);
    		}
    		
    		goldSen.replaceDepTree(maxCand.mpqaSentence.dg);
    		
            exprResult.add(goldSen.compareExpressions(maxCand.mpqaSentence));
            linkResult.add(goldSen.compareLinks(maxCand.mpqaSentence));
    	}
		
	}

	public static void main(String[] argv) {
		if(argv[0].equals("-baseline"))
			baseline(argv);
		else {
			System.err.println("Illegal arguments: " + Arrays.toString(argv));
			System.exit(1);
		}
	}

	
}
