package mpqa_structlearn.synsem;

import java.util.*;

import se.lth.cs.nlp.depsrl.format.*;
import se.lth.cs.nlp.nlputils.core.CollectionUtils;
import se.lth.cs.nlp.nlputils.core.Triple;
import se.lth.cs.nlp.nlputils.depgraph.*;

public class NBestSynsem {

	DepGraph[] dgs;
	List<PAStructure>[] pas;
	double[] synScores;
	double[] semScores;

	public static NBestSynsem next(Scanner sc) {
		if(!sc.hasNextLine())
			return null;
		
		ArrayList<DepGraph> dgs = new ArrayList();
		ArrayList<List<PAStructure>> pas = new ArrayList();
		ArrayList<Double> synScores = new ArrayList();
		ArrayList<Double> semScores = new ArrayList();
		
		while(true) {
			if(!sc.hasNextLine())
				break;
			String line = sc.nextLine();
			if(dgs.size() == 0 && line.equals("-----"))
				line = sc.nextLine();
			if(line.equals("-----"))
				break;
			if(!line.startsWith("Candidate"))
				throw new RuntimeException("illegal format");
			line = sc.nextLine();
			synScores.add(Double.parseDouble(line.substring("Syntactic score: ".length())));
			line = sc.nextLine();
			semScores.add(Double.parseDouble(line.substring("Semantic score: ".length())));

			Triple<DepGraph, DepGraph, List<PAStructure>> g = CoNLL2008Format.readNextGraph(sc);
			dgs.add(g.first);
			pas.add(g.third);
		}

		NBestSynsem out = new NBestSynsem();
		out.synScores = CollectionUtils.toDoubleArray(synScores);
		out.semScores = CollectionUtils.toDoubleArray(semScores);
		out.dgs = dgs.toArray(new DepGraph[0]);
		out.pas = pas.toArray(new ArrayList[0]);		
		return out;
	}
	
	public static HashMap<String, NBestSynsem> readAll(Scanner sc) {
		HashMap<String, NBestSynsem> out = new HashMap();
		NBestSynsem nb = next(sc);
		while(nb != null) {
			String s = nb.dgs[0].toTokenString();
			out.put(s, nb);
			nb = next(sc);
		}
		return out;
	}
	
}
