package mpqa_structlearn.tkpolarity;

import java.io.*;
import java.util.*;

import mpqa_structlearn.io.*;
import mpqa_structlearn.polarity.BaselinePolarityClassifier;
import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.*;
import se.lth.cs.nlp.nlputils.ml_long.*;
import se.lth.cs.nlp.nlputils.pstree.*;
import se.lth.cs.nlp.nlputils.pstree.english.*;

import gnu.trove.*;

public class TKPolarityClassifier {

	public static void main(String[] argv) {
		if(argv[0].equals("-genTrain"))
			genTrain(argv);
		//else if(argv[0].equals("-test"))
		//	test(argv);
		else
			throw new RuntimeException("!!!");
	}
	
	private static String getTokenString(PhraseStructureTree tree) {
		StringBuilder out = new StringBuilder();
		for(TerminalNode tn: tree.getTokens()) {
			if(tn instanceof EmptyNode)
				continue;
			if(out.length() > 0)
				out.append(" ");
			String w = ((TokenNode) tn).getWord();
			w = TokenNode.deescapeBracket(w);
			out.append(w);
		}
		
		//if(out.toString().contains("impugn"))
		//	System.out.println("|" + out.toString() + "|");		
		
		return out.toString();
	}

	public static void genTrain(String[] argv) {
		String dir = argv[1]; // LK annotation for *ALL* texts
		String constFile = argv[2]; // parses for *ALL* sentences
		String subjLexFile = argv[3];
		String docListFile = argv[4];

		int contextLength = Integer.parseInt(argv[5]);

		boolean useLinearFeatures = Boolean.parseBoolean(argv[6]);

		String outPosFile = argv[7];
		String outNeuFile = argv[8];
		String outNegFile = argv[9];

		String outEncFile = null;
		if(useLinearFeatures)
			outEncFile = argv[10];
		
        ModifiedYamadaHeadFinder hf = ModifiedYamadaHeadFinder.instance();
        hf.setConjAsHead(false);
        hf.setPrepAsHead(true);
        hf.setPosAsHead(false);
        hf.setSubAsHead(true);
        hf.setIMAsHead(true);
        hf.setWhAsHead(false);
        ModifiedYamadaHeadFinder.setLogger(new MessageLogger());
        PennRules.setLogger(new MessageLogger());
        
		try {

			BufferedReader constInput = Util.openFileReader(constFile, "UTF-8");
			PennTreeParser tp = new PennTreeParser(constInput, hf);
			HashMap<String, PhraseStructureTree> psMap = new HashMap();
			while(tp.hasMoreTrees()) {
				PhraseStructureTree t = tp.parseTree();
				psMap.put(getTokenString(t), t);			
			}
			System.out.println("Read constituent trees.");

			PrintWriter pwPos = new PrintWriter(new FileWriter(outPosFile)); 
			PrintWriter pwNeu = new PrintWriter(new FileWriter(outNeuFile)); 
			PrintWriter pwNeg = new PrintWriter(new FileWriter(outNegFile)); 

			TLongIntHashMap enc = null;
			if(useLinearFeatures)
				enc = new TLongIntHashMap();
			
			HashSet<String> selectedFiles = new HashSet();
			Util.readLines(docListFile, selectedFiles);
			LKReader r = new LKReader(dir, subjLexFile);
			while(r.hasNext()) {
				MPQAText text = r.next();
				String shortSource = text.source.substring(23);				
				if(selectedFiles.contains(text.source) || selectedFiles.contains(shortSource)) {
					System.out.print(".");
					System.out.flush();
					extractTrainingData(text, psMap, contextLength, 
							pwPos, pwNeu, pwNeg, enc);
				}
			}
			System.out.println();

			pwPos.close();
			pwNeu.close();
			pwNeg.close();

			if(useLinearFeatures) {
				ObjectOutputStream oos 
					= new ObjectOutputStream(new FileOutputStream(outEncFile));
				oos.writeObject(enc);
				oos.close();
			}
			
		} catch(Exception e) {
			e.printStackTrace();
		}
	}
	
	private static String getPolarity(MPQAExpression expr) {
		String pol = expr.polarity;
		if(pol == null)
			return "neutral";
		pol = pol.trim();
		if(pol.equals(""))
			return "neutral";
		if(pol.startsWith("uncertain-"))
			pol = pol.substring("uncertain-".length());
		if(!pol.matches("positive|neutral|negative|both"))
			throw new RuntimeException("pol = " + pol);		
		return pol;
	}
	
	private static void extractTrainingData(MPQAText text,
			HashMap<String, PhraseStructureTree> psMap,
			int contextLength,
			PrintWriter pwPos, PrintWriter pwNeu, PrintWriter pwNeg,
			TLongIntHashMap enc) throws IOException {		
		
		for(MPQASentence sen: text.sentences) {			

			PhraseStructureTree tree = null;			
			for(MPQAExpression expr: sen.exprs) {
				if(expr.type.equals("os"))
					continue;

				if(tree == null)
					tree = psMap.get(sen.dg.toTokenString());
				if(tree == null) {
					System.out.println("Found no pstree");
					System.out.println("|" + sen.dg.toTokenString() + "|");
					System.exit(1);
				}
				
				expr.type = "OP"; // no ds/es

				String rep1 = markNodesPS1(expr, tree, contextLength);
				String rep2 = markNodesPS2(expr, tree, contextLength);
				String rep3 = markNodesPS3(expr, tree, contextLength);			
				String seq1 = markWords(expr, sen.dg, contextLength);
				
				String seq2 = markWordsMD(expr, sen.dg, contextLength, true, true, true);

				String lin = extractLinearFeatures(expr, enc);
				
				String pol = getPolarity(expr);
				
				if(pol.matches("positive|both"))
					pwPos.print("+1");
				else
					pwPos.print("-1");
				if(pol.matches("negative|both"))
					pwNeg.print("+1");
				else
					pwNeg.print("-1");
				if(pol.equals("neutral"))
					pwNeu.print("+1");
				else
					pwNeu.print("-1");

				String line = " |BT| " + seq1 + " |BT| " + rep1 + " |BT| " 
					+ rep2 + " |BT| " + rep3 + " |BT| " + seq2 + " |ET|"
					+ lin + " |EV|";
				
				pwPos.println(line);
				pwNeg.println(line);
				pwNeu.println(line);
			}
		}
		
		// TODO vi borde också märka upp orden i träden -- de är väl informativa?
		// TODO subjlex-info?
		
	}
	
	private static String extractLinearFeatures(MPQAExpression expr,
			TLongIntHashMap enc) {
		// TODO denna metod förutsätter att det är hash-longs i extraktionen
		SparseVector sv1 = BaselinePolarityClassifier.represent(expr, null /* OBS */);
		SparseVector sv2 = new SparseVector();

		for(int i = 0; i < sv1.index; i++) {
			long k = sv1.keys[i];
			int k2 = enc.get(k);
			if(k2 == 0) {
				k2 = enc.size() + 1;
				enc.put(k, k2);
			} else {
				//System.err.println("*");
				//System.err.flush();
			}
			sv2.put(k2, sv1.values[i]);
		}
		sv2.sortIndices();

		//System.err.println("sv1 = " + sv1);
		//System.err.println("sv2 = " + sv2);
		//System.exit(0);
		
		StringBuilder sb = new StringBuilder();
		for(int i = 0; i < sv2.index; i++) {
			sb.append(" ");
			double v = sv2.values[i];
			double vr = Math.round(v);
			String vs;
			if(v - vr < 1e-9 && v - vr > -1e-9)
				vs = String.valueOf((int) v);
			else
				vs = String.valueOf(v);
			sb.append(sv2.keys[i] + ":" + vs);
		}
		return sb.toString();
	}

	static PhraseStructureTree lcClone(PhraseStructureTree tree) {
		PhraseStructureTree clone = tree.clone();
		for(TerminalNode tn: clone.getTokens()) {
			if(tn instanceof EmptyNode)
				continue;
			TokenNode t = (TokenNode) tn;
			String w = t.getWord().toLowerCase();
			w = w.replaceAll("[1-9]", "0");
			t.setWord(w);
		}
		return clone;
	}
	
	/* Mark the first node whose span is included */
	static String markNodesPS1(MPQAExpression expr, PhraseStructureTree tree,
			int contextLength) {		
		PhraseStructureTree clone = lcClone(tree);

		int start = expr.span.get(0).position - 1;
		int end = expr.span.get(expr.span.size()-1).position;
		
		for(Node n: clone.getNodes()) {

			Pair<Integer, Integer> span = n.span();
			if(span.left >= start && span.right <= end) {
				if(n instanceof TokenNode) {
					TokenNode tn = (TokenNode) n;
					tn.setPos(tn.getPos() + "#" + expr.type);
				} else if(n instanceof NonterminalNode) {
					NonterminalNode nt = (NonterminalNode) n;
					nt.setLabel(nt.getLabel() + "#" + expr.type);
				}
				break;
			}
		}
		try {
			int cstart = Math.max(start - contextLength, 0);
			int cend = Math.min(end + contextLength, clone.getTokens().size());
			clone.retainRange(cstart, cend);
			return clone.toStrippedBracketing();
		} catch(Exception e) {
			System.out.println("Error: ");
			System.out.println("Expr: " + expr);
			System.out.println(tree.tabbedOutput());			
			System.out.println(clone.tabbedOutput());
			e.printStackTrace();			
			System.exit(1);
			return null;
		}
	}	

	/* Mark the highest nodes included in the spans */
	static String markNodesPS3(MPQAExpression expr, PhraseStructureTree tree, 
			int contextLength) {
		PhraseStructureTree clone = lcClone(tree);	
		int start = expr.span.get(0).position - 1;                // TODO kolla
		int end = expr.span.get(expr.span.size()-1).position; // TODO kolla
		ArrayList<Node> above = clone.getHighestCovering(start, end);
		for(Node n: above) {
			if(n instanceof NonterminalNode) {
				NonterminalNode nn = (NonterminalNode) n;
				nn.setLabel(nn.getLabel() +  "#" + expr.type);
			} else {
				TokenNode tn = (TokenNode) n;
				tn.setPos(tn.getPos() +  "#" + expr.type);				
			}
		}
		int cstart = Math.max(start - contextLength, 0);
		int cend = Math.min(end + contextLength, clone.getTokens().size());
		clone.retainRange(cstart, cend);
		return clone.toStrippedBracketing();
	}

	/* Mark everything whose head is included in span */
	static String markNodesPS2(MPQAExpression expr, PhraseStructureTree tree, 
			int contextLength) {
		PhraseStructureTree clone = lcClone(tree);	
		int start = expr.span.get(0).position - 1;            // TODO kolla
		int end = expr.span.get(expr.span.size()-1).position; // TODO kolla
		for(Node n: clone.getNodes()) {
			TokenNode h = (TokenNode) n.getHead();
			if(h.getPosition() >= start && h.getPosition() < end) {
				if(n instanceof NonterminalNode) {
					NonterminalNode nt = (NonterminalNode) n;
					nt.setLabel(nt.getLabel() + "#" + expr.type);
				} else {
					String p = h.getPos();
					if(p.equals("("))
						p = "-LRB-";
					else if(p.equals(")"))
						p = "-RRB-";						
					h.setPos(p + "#" + expr.type);
				}
			}
		}
		int cstart = Math.max(start - contextLength, 0);
		int cend = Math.min(end + contextLength, clone.getTokens().size());
		clone.retainRange(cstart, cend);
		return clone.toStrippedBracketing();
	}
	
	static String markWords(MPQAExpression expr, DepGraph dg, int contextLength) {		
		int start = expr.span.get(0).position;                    // TODO kolla
		int end = expr.span.get(expr.span.size()-1).position + 1; // TODO kolla		

		int cstart = Math.max(start - contextLength, 1);
		int cend = Math.min(end + contextLength, dg.nodes.length);
		
		StringBuilder out = new StringBuilder("(BOW ");		
		for(int i = cstart; i < start; i++)
			out.append("(" + TokenNode.escapeBracket(dg.nodes[i].word).toLowerCase() + ")");
		out.append("(OS_START)");
		for(int i = start; i < end; i++)
			out.append("(" + TokenNode.escapeBracket(dg.nodes[i].word).toLowerCase() + ")");
		out.append("(OS_END)");
		for(int i = end; i < cend; i++)
			out.append("(" + TokenNode.escapeBracket(dg.nodes[i].word).toLowerCase() + ")");
		out.append(")");
		return out.toString();
	}
	
	static String markWordsMD(MPQAExpression expr, DepGraph dg, int contextLength,
			boolean useWords, boolean usePOS, boolean useSL) {		

		ArrayList<String> slClues = null;
		if(useSL) {
			slClues = expr.containingSentence.slClues;
			if(slClues == null || slClues.size() != dg.nodes.length)
				throw new RuntimeException("!!!");
		}
		
		int start = expr.span.get(0).position;                    // TODO kolla
		int end = expr.span.get(expr.span.size()-1).position + 1; // TODO kolla		

		int cstart = Math.max(start - contextLength, 1);
		int cend = Math.min(end + contextLength, dg.nodes.length);
		
		StringBuilder out = new StringBuilder("(MDREPR ");		
		for(int i = cstart; i < start; i++) {
			out.append("(NULL");
			if(useWords) out.append(" (F0 " + TokenNode.escapeBracket(dg.nodes[i].pos) + ")");
			if(usePOS) out.append(" (F1 " + TokenNode.escapeBracket(dg.nodes[i].word).toLowerCase() + ")");
			if(useSL) out.append(" (F2 " + slClues.get(i) + ")");
			out.append(")");
		}
			
		out.append("(OS ");
		{   
			int i = start;
		    out.append("(B-OS");
		    if(useWords) out.append(" (F0 " + TokenNode.escapeBracket(dg.nodes[i].pos) + ")");
		    if(usePOS) out.append(" (F1 " + TokenNode.escapeBracket(dg.nodes[i].word).toLowerCase() + ")");
		    if(useSL) out.append(" (F2 " + slClues.get(i) + ")");
			out.append(")");
		}
		for(int i = start + 1; i < end; i++) {
		    out.append("(I-OS");
		    if(useWords) out.append(" (F0 " + TokenNode.escapeBracket(dg.nodes[i].pos) + ")");
		    if(usePOS) out.append(" (F1 " + TokenNode.escapeBracket(dg.nodes[i].word).toLowerCase() + ")");
		    if(useSL) out.append(" (F2 " + slClues.get(i) + ")");
			out.append(")");
		}
		out.append(")");

		for(int i = end; i < cend; i++) {
			out.append("(NULL");
			if(useWords) out.append(" (F0 " + TokenNode.escapeBracket(dg.nodes[i].pos) + ")");
			if(usePOS) out.append(" (F1 " + TokenNode.escapeBracket(dg.nodes[i].word).toLowerCase() + ")");
			if(useSL) out.append(" (F2 " + slClues.get(i) + ")");
			out.append(")");
		}
		out.append(")");
		
		return out.toString();
	}
	
	
	
}
