/*
 AnnotationType.java
 
 Copyright (C) 2005, Richard Johansson (richard@cs.lth.se).
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.
 
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 Created on Nov 25, 2005 by Richard Johansson (richard@cs.lth.se).
 
 $Log: AnnotationType.java,v $
 Revision 1.1  2009/01/16 10:04:53  johansson
 Added to brenta repository.

 Revision 1.8  2007/10/08 12:31:34  richard
 Refactored again to remove atoms.

 Revision 1.7.2.1  2007/10/08 12:18:22  richard
 Added iterateTexts.

 Revision 1.7  2006/12/20 09:22:32  richard
 Added methods to print to stdout.

 Revision 1.6  2006/09/07 17:51:19  richard
 New revision.

 Revision 1.5  2006/04/11 08:10:13  richard
 Added documentation.
 
 Revision 1.4  2006/03/29 13:56:12  richard
 Added more printLayers variants.
 
 Revision 1.3  2005/11/30 13:25:17  richard
 Changed the interface.
 
 Revision 1.2  2005/11/30 12:43:28  richard
 Changed the interface.
 
 Revision 1.1  2005/11/30 09:52:24  richard
 Renamed the annotation types.
 
 Revision 1.2  2005/11/30 09:49:09  richard
 Changed the interface.
 
 Revision 1.1  2005/11/28 15:41:00  richard
 Added the file.
 
 
 */
package se.lth.cs.nlp.nlputils.annotations;

import java.util.*;
import java.io.*;

/**
 * This class is inherited by factories of annotation types.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public abstract class AnnotationType {
    
    /**
     * Reads and parses an AnnotatedText from the given InputStream, using the
     * format of the inheriting AnnotationType.
     *
     * @param in the input stream.
     * @return the AnnotatedText.
     * @throws IOException
     */
    public AnnotatedText createLayers(InputStream in) throws IOException {
        throw new UnsupportedOperationException("Not implemented!");
    }
    
    /**
     * Reads and parses an AnnotatedText from a string, using the
     * format of the inheriting AnnotationType.
     * 
     * @param s the string.
     * @return the AnnotatedText.
     * @throws IOException
     */
    public AnnotatedText createLayers(String s) throws IOException {
        throw new UnsupportedOperationException("Not implemented!");
    }
    
    public Iterator<? extends AnnotatedText> iterateTexts(InputStream in) throws IOException {
        throw new UnsupportedOperationException("Not implemented!");        
    }
    
    /**
     * Prints a list of layers on a text to a PrintWriter using the
     * format of the inheriting AnnotationType.
     *
     * @param out the writer.
     * @param text the text.
     * @param l the list of layers.
     * @throws IOException
     */
    public void printLayers(PrintWriter out, String text, List<AnnotationLayer> l) 
    throws IOException {
        throw new UnsupportedOperationException("Not implemented!");
    }
    
    /**
     * Prints an AnnotatedText to a PrintWriter using the
     * format of the inheriting AnnotationType.
     *
     * @param out the writer.
     * @param text the AnnotatedText.
     * @throws IOException
     */
    public void printLayers(PrintWriter out, AnnotatedText text) 
    throws IOException {
        printLayers(out, text.text, text.layers);
    }
    
    /**
     * Prints a list of layers on a text to an OutputStream using the
     * format of the inheriting AnnotationType.
     * 
     * @param out the output stream.
     * @param text the text.
     * @param l the list of layers.
     * @throws IOException
     */
    public void printLayers(OutputStream out, String text, List<AnnotationLayer> l) 
    throws IOException {
        throw new UnsupportedOperationException("Not implemented!");
    }
    
    /**
     * Prints an AnnotatedText to an OutputStream using the
     * format of the inheriting AnnotationType.
     *
     * @param out the output stream.
     * @param text the AnnotatedText.
     * @throws IOException
     */
    public void printLayers(OutputStream out, AnnotatedText text) 
    throws IOException {
        printLayers(out, text.text, text.layers);
    }
    
    public void printLayers(String text, List<AnnotationLayer> layers) {
        try {
            printLayers(System.out, text, layers);
        } catch(IOException e) {
            throw new RuntimeException(e);
        }
    }
    
    public void printLayers(AnnotatedText text) {
        try {
            printLayers(System.out, text.text, text.layers);
        } catch(IOException e) {
            throw new RuntimeException(e);
        }
    }
    
}