/*  
    ColumnSpanAnnotation.java.

	$Log: ColumnSpanAnnotation.java,v $
	Revision 1.1  2010-06-07 10:54:39  johansson
	Added.


 */

package se.lth.cs.nlp.nlputils.annotations;

import java.io.*;
import java.util.*;

public class ColumnSpanAnnotation extends AnnotationType {

	private ColumnSpanAnnotation() {}

	private static final ColumnSpanAnnotation instance = new ColumnSpanAnnotation();

	public static ColumnSpanAnnotation instance() {
		return instance;
	}

	public void printLayers(PrintWriter pw, AnnotatedText text) throws IOException {
		if(text.tokens == null)
			throw new IllegalArgumentException("must be tokenized");
		
		StringBuilder[] sbs = new StringBuilder[text.tokens.length];
		for(int i = 0; i < text.tokens.length; i++)
			sbs[i] = new StringBuilder(text.tokens[i]);

		//pw.println("nl = " + text.layers.size());
		
		for(AnnotationLayer<Span> l: text.layers) {

			//System.out.println(l);
			pw.println(l);
			
			Collections.sort(l.spans, Span.NestingOrder.instance());
			pad(sbs);

			int index = 0;
			for(Span s: l) {
				if(index > s.tokenStart) {
					pw.println(index);
					pw.println(s);
					pw.flush();
					throw new RuntimeException("!!!");
				}
				//for(int i = index; i < s.tokenStart; i++)
				//	sbs[i].append(" *");

				if(s.tokenEnd <= s.tokenStart)
					// what to do?
					continue;
				
				sbs[s.tokenStart].append("(" + s.label);
				for(int i = s.tokenStart + 1; i < s.tokenEnd; i++)
					sbs[i].append(" -");
				sbs[s.tokenEnd].append(s.label + ")");
				
				index = s.tokenEnd;	
			}
			
			//break;
		}
		
		for(int i = 0; i < sbs.length; i++)
			pw.println(sbs[i]);
		pw.flush();
	}
		
    private static void pad(StringBuilder[] sbs) {
        int max = -1;
        for(int i = 0; i < sbs.length; i++)
            if(sbs[i].length() > max)
                max = sbs[i].length();
        for(int i = 0; i < sbs.length; i++) {
            int l = max - sbs[i].length() + 1;
            for(int j = 0; j < l; j++)
                sbs[i].append(' ');
        }        
    }

	
}
