/*
 FrameNetLUAnnotation.java
  
 Created on Mar 29, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: FrameNetCorpusAnnotation.java,v $
 Revision 1.1  2009/01/16 10:04:53  johansson
 Added to brenta repository.

 Revision 1.13  2008/03/11 13:26:58  richard
 Backwards compatible with 1.5.

 Revision 1.12  2007/10/08 12:31:34  richard
 Refactored again to remove atoms.

 Revision 1.11.2.1  2007/10/08 12:18:31  richard
 Added iterateTexts.

 Revision 1.11  2007/10/05 07:55:44  richard
 Quote bug.

 Revision 1.10  2007/10/04 13:58:10  richard
 Overlaps.

 Revision 1.9  2007/10/04 13:27:00  richard
 Hyphens in comments.

 Revision 1.8  2007/09/13 14:50:30  richard
 Backwards-compatibility.

 Revision 1.7  2007/08/23 10:20:08  richard
 Null checks.

 Revision 1.6  2007/08/23 06:37:56  richard
 Sets itype.

 Revision 1.5  2007/08/17 10:18:28  richard
 Saves all information. Printout.

 Revision 1.4  2007/08/13 12:25:40  richard
 All layers in one text instead of multiple.

 Revision 1.3  2007/08/10 13:29:53  richard
 Disables validation.

 Revision 1.2  2007/04/05 13:24:43  richard
 Handles empty sentences.

 Revision 1.1  2007/03/29 09:02:55  richard
 Added.

 */
package se.lth.cs.nlp.nlputils.annotations;

import java.io.*;
import java.util.*;

import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.*;

import javax.xml.parsers.*;

/* For debugging. */
import se.lth.cs.nlp.nlputils.framenet.*;
import se.lth.cs.nlp.nlputils.core.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class FrameNetCorpusAnnotation extends AnnotationType {

    private static final String ID = "ID";
    
    /* TODO poss, syntax!!! */
    
    private FrameNetCorpusAnnotation() {}

    private static final FrameNetCorpusAnnotation instance = new FrameNetCorpusAnnotation();

    public static FrameNetCorpusAnnotation instance() {
        return instance;
    }

    private int idCounter;
    
    private boolean allowOverlaps = true;
    public void setOverlapsAllowed(boolean b) {
        allowOverlaps = b;
    }
    
    public List<AnnotatedText> createTexts(InputStream in) throws IOException {
        try {
            XMLReader reader = makeXMLReader();
            try {
                reader.setFeature("http://apache.org/xml/features/nonvalidating/load-external-dtd",
                        false);
            } 
            catch (Exception e) {
                System.err.println("could not set parser feature");
            }
            XMLCallback c = new XMLCallback();
            reader.setContentHandler(c);
            reader.parse(new org.xml.sax.InputSource(in));

            return c.texts;
        } catch(IOException e) {
            throw e;
        } catch(Exception e) {
            e.printStackTrace();
            throw new IOException("Couldn't parse!");
        }        
    }

    public Iterator<AnnotatedText> iterateTexts(InputStream in) throws IOException {
        XMLReader reader = null;
        try {
            reader = makeXMLReader();
            reader.setFeature("http://apache.org/xml/features/nonvalidating/load-external-dtd",
                    false);
        } 
        catch (Exception e) {
            throw new IOException(e.getMessage()); // 1.5
        }
        XMLCallback c = new XMLCallback();
        c.texts = null;
        reader.setContentHandler(c);

        new ParserThread(reader, in).start();

        return new TextIterator(c);
    }
    
    private class ParserThread extends Thread {
        private XMLReader reader;
        private InputStream in;
        ParserThread(XMLReader reader, InputStream in) {
            this.reader = reader;
            this.in = in;
        }
        public void run() {
            try {
                reader.parse(new org.xml.sax.InputSource(in));
            } catch(Exception e) {
                throw new RuntimeException(e);
            }
        }
    }
    
    private class TextIterator implements Iterator<AnnotatedText> {

        private XMLCallback cb;
        private AnnotatedText next;
        private boolean end = false;
        
        TextIterator(XMLCallback cb) {
            this.cb = cb;
        }
        
        public AnnotatedText next() {
            if(end)
                throw new NoSuchElementException("No more texts!");
            if(next == null) {                
                next = cb.getCurrent();
                if(next == null) {
                    end = true;
                    throw new NoSuchElementException("No more texts!");
                }
            }
            AnnotatedText out = next;
            next = null;
            return out;
        }
        
        public boolean hasNext() {
            if(end)
                return false;
            if(next != null)
                return true;
            next = cb.getCurrent();
            if(next == null)
                end = true;
            return next != null;
        }
        
        public void remove() {
            throw new UnsupportedOperationException("unsupported!");
        }
    }
    
    
    public AnnotatedText createLayers(InputStream in) throws IOException {
        throw new UnsupportedOperationException("Unsupported!");
    }

    
    private final class XMLCallback extends DefaultHandler implements ContentHandler {

        private boolean pending = false;
        private AnnotatedText currentText;

        private synchronized AnnotatedText getCurrent() {
            while(!pending)
                try {
                    wait();
                } catch(InterruptedException e) {};
            AnnotatedText out = currentText;
            currentText = null;
            //System.out.println("getCurrent returns " + out);
            pending = false;
            notifyAll();
            return out;
        }

        private synchronized void giveCurrent() {
            pending = true;
            notifyAll();
            while(pending)
                try {
                    wait();
                } catch(InterruptedException e) {};
        }
        
        ArrayList<AnnotatedText> texts = new ArrayList<AnnotatedText>(); 

        private String currentString;

        private String currentFrame, currentLU, currentStatus;
        
        private AnnotationLayer currentLayer;
        private StringBuilder currentChars;
        
        private String corpusName, corpusID;
        private String docDesc, docID;
        private String parID, docOrder;
	private String annSetID;

        private String luID, frID;
        
        public void startElement(String namespace, String localname,  
                String type, Attributes attributes) {            
            if(type.equals("sentence")) {                
                currentText = new AnnotatedText();
                currentText.setProperty(ID, attributes.getValue(ID));
                currentText.setProperty("corpusName", corpusName);
                currentText.setProperty("corpusID", corpusID);
                currentText.setProperty("docDesc", docDesc);
                currentText.setProperty("docID", docID);
                currentText.setProperty("parID", parID);
                currentText.setProperty("docOrder", docOrder);                
            } else if(type.equals("text")) {
                currentChars = new StringBuilder();
            } else if(type.equals("corpus")) {
                corpusName = attributes.getValue("name");
                corpusID = attributes.getValue(ID);
            } else if(type.equals("document")) {
                docDesc = attributes.getValue("description");
                docID = attributes.getValue(ID);
            } else if(type.equals("paragraph")) {
                parID = attributes.getValue(ID);
                docOrder = attributes.getValue("documentOrder");
            } else if(type.equals("annotationSet")) {
                currentFrame = attributes.getValue("frameName");
                currentLU = attributes.getValue("luName");
                currentStatus = attributes.getValue("status");                
		annSetID = attributes.getValue(ID);
                luID = attributes.getValue("lexUnitRef");
                frID = attributes.getValue("frameRef");
            } else if(type.equals("layer")) {
                String name = attributes.getValue("name");
                if(name == null)
                    throw new RuntimeException("no name for layer");

                String rs = attributes.getValue("rank");
                int rank = 0;
                if(rs != null)
                    rs = rs.trim();
                if(rs != null && rs.length() > 0)
                    rank = Integer.parseInt(rs);
                if(rank > 1) {
                    for(ListIterator<AnnotationLayer> iter 
                            = currentText.layers.listIterator(currentText.layers.size());
                    iter.hasPrevious(); ) {
                        AnnotationLayer l = iter.previous();
                        if(l.label.equals("Target"))
                            throw new RuntimeException("Rank is " + rank + ", found target before FE.");
                        if(l.label.equals("FE")) {
                            currentLayer = l;
                            break;
                        }
                    }
                    if(currentLayer == null)
                        throw new RuntimeException("Rank is " + rank + ", didn't find FE layer");
                } else {
                    currentLayer = new AnnotationLayer();
                    currentLayer.label = attributes.getValue("name");
                    if(currentLayer.label.equals("Target")) {
                        currentLayer.setProperty("frame", currentFrame);
                        currentLayer.setProperty("lu", currentLU);
                        currentLayer.setProperty("status", currentStatus);
			currentLayer.setProperty("annotationSetID", annSetID);
			currentLayer.setProperty("luID", luID);
                        currentLayer.setProperty("frameID", frID);
                    } else if(currentLayer.label.equals("NER")) {
			currentLayer.setProperty("annotationSetID", annSetID);
                        currentLayer.setProperty("status", currentStatus);
		    }
		    currentLayer.setProperty(ID, attributes.getValue(ID));
                    currentText.layers.add(currentLayer);
                }
            } else if(type.equals("label")) {
                
                String name = attributes.getValue("name");
                if(name == null)
                    throw new RuntimeException("no name for span");
                String start_s = attributes.getValue("start");
                String end_s = attributes.getValue("end");
                String itype = attributes.getValue("itype");

                int start, end;
                
                if(start_s == null) {
                    if(itype == null)
                        throw new RuntimeException("Neither span nor itype declared!");
                    start = 0; //-1;
                    end = 0; //-1;
                } else {
                    if(itype != null) {
                        throw new RuntimeException("Both span and itype declared!");       
                    }
                    start = Integer.parseInt(start_s);
                    end = Integer.parseInt(end_s) + 1;
                }

                Span sp = new Span();
		sp.id = attributes.getValue(ID);
                sp.start = start;
                sp.end = end;
                sp.label = name;
                sp.properties = new HashMap<String, Object>();

                if(itype != null)
                    sp.properties.put("itype", itype);

                if(currentLayer == null)
                    throw new RuntimeException("currentLayer is null");

                if(allowOverlaps || !currentLayer.overlaps(sp))
                    currentLayer.spans.add(sp);              
            } else if(type.equals("text")) {
                currentChars = new StringBuilder();
            }
        }
        
        public void endElement(String namespace, String localname,  
                String type) {
            if(type.equals("annotationSet")) {
                /* TODO sortera h�r */
            } else if(type.equals("layer")) {
                currentLayer = null;
            } else if(type.equals("text")) {
                currentString = currentChars.toString();
                currentChars = null;
            } else if(type.equals("sentence")) {
                currentText.text = currentString;
                FrameNetAnnotationLayers.checkLayers(currentText);
		for(AnnotationLayer l: currentText.layers)
		    if(l.getProperty(ID) == null)
			l.setProperty(ID, "" + idCounter++);
                
		//System.out.println("currentText = " + currentText);
                
		if(texts != null)
		    texts.add(currentText);
                else
                    giveCurrent();
                
                currentText = null;
	    } else if(type.equals("corpus")) {
	        if(texts == null) {
	            giveCurrent();
                }
            }
        }
        
        public void characters(char[] ch, int start, int len) {
            if(currentChars != null)
                currentChars.append(new String(ch, start, len));
        }
        
    }
    
    private XMLReader makeXMLReader() throws SAXException, ParserConfigurationException { 
        javax.xml.parsers.SAXParserFactory saxParserFactory 
        = javax.xml.parsers.SAXParserFactory.newInstance(); 
        javax.xml.parsers.SAXParser saxParser = saxParserFactory.newSAXParser(); 
        org.xml.sax.XMLReader parser = saxParser.getXMLReader(); 
        return parser; 
    }

    private static void checkCoreConsistency(AnnotationLayer<? extends Span> roleLayer,
            String frame,
            FrameNet framenet) {

        Set<Set<String>> coreSets = framenet.getCoreSets(frame);
        
        Set<Pair<String, String>> excludes = framenet.getExcludes(frame);
        Set<Pair<String, String>> requires = framenet.getRequires(frame);
        
        //System.out.println("coreSets = " + coreSets);
        
        for(Set<String> cs: coreSets) {
            boolean seen = false;
            for(Span sp: roleLayer.spans)
                if(cs.contains(sp.label)) {
                    seen = true;
                    break;
                }
            if(!seen) {
                //throw new RuntimeException("Core set unmatched: " + cs);
            }
        }

        for(Span sp1: roleLayer.spans)
            for(Pair<String, String> p: excludes)
                if(p.left.equals(sp1.label))
                    for(Span sp2: roleLayer.spans)
                        if(p.right.equals(sp2.label)) {
                            //throw new RuntimeException(sp1.label + " excludes " + sp2.label);
                        }
        
        for(Span sp1: roleLayer.spans)
            for(Pair<String, String> p: requires)
                if(p.left.equals(sp1.label)) {
                    boolean seen = false;
                    for(Span sp2: roleLayer.spans)
                        if(p.right.equals(sp2.label)) {
                            seen = true;
                            break;
                        }
                    if(!seen) {
                        throw new RuntimeException(sp1.label + " requires " + p.right);      
                    }
                }
                            
        //System.exit(0);
        
    }
    
    public static void main(String[] argv) {
        FrameNetCorpusAnnotation f = FrameNetCorpusAnnotation.instance();
        AnnotatedText t = null;

        try {

            FrameNet fn = new FrameNet("/slask/fn13/frXML/frames.xml",
                    "/slask/fn13/frXML/frRelation.xml");            
            
            for(String file: argv) {
                //System.err.println(file);
                List<AnnotatedText> texts = f.createTexts(new FileInputStream(file));

                if(true)
                for(AnnotatedText text: texts) {
                    //System.out.println(text);
                    t = text;
                    //RawXMLAnnotation.instance().printLayers(System.out, text);

                    /*checkCoreConsistency(text.layers.get(2),
                                         (String) text.getProperty("frame"),
                                         fn);*/
                    
                    //System.out.println(text.text);

		    //if(t.layers.size() > 3)
                    //t.layers = t.layers.subList(0, 3);

                    //RawXMLAnnotation.instance().printLayers(text);

                    System.out.println(text);
                    
                    if(false)
                    for(AnnotationLayer<Span> layer: text.layers)
                        for(Span span: layer.spans) {
                            if(span.start == span.end)
                                continue;
                            
                            if(Character.isWhitespace(text.text.charAt(span.start)))
                                throw new RuntimeException("Whitespace at span start!");
                            if(span.start > 0
                               && !Character.isWhitespace(text.text.charAt(span.start - 1)))
                                throw new RuntimeException("No whitespace before span start!");            
                            if(Character.isWhitespace(text.text.charAt(span.end - 1)))
                                throw new RuntimeException("Whitespace at span end!");
                            if(span.end < text.text.length()
                               && !Character.isWhitespace(text.text.charAt(span.end)))
                                throw new RuntimeException("No whitespace after span end!");     
                        }
                    
                    
                }
            }
        } catch(Exception e) {
            System.err.println(t);
            try {
                RawXMLAnnotation.instance().printLayers(System.err, t);
            } catch(Exception e2) {
            }
            
            e.printStackTrace();
        }
    }

    private static String escape(String s) {
        s = s.replaceAll("&", "____AMP____");
        s = s.replaceAll("\"", "&quot;");

        //s = s.replaceAll("'", "&apos;");
        s = s.replaceAll("<", "&lt;");
        s = s.replaceAll(">", "&gt;");

        s = s.replaceAll("____AMP____", "&amp;");
        
        return s;
    }
    
    private void printLayer(PrintWriter out, AnnotationLayer<? extends Span> layer, 
			    AnnotatedText text) throws IOException {
        out.print("                    <layer id=\"" + layer.getProperty(ID)
		  + "\" name=\"" + layer.label + "\"");
        if(layer.label != null && layer.label.equals("FE"))
            out.print(" rank=\"1\"");
        if(layer.spans.size() == 0) {
            out.println("/>");
        } else {
            out.println(">");
            out.println("                      <labels>");
            for(Span s: layer.spans) {
                out.print("                        <label ");
                out.print("name=\"" + s.label + "\" ");
                out.print("ID=\"" + s.id + "\" ");
		String spanText = null;
                if(s.end > s.start) {
                    out.print("start=\"" + s.start + "\" ");
                    out.print("end=\"" + (s.end-1) + "\" ");
		    spanText = text.text.substring(s.start, s.end);
		    while(spanText.contains("--"))
		        spanText = spanText.replaceAll("--", "__");
                }
                Object itype = s.getProperty("itype");
                if(itype != null)
                    out.print("itype=\"" + itype + "\" ");
                out.print("/>");
		if(spanText != null)
		    out.print(" <!-- " + spanText + " -->");
		out.println();
            }
            out.println("                      </labels>");
            out.println("                    </layer>");
        }
        
    }
    
    public void printLayers(PrintWriter out, AnnotatedText text) throws IOException {
        String sid = (String) text.getProperty("ID");
        out.println("            <sentence ID=\"" + sid + "\">");
        out.println("              <text>" + escape(text.text) + "</text>");
        out.println("              <annotationSets>");

        /* NER. */
        
        Object status = text.layers.get(0).getProperty("status");
        if(status == null)
            status = "UNKNOWN";
        // TODO id och status f�r annotationsets och layers.
        out.println("                <annotationSet ID=\"" 
		    + text.layers.get(0).getProperty("annotationSetID") 
		    + "\" status=\"" + status + "\">");
        out.println("                  <layers>");

        printLayer(out, text.layers.get(0), text);
        
        out.println("                  </layers>");
        out.println("                </annotationSet>");

        if(text.layers.size() > 1) {
            for(int i = 1; i < text.layers.size(); i += 7) {
                AnnotationLayer tl = text.layers.get(i);
                String id = (String) tl.getProperty("annotationSetID");
                String frame = (String) tl.getProperty("frame");
                if(frame == null)
                    continue;
                String lu = (String) tl.getProperty("lu");
                status = tl.getProperty("status");
                if(status == null)
                    status = "UNKNOWN";
                
                out.print("                <annotationSet ID=\"" + id + "\" "); 
                out.print("status=\"" + status + "\" ");
                out.print("frameName=\"" + frame + "\" ");
                out.print("luName=\"" + lu + "\" ");
                if(tl.getProperty("luID") != null)
                    out.print("lexUnitRef=\"" + tl.getProperty("luID") + "\" ");
                if(tl.getProperty("frameID") != null)
                    out.print("frameRef=\"" + tl.getProperty("frameID") + "\" ");
                out.println(">");
                out.println("                  <layers>");                
                
                for(int j = i; j < i + 7; j++)
                    if(text.layers.get(j).label != null)
                        printLayer(out, text.layers.get(j), text);
                    
                out.println("                  </layers>");
                out.println("                </annotationSet>");                
            }
        }
        
        out.println("              </annotationSets>");
        out.println("            </sentence>");
        
    }

    public void printCorpusPreamble(PrintWriter out, AnnotatedText text) {
        String corpusID = (String) text.getProperty("corpusID");
        String corpusName = (String) text.getProperty("corpusName");
        
        out.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
        out.println("<corpus ID=\"" + corpusID + "\" name=\"" + corpusName 
                    + "\" XMLCreated=\"" + new Date() + "\">");
        out.println("  <documents>");
    }

    public void printDocumentPreamble(PrintWriter out, AnnotatedText text) {
        String docID = (String) text.getProperty("docID");
        String docDesc = (String) text.getProperty("docDesc");
        out.println("    <document ID=\"" + docID 
                + "\" description=\"" + docDesc + "\">");        
        out.println("      <paragraphs>");

    }
    
    public void printParagraphPreamble(PrintWriter out, AnnotatedText text) {
        String parID = (String) text.getProperty("parID");
        String docOrder = (String) text.getProperty("docOrder");
        out.println("        <paragraph ID=\"" + parID
                    + "\" documentOrder=\"" + docOrder + "\">");
        out.println("          <sentences>");
    }
    
    public void printParagraphEnd(PrintWriter out) {
        out.println("          </sentences>");
        out.println("        </paragraph>");        
    }
    
    public void printDocumentEnd(PrintWriter out) {
        out.println("      </paragraphs>");
        out.println("    </document>");
    }
    
    public void printCorpusEnd(PrintWriter out) {
        out.println("  </documents>");
        out.println("</corpus>");        
    }
    
}
    