/*
 FrameNetLUAnnotation.java
  
 Created on Nov 9, 2006 by Richard Johansson (richard@cs.lth.se).

 $Log: FrameNetLUAnnotation.java,v $
 Revision 1.1  2009/01/16 10:04:53  johansson
 Added to brenta repository.

 Revision 1.10  2008/03/11 13:27:28  richard
 Backwards compatible with 1.5.

 Revision 1.9  2007/10/08 12:31:34  richard
 Refactored again to remove atoms.

 Revision 1.8.2.1  2007/10/08 12:18:31  richard
 Added iterateTexts.

 Revision 1.8  2007/10/04 13:58:10  richard
 Overlaps.

 Revision 1.7  2007/08/23 06:37:56  richard
 Sets itype.

 Revision 1.6  2007/08/21 07:24:52  richard
 IDs.

 Revision 1.5  2007/08/17 10:18:40  richard
 Saves more information.

 Revision 1.4  2007/08/13 12:25:18  richard
 Restructured layers.

 Revision 1.3  2007/08/10 13:29:53  richard
 Disables validation.

 Revision 1.2  2006/12/22 15:40:58  richard
 Bugs.

 Revision 1.1  2006/11/13 16:21:23  richard
 Added the file.

   
 */
package se.lth.cs.nlp.nlputils.annotations;

import java.io.*;
import java.util.*;

import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.*;

import javax.xml.parsers.*;

/* For debugging. */
import se.lth.cs.nlp.nlputils.framenet.*;
import se.lth.cs.nlp.nlputils.core.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class FrameNetLUAnnotation extends AnnotationType {

    private FrameNetLUAnnotation() {}

    private static final FrameNetLUAnnotation instance = new FrameNetLUAnnotation();

    public static FrameNetLUAnnotation instance() {
        return instance;
    }

    private int idCounter = 0;
    private int parCounter = 0;
    private static final String ID = "ID";

    private boolean allowOverlaps = true;
    public void setOverlapsAllowed(boolean b) {
        allowOverlaps = b;
    }
    
    public List<AnnotatedText> createTexts(InputStream in) throws IOException {
        try {
            XMLReader reader = makeXMLReader(); 
            try {
                reader.setFeature("http://apache.org/xml/features/nonvalidating/load-external-dtd",
                        false);
            } 
            catch (Exception e) {
                System.err.println("could not set parser feature");
            }

            XMLCallback c = new XMLCallback();
            reader.setContentHandler(c);
            reader.parse(new org.xml.sax.InputSource(in));

            return c.texts;
        } catch(IOException e) {
            throw e;
        } catch(Exception e) {
            e.printStackTrace();
            throw new IOException("Couldn't parse!");
        }        
    }
    
    public Iterator<AnnotatedText> iterateTexts(InputStream in) throws IOException {
        XMLReader reader = null;
        try {
            reader = makeXMLReader();
            reader.setFeature("http://apache.org/xml/features/nonvalidating/load-external-dtd",
                    false);
        } 
        catch (Exception e) {
            throw new IOException(e.getMessage()); // 1.5
        }
        XMLCallback c = new XMLCallback();
        c.texts = null;
        reader.setContentHandler(c);

        throw new UnsupportedOperationException("unimplemented");
    }
    
    
    
    public AnnotatedText createLayers(InputStream in) throws IOException {
        throw new UnsupportedOperationException("Unsupported!");
    }

    private final class XMLCallback extends DefaultHandler implements ContentHandler {

        ArrayList<AnnotatedText> texts = new ArrayList<AnnotatedText>(); 

        private String currentFrame;
        private String currentLU;
        private String currentStatus;
        
        private AnnotatedText currentText;
        private AnnotationLayer currentLayer;
        private StringBuilder currentChars;

        private ArrayList<Span> poss;

	private String docID;
	private String docDesc;
	private String parID;
	private String docOrder;
	private String corpusID;
	private String corpusName;

        private String luID;
        
	private String currentAnnotationSet;
        
        public void startElement(String namespace, String localname,  
                String type, Attributes attributes) {            
            if(type.equals("lexunit-annotation")) {                
                currentFrame = attributes.getValue("frame");
                currentLU = attributes.getValue("name");
		corpusID = attributes.getValue(ID);
		corpusName = "LU: " + currentLU + "/" + currentFrame;
		parCounter = 0;
                luID = attributes.getValue(ID);
            } else if(type.equals("annotationSet")) {
                currentText = new AnnotatedText();                
                currentStatus = attributes.getValue("status");
                if(currentStatus == null)
                    System.err.println("No status found!");

		currentText.setProperty("corpusID", corpusID);
		currentText.setProperty("corpusName", corpusName);

		currentText.setProperty("docID", docID);
		currentText.setProperty("docDesc", docDesc);

		currentText.setProperty("parID", parID);
		currentText.setProperty("docOrder", docOrder);
		
		currentAnnotationSet = attributes.getValue(ID);
		
		String sentenceID = "" + idCounter++;
		currentText.setProperty(ID, sentenceID);

            } else if(type.equals("layer")) {
                String name = attributes.getValue("name");
                if(name == null)
                    throw new RuntimeException("no name for layer");
                for(int i = currentText.layers.size() - 1; i >= 0; i--) {
                    AnnotationLayer l = currentText.layers.get(i);
                    if(l.label.equals(name)
                       || FrameNetAnnotationLayers.isPosLayerLabel(l.label)
                          && FrameNetAnnotationLayers.isPosLayerLabel(name)) {
                        currentLayer = l;
                        break;
                    }
                }
                if(currentLayer == null) {
                    currentLayer = new AnnotationLayer(name);
                    if(name.equals("Target")) {
                        currentLayer.setProperty("frame", currentFrame);
                        currentLayer.setProperty("lu", currentLU);
                        currentLayer.setProperty("status", currentStatus);
			currentLayer.setProperty("annotationSetID", currentAnnotationSet);
			currentLayer.setProperty("luID", luID);
                        // TODO par
                    }
                    currentLayer.setProperty(ID, attributes.getValue(ID));
                    currentText.layers.add(currentLayer);
                }
                
            } else if(type.equals("label")) {
                String name = attributes.getValue("name");
                if(name == null)
                    throw new RuntimeException("no name for span");
                String start_s = attributes.getValue("start");
                String end_s = attributes.getValue("end");
                String itype = attributes.getValue("itype");

                int start, end;
                
                if(start_s == null) {
                    if(itype == null)
                        throw new RuntimeException("Neither span nor itype declared!");
                    start = 0; //-1;
                    end = 0; //-1;
                } else {
                    if(itype != null) {
                        throw new RuntimeException("Both span and itype declared!");       
                    }
                    start = Integer.parseInt(start_s);
                    end = Integer.parseInt(end_s) + 1;
                }
                Span sp = new Span();
                sp.start = start;
                sp.end = end;
                sp.label = name;
                if(itype != null)
                    sp.setProperty("itype", itype);
                sp.id = attributes.getValue(ID);
                
                if(allowOverlaps || !currentLayer.overlaps(sp))
                    currentLayer.spans.add(sp);
            } else if(type.equals("sentence")) {
                
            } else if(type.equals("parts-of-speech")) {
                poss = new ArrayList<Span>();
                currentText.setProperty("pos-type", attributes.getValue("type"));
            } else if(type.equals("pos")) {
                Span pos = new Span();
                pos.start = Integer.parseInt(attributes.getValue("start"));
                pos.end = 1 + Integer.parseInt(attributes.getValue("end"));
                pos.label = attributes.getValue("name");
                if(pos.end > pos.start)
                    poss.add(pos);
            } else if(type.equals("text")) {
                currentChars = new StringBuilder();
            } else if(type.equals("subcorpus")) {
		docDesc = attributes.getValue("name");
		docID = "" + idCounter++;
		parID = "" + idCounter++;
		docOrder = "" + ++parCounter;
	    }
        }
        
        public void endElement(String namespace, String localname,  
                String type) {
            if(type.equals("annotationSet")) {
                for(int i = 0; i < currentText.layers.size(); i++) {
                    AnnotationLayer l = currentText.layers.get(i);
                    if(l.label.equals("Target")) {
                        currentText.layers.remove(i);
                        if(currentText.layers.get(0).label.equals("NER"))
                            currentText.layers.add(1, l);
                        else
                            currentText.layers.add(0, l);
                    }
                }
                FrameNetAnnotationLayers.checkLayers(currentText);
		AnnotationLayer nel = currentText.layers.get(0);
		Object neID = nel.getProperty(ID);
		if(neID == null) {
		    nel.setProperty("annotationSetID", "" + idCounter++);
		    nel.setProperty(ID, idCounter++);
		}
		if(texts != null)
		    texts.add(currentText);
                else
                    throw new UnsupportedOperationException("unimplemented");

                currentText = null;
            } else if(type.equals("layer")) {
                currentLayer = null;
            } else if(type.equals("sentence")) {
            } else if(type.equals("parts-of-speech")) {                    
                Collections.sort(poss, Span.NestingOrder.instance());
                currentText.tokens = new String[poss.size()];
                currentText.poss = new String[poss.size()];
                for(int i = 0; i < poss.size(); i++) {
                    Span s = poss.get(i);
                    currentText.tokens[i] = currentChars.substring(s.start, s.end);
                    currentText.poss[i] = s.label;                    
                }
            } else if(type.equals("text")) {
                currentText.text = currentChars.toString();
                currentChars = null;
            }
        }
        
        public void characters(char[] ch, int start, int len) {
            if(currentChars != null)
                currentChars.append(new String(ch, start, len));
        }
        
    }
    
    private XMLReader makeXMLReader() throws SAXException, ParserConfigurationException { 
        javax.xml.parsers.SAXParserFactory saxParserFactory 
        = javax.xml.parsers.SAXParserFactory.newInstance(); 
        javax.xml.parsers.SAXParser saxParser = saxParserFactory.newSAXParser(); 
        org.xml.sax.XMLReader parser = saxParser.getXMLReader(); 
        return parser; 
    }

    private static void checkCoreConsistency(AnnotationLayer<? extends Span> roleLayer,
            String frame,
            FrameNet framenet) {

        Set<Set<String>> coreSets = framenet.getCoreSets(frame);
        
        Set<Pair<String, String>> excludes = framenet.getExcludes(frame);
        Set<Pair<String, String>> requires = framenet.getRequires(frame);
        
        //System.out.println("coreSets = " + coreSets);
        
        for(Set<String> cs: coreSets) {
            boolean seen = false;
            for(Span sp: roleLayer.spans)
                if(cs.contains(sp.label)) {
                    seen = true;
                    break;
                }
            if(!seen) {
                //throw new RuntimeException("Core set unmatched: " + cs);
            }
        }

        for(Span sp1: roleLayer.spans)
            for(Pair<String, String> p: excludes)
                if(p.left.equals(sp1.label))
                    for(Span sp2: roleLayer.spans)
                        if(p.right.equals(sp2.label)) {
                            //throw new RuntimeException(sp1.label + " excludes " + sp2.label);
                        }
        
        for(Span sp1: roleLayer.spans)
            for(Pair<String, String> p: requires)
                if(p.left.equals(sp1.label)) {
                    boolean seen = false;
                    for(Span sp2: roleLayer.spans)
                        if(p.right.equals(sp2.label)) {
                            seen = true;
                            break;
                        }
                    if(!seen) {
                        throw new RuntimeException(sp1.label + " requires " + p.right);      
                    }
                }
                            
        //System.exit(0);
        
    }
    
    public static void main(String[] argv) {
        FrameNetLUAnnotation f = FrameNetLUAnnotation.instance();
        AnnotatedText t = null;

        try {
            FrameNet fn = new FrameNet("/slask/fn13/frXML/frames.xml",
                    "/slask/fn13/frXML/frRelation.xml", true);
            
            for(String file: argv) {
                //System.err.println(file);
                List<AnnotatedText> texts = f.createTexts(new FileInputStream(file));
                for(AnnotatedText text: texts) {
                    //System.out.println(text);
                    t = text;
                    //RawXMLAnnotation.instance().printLayers(System.out, text);

                    /*checkCoreConsistency(text.layers.get(2),
                                         (String) text.getProperty("frame"),
                                         fn);*/
                    
                    System.out.println(text);
                    
                    //RawXMLAnnotation.instance().printLayers(System.out, t);
                    //if(t.poss != null)
                    //    System.out.println(Arrays.asList(t.poss));
                    
                    for(AnnotationLayer<Span> layer: text.layers)
                        for(Span span: layer.spans) {
                            if(span.start == span.end)
                                continue;
                            
                            if(Character.isWhitespace(text.text.charAt(span.start)))
                                throw new RuntimeException("Whitespace at span start!");
                            if(span.start > 0
                               && !Character.isWhitespace(text.text.charAt(span.start - 1)))
                                throw new RuntimeException("No whitespace before span start!");            
                            if(Character.isWhitespace(text.text.charAt(span.end - 1)))
                                throw new RuntimeException("Whitespace at span end!");
                            if(span.end < text.text.length()
                               && !Character.isWhitespace(text.text.charAt(span.end)))
                                throw new RuntimeException("No whitespace after span end!");     
                        }
                    
                    
                }
            }
        } catch(Exception e) {
            System.err.println(t);
            try {
                RawXMLAnnotation.instance().printLayers(System.err, t);
            } catch(Exception e2) {
            }
            
            e.printStackTrace();
        }
    }
}
    