package se.lth.cs.nlp.nlputils.annotations;

import java.io.*;
import java.util.*;

import se.lth.cs.nlp.nlputils.core.Strings;
import se.lth.cs.nlp.nlputils.core.Util;
import se.lth.cs.nlp.nlputils.core.PRFStats;

public class IOBAnnotation {

	private IOBAnnotation() {}
	
	private static final IOBAnnotation instance = new IOBAnnotation();

    public static IOBAnnotation instance() {
        return instance;
    }

    public AnnotatedText createText(BufferedReader br, int tokenColumn, int[] annotationColumns) throws IOException {
    	return createText(br, tokenColumn, annotationColumns, false, false);
    }
    
    private static boolean isOutside(String s) {
    	if(s.equals("O"))
    		return true;
    	if(s.equals("0"))
    		return true;
    	if(s.equals("_"))
    		return true;
    	return false;
    }
    
    public AnnotatedText createText(BufferedReader br, int tokenColumn, 
    		int[] annotationColumns, boolean singleSentence, boolean allowNonIOB) throws IOException {

    	AnnotatedText out = new AnnotatedText();
    	
    	ArrayList<String> tokens = new ArrayList();
    	AnnotationLayer[] layers = new AnnotationLayer[annotationColumns.length];
    	for(int i = 0; i < layers.length; i++)
    		layers[i] = new AnnotationLayer();
    	
    	Span[] openSpans = new Span[annotationColumns.length];
    	
    	String line = br.readLine();
    	if(line == null) {
    		return null;
    	}
    	int index = 0;
    	while(line != null) {
    		line = line.trim();
    		
    		//System.out.print(line);
    		
    		if(!line.equals("")) {
    			String[] ts = line.split("\\s+");

    			tokens.add(ts[tokenColumn]);
    			for(int i = 0; i < annotationColumns.length; i++) {
    				int ix = annotationColumns[i];

    				String tag = ts[ix];
    				tag = decodeSucheta(tag);
    				
    				if(!isOutside(tag) && !tag.startsWith("B-")
    						&& !tag.startsWith("E-") && !tag.startsWith("I-")) {
    					//throw new IllegalArgumentException("Illegal tag: " + tag);
    					if(allowNonIOB)
    						tag = "B-" + tag;
    					else
    						throw new IllegalArgumentException("Illegal tag: " + tag);
    				}
    				
    				if(openSpans[i] != null
    						&& (tag.startsWith("B-") 
    								|| isOutside(tag))) {
    					openSpans[i].tokenEnd = index;
    					layers[i].add(openSpans[i]);
    					openSpans[i] = null;
    				}

    				if(openSpans[i] != null
    						&& (tag.startsWith("I-") 
    								&& !tag.substring(2).equals(openSpans[i].label))) {
    					// erroneous label switch...
    				
    					// close the open span
    					openSpans[i].tokenEnd = index;
    					layers[i].add(openSpans[i]);
    					openSpans[i] = null;

    					// open a new span
    					Span s = new Span();
    					s.label = tag.substring(2);
    					s.tokenStart = index;
    					openSpans[i] = s;    					
    				}
    				
    				if(tag.startsWith("B-")
    				   || openSpans[i] == null && !isOutside(tag)) {
    					Span s = new Span();
    					s.label = tag.substring(2);
    					s.tokenStart = index;
    					openSpans[i] = s;
    				}
    				
    				// E label
    				if(tag.startsWith("E-")) {
    					openSpans[i].tokenEnd = index + 1;
    					layers[i].add(openSpans[i]);
    					openSpans[i] = null;
    				}
    				
    				//System.out.print("\t" + openSpans[i]);
    				
    			}
    			index++;
    		} else {
    			for(int i = 0; i < annotationColumns.length; i++)
    				if(openSpans[i] != null) {
    					openSpans[i].tokenEnd = index;
    					layers[i].add(openSpans[i]);
    					openSpans[i] = null;
    				}
    			if(singleSentence)
    				break;
    		}
    		
    		//System.out.println();
    		
    		line = br.readLine();
    	}
    	
    	out.layers = Arrays.asList(layers);
    	out.tokens = tokens.toArray(new String[0]);
    	
    	return out;
    }
    
    private String decodeSucheta(String tag) {
    	if(tag.equals("Othr"))
    		return "O";
    	if(tag.endsWith("-B"))
    		tag = "B-" + tag.substring(0, tag.length()-2);
    	if(tag.endsWith("-E"))
    		tag = "E-" + tag.substring(0, tag.length()-2);
    	if(tag.endsWith("-I"))
    		tag = "I-" + tag.substring(0, tag.length()-2);
		return tag;
	}

	private static void pad(StringBuilder[] sbs) {
        int max = -1;
        for(int i = 0; i < sbs.length; i++)
            if(sbs[i].length() > max)
                max = sbs[i].length();
        max++;
        for(int i = 0; i < sbs.length; i++) {
            int l = max - sbs[i].length() + 1;
            for(int j = 0; j < l; j++)
                sbs[i].append(' ');
        }        
    }

    public static void printSentenceNiceColumns(String[][] sentence, PrintWriter out) throws IOException {
    	//out.println(Arrays.deepToString(sentence));
    	//out.flush();
    	
    	if(sentence.length == 0)
    		return;
    	int n = sentence[0].length;
    	StringBuilder[] sbs = new StringBuilder[sentence.length];
        for(int i = 0; i < sbs.length; i++)
            sbs[i] = new StringBuilder();
        for(int i = 0; i < n; i++) {
        	for(int j = 0; j < sentence.length; j++)
        		sbs[j].append(sentence[j][i]);
        	if(i < n - 1)
        		pad(sbs);
        }
        for(int i = 0; i < sbs.length; i++)
        	out.println(sbs[i]);
        out.println();
    }
    
    public static void printNiceColumns(String[] argv) {
    	try {
    		BufferedReader br = new BufferedReader(new FileReader(argv[1]));
    		PrintWriter pw = new PrintWriter(System.out);

    		ArrayList<String[]> lines = new ArrayList();
    		
    		String line = br.readLine();
    		while(line != null) {
    			line = line.trim();
    			if(line.isEmpty()) {
    				String[][] lines2 = lines.toArray(new String[0][]);
    				printSentenceNiceColumns(lines2, pw);
    				lines.clear();
    			} else
    				lines.add(line.split("\\s+"));
    			
    			line = br.readLine();
    		}

    		pw.close();
    		
    	} catch(Exception e) {
    		e.printStackTrace();
    	}
    }
    
    private static int[] getTokenInternalIndices(Span s, AnnotatedText t) {
        if(s.tokenEnd - 1 > s.tokenStart)
            return null;
        int start1 = t.tokenStarts[s.tokenStart];
        int end1 = t.tokenEnds[s.tokenEnd-1];
        if(s.start == start1 && s.end >= end1)
            return null;
        return new int[] { s.start-start1, s.end-start1 };
    }
    
    public void printLayers(PrintWriter pw, AnnotatedText text) throws IOException {
		if(text.tokens == null)
			throw new IllegalArgumentException("must be tokenized");
		String[][] layerTags = new String[text.tokens.length][text.layers.size()];
		for(int i = 0; i < text.layers.size(); i++) {
			for(Span s: (AnnotationLayer<Span>) text.layers.get(i)) {
			    layerTags[s.tokenStart][i] = "B-" + s.label;
				for(int j = s.tokenStart + 1; j < s.tokenEnd; j++)
					layerTags[j][i] = "I-" + s.label;
			}
		}
		for(int i = 0; i < text.tokens.length; i++) {
			pw.print(text.tokens[i]);
			for(int j = 0; j < text.layers.size(); j++) {
				pw.print("\t");
				pw.print(layerTags[i][j] == null? "O": layerTags[i][j]);
			}
			pw.println();
		}
		pw.println();
    }
    
    public void printLayers(PrintWriter pw, AnnotatedText text, AnnotationLayer[] layers) throws IOException {
		if(text.tokens == null)
			throw new IllegalArgumentException("must be tokenized");
		String[][] layerTags = new String[text.tokens.length][layers.length];
		for(int i = 0; i < layers.length; i++) {
			for(Span s: (AnnotationLayer<Span>) layers[i]) {
				layerTags[s.tokenStart][i] = "B-" + s.label;
				for(int j = s.tokenStart + 1; j < s.tokenEnd; j++)
					layerTags[j][i] = "I-" + s.label;
			}
		}
		for(int i = 0; i < text.tokens.length; i++) {
			pw.print(text.tokens[i]);
			for(int j = 0; j < layers.length; j++) {
				pw.print("\t");
				pw.print(layerTags[i][j] == null? "O": layerTags[i][j]);
			}
			pw.println();
		}
		pw.println();
    }
    
    public void printLayers(AnnotatedText text) {
		if(text.tokens == null)
			throw new IllegalArgumentException("must be tokenized");
		String[][] layerTags = new String[text.tokens.length][text.layers.size()];
		for(int i = 0; i < text.layers.size(); i++) {
			for(Span s: (AnnotationLayer<Span>) text.layers.get(i)) {
				layerTags[s.tokenStart][i] = "B-" + s.label;
				for(int j = s.tokenStart + 1; j < s.tokenEnd; j++)
					layerTags[j][i] = "I-" + s.label;
			}
		}
		for(int i = 0; i < text.tokens.length; i++) {
			System.out.print(text.tokens[i]);
			for(int j = 0; j < text.layers.size(); j++) {
				System.out.print("\t");
				System.out.print(layerTags[i][j] == null? "O": layerTags[i][j]);
			}
			System.out.println();
		}
		System.out.println();
    }
    
    public void printLayers(PrintWriter pw, AnnotatedText text, boolean spaces) throws IOException {
        if(!spaces)
            printLayers(pw, text);
        if(text.tokens == null)
            throw new IllegalArgumentException("must be tokenized");
        String[][] layerTags = new String[text.tokens.length][text.layers.size()+1];
        
        for(int i = 0; i < text.layers.size(); i++) {
            for(Span s: (AnnotationLayer<Span>) text.layers.get(i)) {

                int[] internal = getTokenInternalIndices(s, text);

                String ioblabel = null;
                if(internal != null)
                    ioblabel = "X-";
                else if(s.tokenStart == s.tokenEnd-1)
                    ioblabel = "U-";
                else
                    ioblabel = "B-";
                    
                layerTags[s.tokenStart][i+1] = ioblabel + s.label;

                if(internal != null)
                    layerTags[s.tokenStart][i+1] += Arrays.toString(internal); 

                for(int j = s.tokenStart + 1; j < s.tokenEnd; j++)
                    layerTags[j][i+1] = "I-" + s.label;
            }
        }
        
        for(int i = 0; i < layerTags.length; i++) {
            layerTags[i][0] = text.tokens[i];
            for(int j = 0; j < layerTags[i].length; j++)
                if(layerTags[i][j] == null)
                    layerTags[i][j] = "O";
        }

        Strings.printColumns(layerTags, spaces, pw);
        pw.println();
    }
    
    public void printLayers(AnnotatedText text, boolean spaces) {
        if(!spaces)
            printLayers(text);
        if(text.tokens == null)
            throw new IllegalArgumentException("must be tokenized");
        String[][] layerTags = new String[text.tokens.length][text.layers.size()];
        
        for(int i = 0; i < text.layers.size(); i++) {
            for(Span s: (AnnotationLayer<Span>) text.layers.get(i)) {
                layerTags[s.tokenStart][i] = "B-" + s.label;
                for(int j = s.tokenStart + 1; j < s.tokenEnd; j++)
                    layerTags[j][i] = "I-" + s.label;
            }
        }
        
        for(int i = 0; i < layerTags.length; i++)
            for(int j = 0; j < layerTags[i].length; j++)
                if(layerTags[i][j] == null)
                    layerTags[i][j] = "O";

        Strings.printColumns(layerTags, spaces);
        System.out.println();
    }
    
    private static PRFStats compare(AnnotatedText text, 
    		AnnotationLayer l1, AnnotationLayer l2,
    		String type, boolean trimPunctuation) {
    	if(type != null) {
			for(Iterator<Span> it = l1.spans.iterator(); it.hasNext(); )
				if(!it.next().label.equals(type))
					it.remove();
			for(Iterator<Span> it = l2.spans.iterator(); it.hasNext(); )
				if(!it.next().label.equals(type))
					it.remove();    				
		}
		if(trimPunctuation) {
			l1.trimPunctuationTokens(text);
			l2.trimPunctuationTokens(text);
		}
		return l1.compareApproxPRF(l2);
    }
    
    public static void eval(String[] argv) {
    	try {    		
    		BufferedReader br = new BufferedReader(new FileReader(argv[1]));

    		String line = br.readLine();
    		String[] tokens = line.split("\\s+");
    		int ncols = tokens.length;

    		boolean trimPunctuation = false;    		
    		String sigTest = null;
    		String compareTo = null;
    		String type = null;
    		for(int i = 2; i < argv.length; i++) {
    			if(argv[i].equals("-trim"))
    				trimPunctuation = true;
    			else if(argv[i].equals("-conf"))
    				sigTest = argv[++i];
    			else if(argv[i].equals("-compareTo"))
    				compareTo = argv[++i];   			
    			else if(type == null)
    				type = argv[i];
    			else
    				throw new RuntimeException("Illegal argument: " + argv[i]);
    		}    				

    		//br = new BufferedReader(new FileReader(argv[1]));
    		br = Util.openFileReader(argv[1]);

    		BufferedReader br2 = null;
    		int ncols2 = -1;
    		if(compareTo != null) {
    			br2 = Util.openFileReader(compareTo);
    			line = br2.readLine();
    			tokens = line.split("\\s+");
        		ncols2 = tokens.length;
        		br2.close();
        		br2 = Util.openFileReader(compareTo);
    		}
    		
    		if(true) {
    			PRFStats stats = new PRFStats(sigTest != null);
    			PRFStats stats2 = null;
    			if(compareTo != null)
    				stats2 = new PRFStats(sigTest != null);
    			
    			AnnotatedText text = instance.createText(br, 0, new int[] { ncols-2, ncols-1 }, true, true);
    			int ntokens1 = 0, ntokens2 = 0;
    			while(text != null) {
    				AnnotationLayer l1 = text.layers.get(0);
    				AnnotationLayer l2 = text.layers.get(1);
    				stats.add(compare(text, l1, l2, type, trimPunctuation));

    				ntokens1 += text.tokens.length + 1;
    				
    				if(compareTo != null) {
    					AnnotatedText text2 = instance.createText(br2, 0, new int[] { ncols2-2, ncols2-1 }, true, true);
    					if(text2 == null) {
    						throw new RuntimeException("!!!");
    					}
        				ntokens2 += text2.tokens.length + 1;
        				if(ntokens1 != ntokens2)
    						throw new RuntimeException("!!!!");
        				
    					PRFStats st2 = compare(text2, 
        						text2.layers.get(0), text2.layers.get(1),
        						type, trimPunctuation);
    					stats2.add(st2);
    				}
    				
    				text = instance.createText(br, 0, new int[] { ncols-2, ncols-1 }, true, true);
    			}
    			System.out.println("Number of sentences: " + stats.size());
    			System.out.format("Correct: %d, guessed: %d, in gold: %d\nProp guessed: %f, prop found: %f, overlaps: %d\n",
    					(int) stats.nCorrect, (int) stats.nGuesses, (int) stats.nInGold,
    					stats.propGuessed, stats.propFound, (int) stats.nOverlap);
    			stats.print();
    			if(compareTo != null) {
    				System.out.println("Compared to:");
    				stats2.print();
    			}
    			if(false) {
    				PrintWriter pw = new PrintWriter(new FileWriter("iobeval.prf"));	    			
    				stats.printSamples(pw);
    				pw.close();
    			}
    			if(sigTest != null) {
    				if(compareTo == null)
    					PRFStats.testSignificance1(stats, sigTest);
    				else {
    					if(sigTest.startsWith("perm_")) {
    						sigTest = sigTest.substring(5);
    						System.out.println("Testing significance of difference using a permutation test.");
    						PRFStats.testSignificance2(stats, stats2, sigTest, true);
    					} else {
    						System.out.println("Testing significance of difference using a bootstrapped confidence interval of the difference.");
    						PRFStats.testSignificance2(stats, stats2, sigTest, false);
    					}
    				}
    			}
    		} else {
    			AnnotatedText text = instance.createText(br, 0, new int[] { ncols-2, ncols-1 });    			
    			
    			AnnotationLayer l1 = text.layers.get(0);
				AnnotationLayer l2 = text.layers.get(1);
    			
				/*double[] result = l1.compareApprox(l2);
        		double ep = result[3] / result[4];
        		double er = result[3] / result[5];
        		double ef = 2*ep*er/(ep+er);
        		
        		double pp = result[6] / result[4];
        		double pr = result[7] / result[5];
        		double pf = 2*pp*pr/(pp+pr);
        		
        		double op = result[8] / result[4];
        		double or = result[8] / result[5];
        		double of = 2*op*or/(op+or);
        		
        		System.out.format("%f\t%f\t%f\n", ep, er, ef);*/
								
				PRFStats stats = l1.compareApproxPRF(l2);
				stats.print();
        		
        		//System.out.println(l1.compare(l2));
    		}

    	} catch(Exception e) {
    		e.printStackTrace();
    	}
    }

    private static final int MAX_BY_EXACT = 0;
    private static final int MAX_BY_PARTIAL = 1;
    private static final int MAX_BY_OVERLAP = 2;
    
    public static void evaloracle(String[] argv) {
    	try {    		
    		BufferedReader br = Util.openFileReader(argv[1]);

    		br.readLine();
    		String line = br.readLine();
    		String[] tokens = line.split("\\s+");
    		int ncols = tokens.length;

    		boolean trimPunctuation = false;
    		String printFile = null;
    		int N = Integer.MAX_VALUE;
    		String type = null;
    		int maxBy = MAX_BY_EXACT;
    		for(int i = 2; i < argv.length; i++) {
    			if(argv[i].equals("-trim"))
    				trimPunctuation = true;
    			else if(argv[i].equals("-nbest"))
    				N = Integer.parseInt(argv[++i]);
    			else if(argv[i].equals("-print"))
    				printFile = argv[++i];
    			else if(argv[i].equals("-maxby")) {
    				String m = argv[++i];
    				if(m.equals("exact"))
    					maxBy = MAX_BY_EXACT; 
    				else if(m.equals("partial"))
    					maxBy = MAX_BY_PARTIAL;
    				else if(m.equals("overlap"))
    					maxBy = MAX_BY_OVERLAP;
    				else
    					throw new RuntimeException("-maxby must be exact|partial|overlap");
    			}   				
    			else if(type == null)
    				type = argv[i];
    			else 
    				throw new RuntimeException("Illegal argument: " + argv[i]);
    		}    				
    		
    		br = Util.openFileReader(argv[1]);
    		PrintWriter pw = null;
    		if(printFile != null)
    			pw = new PrintWriter(new FileWriter(printFile));
    		
    		if(true) {
    			PRFStats stats = new PRFStats(true);

    			PRFStats selection = null;    			
				AnnotatedText maxText = null;
    			
    			int n = 0;
    			
    			while(true) {
    				line = br.readLine();
    				while(line != null && (line.isEmpty() || line.equals("\\n")))
    					line = br.readLine();
    				if(selection != null && (line == null || line.startsWith("# 0 "))) {
    					stats.add(selection);
    					selection = null;
    					n++;
    					if(n % 100 == 0)
    						System.out.println(n);
    					if(pw != null) {
    						instance.printLayers(pw, maxText);
    					}
    					maxText = null;
    				}
    				if(line == null)
    					break;
    				if(!line.startsWith("#"))
    					throw new RuntimeException("illegal format: |" + line + "|");
    				String[] ss = line.split(" ");
    				int i = Integer.parseInt(ss[1]);

    				//System.out.println("i = " + i);
    				
    				AnnotatedText text = instance.createText(br, 0, new int[] { ncols-2, ncols-1 }, true, false);
    				AnnotationLayer l1 = text.layers.get(0);
    				AnnotationLayer l2 = text.layers.get(1);

    				if(i < N) {
    					if(type != null) {
    						for(Iterator<Span> it = l1.spans.iterator(); it.hasNext(); )
    							if(!it.next().label.equals(type))
    								it.remove();
    						for(Iterator<Span> it = l2.spans.iterator(); it.hasNext(); )
    							if(!it.next().label.equals(type))
    								it.remove();    				
    					}
    					if(trimPunctuation) {
    						l1.trimPunctuationTokens(text);
    						l2.trimPunctuationTokens(text);
    					}

    					PRFStats res = l1.compareApproxPRF(l2);
    					if(selection == null) {
    						selection = res;    						
    						maxText = text;
    					} else {
    						boolean isMax = false;
    						switch(maxBy) {
    						case MAX_BY_EXACT:
    							isMax = res.getF1() > selection.getF1();
    							break;
    						case MAX_BY_PARTIAL:
    							isMax = res.getPartialF1() > selection.getPartialF1();
    							break;
    						case MAX_BY_OVERLAP:
    							isMax = res.getOverlapF1() > selection.getOverlapF1();
    							break;
    						}
    						if(isMax) {
    							selection = res;
    							maxText = text;
    						}
    					}
    				}
    			}

    			System.out.println("Number of sentences: " + stats.size());
    			System.out.format("Correct: %d, guessed: %d, in gold: %d\nProp guessed: %f, prop found: %f, overlaps: %d\n",
    					(int) stats.nCorrect, (int) stats.nGuesses, (int) stats.nInGold,
    					stats.propGuessed, stats.propFound, (int) stats.nOverlap);
    			stats.print();
    			if(pw != null)
    				pw.close();
    			
    			if(false) {
    				pw = new PrintWriter(new FileWriter("iobeval.prf"));	    			
    				stats.printSamples(pw);
    				pw.close();
    			}
    		}

    	} catch(Exception e) {
    		e.printStackTrace();
    	}
    }
    
    public static void main(String[] argv) {
    	if(argv.length == 0) {
    		System.err.println("no argument given");
    		System.exit(0);
    	}
    	if(argv[0].equals("-eval"))
    		eval(argv);
    	else if(argv[0].equals("-oracle"))
    		evaloracle(argv);
    	else if(argv[0].equals("-print"))
    		printNiceColumns(argv);
    	else {
    		System.err.println("unknown mode " + argv[0]);
    		System.exit(0);    		
    	}
    }
    
}
