/*
 LineAnnotation.java
  
 Created on Aug 17, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: LineAnnotation.java,v $
 Revision 1.1  2009/01/16 10:04:53  johansson
 Added to brenta repository.

 Revision 1.3  2007/09/13 11:48:26  richard
 Skips empty lines.

 Revision 1.2  2007/09/05 14:35:14  richard
 Added createText.

 Revision 1.1  2007/08/17 10:17:28  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.annotations;

import java.io.*;
import java.util.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class LineAnnotation extends AnnotationType {

    private LineAnnotation() {}

    private static final LineAnnotation instance = new LineAnnotation();

    public static LineAnnotation instance() {
        return instance;
    }
    
    private static final String ENCODING = "UTF-8";
    
    public List<AnnotatedText> createTexts(InputStream in) throws IOException {
        BufferedReader r = new BufferedReader(new InputStreamReader(in, ENCODING));
        ArrayList<AnnotatedText> out = new ArrayList();
        String line = r.readLine();
        while(line != null) {
            line = line.trim();
            if(line.length() > 0) {
                AnnotatedText t = new AnnotatedText();
                t.text = line;
                out.add(t);
            }
            line = r.readLine();
        }
        return out;
    }
    
    public Iterator<AnnotatedText> iterateTexts(InputStream in) throws IOException {
    	BufferedReader r = new BufferedReader(new InputStreamReader(in, ENCODING));
    	return new LineIterator(r);
    }

    private class LineIterator implements Iterator<AnnotatedText> {

    	BufferedReader br;

    	private String currentLine;
    	private boolean ended = false;
    	
    	LineIterator(BufferedReader br) {
    		this.br = br;
    	}
    	
		public synchronized boolean hasNext() {
			if(ended)
				return false;
			if(currentLine != null)
				return true;
			readNextLine();
			return !ended;
		}

		public synchronized AnnotatedText next() {
			if(ended)
				throw new IllegalStateException("no next element");
			if(currentLine == null)
				readNextLine();
			if(ended)
				throw new IllegalStateException("no next element");
			AnnotatedText out = createText(currentLine);
			currentLine = null;
			return out;
		}

		private synchronized void readNextLine() {
			if(currentLine != null)
				throw new RuntimeException("!");
			try {
				currentLine = br.readLine();
			} catch(IOException e) {
				throw new RuntimeException(e);
			}
			if(currentLine == null)
				ended = true;
		}
		
		public void remove() {
			throw new UnsupportedOperationException("unsupported");
		}
    	
    }
    
    public AnnotatedText createText(String s) {
        AnnotatedText t = new AnnotatedText();
        t.text = s;
        return t;
    }
    
}
