/*
 ProcessedFNAnnotation.java
 
 Copyright (C) 2005, Richard Johansson (richard@cs.lth.se).
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.
 
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 Created on Nov 25, 2005 by Richard Johansson (richard@cs.lth.se).
 
 $Log: ProcessedFNAnnotation.java,v $
 Revision 1.1  2009/01/16 10:04:53  johansson
 Added to brenta repository.

 Revision 1.8  2007/08/17 10:17:45  richard
 Deprecated.

 Revision 1.7  2006/09/07 17:51:19  richard
 New revision.

 Revision 1.6  2006/04/11 08:40:17  richard
 Added documentation.
 
 Revision 1.5  2006/04/05 12:08:38  richard
 Sets token indices.
 
 Revision 1.4  2006/03/29 13:56:51  richard
 Put the null-instatiated spans at (0,0).
 
 Revision 1.3  2005/11/30 13:31:11  richard
 Implemented the import.
 
 Revision 1.2  2005/11/30 12:44:39  richard
 Changed the interface.
 
 Revision 1.1  2005/11/30 09:52:25  richard
 Renamed the annotation types.
 
 Revision 1.2  2005/11/30 09:49:09  richard
 Changed the interface.
 
 Revision 1.1  2005/11/28 15:41:00  richard
 Added the file.
 
 
 */
package se.lth.cs.nlp.nlputils.annotations;

import java.util.*;
import java.io.*;

/**
 * An AnnotationType for the simplified FrameNet annotation.
 * @deprecated
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class ProcessedFNAnnotation extends AnnotationType {
    
    private ProcessedFNAnnotation() {}
    
    private static final ProcessedFNAnnotation instance = new ProcessedFNAnnotation();
    
    /**
     * Returns the instance of this AnnotationType.
     * 
     * @return the instance of this AnnotationType.
     */
    public static ProcessedFNAnnotation instance() {
        return instance;
    }
    
    public AnnotatedText createLayers(InputStream in) throws IOException {
        BufferedReader r = new BufferedReader(new InputStreamReader(in));
        String line = r.readLine();
        if(line == null)
            return null;
        else
            return createLayers(line);
    }
    
    public AnnotatedText createLayers(String line1) {
        AnnotatedText out = new AnnotatedText();
        
        int index0 = line1.indexOf(',');
        String frame = line1.substring(0, index0).trim();
        index0 += 1;
        int index1 = line1.indexOf(',', index0);
        String lemma_pos = line1.substring(index0, index1).trim();
        index1 += 1;
        int index2 = line1.indexOf(',', index1);
        String status = line1.substring(index1, index2).trim();
        index2 += 1;
        int index3 = line1.indexOf(']', index2) + 1;
        String roleAnn = line1.substring(index2, index3).trim();
        index3 = line1.indexOf(',', index3) + 1;
        int index4 = line1.indexOf(']', index3) + 1;
        String otherAnn = line1.substring(index3, index4).trim();
        index4 = line1.indexOf(',', index4) + 1;
        int index5 = line1.indexOf(']', index4) + 1;
        String targetAnn = line1.substring(index4, index5).trim();
        index5 = line1.indexOf(':', index5) + 1;
        
        out.text = line1.substring(index5).trim();
        out.properties = new HashMap<String, Object>();
        
        out.properties.put("frame", frame);
        out.properties.put("lemma_pos", lemma_pos);
        out.properties.put("status", status);
        
        out.layers = new ArrayList<AnnotationLayer>();
        out.layers.add(buildLayer(roleAnn, true));
        out.layers.add(buildLayer(otherAnn, false));
        out.layers.add(buildLayer(targetAnn, false));
        
        out.tokens = out.text.split(" ");
        
        for(AnnotationLayer layer: out.layers)
            layer.findTokenIndices(out.tokens);
        
        return out;
    }
    
    private static AnnotationLayer buildLayer(String annotation, boolean isRoleLayer) {
        annotation = annotation.substring(1, annotation.length() - 1).trim();
        AnnotationLayer l = new AnnotationLayer();
        if(annotation.equals(""))
            return l;
        String[] ss = annotation.split(" ");
        for(int i = 0; i < ss.length; i++) {
            Span sp = buildSpan(ss[i], isRoleLayer);
            sp.id = "s" + l.spans.size();
            l.spans.add(sp);
        }
        return l;
    }
    
    private static Span buildSpan(String s, boolean isRoleLayer) {
        s = s.substring(1, s.length() - 1).trim();
        String[] ss = s.split("\\,");
        Span sp = new Span();
        
        if(isRoleLayer) {
            sp.label = ss[0].intern();
            sp.properties = new HashMap<String, Object>();
            //sp.properties.put("role", ss[0].intern());
            sp.properties.put("status", ss[1].intern());
        } else
            sp.label = ss[0].intern();
        
        try {
            sp.start = Integer.parseInt(ss[2]);
            sp.end = Integer.parseInt(ss[3]);
            if(sp.end != -1)
                sp.end += 1;
            else
                sp.start = sp.end = 0;
        } catch(Exception e) {
            e.printStackTrace();
            throw new RuntimeException("Got exception.");
        }
        return sp;
    }
    
}
