/*
 Copyright (2009) Richard Johansson (richard.johansson@gu.se)
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at 
 
 http://www.apache.org/licenses/LICENSE-2.0
 
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License. 
*/

/**
 * 
 */
package se.lth.cs.nlp.nlputils.core;

import java.util.Iterator;
import java.util.NoSuchElementException;

import se.lth.cs.nlp.nlputils.core.CollectionUtils;

public class ArrayQueue<C extends Comparable> implements Iterable<C> {

    private C[] q;
    private int len;
    public ArrayQueue(int N) {
        q = (C[]) new Comparable[N];
        len = 0;
    }
    
    public void add(C item) {
        int pos = len;
        for(int i = len-1; i >= 0; i--) {
            if(q[i].compareTo(item) > 0)
                pos--;
            else
                break;
        }
        if(pos == q.length)
            return;
        if(len == q.length)
            for(int i = q.length-1; i > pos; i--)
                q[i] = q[i-1];                
        else {
            for(int i = len; i > pos; i--)
                q[i] = q[i-1];
            len++;
        }
        q[pos] = item;
    }
    public Iterator<C> iterator() {
        //return new ArrayIterator(q);
        return CollectionUtils.iterator(q);
    }
    public C first() {
        if(len == 0)
            throw new NoSuchElementException("len = 0");
        return q[0];
    }
    public int size() {
        return len;
    }
    public C get(int index) {
        if(index >= len)
            throw new ArrayIndexOutOfBoundsException("length");
        return q[index];
    }
    
}