/*
 BigArray.java
  
 Created on May 10, 2010 by Richard Johansson (johansson@disi.unitn.it).

 $Log: BigArray.java,v $
 Revision 1.1  2010-06-07 11:02:25  johansson
 Added.

   
 */
package se.lth.cs.nlp.nlputils.core;

import java.io.Serializable;
import java.util.Arrays;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class BigArray<V> implements Serializable {

    private static final int DEFAULT_CHUNK = 1 << 24;
    
    private static final long serialVersionUID = 0L;

    private V[][] arrs;
    private int size;

    private int chunkSizeBits;
    private int chunkIndexMask;
    
    public BigArray(int size, int chunkSize) {
        this.size = size;
        int bit = Util.msb(chunkSize) - 1;
        if((chunkSize & ~(1 << bit)) != 0)
            throw new IllegalArgumentException("Chunk size must be a power of 2.");                
        chunkSizeBits = bit;        
        chunkIndexMask = (1 << bit) - 1;        
        int nChunks = (int) Math.ceil((double) size / chunkSize);
        arrs = (V[][]) new Object[nChunks][];
        int lastSize = size - chunkSize*(nChunks - 1);
        for(int i = 0; i < nChunks - 1; i++)
            arrs[i] = (V[]) new Object[chunkSize];        
        arrs[nChunks-1] = (V[]) new Object[lastSize];
    }    
    
    public BigArray(int size) {
        this(size, DEFAULT_CHUNK);
    }
    
    public int length() {
        return size;
    }
    
    public void set(int index, V value) {
        if(index < 0 || index >= size)
            throw new ArrayIndexOutOfBoundsException("illegal index: " + index);
        int index1 = index >>> chunkSizeBits;
        int index2 = index & chunkIndexMask;
        arrs[index1][index2] = value;
    }

    public void fill(V value) {
        for(int i = 0; i < arrs.length; i++)
            Arrays.fill(arrs[i], value);
    }
    
    public V get(int index) {
        if(index < 0 || index >= size)
            throw new ArrayIndexOutOfBoundsException("illegal index: " + index);
        int index1 = index >>> chunkSizeBits;
        int index2 = index & chunkIndexMask;
        return arrs[index1][index2];
    }
    
}
