/*
 BigIntArray.java
  
 Created on Apr 8, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: BigIntArray.java,v $
 Revision 1.2  2009/11/16 11:36:04  johansson
 Renamed for compatibility with Windows systems.

 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.3  2008/04/08 11:43:43  richard
 More methods.

 Revision 1.2  2008/04/08 09:17:04  richard
 Serializable.

 Revision 1.1  2008/04/08 08:41:24  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.core;

import java.io.Serializable;
import java.util.Arrays;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class BigIntArray implements Serializable {

    private static final int DEFAULT_CHUNK = 1 << 24;
    
    private static final long serialVersionUID = 0L;

    private int[][] arrs;
    private int size;

    private int chunkSizeBits;
    private int chunkIndexMask;
    
    public BigIntArray(int size, int chunkSize) {
        this.size = size;
        int bit = Util.msb(chunkSize) - 1;
        if((chunkSize & ~(1 << bit)) != 0)
            throw new IllegalArgumentException("Chunk size must be a power of 2.");                
        chunkSizeBits = bit;        
        chunkIndexMask = (1 << bit) - 1;        
        int nChunks = (int) Math.ceil((double) size / chunkSize);
        arrs = new int[nChunks][];
        int lastSize = size - chunkSize*(nChunks - 1);
        for(int i = 0; i < nChunks - 1; i++)
            arrs[i] = new int[chunkSize];        
        arrs[nChunks-1] = new int[lastSize];
    }    
    
    public BigIntArray(int size) {
        this(size, DEFAULT_CHUNK);
    }
    
    public int length() {
        return size;
    }
    
    public void set(int index, int value) {
        if(index < 0 || index >= size)
            throw new ArrayIndexOutOfBoundsException("illegal index: " + index);
        int index1 = index >>> chunkSizeBits;
        int index2 = index & chunkIndexMask;
        arrs[index1][index2] = value;
    }

    public void fill(int value) {
        for(int i = 0; i < arrs.length; i++)
            Arrays.fill(arrs[i], value);
    }
    
    public int modify(int index, int value) {
        if(index < 0 || index >= size)
            throw new ArrayIndexOutOfBoundsException("illegal index: " + index);
        int index1 = index >>> chunkSizeBits;
        int index2 = index & chunkIndexMask;
        arrs[index1][index2] += value;
        return arrs[index1][index2];
    }

    public int get(int index) {
        if(index < 0 || index >= size)
            throw new ArrayIndexOutOfBoundsException("illegal index: " + index);
        int index1 = index >>> chunkSizeBits;
        int index2 = index & chunkIndexMask;
        return arrs[index1][index2];
    }
    
}
