package se.lth.cs.nlp.nlputils.core;

import java.util.Arrays;
import java.util.Random;

public class BloomHistogram {

	/* TODO implement Wang et al instead */
	
	private static final int LOG_INT_SIZE = 5;
	private static final int INT_LOWER_BITS = (1 << LOG_INT_SIZE) - 1; 
	private int[][] stored; 

	private int[] hist;
	
	private int nhashes, mod, mod2;
	
	public BloomHistogram(int size, int nhashes) {
		int logsize = Util.msb(Math.max(0, size - 1));
		if(size > 1 << logsize)
			logsize++;
		int len = 1 << Math.max(0, logsize - LOG_INT_SIZE);		
		stored = new int[len][nhashes];		
		mod = len - 1;
		this.nhashes = nhashes;
		this.hist = new int[1 << logsize];
		this.mod2 = (1 << logsize) - 1;
	}

	public void add(String k) {
		int h = 0;
		for(int i = 0; i < nhashes; i++) {
			h = i == 0? ihash(k, i): ihashNext(h);			
			int ix = (h >>> LOG_INT_SIZE) & mod;
			int flag = 1 << (h & INT_LOWER_BITS);			
			stored[ix][i] |= flag;
		} 
		h = ihashNext(h);
		int ix = h & mod2;
		hist[ix]++;
	}
	
	
	public int getFrequency(String k) {
		int h = 0;
		for(int i = 0; i < nhashes; i++) {
			h = i == 0? ihash(k, i): ihashNext(h);	
			int ix = (h >>> LOG_INT_SIZE) & mod;
			int flag = 1 << (h & INT_LOWER_BITS);			
			if((stored[ix][i] & flag) == 0)
				return 0;
		}
		h = ihashNext(h);
		int ix = h & mod2;
		return hist[ix];
	}
	
	private static int ihash(String s, int start) {
		int hash = start;
		int l = s.length();
		for(int i = 0; i < l; i++) {			
			hash += s.charAt(i);
			hash += (hash << 10L);
			hash ^= (hash >> 6L);
		}
		hash += (hash << 3L);
		hash ^= (hash >> 11L);
		hash += (hash << 15L);
		return hash;
	}
	
	private static int ihashNext(int hash) {
		hash += (hash << 3L);
		hash ^= (hash >> 11L);
		hash += (hash << 15L);
		return hash;
	}
	
	private static long lhash(String s, long start) {
		long hash = start;
		int l = s.length();
		for(int i = 0; i < l; i++) {			
			hash += s.charAt(i);
			hash += (hash << 10L);
			hash ^= (hash >> 6L);
		}
		hash += (hash << 3L);
		hash ^= (hash >> 11L);
		hash += (hash << 15L);
		return hash;
	}
	
	public static void main(String[] argv) {
		
		BloomHistogram bf = new BloomHistogram((1 << 12)*8, 8);

		System.out.println(bf.stored.length * bf.nhashes + bf.hist.length);
		
		Histogram<String> hist = new Histogram();
		
		int niter = 100000;
		int nvals = 1000000;
		
		Random rand = new Random(0);
		
		for(int i = 0; i < niter; i++) {
			String s = "" + rand.nextInt(nvals);
			hist.add(s);
			bf.add(s);			
		}

		for(IntObjPair<String> p: hist.asSortedList()) {
			System.out.println(p.right + "\t" + p.left + "\t" + bf.getFrequency(p.right));
		}
		
	}
	
}
