package se.lth.cs.nlp.nlputils.core;

import java.util.*;

public class ConfusionMatrix<T> {

	private HashMap<T, Histogram<T>> m = new HashMap();
	private HashSet<T> all = new HashSet();
	
	private int ncorr, ntotal;
	
	public ConfusionMatrix() {		
	}

	public void add(T correct, T guess) {
		Histogram<T> h = m.get(correct);
		if(h == null) {
			h = new Histogram();
			m.put(correct, h);
		}
		h.add(guess);
		all.add(correct);
		all.add(guess);
		
		if(correct.equals(guess))
			ncorr++;
		ntotal++;		
	}

	public double[] getPRF(T cls) {
		int nguess = 0, ntotal = 0, ncorr = 0;			
		for(T t: m.keySet()) {
			Histogram<T> h = m.get(t);
			if(t.equals(cls)) {
				for(T t2: all) {
					int f = h.getFrequency(t2);
					ntotal += f;
					if(t2.equals(cls)) {
						ncorr += f;
						nguess += f;
					}
				}
			} else {
				for(T t2: all) {
					if(t2.equals(cls)) {
						int f = h.getFrequency(t2);
						nguess += f;
					}
				}
			}
		}
		double p, r, f;
		
		if(nguess == 0)
			p = 1.0;
		else
			p = (double) ncorr / nguess;
		
		if(ntotal == 0)
			r = 1.0;
		else
			r = (double) ncorr / ntotal;

		if(p + r > 0)
			f = 2*p*r/(p+r);
		else
			f = 0.0;
		return new double[] { p, r, f, ncorr, nguess, ntotal };
	}
	
	public double getMicroAccuracy() {
		return (double) ncorr / ntotal;
	}

	public double[] getMacroPRF() {
		double p = 0, r = 0, ncorr = 0, nguess = 0, ntotal = 0;
		int n = 0;
		for(T t: m.keySet()) {
			n++;
			double[] prf = getPRF(t);
			p += prf[0];
			r += prf[1];
			ncorr += prf[3];
			nguess += prf[4];
			ntotal += prf[5];
		}
		p /= n;
		r /= n;
		double f = 2*p*r/(p+r);
		return new double[] { p, r, f, ncorr, nguess, ntotal };		
	}

	
	public String toString() {
		ArrayList<Comparable> l = new ArrayList(all);
		try {
			Collections.sort(l);
		} catch(Exception e) {			
		}
		
		StringBuilder out = new StringBuilder();

		out.append("Micro accuracy: " + ncorr + " / " + ntotal + " = ");
		out.append((double) Math.round(getMicroAccuracy() * 100000) / 100000);
		out.append("\n");
		
		out.append("Macro: " + Arrays.toString(getMacroPRF()) + "\n");
		
		for(int i = 0; i < l.size(); i++)
			out.append(i + ". " + l.get(i) + "\n");
		out.append("\n");
		int count = 0;
		for(Object k: l) {
			Histogram<T> h = m.get((T) k);
			out.append(count + ": ");
			if(h == null) {
				for(int i = 0; i < l.size(); i++)
					out.append(" 0");
			} else {
				for(Object k2: l)
					out.append(" " + h.getFrequency((T) k2));
			}
			out.append("\n");
			count++;
		}
		
		out.append("\n");		

		count = 0;
		for(Object k: l) {
			out.append(count + ": " + Arrays.toString(getPRF((T) k)) + "\n");
			count++;
		}
		
		return out.toString();
	}
	
}
