/*
 Copyright (2008) Richard Johansson (richard.johansson@gu.se)
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at 
 
 http://www.apache.org/licenses/LICENSE-2.0
 
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License. 
*/

/*
 DoubleIntPair.java
  
 Created on Apr 23, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: DoubleIntPair.java,v $
 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.1  2008/04/23 10:51:41  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.core;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class DoubleIntPair {

    private static final long serialVersionUID = 0;

    public double left;
    public int right;
    
    public DoubleIntPair(double left, int right) {
        this.left = left;
        this.right = right;
    }
    
    public String toString() {
        return "(" + left + ", " + right + ")";
    }
    
    public boolean equals(Object o) {
        if(!(o instanceof DoubleIntPair))
            return false;
        DoubleIntPair p = (DoubleIntPair) o;
        return left == p.left && right == p.right;
    }
    
    private static final int HASH_OFFSET = 100003;

    private static final long LOWER = (1L << 32L) - 1L;
    private static final long UPPER = LOWER << 32;
    
    public int hashCode() {
        long l = Double.doubleToRawLongBits(left);
        int lefthash = (int) (l & LOWER) + (int) (l & UPPER >>> 32L);
        return HASH_OFFSET*lefthash + right;
    }
 
    public DoubleIntPair clone() {
        return new DoubleIntPair(left, right);
    }
    
    public static final Comparator<DoubleIntPair> BY_LEFT = new Comparator<DoubleIntPair>() {
        public int compare(DoubleIntPair p1, DoubleIntPair p2) {
            return Double.compare(p1.left, p2.left);
        }
    };
    
    public static final Comparator<DoubleIntPair> BY_RIGHT = new Comparator<DoubleIntPair>() {
        public int compare(DoubleIntPair p1, DoubleIntPair p2) {
            return p1.right - p2.right;
        }
    };
    
	public static Collection<Double> lefts(Collection<DoubleIntPair> coll) {
		Collection<Double> out = (Collection) CollectionUtils.newCollection(coll);
		for(DoubleIntPair p: coll) {
			out.add(p.left);
		}
		return out;
	}
	
	public static Collection<Integer> rights(Collection<DoubleIntPair> coll) {
		Collection<Integer> out = (Collection) CollectionUtils.newCollection(coll);
		for(DoubleIntPair p: coll) {
			out.add(p.right);
		}
		return out;
	}
    
	public static ArrayList<Double> lefts(ArrayList<DoubleIntPair> coll) {
		ArrayList<Double> out = new ArrayList();
		for(DoubleIntPair p: coll) {
			out.add(p.left);
		}
		return out;
	}
	
	public static ArrayList<Integer> rights(ArrayList<DoubleIntPair> coll) {
		ArrayList<Integer> out = new ArrayList();
		for(DoubleIntPair p: coll) {
			out.add(p.right);
		}
		return out;
	}
    
	public static double[] leftsArray(Collection<DoubleIntPair> coll) {
		double[] out = new double[coll.size()];
		int index = 0;
		for(DoubleIntPair p: coll) {
			out[index++] = p.left;
		}
		return out;
	}
	
	public static int[] rightsArray(Collection<DoubleIntPair> coll) {
		int[] out = new int[coll.size()];
		int index = 0;
		for(DoubleIntPair p: coll) {
			out[index++] = p.right;
		}
		return out;
	}
	
}
