/*
 Copyright (2007) Richard Johansson (richard.johansson@gu.se)
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at 
 
 http://www.apache.org/licenses/LICENSE-2.0
 
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License. 
*/

/*
 IntObjPair.java
  
 Created on Aug 14, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: DoubleObjPair.java,v $
 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.2  2008/04/21 08:39:10  richard
 Added comparator.

 Revision 1.1  2007/09/28 14:45:10  richard
 Added.

 Revision 1.1  2007/08/17 10:20:11  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.core;

import java.io.Serializable;
import java.util.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class DoubleObjPair<T> implements Serializable, Comparable<DoubleObjPair<T>> {

    private static final long serialVersionUID = 0;

    public double left;
    public T right;
    
    public DoubleObjPair(double left, T right) {
        this.left = left;
        this.right = right;
    }
    
    public String toString() {
        return "(" + left + ", " + right + ")";
    }
    
    public boolean equals(Object o) {
        if(!(o instanceof DoubleObjPair))
            return false;
        DoubleObjPair p = (DoubleObjPair) o;
        return left == p.left && right.equals(p.right);
    }
    
    private static final int HASH_OFFSET = 100003;

    private static final long LOWER = (1L << 32L) - 1L;
    private static final long UPPER = LOWER << 32;
    
    public int hashCode() {
        long l = Double.doubleToRawLongBits(left);
        int lefthash = (int) (l & LOWER) + (int) (l & UPPER >> 32);
        return HASH_OFFSET*lefthash + right.hashCode();
    }
 
    public DoubleObjPair clone() {
        return new DoubleObjPair(left, right);
    }
    
    public static final Comparator<DoubleObjPair> BY_LEFT = new Comparator<DoubleObjPair>() {
        public int compare(DoubleObjPair p1, DoubleObjPair p2) {
            return Double.compare(p1.left, p2.left);
        }
    };
    
    public static final Comparator<DoubleObjPair> BY_RIGHT = new Comparator<DoubleObjPair>() {
        public int compare(DoubleObjPair p1, DoubleObjPair p2) {
            return ((Comparable) p1.right).compareTo(p2.right);
        }
    };
    
	public static <T> Collection<Double> lefts(Collection<DoubleObjPair<T>> coll) {
		Collection<Double> out = (Collection) CollectionUtils.newCollection(coll);
		for(DoubleObjPair p: coll) {
			out.add(p.left);
		}
		return out;
	}
	
	public static <T> double[] leftsArray(Collection<DoubleObjPair<T>> coll) {
		double[] out = new double[coll.size()];
		int index = 0;
		for(DoubleObjPair<T> p: coll) {
			out[index++] = p.left;
		}
		return out;
	}
	
	public static <T> Collection<T> rights(Collection<DoubleObjPair<T>> coll) {
		Collection<T> out = (Collection) CollectionUtils.newCollection(coll);
		for(DoubleObjPair<T> p: coll) {
			out.add(p.right);
		}
		return out;
	}
    
	public static <T> ArrayList<Double> lefts(ArrayList<DoubleObjPair<T>> coll) {
		ArrayList<Double> out = new ArrayList();
		for(DoubleObjPair<T> p: coll) {
			out.add(p.left);
		}
		return out;
	}
	
	public static <T> ArrayList<T> rights(ArrayList<DoubleObjPair<T>> coll) {
		ArrayList<T> out = new ArrayList();
		for(DoubleObjPair<T> p: coll) {
			out.add(p.right);
		}
		return out;
	}
	
	public static <T> T[] rightsArray(Collection<DoubleObjPair<T>> coll) {
		T[] out = (T[]) new Object[coll.size()];
		int index = 0;
		for(DoubleObjPair<T> p: coll) {
			out[index++] = p.right;
		}
		return out;
	}

    @Override
    public int compareTo(DoubleObjPair<T> o) {
        return Double.compare(left, o.left);
    }
    
	
}
