/*
 GenerationalLongDoubleMap.java
  
 Created on June 10, 2009 by Richard Johansson (johansson@disi.unitn.it).

 $Log: GenerationalLongDoubleMap.java,v $
 Revision 1.1  2009/11/16 11:36:31  johansson
 Added.

   
 */
package se.lth.cs.nlp.nlputils.core;

import java.io.Serializable;

import java.util.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class GenerationalLongDoubleMap implements Serializable {

    private static final long serialVersionUID = 0L;
    
    //private BigLongArray keys;
    //private BigDoubleArray values;
    //public GenerationalLongDoubleMap gen;

    private long[] keys;
    private double[] values;
    private int[] gen;
    
    private int size;

    private int currentGeneration = 1;
    
    int nreported = 0;
    
    public GenerationalLongDoubleMap(int size) {
        keys = new long[size]; //BigLongArray(size);
        values = new double[size];
        gen = new int[size];
    }

    private int hash(long l) {
        int hash = 0;
        for(int i = 0; i < 8; i++) {
            hash += (l & 255);
            l >>= 8;
            hash += (hash << 10);
            hash ^= (hash >> 6);
        }
        hash += (hash << 3);
        hash ^= (hash >> 11);
        hash += (hash << 15);
        return hash;
    }
    
    private int maxOffset = 0;
    
    public double put(long key, double value) {
        if(size == values.length)
            throw new RuntimeException("Map is full!");
        int index = hash(key) % keys.length;

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            long k = keys[index];
            int g = gen[index];
            if(g != currentGeneration) {
                keys[index] = key;
                values[index] = value;
                gen[index] = currentGeneration;
                size++;

                return 0;
            } else if(k == key) {
                double oldValue = values[index];
                values[index] = value;
                return oldValue;
            }
            offset++;
            index += offset * offset;
            if(index < 0)
                index = -index;
            index %= keys.length;
        }        
    }
    
    public double get(long key) {
        int index = hash(key) % keys.length;

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            long k = keys[index];
            int g = gen[index];
            if(g != currentGeneration)
                return Double.NaN;
            if(k == key)
                return values[index];
            offset++;
            index += offset * offset;
            index %= keys.length;
            //
            if(index < 0)
                index = -index;

        }        
    }
    
    public boolean containsKey(long key) {
        if(key == 0)
            return false;
        int index = hash(key) % keys.length;

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            long k = keys[index];
            int g = gen[index];
            if(g != currentGeneration)
                return false;
            if(k == key)
                return true;
            offset++;
            index += offset * offset;
            index %= keys.length;
            //
            if(index < 0)
                index = -index;

        }        
    }
    
    public long[] getKeys() {
        //return keys;
        throw new RuntimeException("Unimplemented!");
    }

    public double[] getValues() {
        //return values;
        throw new RuntimeException("Unimplemented!");
    }
 
    public int size() {
        return size;
    }
    
    public void clear() {
    	size = 0;
    	currentGeneration++;
    }
    
    public static void main(String[] argv) {

    	Random rand = new Random(0);

    	int N = 2000000;
    	
    	GenerationalLongDoubleMap m1 = new GenerationalLongDoubleMap(10*N);
    	HashMap<Long, Double> m2 = new HashMap();
    	
    	for(int i = 0; i < N; i++) {
    		long l = rand.nextInt();
    		double d = rand.nextDouble();
    		
    		m1.put(l, d);    		
    	}
    	
    	//m1.clear();

    	for(int i = 0; i < N; i++) {
    		long l = rand.nextInt();
    		double d = rand.nextDouble();
    		    		
    		m1.put(l, d);
    		m2.put(l, d);
    	}
    	
    	for(long k: m2.keySet()) {
    		Double d = m2.get(k);
    		if(d == null)
    			throw new RuntimeException("1");
    		double dd = d;
    		if(!m2.containsKey(k))
    			throw new RuntimeException("2");
    		double d2 = m2.get(k);
    		if(Math.abs(d2 - dd) > 1e-6)
    			throw new RuntimeException("3");
    	}
    	
    	System.out.println(m1.size());
    	System.out.println(m2.size());
    	
    }
    
}
