/*
 Histogram.java
  
 Created on Nov 1, 2006 by Richard Johansson (richard@cs.lth.se).

 $Log: Histogram.java,v $
 Revision 1.2  2011-07-11 11:17:59  johansson
 More methods.

 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.10  2008/12/03 15:08:44  richard
 Types.

 Revision 1.9  2007/11/06 09:56:38  richard
 Updated trove. SetFrequency.

 Revision 1.8  2007/11/01 10:06:25  richard
 Bug in asSortedList.

 Revision 1.7  2007/10/22 07:46:24  richard
 Reimplemented two operations.

 Revision 1.6  2007/09/28 14:45:27  richard
 Uses IntObjPair.

 Revision 1.5  2007/03/28 07:44:47  richard
 AsList returns the pairs in descending order.

 Revision 1.4  2006/11/16 21:18:03  richard
 New revision.

 Revision 1.3  2006/11/14 11:54:39  richard
 Added getMostFrequent.

 Revision 1.2  2006/11/08 10:29:09  richard
 Serializable.

 Revision 1.1  2006/11/06 08:11:04  richard
 Added the file.

   
 */
package se.lth.cs.nlp.nlputils.core;

import gnu.trove.*;

import java.util.*;
import java.io.Serializable;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class Histogram<K> implements Serializable {

    private static final long serialVersionUID = 0;

    public Histogram() {
    }
    
    public Histogram(Collection<K> keys) {
        addAll(keys);
    }
    
    private TObjectIntHashMap<K> map = new TObjectIntHashMap();

    public void add(K key) {
        map.adjustOrPutValue(key, 1, 1);
    }
    
    public void addAll(Collection<K> keys) {
        for(K k: keys)
            add(k);
    }

    public void remove(K key) {
    	map.remove(key);
    }

    public boolean containsKey(K key) {
    	return map.containsKey(key);
    }
    
    public int getFrequency(K key) {
        return map.get(key);
    }
    
    public void setFrequency(K key, int value) {
        map.put(key, value);
    }

    public ArrayList<IntObjPair<K>> asSortedList() {
    	//System.out.println("map.size = " + map.size());
    	
    	ArrayList<IntObjPair<K>> out = new ArrayList<IntObjPair<K>>();
        TObjectIntIterator<K> it = map.iterator();
        
        // VERY STRANGE!!!!!
        // Bug in trove?
        // TODO check with new version of trove
        
        /*for(int i = map.size(); i-- > 0; ) {
    	System.out.println("i = " + i);
    	System.out.println(it.hasNext());
    	it.advance();
        out.add(new IntObjPair<K>(it.value(), it.key()));
        System.out.println(out);
    	}*/
        
        //out.add(new IntObjPair<K>(it.value(), it.key()));
        
        while(it.hasNext()) { 
        	it.advance();
            out.add(new IntObjPair<K>(it.value(), it.key()));
            //System.out.println(out);
        }

        /*if(out.size() != map.size()) {
        	throw new RuntimeException("!!!");
        }*/
        
        //System.out.println("out.size = " + out.size());        
        //System.exit(0);
        
        Collections.sort(out, new NegatedComparator(IntObjPair.BY_LEFT));
        return out;
    }
    
    public String toString() {
        return asSortedList().toString();
    }

    public K getMostFrequent() {
        TObjectIntIterator<K> it = map.iterator();
        int max = -1;
        K maxKey = null;
        for(int i = map.size(); i-- > 0; ) {
            it.advance();
            if(it.value() > max) {
                max = it.value();
                maxKey = it.key();
            }
        }
        return maxKey;
    }

    public Set<K> threshold(int t) {
        throw new UnsupportedOperationException("Unimplemented!");
    }
    
    public int size() {
        return map.size();
    }

    public void pruneByThreshold(int threshold) {

    	TObjectIntIterator<K> it = map.iterator();
    	/*for(int i = map.size(); i-- > 0; ) {
            it.advance();
            if(it.value() <= threshold)
            	it.remove();
        }*/
    	 
        while(it.hasNext()) { 
        	it.advance();
            if(it.value() <= threshold)
            	it.remove();
        }
    }
    
    public List<K> keys() {
    	return (List) Arrays.asList(map.keys());
    }
    
}
