/*
 HistogramMap.java
  
 Created on Nov 1, 2006 by Richard Johansson (richard@cs.lth.se).

 $Log: HistogramMap.java,v $
 Revision 1.2  2011-07-11 11:17:59  johansson
 More methods.

 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.5  2007/09/28 14:45:41  richard
 Histogram uses IntObjPair.

 Revision 1.4  2006/11/16 21:18:03  richard
 New revision.

 Revision 1.3  2006/11/14 11:55:20  richard
 Added getMostFrequentAssociations.

 Revision 1.2  2006/11/08 10:29:09  richard
 Serializable.

 Revision 1.1  2006/11/06 08:17:07  richard
 Added the file.

   
 */
package se.lth.cs.nlp.nlputils.core;

import java.util.*;
import java.io.Serializable;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class HistogramMap<K, V> implements Serializable {

    private static final long serialVersionUID = 0;
    
    private HashMap<K, Histogram<V>> table = new HashMap<K, Histogram<V>>();
    
    public void add(K key, V value) {
        Histogram<V> histogram = table.get(key);
        if(histogram == null) {
            histogram = new Histogram<V>();
            table.put(key, histogram);
        }
        histogram.add(value);
    }

    public void remove(K key) {
    	table.remove(key);
    }
    
    public int getFrequency(K key, V value) {
        Histogram<V> histogram = table.get(key);
        if(histogram == null)
            return -1;
        return histogram.getFrequency(value);
    }

    public Histogram<V> getFrequencies(K key) {
        return table.get(key);
    }

    public ArrayList<IntObjPair<V>> getFrequenciesAsSortedList(K key) {
        Histogram<V> histogram = table.get(key);
        if(histogram == null)
            return null;
        return histogram.asSortedList();
    }
    
    public String toString() {
        return table.toString();
    }
    
    public Set<K> keySet() {
        return table.keySet();
    }

    public List<V> getValues(K key) {
    	Histogram<V> histogram = table.get(key);
        if(histogram == null)
            return null;
        return histogram.keys();
    }
    
    public boolean containsKey(K key) {
    	return table.containsKey(key);
    }
    
    public boolean containsPair(K key, V value) {
    	Histogram<V> h = table.get(key);
    	if(h == null)
    		return false;
    	return h.containsKey(value);
    }
    
    public HashMap<K, V> getMostFrequentAssociations() {
        HashMap<K, V> out = new HashMap<K, V>();
        
        for(Map.Entry<K, Histogram<V>> e: table.entrySet()) {
            K key = e.getKey();
            Histogram<V> h = e.getValue();
            out.put(key, h.getMostFrequent());
        }        
        
        return out;
    }

    public void pruneByThreshold(int threshold) {
    	for(Iterator<Map.Entry<K, Histogram<V>>> it = table.entrySet().iterator(); it.hasNext(); ) {
    		Map.Entry<K, Histogram<V>> e = it.next();
    		Histogram h = e.getValue();
    		h.pruneByThreshold(threshold);
    		if(h.size() == 0)
    			it.remove();
    	}
    }
    
    public int size() {
        return table.size();
    }
    
}
