/*
 IntHistogram.java
  
 Created on Oct 22, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: IntHistogram.java,v $
 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.2  2007/11/06 09:56:55  richard
 Updated trove. SetFrequency.

 Revision 1.1  2007/10/22 07:43:19  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.core;

import gnu.trove.*;

import java.util.*;
import java.io.Serializable;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class IntHistogram implements Serializable {

    private static final long serialVersionUID = 0;
    
    public IntHistogram() {
    }
    
    public IntHistogram(Collection<Integer> keys) {
        addAll(keys);
    }
    
    public IntHistogram(int[] keys) {
        addAll(keys);
    }
    
    private TIntIntHashMap map = new TIntIntHashMap();

    public void add(int key) {
        map.adjustOrPutValue(key, 1, 1);
    }

    public void addAll(Collection<Integer> keys) {
        for(Integer k: keys)
            add(k);
    }
    
    public void addAll(int[] keys) {
        for(int k: keys)
            add(k);
    }
    
    public void addToValue(int key, int value) {
        map.adjustOrPutValue(key, value, value);
    }
    
    public void setFrequency(int key, int value) {
        map.put(key, value);
    }
    
    public int getFrequency(int key) {
        return map.get(key);
    }

    /** Left is frequency, right is value. */
    public ArrayList<IntPair> asSortedList() {
        ArrayList<IntPair> out = new ArrayList();

        TIntIntIterator it = map.iterator();
        for(int i = map.size(); i-- > 0; ) {
            it.advance();
            out.add(new IntPair(it.value(), it.key()));
        }

        Collections.sort(out, new NegatedComparator(IntPair.BY_LEFT));
        return out;
    }    
    
    public String toString() {
        return asSortedList().toString();
    }

    public int getMostFrequent() {

        TIntIntIterator it = map.iterator();
        int max = -1;
        int maxKey = 0;
        for(int i = map.size(); i-- > 0; ) {
            it.advance();
            if(it.value() > max) {
                max = it.value();
                maxKey = it.key();
            }
        }
        return maxKey;
    }

    public TIntHashSet threshold(int t) {
        throw new UnsupportedOperationException("Unimplemented!");
    }
    
    public int size() {
        return map.size();
    }
    
}
