/*
 Copyright (2007) Richard Johansson (richard.johansson@gu.se)
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at 
 
 http://www.apache.org/licenses/LICENSE-2.0
 
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License. 
*/

/*
 IntObjPair.java
  
 Created on Aug 14, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: IntObjPair.java,v $
 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.2  2007/09/28 14:46:02  richard
 Changed hashing. Comparator.

 Revision 1.1  2007/08/17 10:20:11  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.core;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class IntObjPair<T> {

    private static final long serialVersionUID = 0;

    public int left;
    public T right;
    
    public IntObjPair(int left, T right) {
        this.left = left;
        this.right = right;
    }
    
    public String toString() {
        return "(" + left + ", " + right + ")";
    }
    
    public boolean equals(Object o) {
        if(!(o instanceof IntObjPair))
            return false;
        IntObjPair p = (IntObjPair) o;
        return left == p.left && right.equals(p.right);
    }
    
    private static final int HASH_OFFSET = 100003;
    
    public int hashCode() {
        return HASH_OFFSET*left + right.hashCode();
    }
 
    public IntObjPair clone() {
        return new IntObjPair(left, right);
    }
    
    public static final Comparator<IntObjPair> BY_LEFT = new Comparator<IntObjPair>() {
        public int compare(IntObjPair p1, IntObjPair p2) {
            return p1.left - p2.left;
        }
    };
 
	public static <T> Collection<Integer> lefts(Collection<IntObjPair<T>> coll) {
		Collection<Integer> out = (Collection) CollectionUtils.newCollection(coll);
		for(IntObjPair p: coll) {
			out.add(p.left);
		}
		return out;
	}
	
	public static <T> int[] leftsArray(Collection<IntObjPair<T>> coll) {
		int[] out = new int[coll.size()];
		int index = 0;
		for(IntObjPair<T> p: coll) {
			out[index++] = p.left;
		}
		return out;
	}
	
	public static <T> Collection<T> rights(Collection<IntObjPair<T>> coll) {
		Collection<T> out = (Collection) CollectionUtils.newCollection(coll);
		for(IntObjPair<T> p: coll) {
			out.add(p.right);
		}
		return out;
	}
    
	public static <T> ArrayList<Integer> lefts(ArrayList<IntObjPair<T>> coll) {
		ArrayList<Integer> out = new ArrayList();
		for(IntObjPair<T> p: coll) {
			out.add(p.left);
		}
		return out;
	}
	
	public static <T> ArrayList<T> rights(ArrayList<IntObjPair<T>> coll) {
		ArrayList<T> out = new ArrayList();
		for(IntObjPair<T> p: coll) {
			out.add(p.right);
		}
		return out;
	}
	
	public static <T> T[] rightsArray(Collection<IntObjPair<T>> coll) {
		T[] out = (T[]) new Object[coll.size()];
		int index = 0;
		for(IntObjPair<T> p: coll) {
			out[index++] = p.right;
		}
		return out;
	}
    
}
