/*
 Copyright (2007) Richard Johansson (richard.johansson@gu.se)
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at 
 
 http://www.apache.org/licenses/LICENSE-2.0
 
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License. 
*/

/*
 IntPair.java
  
 Created on Mar 19, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: IntPair.java,v $
 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.4  2008/04/02 11:27:16  richard
 Sort and toArray.

 Revision 1.3  2007/10/22 07:47:16  richard
 Added comparator.

 Revision 1.2  2007/09/24 11:23:01  richard
 New offset value.

 Revision 1.1  2007/03/28 07:45:10  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.core;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class IntPair implements java.io.Serializable {
    private static final long serialVersionUID = 0;

    public int left, right;
    
    public IntPair(int left, int right) {
        this.left = left;
        this.right = right;
    }
    
    public String toString() {
        return "(" + left + ", " + right + ")";
    }
    
    public boolean equals(Object o) {
        if(!(o instanceof IntPair))
            return false;
        IntPair p = (IntPair) o;
        return left == p.left && right == p.right;
    }
    
    private static final int HASH_OFFSET = 100003;
    
    public int hashCode() {
        return HASH_OFFSET*left + right;
    }
 
    public IntPair clone() {
        return new IntPair(left, right);
    }
    
    public static final Comparator<IntPair> BY_LEFT = new Comparator<IntPair>() {
        public int compare(IntPair p1, IntPair p2) {
            return p1.left - p2.left;
        }
    };
    
    public void sort() {
        int tmpl = left;
        if(right < left) {
            tmpl = left;
            left = right;
            right = tmpl;
        }
    }
    
    public int[] toArray() {
        return new int[] { left, right };
    }

	public static Collection<Integer> lefts(Collection<IntPair> coll) {
		Collection<Integer> out = (Collection) CollectionUtils.newCollection(coll);
		for(IntPair p: coll) {
			out.add(p.left);
		}
		return out;
	}
	
	public static Collection<Integer> rights(Collection<IntPair> coll) {
		Collection<Integer> out = (Collection) CollectionUtils.newCollection(coll);
		for(IntPair p: coll) {
			out.add(p.right);
		}
		return out;
	}
    
	public static ArrayList<Integer> lefts(ArrayList<IntPair> coll) {
		ArrayList<Integer> out = new ArrayList();
		for(IntPair p: coll) {
			out.add(p.left);
		}
		return out;
	}
	
	public static ArrayList<Integer> rights(ArrayList<IntPair> coll) {
		ArrayList<Integer> out = new ArrayList();
		for(IntPair p: coll) {
			out.add(p.right);
		}
		return out;
	}
    
	public static int[] leftsArray(Collection<IntPair> coll) {
		int[] out = new int[coll.size()];
		int index = 0;
		for(IntPair p: coll) {
			out[index++] = p.left;
		}
		return out;
	}
	
	public static int[] rightsArray(Collection<IntPair> coll) {
		int[] out = new int[coll.size()];
		int index = 0;
		for(IntPair p: coll) {
			out[index++] = p.right;
		}
		return out;
	}
    
}

