/*
 IntervalHistogram.java
  
 Created on Feb 20, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: IntervalHistogram.java,v $
 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.1  2007/02/20 08:48:48  richard
 Added the file.

   
 */
package se.lth.cs.nlp.nlputils.core;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class IntervalHistogram {

    private double low, high, step;

    private int[] hist;
    private int n;
    
    public IntervalHistogram(double low, double high, int n) {
        this.low = low;
        this.high = high;
        step = (high - low) / n;
        hist = new int[n + 2];
    }

    public DoubleIntPair[] toList() {
        DoubleIntPair[] out = new DoubleIntPair[hist.length];
        for(int i = 0; i < hist.length; i++)
            out[i] = new DoubleIntPair(low + i*step, hist[i]);
        return out;
    }
    
    public void add(double d) {
        n++;
        if(d < low)
            hist[0]++;
        else if(d > high)
            hist[hist.length - 1]++;
        else {
            int index = (int) ((d - low) / step);
            hist[index + 1]++;
        }
    }
        
    public double getQuantile(double p) {
        int n2 = (int) (p*n);
        if(n2 <= hist[hist.length-1])
            return high;
        n2 -= hist[hist.length-1];
        
        for(int i = hist.length - 2; i > 0; i--) {
            if(n2 < hist[i]) {
                double h = low + step*i;
                return h - step * ((double) n2 / hist[i]);
            }
            n2 -= hist[i];
        }

        return low;
    }

    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("[ < " + low + ": " + hist[0]);
        double d = low;
        for(int i = 1; i < hist.length - 1; i++) {
            sb.append(", " + d + " <= x < ");
            d += step;
            sb.append(d + ": " + hist[i]);
        }
        sb.append(", >= " + high + ": " + hist[hist.length-1] + "]");
        return sb.toString();
    }

    public static void main(String[] argv) {
        IntervalHistogram h = new IntervalHistogram(-5, 5, 1000);
        java.util.Random rand = new java.util.Random(10);
        for(int i = 0; i < 1000000; i++)
            h.add(rand.nextGaussian());
        
        System.out.println(h.getQuantile(0.45));
        
        
    }
    
}
