/*
 Copyright (2005) Richard Johansson (richard.johansson@gu.se)
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at 
 
 http://www.apache.org/licenses/LICENSE-2.0
 
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License. 
*/

/*
 ListMap.java
 
 Copyright (C) 2005, Richard Johansson (richard@cs.lth.se).
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.
 
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 Created on Nov 29, 2005 by Richard Johansson (richard@cs.lth.se).
 
 $Log: ListMap.java,v $
 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.6  2006/11/20 10:21:18  richard
 Added keySet.

 Revision 1.5  2006/11/16 21:18:03  richard
 New revision.

 Revision 1.4  2006/11/16 09:45:19  richard
 Added putAll.

 Revision 1.3  2006/11/08 10:29:09  richard
 Serializable.

 Revision 1.2  2006/11/07 10:37:24  richard
 Added toString.

 Revision 1.1  2006/11/07 10:21:13  richard
 Renamed.

 Revision 1.1  2006/11/06 08:24:21  richard
 Renamed.

 Revision 1.2  2006/09/07 17:51:19  richard
 New revision.

 Revision 1.1  2006/05/16 13:38:13  richard
 Re-added because of CVS problems.
 
 Revision 1.2  2006/04/07 14:46:08  richard
 Added documentation.
 
 Revision 1.1  2005/11/29 13:29:20  richard
 Added the file.
 
 
 */
package se.lth.cs.nlp.nlputils.core;

import java.util.*;
import java.io.Serializable;

/**
 * A map from a key type to a list of the value type.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class ListMap<K, V> implements Serializable {
    
    private static final long serialVersionUID = 0;
    
    private static final int START_SIZE = 2;
    
    private HashMap<K, ArrayList<V>> map = new HashMap<K, ArrayList<V>>();
    
    /**
     * Adds <code>value</code> to the association list of <code>key</code>.
     * @param key the key.
     * @param value the value.
     */
    public void put(K key, V value) {
        ArrayList<V> l = map.get(key);
        if(l == null) {
            l = new ArrayList<V>(START_SIZE);
            map.put(key, l);
        }
        l.add(value);
    }
    
    /**
     * Returns the association list for <code>key</code>.
     * @param key the key.
     * @return the list.
     */
    public ArrayList<V> get(K key) {
        return map.get(key);
    }
    
    /**
     * Checks whether the map contains a given key.
     * @param key the key.
     */
    public boolean containsKey(K key) {
        return map.containsKey(key);
    }

    public String toString() {
        return map.toString();
    }

    public void putAll(K key, Collection<V> values) {
        ArrayList<V> l = map.get(key);        
        if(l == null) {
            l = new ArrayList<V>(START_SIZE);
            map.put(key, l);
        }
        l.addAll(values);
    }
    
    public void putAll(K key, V[] values) {
        putAll(key, Arrays.asList(values));
    }
    
    public int size() {
        return map.size();
    }

    public Set<K> keySet() {
        return map.keySet();
    }
    
}
