package se.lth.cs.nlp.nlputils.core;

import java.util.ArrayList;
import java.util.Arrays;

public class ListParser {

    
    public static Object[] parse(String s0) {
        String s = s0.trim();
        if(!s.startsWith("["))
            throw new IllegalArgumentException("Illegal format: " + s0);
        if(!s.endsWith("]"))
            throw new IllegalArgumentException("Illegal format: " + s0);
        IntObjPair<Object[]> ip = nextArr(s, 0, false);
        return ip.right;
    }
    
    public static String[] parseFlat(String s0) {
        String s = s0.trim();
        if(!s.startsWith("["))
            throw new IllegalArgumentException("Illegal format: " + s0);
        if(!s.endsWith("]"))
            throw new IllegalArgumentException("Illegal format: " + s0);
        IntObjPair<Object[]> ip = nextArr(s, 0, true);
        String[] out = new String[ip.right.length];
        System.arraycopy(ip.right, 0, out, 0, out.length);
        return out;
    }
    
    public static IntObjPair next(String s, int start) {
        return null;
    }

    public static IntObjPair<String> nextSQ(String s, int start) {
        int end = start + 1;
        while(true) {
            end = s.indexOf('\'', end);
            if(end == -1)
                throw new IllegalArgumentException("Unmatched single quote: " + s);
            int count = 0;
            while(end-count-1 >= 0 && s.charAt(end-count-1) == '\\')
                count++;
            if(count % 2 == 0)
                break;
            end += 1;
        }
        return new IntObjPair<String>(end+1, s.substring(start, end+1));
    }
    
    public static IntObjPair<String> nextDQ(String s, int start) {
        int end = start + 1;
        while(true) {
            end = s.indexOf('\"', end);
            if(end == -1)
                throw new IllegalArgumentException("Unmatched double quote: " + s);
            int count = 0;
            while(end-count-1 >= 0 && s.charAt(end-count-1) == '\\')
                count++;
            if(count % 2 == 0)
                break;
            end += 1;
        }
        return new IntObjPair<String>(end+1, s.substring(start, end+1));
    }

    private static IntObjPair<String> nextOther(String s, int start, boolean flat) {
        int len = s.length();
        if(flat) {
            int ix = s.lastIndexOf(']');
            if(ix == -1)
                throw new IllegalArgumentException("Can't process this string: |" + s + "|");
            len = ix+1;
        }

        for(int i = start; i < len; i++) {
            char c = s.charAt(i);
            if(c == ']') {
                if(!flat || i == len-1)
                    return new IntObjPair<String>(i, s.substring(start, i));
            }
            if(Character.isSpaceChar(c) || c == ',')
                return new IntObjPair<String>(i, s.substring(start, i));
        }
        throw new IllegalArgumentException("Can't process this string: |" + s + "|");
    }
    
    private static IntObjPair<Object[]> nextArr(String s, int start, boolean flat) {
        if(s.charAt(start) != '[')
            throw new IllegalArgumentException("Illegal format: " + s);
        ArrayList l = new ArrayList();        
        int ix = eatSpace(s, start + 1);
        if(s.charAt(ix) == ']')
            return new IntObjPair(ix+1, new Object[0]);
        while(true) {
            char c = s.charAt(ix);
            //if(c == ']')
            //    break;
            IntObjPair ip = null;
            switch(c) {
            case '\'':
                ip = nextSQ(s, ix);
                break;
            case '"':
                ip = nextDQ(s, ix);
                break;
            case '[':
                if(!flat) {
                    ip = nextArr(s, ix, false);
                    break;
                }
            default:
                ip = nextOther(s, ix, flat); //false /*depth == 0*/);
            }
            //System.out.println("ip = " + ip);
            l.add(ip.right);
            //if(l.size() > 5)
            //    System.exit(1);
            ix = eatSpace(s, ip.left);
            if(s.charAt(ix) == ']')
                break;
            if(s.charAt(ix) != ',')
                throw new IllegalArgumentException("Illegal format: |" + s + "|, c = " + s.charAt(ix));
            ix = eatSpace(s, ix + 1);
        }
        return new IntObjPair(ix+1, l.toArray(new Object[0]));
    }
    
    private static int eatSpace(String s, int start) {
        int len = s.length();
        for(int i = start; i < len; i++) {
            char c = s.charAt(i);
            if(!Character.isSpaceChar(c))
                return i;
        }
        throw new IllegalArgumentException("Can't process this string: |" + s + "|");
    }
    
    public static void main(String[] argv) {
        String testStr = "[[q],abc, [x,y], \"def,'ghi  ' \",[   ]]";
        System.out.println("|" + testStr + "|");
        System.out.println(Arrays.deepToString(parse(testStr)));
    }
    
}
