/*
 PartialOrders.java
  
 Created on Nov 21, 2006 by Richard Johansson (richard@cs.lth.se).

 $Log: PartialOrders.java,v $
 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.2  2006/11/21 14:00:13  richard
 Fixed bug in topsort.

 Revision 1.1  2006/11/21 13:49:08  richard
 Added the file.

   
 */
package se.lth.cs.nlp.nlputils.core;

import java.util.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class PartialOrders {

    public static <T> HashMap<T, Integer> topologicalOrder(Collection<? extends T> c,
            PartialComparator<T> comp) {
        HashMap<T, Integer> done = new HashMap<T, Integer>();
        HashSet<T> remaining = new HashSet<T>(c);
        int n = remaining.size();        
        ArrayList<T> tmp = new ArrayList<T>();
        int index = 0;
        while(done.size() < n) {
            for(T t1: remaining) {
                boolean preceded = false;
                for(T t2: remaining)
                    if(comp.compare(t1, t2) == PartialComparator.GT) {
                        preceded = true;
                        break;
                    }
                if(!preceded)
                    tmp.add(t1);
            }
            for(T t: tmp)
                done.put(t, index);
            remaining.removeAll(tmp);
            index++;
            tmp.clear();            
        }        
        return done;
    }
    
    public static <T> HashMap<T, Integer> topologicalOrder(Collection<? extends T> c,
            Comparator<T> comp) {
        return topologicalOrder(c, fromComparator(comp));
    }
    
    public static <T> HashMap<T, Integer> topologicalOrder(Collection<? extends PartiallyComparable> c) {
        return null;
    }

    private static class Comp2PartComp<T> implements PartialComparator<T> {
        Comparator<T> comp;
        Comp2PartComp(Comparator<T> comp) {
            this.comp = comp;
        }
        public int compare(T o1, T o2) {
            int c = comp.compare(o1, o2);
            if(c < 0)
                return PartialComparator.LT;
            if(c > 0)
                return PartialComparator.GT;
            return PartialComparator.EQ;
        }
    }
    
    public static <T> PartialComparator<T> fromComparator(Comparator<T> comp) {
        return new Comp2PartComp<T>(comp);
    }

}
