/*
 Partition.java
  
 Created on Jul 24, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: Partition.java,v $
 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.1  2008/08/14 08:11:58  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.core;

import java.util.Arrays;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class Partition {

    private int[] s;
    
    public Partition(int n) {
        s = new int[n];
        Arrays.fill(s, -1);
    }

    public void union(int root1, int root2) {
        assertIsRoot(root1);
        assertIsRoot(root2);
        if(root1 == root2)
            throw new IllegalArgumentException("root1 == root2");
        if(s[root2] < s[root1])
            s[root1] = root2;
        else {
            if(s[root1] == s[root2])
                s[root1]--;
            s[root2] = root1;
        }            
    }
    
    public int find(int x) {
        assertIsItem(x);
        if(s[x] < 0)
            return x;
        else {
            s[x] = find(s[x]);
            return s[x];
        }            
    }
    
    private void assertIsRoot(int root) {
        assertIsItem(root);
        if(s[root] > 0)
            throw new IllegalArgumentException(root + " is no root");        
    }
    
    private void assertIsItem(int x) {
        if(x < 0 || x >= s.length)
            throw new IllegalArgumentException(x + " is no item");
    }
    
}
