
/*
 Predicate.java
 
 Copyright (C) 2005, Richard Johansson (richard@cs.lth.se).
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.
 
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 Created in 2003 by Richard Johansson.
 
 $Log: Predicate.java,v $
 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.2  2006/09/07 17:51:19  richard
 New revision.

 Revision 1.1  2006/05/16 13:38:13  richard
 Re-added because of CVS problems.
 
 Revision 1.5  2006/04/11 14:36:49  richard
 Added more documentation.
 
 Revision 1.4  2006/04/07 14:46:08  richard
 Added documentation.
 
 Revision 1.3  2006/02/23 09:55:51  richard
 Added constant true and false predicates.
 
 Revision 1.2  2006/02/23 09:47:10  richard
 Removed the TruePredicate class and implemented similar funcionality in Predicate.
 
 Revision 1.1  2005/11/28 15:41:00  richard
 Added the file.
 
 
 */

package se.lth.cs.nlp.nlputils.core;

/**
 * A callable one-argument predicate.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public abstract class Predicate<T> {
    
    /**
     * A predicate that always returns true.
     */
    public static final Predicate TRUE = new Predicate() {
        public boolean apply(Object o) {
            return true;
        }
    };
    
    /**
     * A predicate that always returns false.
     */
    public static final Predicate FALSE = new Predicate() {
        public boolean apply(Object o) {
            return false;
        }
    };
    
    /**
     * Initiation parameter.
     */
    protected Object in;
    
    /**
     * Initiates the predicate.
     */
    public Predicate() {
    }
    
    /**
     * Initiates the predicate with a parameter.
     * 
     * @param in an initiation parameter.
     */
    public Predicate(Object in) {
        this.in = in;
    }
    
    /**
     * Calls this predicate.
     * 
     * @param o the object to test.
     * @return the value of the predicate on the object <code>o</code>.
     */
    public abstract boolean apply(T o);
    
}
