package se.lth.cs.nlp.nlputils.core;

import java.util.TimerTask;

public class ReschedulableTimer {

    private TimerTask task;
    private long wakeup;
    
    public synchronized void schedule(TimerTask task, long delay) {
        this.task = task;
        wakeup = System.currentTimeMillis() + delay;
        new TimerThread().start();
    }

    public synchronized void reschedule(long delay) {
        wakeup = System.currentTimeMillis() + delay;
    }
    
    private class TimerThread extends Thread {        
        public void run() {
            synchronized(ReschedulableTimer.this) {
                long l0 = System.currentTimeMillis();
                while(wakeup > l0)
                    try {
                        ReschedulableTimer.this.wait(wakeup - l0);
                        l0 = System.currentTimeMillis();
                    } catch(InterruptedException e) {
                        return;
                    }
            }
            task.run();
        }
    }
    
}
