/*
 Copyright (2006) Richard Johansson (richard.johansson@gu.se)
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at 
 
 http://www.apache.org/licenses/LICENSE-2.0
 
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License. 
*/

/*
 SetMap.java
 
 Copyright (C) 2006, Richard Johansson (richard@cs.lth.se).
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.
 
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 Created on Nov 7, 2006 by Richard Johansson (richard@cs.lth.se).
 
 $Log: SetMap.java,v $
 Revision 1.2  2010-06-07 11:04:21  johansson
 Added containsPair.

 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.5  2006/11/20 10:21:18  richard
 Added keySet.

 Revision 1.4  2006/11/16 21:18:03  richard
 New revision.

 Revision 1.3  2006/11/16 09:45:19  richard
 Added putAll.

 Revision 1.2  2006/11/08 10:29:09  richard
 Serializable.

 Revision 1.1  2006/11/07 10:37:03  richard
 Added.

 
 */
package se.lth.cs.nlp.nlputils.core;

import java.util.*;
import java.io.Serializable;

/**
 * A map from a key type to a set of the value type.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class SetMap<K, V> implements Serializable {
    
    private static final long serialVersionUID = 0;
    
    private static final int START_SIZE = 2;
    
    private HashMap<K, HashSet<V>> map = new HashMap<K, HashSet<V>>();
    
    /**
     * Adds <code>value</code> to the association list of <code>key</code>.
     * @param key the key.
     * @param value the value.
     */
    public void put(K key, V value) {
        HashSet<V> l = map.get(key);
        if(l == null) {
            l = new HashSet<V>(START_SIZE);
            map.put(key, l);
        }
        l.add(value);
    }

    public void putAll(K key, Collection<V> values) {
        HashSet<V> l = map.get(key);        
        if(l == null) {
            l = new HashSet<V>(START_SIZE);
            map.put(key, l);
        }
        l.addAll(values);
    }
    
    public void putAll(K key, V[] values) {
        putAll(key, Arrays.asList(values));
    }
    
    /**
     * Returns the association list for <code>key</code>.
     * @param key the key.
     * @return the list.
     */
    public HashSet<V> get(K key) {
        return map.get(key);
    }
    
    /**
     * Checks whether the map contains a given key.
     * @param key the key.
     */
    public boolean containsKey(K key) {
        return map.containsKey(key);
    }
    
    public boolean containsPair(K key, V value) {
    	HashSet<V> s = get(key);
    	if(s == null)
    		return false;
    	else
    		return s.contains(value);
    }
    
    public String toString() {
        return map.toString();
    }
 
    public int size() {
        return map.size();
    }

    public Set<K> keySet() {
        return map.keySet();
    }

	public Set<Map.Entry<K, HashSet<V>>> entrySet() {
		return map.entrySet();
	}

}
