/*
 Strings.java
  
 Created on Nov 14, 2006 by Richard Johansson (richard@cs.lth.se).

 $Log: Strings.java,v $
 Revision 1.3  2011-02-15 16:02:42  johansson
 Hopefully more efficient xml decoding.

 Revision 1.2  2009/11/16 11:37:51  johansson
 count encodeXML

 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.4  2007/12/18 15:14:44  richard
 Added count.

 Revision 1.3  2006/11/16 21:18:03  richard
 New revision.

 Revision 1.2  2006/11/14 10:51:45  richard
 Joining an array.

 Revision 1.1  2006/11/14 10:49:00  richard
 Added the file.

   
 */
package se.lth.cs.nlp.nlputils.core;

import java.io.PrintWriter;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class Strings {

    public static String join(Collection coll, String separator) {
        StringBuilder sb = new StringBuilder();
        Iterator it = coll.iterator();
        if(!it.hasNext())
            return "";
        Object item = it.next();
        sb.append(item.toString());
        while(it.hasNext()) {
            sb.append(separator);
            item = it.next();
            sb.append(item.toString());
        }
        return sb.toString();
    }
    
    public static String join(Object[] coll, String separator) {
        return join(Arrays.asList(coll), separator);
    }

    public static String join(float[] fs, String separator) {
        StringBuilder sb = new StringBuilder();
        if(fs.length == 0)
            return "";
        sb.append(fs[0]);
        for(int i = 1; i < fs.length; i++) {
            sb.append(separator);
            sb.append(fs[i]);
        }
        return sb.toString();
    }
    
    public static String join(double[] ds, String separator) {
        StringBuilder sb = new StringBuilder();
        if(ds.length == 0)
            return "";
        sb.append(ds[0]);
        for(int i = 1; i < ds.length; i++) {
            sb.append(separator);
            sb.append(ds[i]);
        }
        return sb.toString();
    }
    
    public static String join(int[] is, String separator) {
        StringBuilder sb = new StringBuilder();
        if(is.length == 0)
            return "";
        sb.append(is[0]);
        for(int i = 1; i < is.length; i++) {
            sb.append(separator);
            sb.append(is[i]);
        }
        return sb.toString();
    }
    
    public static int count(String s, char ch) {
        int out = 0;
        for(int i = 0; i < s.length(); i++)
            if(s.charAt(i) == ch)
                out++;
        return out;
    }
 
    public static int count(String s, String substr) {
    	int l = s.length();
    	int ls = substr.length();
    	if(l < ls)
    		return 0;    	
    	int out = 0;
    	for(int i = 0; i <= l - ls; i++) {
    		if(s.substring(i, i + ls).equals(substr))
    			out++;
    	}
    	return out;
    }

    private static Pattern XML_ENTITY_PATTERN, XML_AMP_PATTERN;
    
    public static String decodeXML(String s) {
    	if(XML_ENTITY_PATTERN == null) {
    		XML_ENTITY_PATTERN = Pattern.compile("&(#?)[a-z0-9]+;");
    		XML_AMP_PATTERN = Pattern.compile("&amp;");
    	}
    	Matcher m = XML_ENTITY_PATTERN.matcher(s);
    	StringBuffer sb = new StringBuffer();
    	while(m.find()) {
    		String entity = m.group();
    		if(entity.equals("&amp;"))
    			m.appendReplacement(sb, "&amp;");
            else if(entity.equals("&apos;"))
                m.appendReplacement(sb, "'");
            else if(entity.equals("&quot;"))
                m.appendReplacement(sb, "\"");
    		else if(entity.equals("&lt;"))
    			m.appendReplacement(sb, "<");
            else if(entity.equals("&gt;"))
                m.appendReplacement(sb, ">");
            else if(entity.startsWith("&#")) {
                char c;
                if(entity.startsWith("&#0x"))
                    c = (char) Integer.parseInt(entity.substring(4, entity.length()-1), 16);
                else
                    c = (char) Integer.parseInt(entity.substring(2, entity.length()-1), 10);
                //System.out.println("entity = " + entity);
                //System.out.println(c);
                //System.exit(0);
                m.appendReplacement(sb, "" + c);
            } 
            else if(entity.equals("&ccaron;"))
                m.appendReplacement(sb, "č");
            else if(entity.equals("&ecaron;"))
                m.appendReplacement(sb, "ě");
            else if(entity.equals("&ncaron;"))
                m.appendReplacement(sb, "ň");
            else if(entity.equals("&rcaron;"))
                m.appendReplacement(sb, "ř");
            else if(entity.equals("&scaron;"))
                m.appendReplacement(sb, "š");
            else if(entity.equals("&tcaron;"))
                m.appendReplacement(sb, "ť");
            else if(entity.equals("&zcaron;"))
                m.appendReplacement(sb, "ž");
            else if(entity.equals("&mgr;"))
                m.appendReplacement(sb, "μ");
            else if(entity.equals("&ohgr;"))
                m.appendReplacement(sb, "ω");
            else if(entity.equals("&pgr;"))
                m.appendReplacement(sb, "π");
    		else
    			throw new IllegalArgumentException("Unknown entity: " + entity);
    	}   	
    	m.appendTail(sb);
    	//System.out.println("Stage 1: |" + sb + "|");    	
    	return XML_AMP_PATTERN.matcher(sb.toString()).replaceAll("&");
    }

    public static String encodeXML(String s) {
    	s = s.replaceAll("&", "&amp;");
    	s = s.replaceAll("\"", "&quot;");
    	s = s.replaceAll("<", "&lt;");
    	s = s.replaceAll(">", "&gt;");
    	return s;
    }
    
    public static String extractBetween(String s, char c1, char c2, int start) {
    	int ix1 = s.indexOf(c1, start);
    	if(ix1 == -1)
    		return null;
    	int ix2 = s.indexOf(c2, ix1 + 1);
    	if(ix2 == -1)
    		return null;
    	return s.substring(ix1+1, ix2);
    }

	public static String extractBetween(String s, char c1, char c2) {
		return extractBetween(s, c1, c2, 0);
	}

	public static boolean startsWithCap(String s) {
		if(s.isEmpty())
			return false;
		return Character.isUpperCase(s.charAt(0));
	}
    
	public static String capitalizeFirst(String s) {
		int l = s.length();
		if(l > 0)
			return s.substring(0, 1).toUpperCase() + s.substring(1, l);
		else
			return s;
	}
	
	public static String toCapitalizedLower(String s) {
		int l = s.length();
		if(l > 0)
			return s.substring(0, 1).toUpperCase() + s.substring(1, l).toLowerCase(); 
		else
			return s;
	}

	private static Pattern XML_FEATURE_PATTERN;	
	public static HashMap<String, String> getXMLFeatures(String s) {
		if(XML_FEATURE_PATTERN == null) {
    		XML_FEATURE_PATTERN = Pattern.compile("([A-Za-z_][a-zA-Z0-9_]*)=\"([^\"]+)\"");
    	}
		HashMap<String, String> out = new HashMap();
		Matcher m = XML_FEATURE_PATTERN.matcher(s);
		while(m.find()) {
			String k = m.group(1);
			String v = m.group(2);
			out.put(k, v);
		}
		return out;
	}
	
	private static Pattern XML_TEXTNODE_PATTERN;
	public static String getXMLNodeTextContent(String s) {
		if(XML_TEXTNODE_PATTERN == null) {
    		XML_TEXTNODE_PATTERN = Pattern.compile(">([^<>]*)<");
    	}
		Matcher m = XML_TEXTNODE_PATTERN.matcher(s);
		if(!m.find())
			return null;
		return decodeXML(m.group(1));
	}

	public static boolean containsNumber(String h2) {
		int l = h2.length();
		for(int i = 0; i < l; i++)
			if(Character.isDigit(h2.charAt(i)))
				return true;				
		return false;
	}

    public static String getCommonPrefix(String s1, String s2) {
        int l1 = s1.length();
        int l2 = s2.length();
        for(int i = 0; i < l1 || i < l2; i++) {
            if(i == l1)
                return s1;
            if(i == l2)
                return s2;
            if(s1.charAt(i) != s2.charAt(i))
                return s1.substring(0, i);
        }
        return s1;
    }

    public static boolean noLowerCase(String word) {
        int l = word.length();
        for(int i = 0; i < l; i++)
            if(Character.isLowerCase(word.charAt(i)))
                return false;
        return true;
    }
    
    public static void printColumns(String[][] table, boolean spaces) {
        StringBuilder[] sbs = new StringBuilder[table.length];

        int maxcol = 0;
        for(int i = 0; i < sbs.length; i++) {
            sbs[i] = new StringBuilder();
            if(table[i].length > maxcol)
                maxcol = table[i].length;
        }

        for(int j = 0; j < maxcol; j++) {
            if(j > 0)
                if(!spaces) {
                    for(int i = 0; i < sbs.length; i++)
                        if(j < table[i].length)
                            sbs[i].append("\t");
                } else {
                    int maxlen = 0;
                    for(int i = 0; i < sbs.length; i++)
                        if(j < table[i].length && sbs[i].length() > maxlen)
                            maxlen = sbs[i].length();
                    for(int i = 0; i < sbs.length; i++)
                        if(j < table[i].length) {
                            for(int k = sbs[i].length(); k < maxlen + 1; k++)
                                sbs[i].append(" ");
                            sbs[i].append(table[i][j]);
                        }
                }
        }
        
        for(int i = 0; i < table.length; i++)
            System.out.println(sbs[i]);        
    }
    
    public static void printColumns(String[][] table, boolean spaces, PrintWriter pw) {
        StringBuilder[] sbs = new StringBuilder[table.length];

        int maxcol = 0;
        for(int i = 0; i < sbs.length; i++) {
            sbs[i] = new StringBuilder();
            if(table[i].length > maxcol)
                maxcol = table[i].length;
        }

        for(int j = 0; j < maxcol; j++) {
            if(j > 0)
                if(!spaces) {
                    for(int i = 0; i < sbs.length; i++)
                        if(j < table[i].length)
                            sbs[i].append("\t");
                } else {
                    int maxlen = 0;
                    for(int i = 0; i < sbs.length; i++)
                        if(j < table[i].length && sbs[i].length() > maxlen)
                            maxlen = sbs[i].length();
                    for(int i = 0; i < sbs.length; i++)
                        if(j < table[i].length) {
                            for(int k = sbs[i].length(); k < maxlen + 1; k++)
                                sbs[i].append(" ");
                        }
                }
            for(int i = 0; i < sbs.length; i++)
                if(j < table[i].length)
                    sbs[i].append(table[i][j]);
        }
        
        for(int i = 0; i < table.length; i++)
            if(pw == null)
                System.out.println(sbs[i]);
            else
                pw.println(sbs[i]);
    }
    
    
//    public static void main(String[] argv) {
//        System.out.println(count(argv[0], argv[1]));
//    }

    public static String toString(double d, boolean sci, int prec) {
        String s;
        if(sci)
            s = String.format((Locale) null, "%." + prec + "g", d);
        else
            s = String.format((Locale) null, "%." + prec + "f", d);
        return s.replaceFirst("\\.?0+(e|$)", "$1");
    }
    
    public static String toString(double d) {
        return toString(d, false, 6);
    }
    
    public static void main(String[] argv) {
        //System.out.println(getCommonPrefix(argv[0], argv[1]));
        System.out.println(toString(Double.parseDouble(argv[0])));
    }

}
