/*
 ChuLiuEdmonds.java
  
 Created on Dec 19, 2005 by Richard Johansson (richard@cs.lth.se).

 $Log: ChuLiuEdmonds.java,v $
 Revision 1.1  2007/03/07 14:54:44  richard
 Moved to common.

 Revision 1.1  2007/02/13 08:23:18  richard
 Int representation. Refactored packages.

 Revision 1.1  2006/03/07 09:38:36  richard
 Added the file.

   
 */
package se.lth.cs.nlp.nlputils.depgraph;

import java.util.*;

import se.lth.cs.nlp.nlputils.core.Pair;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class ChuLiuEdmonds {

	private static final boolean LOG_SCORES = false; 
	
	public static void main(String[] argv) {
		double[][] scores = { { 9, 30, 11 }, { 10, 20, 0 }, { 9, 3, 30 } };
		int[][] heads =  { { 0, 2, 3 }, { 0, 1, 3 }, { 0, 1, 2 } };
		
        //double[][] scores = { { 5, 8 }, { 4, 10, 7 }, { 6, 8 }, { 1, 10 } };
        //int[][] heads =  { { 0, 2 }, { 0, 1, 3 }, { 2, 4 }, { 2, 3 } };
	    
	    //double[][] scores = { { 5, 8 }, { 4, 10, 7 }, { 6, 8 }, { 1, 10 } };
        //int[][] heads =  { { 0, 2 }, { 0, 1, 3 }, { 2, 4 }, { 2, 3 } };
        List<Pair<Double, int[]>> l = findNMaxSums(scores, heads, 1);

        for(int i = 0; i < l.size(); i++) {
            Pair<Double, int[]> p = l.get(i);
            System.out.println(p.left + ": " + Arrays.toString(p.right));
        }
        
	    /*
		for(int k = 0; k <= 0; k++) {

			System.out.println("***** k = " + k);
			
			Random rand = new Random(k);
			
			int N = 300; // dvs N + 1
			
			double[][] scores = new double[N][N];
			int[][] heads = new int[N][N];
			
			for(int i = 0; i < N; i++) {
				for(int j = 0; j < i+1; j++) {
					scores[i][j] = rand.nextDouble();
					heads[i][j] = j;
				}
				for(int j = i+2; j < N+1; j++) {
					scores[i][j-1] = rand.nextDouble();
					heads[i][j-1] = j;
				}
				
				System.out.println(Arrays.toString(heads[i]));
				System.out.println(Arrays.toString(scores[i]));
				
			}
			
			findNMaxSums(scores, heads, 1);
		}*/
	}
	
	public static List<Pair<Double, int[]>> findNMaxSums(double[][] scores,
			int[][] heads, int nbest) {

		//System.out.println("Entering findNMaxSums, scores.length = " + scores.length);
//		try {
//			System.in.read();
//		} catch(Exception e) {
//			e.printStackTrace();
//		}
		
		Node.count = 0;
		
		ArrayList<Node> nodes = new ArrayList<Node>();
		Node[] ns = new Node[heads.length + 1];
		for(int i = 0; i < scores.length+1; i++)
			ns[i] = new Node();
		for(int i = 0; i < scores.length; i++) {
			Node n = ns[i+1];
			for(int j = 0; j < scores[i].length; j++) {
				Edge e = new Edge();
				Node h = ns[heads[i][j]]; 

				e.score = LOG_SCORES? Math.log(scores[i][j]): scores[i][j];
					
				e.startPoint = h;
				e.endPoint = n;
				h.outEdges.add(e);
				n.inEdges.add(e);
			}
		}
		nodes.addAll(Arrays.asList(ns));

		callCount = 0;
		maxCallCount = nodes.size();
		
		chuLiuEdmonds(nodes, ns[0]);

		double sum = 0;
		int[] finalHeads = new int[scores.length];
		for(int i = 1; i <= scores.length; i++) {
			sum += nodes.get(i).headEdge.score;
			finalHeads[i-1] = nodes.get(i).headEdge.startPoint.id;
			if(finalHeads[i-1] > scores.length)
				throw new RuntimeException("Illegal final head!");
//			System.out.print("Final head edge of ");
//			System.out.print(nodes.get(i) + " = ");
//			System.out.println(nodes.get(i).headEdge);
		}
		
		ArrayList<Pair<Double, int[]>> out = new ArrayList<Pair<Double, int[]>>();
		
		out.add(new Pair<Double, int[]>(Math.exp(sum), finalHeads));
		
		return out;
	}

	private static class Edge {
		
		double score;

		ArrayList<Edge> components;
		
		Node startPoint;
		Node endPoint;

		Edge replaced;     /* Ersatt b�ge fr�n nod i cykel. */ 
		
		boolean isMaximal = false; /* Visar om b�gen �r maximal inkommande. */

		public String toString() {
			return "[Edge: " + startPoint.id + "->" + endPoint.id + ", " + score + "]";
		}
		
	}

	private static class Node {

		static int count = 0;
		private int id;
		
		Node() {
			id = count++;
		}
		
		ArrayList<Edge> outEdges = new ArrayList<Edge>();
		ArrayList<Edge> inEdges = new ArrayList<Edge>();

		int roundId;
		Edge headEdge;    /* Maximala inkommande b�ge. */

		boolean isActive = true; /* Visar om noden anv�nds aktivt, dvs inte kontrakterad. */
		boolean inCycle = false; /* Visar om noden tillh�r en cykel. */

		public String toString() {
			StringBuffer out = new StringBuffer();
			out.append("[Node " + id + ": ");
			out.append("]");
			return out.toString();
		}
		
	}

	static int callCount;
	static int maxCallCount;
	
	private static void chuLiuEdmonds(ArrayList<Node> nodes, Node root) {

		if(callCount > maxCallCount) {
			throw new RuntimeException("Call count exceeded!");
		}
			
		callCount++;
		
//		System.out.println("Entering chuLiuEdmonds");
		
		findBestEdges(nodes);
		
		ArrayList<Edge> C = new ArrayList<Edge>();
		if(!findCycle(nodes, C)) {
			//System.out.println("No cycle found, winding up");
			return;
		}

		Collections.reverse(C);
		
//		System.out.println("Found cycle: " + C);

		Node c = contract(nodes, C);

		for(Node n: nodes)
			if(n.inCycle)
				n.isActive = false;
		
//		System.out.println("Contracted node: " + c);

		//System.exit(0);
		
		nodes.add(c);
		
		chuLiuEdmonds(nodes, root);

//		System.out.println(nodes);
		
		for(Edge outEdge: c.outEdges) {
			if(outEdge.endPoint.headEdge == outEdge) {
//				System.out.println("Replacing " + outEdge + " with " + outEdge.replaced);
				outEdge.endPoint.headEdge = outEdge.replaced;
			}
		}

		/* Ta v�gen in i c. */
//		System.out.println(c.headEdge);
//		System.out.println(c.headEdge.replaced);
		Node n = c.headEdge.replaced.endPoint;
		n.headEdge = c.headEdge.replaced;
		

		//System.exit(0);


			
		//Edge headEdge = c.headEdge;

	}

	private static void findBestEdges(ArrayList<Node> nodes) {
		/* F�r varje nod: s�tt b�sta inkommande b�ge till maximal. */
		for(Node n: nodes) {
//			System.out.println("n = " + n);
			if(!n.isActive || n.inEdges.isEmpty())
				continue;
			double maxScore = Double.NEGATIVE_INFINITY;
			Edge maxEdge = null;
			for(Edge e: n.inEdges) {
				//System.out.println(" edge " + e);
				if(!e.startPoint.isActive)
					continue;
				e.isMaximal = false;
				if(e.score > maxScore) {
					maxScore = e.score;
					maxEdge = e;
				}
			}
			maxEdge.isMaximal = true;
			maxEdge.endPoint.headEdge = maxEdge;
//			System.out.println("Best edge: " + maxEdge);
		}
	}

	private static Node contract(ArrayList<Node> allNodes, ArrayList<Edge> cycle) {
		Node c = new Node();

		double sC = 0;
		for(Edge e: cycle) {
			sC += e.score;
		}

//		System.out.println("cycle score: " + sC);
		
		for(Node n: allNodes) {
			if(n.isActive && !n.inCycle) {

				/* n �r en aktiv nod utanf�r cykeln. */

				/* Finn nod i cykeln med maximal inb�ge till n. */
				Edge maxInEdge = null;
				double maxInEdgeScore = Double.NEGATIVE_INFINITY;
				for(Edge inEdge: n.inEdges) {
					Node h = inEdge.startPoint;
					if(h.isActive && h.inCycle && inEdge.score > maxInEdgeScore) {
						maxInEdge = inEdge;
						maxInEdgeScore = inEdge.score;
					}
				}
				if(maxInEdge != null) {
					Edge ne = new Edge();
					ne.score = maxInEdgeScore;
					ne.startPoint = c;
					ne.replaced = maxInEdge;
					ne.endPoint = n;

					c.outEdges.add(ne);
					n.inEdges.add(ne);
					
//					System.out.println("new edge from cycle: " + ne);
				}
				
				/* Finn b�sta v�g in i cykeln. */
				Edge maxOutEdge = null;
				double maxOutEdgeScore = Double.NEGATIVE_INFINITY;
				for(Edge outEdge: n.outEdges) {
					Node d = outEdge.endPoint;
					if(d.isActive && d.inCycle) {
						double sc = outEdge.score - d.headEdge.score;
						if(sc > maxOutEdgeScore) {
							maxOutEdgeScore = sc;
							maxOutEdge = outEdge;
						}
					}
				}
				if(maxOutEdge != null) {
					Edge ne = new Edge();
					ne.score = sC + maxOutEdgeScore;
					ne.startPoint = n;
					ne.endPoint = c;
					ne.replaced = maxOutEdge;
					
					c.inEdges.add(ne);
					n.outEdges.add(ne);
					
//					System.out.println("new edge into cycle: " + ne);
				}
				
			}
		}
		
		return c;
	}
	
	static Node cycleEnd;
	static boolean finished;
	static int roundId;
	
	private static boolean findCycle(ArrayList<Node> ns, ArrayList<Edge> cycle) {
		for(Node n: ns) {
			roundId++;
//			System.out.println("(trying " + n + ")");
			if(n.isActive && n.roundId != roundId && findCycle(n, cycle))
				return true;
		}
		return false;
	}
	
	private static boolean findCycle(Node n, ArrayList<Edge> cycle) {
//		System.out.println("Entering findCycle, n = " + n);
		n.roundId = roundId;
		for(Edge e: n.outEdges) {
			//System.out.println("trying edge " + e);
			if(!e.isMaximal)
				continue;
			//System.out.println("2");
			Node n2 = e.endPoint;
			if(!n2.isActive)
				continue;
			if(n2.roundId == roundId) {
//				System.out.println("cycle found at " + n2);
				
				cycleEnd = n2;
				finished = false;
				cycle.add(e);
				n.inCycle = true;
				//n2.cyclePred = e;
				return true;
			}

			if(findCycle(n2, cycle)) {
//				System.out.println("findCycle returned true");
				
				if(!finished) {
					cycle.add(e);
					n.inCycle = true;
				}
				//n2.cyclePred = e;
				if(cycleEnd == n)
					finished = true;
				return true;
			}
		}
		return false;
	}

	
	
}
