package se.lth.cs.nlp.nlputils.depgraph;

import java.io.BufferedReader;

import se.lth.cs.nlp.nlputils.core.*;

public class DepEvaluator {

    public static void main(String[] argv) {
        boolean nopunc = Boolean.parseBoolean(argv[0]);
        boolean perm = Boolean.parseBoolean(argv[1]);
        String goldfile = argv[2];
        String file1 = argv[3];
        String file2 = null;
        if(argv.length > 4)
            file2 = argv[4];

        //System.out.println(Arrays.toString(argv));
        
        try {
            BufferedReader brg = Util.openFileReader(goldfile, "UTF-8");
            BufferedReader br1 = Util.openFileReader(file1, "UTF-8");
            BufferedReader br2 = file2 == null? null: Util.openFileReader(file2, "UTF-8");

            PRFStats stats1 = new PRFStats(true);
            PRFStats stats2 = new PRFStats(true);
            
            while(true) {
                DepGraph dgg = CoNLLFormat.readNextGraph(brg);
                if(dgg == null)
                    break;
                DepGraph dg1 = CoNLLFormat.readNextGraph(br1);
                if(dg1 == null)
                    throw new RuntimeException("no graph in first file");                
                stats1.add(dgg.eval(dg1, nopunc));
                if(br2 != null) {
                    DepGraph dg2 = CoNLLFormat.readNextGraph(br2);
                    if(dg2 == null)
                        throw new RuntimeException("no graph in second file");
                    stats2.add(dgg.eval(dg2, nopunc));
                }
                //System.out.println("g: " + dgg.nodes.length + "; 1: " + dg1.nodes.length + "; 2: " + dg2.nodes.length);
            }
                       
            System.out.println("* Parser 1 *");
            stats1.print();
            System.out.println();

            if(br2 == null) {
                if(!Util.equal(stats1.getF1(), 0.0)) {
                    System.out.println("* Confidence interval (labeled) *");
                    PRFStats.testSignificance1(stats1, "f1");
                    System.out.println();
                }
                System.out.println("* Confidence interval (unlabeled) *");
                PRFStats.testSignificance1(stats1, "of1");                
            } else {            
                System.out.println("* Parser 2 *");
                stats2.print();
                System.out.println();

                if(!Util.equal(stats1.getF1(), 0.0) && !Util.equal(stats2.getF1(), 0.0)) {
                    System.out.println("* Significance test (labeled) *");
                    PRFStats.testSignificance2(stats1, stats2, "f1", perm);
                    System.out.println();
                }
                System.out.println("* Significance test (unlabeled) *");
                PRFStats.testSignificance2(stats1, stats2, "of1", perm);
            }
            
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
    
}
