package se.lth.cs.nlp.nlputils.depgraph;

import java.io.*;

import se.lth.cs.nlp.nlputils.core.Util;

public class PrintNonproj {

    public static void main1(String[] argv) {
        try {
            PrintWriter pw = new PrintWriter(new OutputStreamWriter(System.out, "UTF-8"));
            BufferedReader br = new BufferedReader(new InputStreamReader(System.in, "UTF-8"));
            
            DepGraph dg = CoNLLFormat.readNextGraph(br);
            while(dg != null) {
                if(!dg.isProjective())
                    CoNLLFormat.printGraph(pw, dg);
                dg = CoNLLFormat.readNextGraph(br);
            }
            
            pw.close();
            
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(0);
        }
    }
    
    public static void main2(String[] argv) {
        try {

            BufferedReader brG = Util.openFileReader(argv[0]);
            BufferedReader brS = Util.openFileReader(argv[1]);

            int nCorrNPGold = 0, nNPGold = 0;
            int nCorrNPPred = 0, nNPPred = 0;
            
            DepGraph dgG = CoNLLFormat.readNextGraph(brG);
            while(dgG != null) {
                DepGraph dgS = CoNLLFormat.readNextGraph(brS);
                for(int i = 1; i < dgG.nodes.length; i++) {
                    DepNode nG = dgG.nodes[i];
                    DepNode nS = dgS.nodes[i];
                    DepNode pG = nG.parents[0];
                    DepNode pS = nS.parents[0];
                    boolean corrLink = pG.position == pS.position;
                    if(!dgG.isProjectiveSpan(i, pG.position)) {
                        nNPGold++;
                        if(corrLink)
                            nCorrNPGold++;
                    }
                    if(!dgS.isProjectiveSpan(i, pS.position)) {
                        nNPPred++;
                        if(corrLink)
                            nCorrNPPred++;
                    }
                }
                dgG = CoNLLFormat.readNextGraph(brG);
            }
            
            double pr = (double) nCorrNPPred / nNPPred;
            double re = (double) nCorrNPGold / nNPGold;
            
            double f1 = 2*pr*re/(pr+re);

            System.out.println("pr = " + nCorrNPPred + "/" + nNPPred + " = " + pr);
            System.out.println("re = " + nCorrNPGold + "/" + nNPGold + " = " + re);
            
            System.out.println("f1 = " + f1);
            
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(0);
        }
    }
    
    public static void main(String[] argv) {
        main2(argv);
    }
    
}
