package se.lth.cs.nlp.nlputils.math;

public class CubicInterpolation {

    public static double compute(double[][] samples, double x) {

        if(samples.length < 4)
            throw new IllegalArgumentException("At least 4 samples needed");

        int n = samples.length;
        
        if(x < samples[0][0])
            throw new IllegalArgumentException("x is lower than lowest sample");
        if(x > samples[n-1][0])
            throw new IllegalArgumentException("x is higher than highest sample");
        
        if(x == samples[0][0])
            return samples[0][1];
        if(x == samples[1][0])
            return samples[1][1];
        if(x == samples[n-1][0])
            return samples[n-1][1];
        if(x == samples[n-2][0])
            return samples[n-2][1];

        if(n == 4 || x > samples[0][0] && x < samples[1][0])
            return computeCubic(samples, 0, x);
        if(x > samples[n-2][0] && x < samples[n-1][0])
            return computeCubic(samples, n - 4, x);
        
        int low = 0, mid = samples.length / 2, high = samples.length - 1;

        while(low < mid && mid < high) {
        	//System.out.printf("%d %d %d\n", low, mid, high);
        	//System.out.printf("%f %f %f\n", 
        	//		samples[low][0], samples[mid][0], samples[high][0]);
        	double m = samples[mid][0];
            if(x == m)
                return samples[mid][1];
            else if(x < m)
            	high = mid;
            else
            	low = mid;
        	mid = (low + high) / 2;
        }
        
    	//System.out.printf("%d %d %d\n", low, mid, high);
    	//System.out.printf("%f %f %f\n",         
    	//		samples[low][0], samples[mid][0], samples[high][0]);
    			
        if(mid == low)
        	return computeCubic(samples, low - 1, x);
        else
        	return computeCubic(samples, low - 2, x);
        
        //System.out.println("found position: " + low);       
        //return 0;
    }
 
    private static double computeCubic(double[][] samples, int start, double x) {
        
        DMatrix m = new DMatrix(4, 4);
        DVector v = new DVector(4);
        
        for(int i = 0; i < 4; i++) {
            double d = 1.0;
            m.set(i, 0, d);
            for(int j = 1; j < 4; j++) {
                d *= samples[i+start][0];
                m.set(i, j, d);
            }
            v.set(i, samples[i+start][1]);
        }
        
        //System.out.println("Before");
        //System.out.println(m);
        //System.out.println(v);
        
        /* Solve linear system. */

        for(int i = 3; i >= 0; i--) {

            double d = m.get(i, i);
            if(d == 0)
                throw new RuntimeException("Too simplistic, can't solve this system...");
            
            m.scaleRow(i, 1.0 / d);
            v.set(i, v.get(i) / d);

            //for(int j = i - 1; j >= 0; j--) {
            for(int j = 3; j >= 0; j--) {
            	if(j != i) {
            		double f = m.get(j, i);
            		m.rowDaxpy(i, j, -f);
            		v.set(j, v.get(j) - f*v.get(i));
            	}
            }
            
            //for(int j = i + 1; j <= 3; j++) {
            //	double f = m.get(i, j)
            //}
            
        }

        //System.out.println("After");        
        //System.out.println(m);
        //System.out.println(v);
        
        double xm = x;
        double out = v.get(0);
        out += xm * v.get(1);
        xm *= x;
        out += xm * v.get(2);
        xm *= x;        
        out += xm * v.get(3);

        return out;
    }
    
    private static double[][] PHI_INV = {
        { 0.5, 0 },
        { 0.6, 0.253347103135800 },
        { 0.7, 0.524400512708041 },
        { 0.75, 0.674489750196082 },
        { 0.8, 0.841621233572914 },
        { 0.85, 1.03643338949379 },
        { 0.9, 1.28155156554460 },
        { 0.95, 1.64485362695147 },
        { 0.975, 1.95996398454005 },
        { 0.99, 2.32634787404084 },
        { 0.995, 2.57582930354890 },
        { 0.999, 3.09023230616781 },
        { 0.9995, 3.29052673149194 },
        { 0.9999, 3.71901648545567 },
        { 0.99995, 3.89059188641317 },
        { 0.99999, 4.26489079392334 },
        { 0.999995, 4.41717341346829 },
        { 0.999999, 4.75342430881759 },
};
    
    public static void main(String[] argv) {
    	//double[][] d = new double[][] {
        //        { 0, 0 }, { 0.5, 0.48 }, { 1, 0.84 }, {1.5, 1}, {2, 0.91}, {2.5, 0.6} 
        //};
        
        System.out.println(compute(PHI_INV, Double.parseDouble(argv[0])));
        
    }

}
