/*
 DVector.java
  
 Created on May 11, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: DVector.java,v $
 Revision 1.1  2009/01/16 10:06:31  johansson
 Added to brenta repository.

 Revision 1.5  2007/06/14 12:12:11  richard
 Removed factor.

 Revision 1.4  2007/05/29 07:07:24  richard
 Dimension check for transposed matrix.

 Revision 1.3  2007/05/22 09:09:24  richard
 Length checks and toString.

 Revision 1.2  2007/05/16 14:42:29  richard
 Added addElements, removeLastElements, add, and set.

 Revision 1.1  2007/05/11 14:56:55  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.math;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class DVector {

    private static final int INCREMENT = 10;
    
    public double[] data;
    public int length;

    public DVector(int length) {
        this.length = length;
        data = new double[length];
    }
    
    public double get(int i) {
        if(i < 0 || i >= length)
            throw new ArrayIndexOutOfBoundsException("DVector: i = " + i);
        return data[i];
    }

    public void set(int i, double d) {
        if(i < 0 || i >= length)
            throw new ArrayIndexOutOfBoundsException("DVector: i = " + i);
        data[i] = d;
    }
    
    public double dot(DVector v) {
        if(length != v.length)
            throw new IllegalArgumentException("length != v.length");
        return BLAS.ddot(length, data, 1, v.data, 1, 0, 0);
    }

    public void scale(double d) {
        BLAS.dscal(length, d, data, 1, 0);
    }
    
    public void copy(DVector x) {
        if(length != x.length)
            throw new IllegalArgumentException("length = " + length + ", x.length = " + x.length);
        BLAS.dcopy(length, x.data, 1, data, 1, 0, 0);
    }
    
    public void addAX(DMatrix a, DVector x, double alpha, double beta) {
        if(!a.trans) {
            if(length != a.m)
                throw new IllegalArgumentException("length = " + length + ", a.m = " + a.m);
            if(x.length != a.n)
                throw new IllegalArgumentException("x.length != a.n");
        } else {
            if(x.length != a.m)
                throw new IllegalArgumentException("x.length = " + length + ", a.m = " + a.m);
            if(length != a.n)
                throw new IllegalArgumentException("length != a.n");
        }
        
        BLAS.dgemv(a.trans, a.m, a.n, alpha, a.data, a.lda, 
                x.data, 1, beta, data, 1, 0, 0, 0, 0);
    }
    
    public void add(DVector x, double alpha) {
        if(length != x.length)
            throw new IllegalArgumentException("length != x.length");
        BLAS.daxpy(length, alpha, x.data, 1, data, 1, 0, 0);
    }
    
    public void addElements(int n) {
        if(length + n > data.length) {
            int nl = length + n + INCREMENT;
            double[] tmp = new double[nl];
            System.arraycopy(data, 0, tmp, 0, length);
            data = tmp;
        }
        length += n;
    }
    
    public void removeLastElements(int n) {
        length -= n;
    }
    
    public String toString() {
        StringBuilder sb = new StringBuilder("[");        
        if(length > 0) {
            for(int i = 0; i < length - 1; i++)
                sb.append(data[i] + ", ");
            sb.append(data[length - 1]);
        }
        sb.append("]");
        return sb.toString();
    }
    
}