package se.lth.cs.nlp.nlputils.math;

import gnu.trove.*;
import java.util.*;
import se.lth.cs.nlp.nlputils.core.*;

public class PermutationTest {

	private static double estimate(DoubleFunction<int[]> estimator,
			Random rand, int sampleLength) {
		//TIntArrayList flips = new TIntArrayList();
		int[] out = new int[sampleLength];
		int index = 0;
		for(int i = 0; i < sampleLength; i++)
			if(rand.nextBoolean())
				out[index++] = i;
				//flips.add(i);
		if(index < sampleLength)
			out[index] = -1;		
		
		//System.out.println(flips.size() + ",");
		
		//return estimator.apply(flips.toNativeArray());
		return estimator.apply(out);
	}
	
	public static Statistics permutationTest(DoubleFunction<int[]> estimator,
			int sampleLength, Random rand, double tolerance, int minNIter,
			int maxNIter) {

		double estMeanDiff = estimator.apply(new int[0]);
		
		//TDoubleArrayList resl = new TDoubleArrayList();
		double[] resl = new double[maxNIter];

		double result;
		double resSum = 0;
		double resAvg = Double.MAX_VALUE;
		double prevAvg;
		int niter = 0;
		do {
			prevAvg = resAvg;
			result = estMeanDiff + estimate(estimator, rand, sampleLength);
			resSum += result;

			resl[niter] = result;
			
			niter++;
			resAvg = resSum / niter;
			//resl.add(result);
		} while(niter < maxNIter && (niter < minNIter || !Util.equal(resAvg, prevAvg, tolerance)));

		//double[] results = resl.toNativeArray();
		double[] results = new double[niter];
		System.arraycopy(resl, 0, results, 0, niter);
		resl = null;
		
		Arrays.sort(results);

		Statistics out = new Statistics();
		out.size = niter;
		out.mean = resAvg;
		out.std = Vectors.std(results);
		if(niter % 2 == 1) {
			out.q01 = results[niter/100];
			out.q025 = results[niter/40];
			out.q05 = results[niter/20];
			out.median = results[niter/2];
			out.q95 = results[niter - niter/20 - 1];
			out.q975 = results[niter - niter/40 - 1];
			out.q99 = results[niter - niter/100 - 1];
		} else {
			out.q01 = 0.5 * (results[niter/100-1] + results[niter/100]);
			out.q025 = 0.5 * (results[niter/40-1] + results[niter/40]);
			out.q05 = 0.5 * (results[niter/20-1] + results[niter/20]);
			out.median = 0.5 * (results[niter/2-1] + results[niter/2]);
			out.q95 = 0.5 * (results[niter-niter/20-2] + results[niter-niter/20-1]);
			out.q975 = 0.5 * (results[niter-niter/40-2] + results[niter-niter/40-1]);
			out.q99 = 0.5 * (results[niter-niter/100-2] + results[niter-niter/100-1]);
		}
		out.data = results;
		return out;
	
	}
	
}
