package se.lth.cs.nlp.nlputils.math;

import java.util.*;

import se.lth.cs.nlp.nlputils.core.*;

public class Resampler {

	private static double estimate(DoubleFunction<int[]> estimator,
			Random rand, int range, int samplesize) {
		int[] selected = new int[samplesize];
		for(int i = 0; i < samplesize; i++)
			selected[i] = rand.nextInt(range);
		return estimator.apply(selected);
	}

	public static Statistics resamplingTest(DoubleFunction<int[]> estimator,
			int range, int samplesize, Random rand, double tolerance, int minNIter,
			int maxNIter) {

		//TDoubleArrayList resl = new TDoubleArrayList();
		double[] resl = new double[maxNIter];
		double result;
		double resSum = 0;
		double resAvg = Double.MAX_VALUE;
		double prevAvg;
		int niter = 0;
		do {
			prevAvg = resAvg;
			result = estimate(estimator, rand, range, samplesize);
			resSum += result;

			resl[niter] = result;
			
			niter++;
			resAvg = resSum / niter;
			
			//resl.add(result);

			
		} while(niter < maxNIter && (niter < minNIter || !Util.equal(resAvg, prevAvg, tolerance)));

		Arrays.sort(resl, 0, niter);
		
		double[] results = new double[niter];
		System.arraycopy(resl, 0, results, 0, niter);
		resl = null;
		
		//double[] results = resl.toNativeArray();
		//Arrays.sort(results);

		Statistics out = new Statistics();
		out.size = niter;
		out.mean = resAvg;
		
		out.std = Vectors.std(results);

		if(niter % 2 == 1) {
			out.q01 = results[niter/100];
			out.q025 = results[niter/40];
			out.q05 = results[niter/20];
			out.median = results[niter/2];
			out.q95 = results[niter - niter/20 - 1];
			out.q975 = results[niter - niter/40 - 1];
			out.q99 = results[niter - niter/100 - 1];
		} else {
			out.q01 = 0.5 * (results[niter/100-1] + results[niter/100]);
			out.q025 = 0.5 * (results[niter/40-1] + results[niter/40]);
			out.q05 = 0.5 * (results[niter/20-1] + results[niter/20]);
			out.median = 0.5 * (results[niter/2-1] + results[niter/2]);
			out.q95 = 0.5 * (results[niter-niter/20-2] + results[niter-niter/20-1]);
			out.q975 = 0.5 * (results[niter-niter/40-2] + results[niter-niter/40-1]);
			out.q99 = 0.5 * (results[niter-niter/100-2] + results[niter-niter/100-1]);
		}
		out.data = results;
		
		return out;
	}
	
	private static class MeanEstimator extends DoubleFunction<int[]> {
		double[] xs;
		MeanEstimator(double[] xs) {
			this.xs = xs;
		}
		public double apply(int[] selected) {
			double sum = 0;
			for(int ix: selected)
				sum += xs[ix];
			return sum / selected.length;
		}
	}

	private static class StdEstimator extends DoubleFunction<int[]> {
		double[] xs;
		StdEstimator(double[] xs) {
			this.xs = xs;
		}
		public double apply(int[] selected) {
			double sum = 0;
			double qsum = 0;
			for(int ix: selected) {
				sum += xs[ix];
				qsum += xs[ix] * xs[ix];
			}
			int n = selected.length;
			return Math.sqrt((qsum - sum*sum/n) / (n-1));
		}
	}
	
	private static double nextExp(Random rand, double lambda) {
		return -Math.log(rand.nextDouble()) / lambda;
	}
	
	public static void main(String[] argv) {
		int N = 100;
		double[] d = new double[N];
		Random rand = new Random(0);
		for(int i = 0; i < N; i++)
			//d[i] = rand.nextGaussian();
			//d[i] = rand.nextDouble();
			//d[i] = rand.nextInt()
			d[i] = nextExp(rand, 0.25);

		System.out.println("Sample mean: " + Vectors.mean(d));
		System.out.println("Sample std:  " + Vectors.std(d));

		System.out.println("Mean standard error: " + Vectors.std(d) / Math.sqrt(N));
		
		StdEstimator est = new StdEstimator(d);			
		//MeanEstimator est = new MeanEstimator(d);		
		System.out.println(resamplingTest(est, d.length, d.length, rand, 1e-12, 10000, Integer.MAX_VALUE));
	}
	
}
