package se.lth.cs.nlp.nlputils.math;

import se.lth.cs.nlp.nlputils.core.CollectionUtils;

public class Vectors {

	public static double sum(double[] d) {
		double out = 0;
		for(int i = 0; i < d.length; i++)
			out += d[i];
		return out;
	}
	
	public static double sum(double[] d, int from, int to) {
		double out = 0;
		for(int i = from; i < to; i++)
			out += d[i];
		return out;
	}
	
	public static double mean(double[] d) {
		return sum(d) / d.length;
	}

	public static double var(double[] d) {
		if(d.length == 1)
			return 0;
		if(d.length == 0)
			throw new IllegalArgumentException("Empty vector");
		double sum = sum(d);
		double sqsum = 0;
		for(int i = 0; i < d.length; i++)
			sqsum += d[i]*d[i];
		double q = sqsum - sum*sum/d.length;
		return q / (d.length - 1);
	}
	
	public static double var(double[] d, int from, int to) {
		int len = to - from;
		if(len == 1)
			return 0;
		if(len <= 0)
			throw new IllegalArgumentException("Empty vector");
		double sum = sum(d, from, to);
		double sqsum = sqsum(d, from, to);
		//for(int i = from; i < to; i++)
		//	sqsum += d[i]*d[i];
		double q = sqsum - sum*sum/len;
		return q / (len - 1);
	}

	public static double std(double[] d) {
		return Math.sqrt(var(d));
	}
	
	public static double std(double[] d, int from, int to) {
		return Math.sqrt(var(d, from, to));
	}
		
	public static void normalizeLogProbs(double[] scores, int length, double scale, double offset) {
		double m = CollectionUtils.max(scores, length);
		m = scale*m + offset;
		double expSum = 0;
		for(int i = 0; i < length; i++) {
			scores[i] = scale*scores[i] + offset;
			expSum += Math.exp(scores[i] - m);
		}
		double logExpSum = m + Math.log(expSum);
		for(int i = 0; i < length; i++)
			scores[i] -= logExpSum;
	}
	
	public static void normalizeLogProbs(double[] scores) {
		normalizeLogProbs(scores, scores.length, 1.0, 0.0);
	}

	public static double sqsum(double[] d, int from, int to) {
		double out = 0.0;
		for(int i = from; i < to; i++)
			out += d[i]*d[i];
		return out;
	}
	
	public static double sqsum(double[] d) {
		return sqsum(d, 0, d.length);
	}
	
}
