
/*
 	
 	AlgorithmFactory.java
  
    Created on Sep 8, 2006 by Richard Johansson (richard@cs.lth.se).
 	
 	$Log: AlgorithmFactory.java,v $
 	Revision 1.2  2009/01/16 10:08:13  johansson
 	Added to brenta repository.
 	
 	Revision 1.1  2009/01/16 10:06:48  johansson
 	Added to brenta repository.
 	
 	Revision 1.1  2008/05/16 12:16:04  richard
 	Branched.
 	
 	Revision 1.2  2008/04/02 11:22:18  richard
 	Verbosity.
 	
 	Revision 1.1  2007/09/28 14:47:02  richard
 	Moved to ml.
 	
 	Revision 1.7  2007/04/05 13:32:29  richard
 	Offset.
 	
 	Revision 1.6  2007/03/01 09:20:37  richard
 	Indented.
 	
 	Revision 1.5  2006/09/27 19:02:49  richard
 	Added a second default path.
 	
 	Revision 1.4  2006/09/22 08:07:53  richard
 	Renamed specification to definition.
 	
 	Revision 1.3  2006/09/19 08:32:06  richard
 	Fixed argument string bug.
 	
 	Revision 1.2  2006/09/18 15:17:49  richard
 	Trims arg string.
 	
 	Revision 1.1  2006/09/08 18:17:02  richard
 	Added the file.
 	
 
 */

package se.lth.cs.nlp.nlputils.ml_long;

import java.lang.reflect.*;

/**
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class AlgorithmFactory<L extends ProblemDefinition<I, V>, I, V> {

    private static final String DEFAULT_PATH_1 = "se.lth.cs.nlp.nlputils.ml_long.algorithms.";
    private static final String DEFAULT_PATH_2 = "se.lth.cs.nlp.nlputils.ml_long.glue.";

    public LearningAlgorithm<L, I, V> create(String name, String argString) {
        argString = argString.trim();
        if(argString.equals(""))
            return create(name, new String[0], 0);
        else {
            String[] argv = argString.split("[ =]+");
            return create(name, argv, 0);
        }
    }

    private static boolean quiet = false;
    
    public LearningAlgorithm<L, I, V> create(String name, String[] argv, int offset) {
        try {
            Class cl;
            try {
                cl = Class.forName(name);
            } catch(ClassNotFoundException e) {
                try {
                    cl = Class.forName(DEFAULT_PATH_1 + name);
                } catch(ClassNotFoundException e2) {
                    cl = Class.forName(DEFAULT_PATH_2 + name);
                }
            }

            if(!quiet)
                System.out.println("Loaded class (" + cl.getName() + ").");

            Constructor con = cl.getConstructor(new Class[] { String[].class, int.class });

            if(!quiet) {
                System.out.println("Loaded constructor (" + con + ").");
                System.out.println("argv = " + java.util.Arrays.toString(argv));
            }

            return (LearningAlgorithm) con.newInstance(new Object[] {argv, offset});
        }
        catch(ClassNotFoundException e) {
            throw new RuntimeException("Couldn't load class " + name);
        }
        catch(InvocationTargetException e) {
            throw new RuntimeException("Illegal arguments to " + name + ": " + e.getTargetException().getMessage());
        }
        catch(NoSuchMethodException e) {
            throw new RuntimeException("The class " + name + " does not have a string array constructor.");
        }
        catch(Exception e) {
            e.printStackTrace();
            throw new RuntimeException("Couldn't create!"); 
        }
    }
    
    public static void setVerbosity(int level) {
        if(level <= 0)
            quiet = true;
    }

}
