/*
 Classifier.java
 
 Created on Dec 22, 2005 by Richard Johansson (richard@cs.lth.se).
 
 $Log: Classifier.java,v $
 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.14  2008/04/10 08:33:02  richard
 Added getModelDensity.

 Revision 1.13  2008/04/10 06:59:28  richard
 Message.

 Revision 1.12  2008/03/14 08:54:51  richard
 Serial number.

 Revision 1.11  2008/03/14 08:45:31  richard
 Added nbest.

 Revision 1.10  2007/09/28 14:47:53  richard
 Refactored to atom-based.

 Revision 1.9  2006/11/20 12:43:29  richard
 Renamed to classifyByFeaturesRestricted.

 Revision 1.8  2006/09/19 13:27:24  richard
 Comment.

 Revision 1.7  2006/09/08 15:32:32  richard
 Parametrized FeatureList.

 Revision 1.6  2006/09/07 17:51:19  richard
 New revision.

 Revision 1.5  2006/04/11 09:31:16  richard
 Changed the API of the FeatureExtractor.extractFeatures method.
 
 Revision 1.4  2006/04/10 09:00:10  richard
 Added documentation.
 
 Revision 1.3  2006/04/07 14:00:46  richard
 Added restricted classification.
 
 Revision 1.2  2006/04/05 11:39:56  richard
 Implements serializable. Added setFeatureExtractor.
 
 Revision 1.1  2006/01/13 13:19:37  richard
 Added the file.
 
 
 */
package se.lth.cs.nlp.nlputils.ml_long;

import java.util.*;
import java.io.Serializable;

/**
 * Abstract class representing a classifier.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public abstract class Classifier<I, V> implements Serializable {
    
    private static final long serialVersionUID = 1L;
    
    /**
     *
     * @param item the item to classify.
     */
    public V classify(I instance) {
        throw new UnsupportedOperationException("Not supported.");        
    }
        
    /**
     * Optional method that makes sure that restricts the range of the output 
     * value.
     * 
     * @param features the feature list.
     * @param classes a collection of allowed classes.
     * @return the result of the classification.
     * @throws UnsupportedOperationException if the method is not implemented.
     */
    public V classifyRestricted(I instance, Collection<V> classes) {
        throw new UnsupportedOperationException("Restricted classification not supported ("
                + this.getClass().getName() + ")");
    }
    
    /**
     *
     * @param item the item to classify.
     */
    public V classifyByFeatures(FeatureList features) {
        throw new UnsupportedOperationException("Classification by features not supported.");       
    }

    public void setFeatureExtractor(FeatureExtractor<I> fe) {
        throw new UnsupportedOperationException("Not supported.");
    }
    
    /**
     *
     * @param item the item to classify.
     */
    public V classifyByFeaturesRestricted(FeatureList features, Collection<V> classes) {
        throw new UnsupportedOperationException("Restricted classification by features not supported.");       
    }
    

   /**
    * TODO order
    *
    * @param item the item to classify.
    */
   public void computeScores(I instance, double[] scores) {
       throw new UnsupportedOperationException("Scoring not supported.");        
   }
       
   /**
    * Optional method that makes sure that restricts the range of the output 
    * value.
    * 
    * @param features the feature list.
    * @param classes a collection of allowed classes.
    * @return the result of the classification.
    * @throws UnsupportedOperationException if the method is not implemented.
    */
   public void computeScoresRestricted(I instance, Collection<V> classes, double[] scores) {
       throw new UnsupportedOperationException("Restricted scoring not supported.");
   }
   
   /**
    *
    * @param item the item to classify.
    */
   public void computeScoresByFeatures(FeatureList features, double[] scores) {
       throw new UnsupportedOperationException("Scoring by features not supported.");       
   }

   /**
    *
    * @param item the item to classify.
    */
   public void computeScoresByFeaturesRestricted(FeatureList features, Collection<V> classes, double[] scores) {
       throw new UnsupportedOperationException("Restricted scoring by features not supported.");       
   }
    
   public int nbest(int n, I instance, V[] values, double[] scores) {
       throw new UnsupportedOperationException("nbest not supported.");
   }
   
   public int nbest(int n, I instance, V[] values, double[] scores, SparseVector[] reps) {
       throw new UnsupportedOperationException("nbest with reps not supported.");
   }

   public double getModelDensity() {
       throw new UnsupportedOperationException("getModelDensity is not supported.");       
   }
   
}
