package se.lth.cs.nlp.nlputils.ml_long;

import se.lth.cs.nlp.nlputils.core.*;
import gnu.trove.*;
import java.util.*;

import java.io.*;

public class InfoGain {

	public static class InfoStats {
		int feature;
		double infogain;
		double fentr;		
		public InfoStats(double infogain, double fentr, int feature) {
			this.feature = feature;
			this.infogain = infogain;
			this.fentr = fentr;
		}
		public String toString() {
			return "(" + feature + ", " + infogain + ", " + fentr + ")";
		}
	}
	
    public static final Comparator<InfoStats> BY_GAIN = new Comparator<InfoStats>() {
        public int compare(InfoStats p1, InfoStats p2) {
            return -Double.compare(p1.infogain, p2.infogain);
        }
    };
    
    public static final Comparator<InfoStats> BY_FENTR = new Comparator<InfoStats>() {
        public int compare(InfoStats p1, InfoStats p2) {
            return -Double.compare(p1.fentr, p2.fentr);
        }
    };
    
    public static final Comparator<InfoStats> BY_FEATURE = new Comparator<InfoStats>() {
        public int compare(InfoStats p1, InfoStats p2) {
            return p1.feature - p2.feature;
        }
    };
    
	
	public static <T> ArrayList<InfoStats> 
	infoGain(ArrayList<Pair<SparseVector, T>> ts) {
		// for now only presence
		// also, we truncate the longs to int for the moment...

		HashMap<T, IntHistogram> m = new HashMap();

		Histogram<T> clHist = new Histogram();
		IntHistogram fhist = new IntHistogram();

		int N = ts.size();

		for(Pair<SparseVector, T> p: ts) {
			IntHistogram hist = m.get(p.right); // effektivare?
			if(hist == null) {
				hist = new IntHistogram();
				m.put(p.right, hist);
			}
			SparseVector sv = p.left;
			for(int i = 0; i < sv.index; i++) {
				if(sv.keys[i] > (long) Integer.MAX_VALUE || sv.keys[i] < (long) Integer.MIN_VALUE)
					throw new RuntimeException("Only int range for the moment");
				if(!Util.equal(sv.values[i], 0.0)) {
					int k = (int) sv.keys[i];
					hist.add(k);
					fhist.add(k);
				}
			}
			clHist.add(p.right);
		}

		ArrayList<IntObjPair<T>> cls = clHist.asSortedList();
		IntHistogram[] hists = new IntHistogram[cls.size()];
		
		for(int i = 0; i < cls.size(); i++)
			hists[i] = m.get(cls.get(i).right);
		
		System.out.println("classes: " + cls);

		double entropy = 0;
		for(int i = 0; i < cls.size(); i++)
			entropy += entropy((double) cls.get(i).left / N);
		
		System.out.println("Entropy: " + entropy);
		
		// H(Y) = sum(y=yval) -p log p, där p = P(Y=yval)
		// H(Y|X=xval) = sum(y=yval) -p log p, där p = P(Y=yval|X=xval)

		// TODO kolla H(Y) - H(Y|X): korrekt med minus?
		// Infogain för en feature v (true eller false)
		// - P(v=true) * H(Y | v=true) - P(v=false) * H(Y | v=false)
		// = - #(v=true)/N * H(Y | v=true) + ...

		// TODO verkar som om det blir 0 för de som bara kommer 1 gång
		// stämmer detta? det är ju så att den räknas som 0 alla gånger
		// den inte syns!
		// kolla också relative IG?

		ArrayList<InfoStats> out = new ArrayList();

		for(IntPair p: fhist.asSortedList()) { // TODO bättre iter
			// p: freq, val
			int feature = p.right;

			int n_pos = p.left;
			int n_neg = N - n_pos;
			
			// p_pos = P(this feature is set)
			double p_pos = (double) n_pos / N;
			// p_neg = P(this feature is not set)
			double p_neg = 1.0 - p_pos;

			double fentr = entropy(p_pos) + entropy(p_neg);		
			double condentr = 0;

			for(int i = cls.size() - 1; i >= 0; i--) {
				IntHistogram hist = hists[i]; 

				// frequency of this output value
				int n_tval = cls.get(i).left; 

				// frequency of this feature for this output value
				int n_tval_pos = hist.getFrequency(feature);
				//P(Y=tval|this feature is set)
				double p_tval_pos = (double) n_tval_pos / n_pos;
				condentr += p_pos * entropy(p_tval_pos);

				int n_tval_neg = n_tval - n_tval_pos;
				if(n_neg > 0) {
					//P(Y=tval|this feature is not set)
					double p_tval_neg = (double) n_tval_neg / n_neg;
					condentr += p_neg * entropy(p_tval_neg);
				}

			}
			out.add(new InfoStats(entropy - condentr, fentr, feature));
		}

		Collections.sort(out, BY_FEATURE);
		Collections.sort(out, BY_FENTR);
		Collections.sort(out, BY_GAIN);
		return out;
	}

	private static final double LOG2_INV = 1.0 / Math.log(2);
	private static double log2(double x) {
		return Math.log(x) * LOG2_INV;
	}
	
	private static double entropy(double p) {
		if(p < 0 || p > 1)
			throw new IllegalArgumentException("Illegal p value");
		if(p < 1e-12 || p > 1.0 - 1e-12)
			return 0;
		return -p * log2(p);
	}
	
	public static void main(String[] argv) {
		try {
			BufferedReader br = Util.openFileReader(argv[0]);

			ArrayList<Pair<SparseVector, Integer>> ts = new ArrayList();

			while(true) {
				String line = br.readLine();
				if(line == null)
					break;
				String[] ss = line.split("\\s+");

				int cl = ss[0].startsWith("+")? 
						Integer.parseInt(ss[0].substring(1)):
							Integer.parseInt(ss[0]);

						SparseVector sv = new SparseVector();
						for(int i = 1; i < ss.length; i++) {
							int ix = ss[i].indexOf(':');
							long k = Long.parseLong(ss[i].substring(0, ix));
							double v = Double.parseDouble(ss[i].substring(ix+1));
							sv.put(k, v);
						}
						sv.trim();
						ts.add(new Pair(sv, cl));
			}
			System.out.println("Read data.");

			ArrayList<InfoStats> out = infoGain(ts);

			for(InfoStats p: out) {
				System.out.println(p);
			}

		} catch(Exception e) {
			e.printStackTrace();
		}
	}


}
