/*
 IntPairEncoder.java
  
 Created on Oct 5, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: IntPairEncoder2.java,v $
 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.7  2008/04/10 07:43:19  richard
 Added size.

 Revision 1.6  2008/02/22 13:17:33  richard
 Added decode.

 Revision 1.5  2007/11/07 15:36:59  richard
 Added.

 Revision 1.5  2007/11/06 09:57:23  richard
 Changed remapLeft.

 Revision 1.4  2007/10/22 07:47:00  richard
 Uses IntHistogram.

 Revision 1.3  2007/10/22 07:31:35  richard
 Renamed IntPairEncoder2 to IntPairEncoder.

 Revision 1.3  2007/10/22 07:27:37  richard
 Added remapLeft.

 Revision 1.2  2007/10/20 14:46:01  richard
 Added remap and freeze.

 Revision 1.1  2007/10/19 13:50:31  richard
 Added.

 Revision 1.2  2007/10/08 12:31:34  richard
 Refactored again to remove atoms.

 Revision 1.1.2.1  2007/10/08 12:19:04  richard
 Refactored again to remove atoms.

   
 */
package se.lth.cs.nlp.nlputils.ml_long;

import java.io.Serializable;
import se.lth.cs.nlp.nlputils.core.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class IntPairEncoder2 implements Serializable {

    /* TODO r det verkligen effektivt att ha en long->int-map? */
    
    private static final long serialVersionUID = 0L;
    private ClosedLongIntMap encoding;
    private int nextIndex;
    private boolean frozen = false;
    
    public IntPairEncoder2(int size) {
        encoding = new ClosedLongIntMap(size);
        nextIndex = 1;
    }
    
    public IntPairEncoder2(int size, double lf, double gf) {
        encoding = new ClosedLongIntMap(size, lf, gf);
        nextIndex = 1;
    }
    
    public int encode(int c, int a) {
        long l = a | ((long) c << 32L);
        if(!frozen) {   
            int encoded = encoding.getOrSet(l, nextIndex);
            if(encoded == nextIndex)
                nextIndex++;
            return encoded;
        } else
            return encoding.get(l);
    }

    public void remapLeft(int ntop) {
        
    }
    
    public void freeze() {
        frozen = true;
        nextIndex = 0;
    }
    
    public int size() {
        return encoding.size();
    }
    
    public IntPair decode(int encoded) {
        long p = encoding.getKey(encoded);
        if(p == 0)
            return null;
        int lower = (int) (p & ((1L << 32L) - 1));
        int upper = (int) (p >> 32L);
        return new IntPair(upper, lower);
    }
    
}
