/*
 IntPairEncoder3.java
  
 Created on Oct 5, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: IntPairEncoder3.java,v $
 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.1  2008/03/03 12:19:33  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.ml_long;

import java.io.Serializable;
import se.lth.cs.nlp.nlputils.core.*;


/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class IntPairEncoder3 implements Serializable {

    private static final long serialVersionUID = 0L;

    private ClosedLongIntMap encoding2;
    private ClosedIntIntMap encoding;

    private int size = 0;

    private int N_F_LOW_BITS;
    private int N_C_LOW_BITS;

    private int LOW_PART_F;
    private int LOW_PART_C;

    private boolean frozen = false;
    
    public IntPairEncoder3(int halfbits, int size1, int size2) {
        if(halfbits > 31 || halfbits < 1)
            throw new IllegalArgumentException("0 < halfbits < 32");
        N_F_LOW_BITS = halfbits;
        N_C_LOW_BITS = 32 - halfbits; // /* hr */ - 1;
        LOW_PART_F = 1 << N_F_LOW_BITS;
        LOW_PART_C = 1 << N_C_LOW_BITS;        
        encoding2 = new ClosedLongIntMap(size2);
        encoding = new ClosedIntIntMap(size1);        
    }
    
    public int encode(int c, int a) {
        int encoded;
        if(a < LOW_PART_F && c < LOW_PART_C) {
            int l = a + (c << N_F_LOW_BITS);
            encoded = encoding.get(l);
            if(encoded == 0 && !frozen) {
                encoded = ++size;
                encoding.put(l, encoded);
            }
        } else {
            long l = a + ((long) c << 32L);
            encoded = encoding2.get(l);
            if(encoded == 0 && !frozen) {
                encoded = ++size;
                encoding2.put(l, encoded);
            }
        }

        if(encoded == 0) {
            //System.out.println("(" + c + ", " + a + ") -> " + encoded);
            //System.exit(1);
        }
        
        return encoded;
    }
    
    public void freeze() {
        frozen = true;
    }
    
    public int size() {
        return encoding.size() + encoding2.size();
    }
    
    public IntPair decode(int encoded) {
        return null;
    }
    
}
