
/* KernelModel.java

   Created on Sep 22, 2006 by Richard Johansson (richard@cs.lth.se).

   $Log: KernelModel.java,v $
   Revision 1.1  2009/01/16 10:06:48  johansson
   Added to brenta repository.

   Revision 1.1  2008/05/16 12:16:04  richard
   Branched.

   Revision 1.8  2007/02/26 08:02:39  richard
   Added factor, refresh, and clear.

   Revision 1.7  2007/02/15 08:51:31  richard
   Copies support vectors.

   Revision 1.6  2007/01/12 11:49:08  richard
   Repaired dot.

   Revision 1.5  2006/10/27 13:04:18  richard
   Bug message.

   Revision 1.4  2006/10/25 07:07:39  richard
   SparseVector as a sequence.

   Revision 1.3  2006/10/05 07:28:42  richard
   Removed import.

   Revision 1.2  2006/10/05 06:58:38  richard
   Serious implementation.

   Revision 1.1  2006/09/22 10:12:38  richard
   Added the file.


 */

package se.lth.cs.nlp.nlputils.ml_long;

import java.io.Serializable;
import java.util.Arrays;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class KernelModel extends Model implements Serializable {

    static {
        //if(true)
        //    throw new RuntimeException("This class is buggy!");
        //System.err.println("WARNING! The KernelModel class is buggy!");
    }
    
    private static final long serialVersionUID = 0L;

    public static final int LINEAR_KERNEL = 0;
    public static final int POLYNOMIAL_KERNEL = 1;
    public static final int GAUSSIAN_KERNEL = 2;
    public static final int SIGMOID_KERNEL = 3;
        
    private int kernelType = 0;
    private int polyDegree;
    private double gamma;
    private double s;
    private double c;
    
    protected double[] alphas;
    protected SparseVector[] svs;
    protected int m = 0;

    public KernelModel(int kernelType, int polyDegree, double gamma,
            double s, double c) {
        this.kernelType = kernelType;
        this.polyDegree = polyDegree;
        this.gamma = gamma;
        this.s = s;
        this.c = c;

        this.alphas = new double[5000];
        this.svs = new SparseVector[5000];
    }
    
    public void update(double d, SparseVector rep) {
    	alphas[m] = d;
    	svs[m] = rep.copy();
    	m++;
    }

    public void updateDiff(double d, SparseVector r1, SparseVector r2, boolean sorted) {
        throw new RuntimeException("Unimplemented");
    }
    
    void refresh() {
    }
    
    public void clear() {
        Arrays.fill(svs, null);
        m = 0;
    }
    
   /**
     * Erases existing support vectors.
     * @param n
     */
    public void ensureCapacity(int n) {
        if(alphas == null || alphas.length < n) {
            alphas = new double[n];
            svs = new SparseVector[n];
        }
        m = 0;
    }
    
    public double score(SparseVector rep) {
        double score = 0;
        prepareDot(rep);
        for(int i = 0; i < m; i++)
            score += alphas[i]*dot(svs[i]);
        return score;
    }

    private double cachedSqrY1;
    private SparseVector cached;

    private double oneDot(SparseVector y1, SparseVector y2) {

        double y12 = y1.dot(y2);
        double f = y1.factor * y2.factor;
        
        switch(kernelType) {
        case LINEAR_KERNEL:
            return f*y12;
        case POLYNOMIAL_KERNEL:
            return f*pow(s * y12 + c, polyDegree);
        case GAUSSIAN_KERNEL: {
            if(true)
                throw new RuntimeException("Gaussian kernel unimplemented!");
            //return Math.exp(-gamma * (cachedSqrY1 - 2*y12 + y22));
        }
        case SIGMOID_KERNEL:
            return f*Math.tanh(s * y12 + c);
        default:
            throw new RuntimeException("Unknown kernel type!");
        }
    }
    
    public void prepareDot(SparseVector y1) {
    	//SparseVector.prepareDot(y1);
    	for(SparseVector sv = y1; sv != null; sv = sv.next)
    		sv.prepareDot();
    	cached = y1;

        //if(kernelType == GAUSSIAN_KERNEL)
        //	cachedSqrY1 = fullDot(y1, y1);
        	//cachedSqrY1 = //y1.dot(y1);
    }

    private double pow(double x, int n) {
        double out = x;
        for(int i = 1; i < n; i++)
            out *= x;
        return out;
    }
    
    public double dot(SparseVector y2) {        
        double d = 0;
        for(SparseVector sv1 = cached; sv1 != null; sv1 = sv1.next)
                for(SparseVector sv2 = y2; sv2 != null; sv2 = sv2.next)
                        d += oneDot(sv1, sv2);
        return d;
    }

    public double dot(SparseVector y1, SparseVector y2) {
    	prepareDot(y1);
    	return dot(y2);
    }

    public double sqrDiff(SparseVector y1, SparseVector y2) {        
        return sqrLength(y1) + sqrLength(y2) - 2*dot(y1, y2);
    }

    public double sqrLength(SparseVector y) {
        return dot(y, y);
    }

}
