/*
 Model.java
  
 Created on Sep 8, 2006 by Richard Johansson (richard@cs.lth.se).

 $Log: Model.java,v $
 Revision 1.4  2010-07-05 14:04:22  johansson
 Added set.

 Revision 1.3  2009/04/16 12:03:35  johansson
 Added update on index.

 Revision 1.2  2009/03/06 13:24:40  johansson
 Added sortedSqrLen.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.5  2007/02/26 08:06:11  richard
 Added clear.

 Revision 1.4  2007/02/20 09:44:07  richard
 Added prune.

 Revision 1.3  2006/09/09 16:18:39  richard
 Uses the new SparseVector.

 Revision 1.2  2006/09/08 16:03:01  richard
 Moved the operations on representations to this class.

 Revision 1.1  2006/09/08 12:01:26  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.ml_long;

import java.util.HashMap;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public abstract class Model {

    public abstract void update(double d, SparseVector rep);
    public abstract void clear();
    public abstract double score(SparseVector rep);
 
    public abstract void updateDiff(double d, SparseVector r1, SparseVector r2, boolean sorted);
    
    public abstract void prepareDot(SparseVector r1);
    public abstract double dot(SparseVector r2);
    public abstract double dot(SparseVector r1, SparseVector r2);
    public abstract double sqrLength(SparseVector y);
    public abstract double sqrDiff(SparseVector y1, SparseVector y2);

    public double modelSqrNorm() {
        throw new UnsupportedOperationException("unimplemented");
    }
    
    public void update(long index, double delta) {
        throw new UnsupportedOperationException("unimplemented");
    }
    
    public void set(long index, double value) {
        throw new UnsupportedOperationException("unimplemented");
    }
    
    public int getMostProminentFeature() {
        throw new UnsupportedOperationException("unimplemented");
    }
    
    public void prune() {
    }
    
    public static <K> void addToMap(HashMap<K, Double> m, K k, double val) {
        Double d = m.get(k);
        m.put(k, d == null? val: (d + val));
    }
}
