/*
 MulticlassDefinition.java

 Created on Sep 1, 2006 by Richard Johansson (richard@cs.lth.se).

 $Log: MulticlassDefinition.java,v $
 Revision 1.5  2011-05-26 09:28:15  johansson
 AUC support.

 Revision 1.4  2010-06-08 09:50:03  johansson
 Bug: forgot sv.clear in argmax/maxLoss loops.

 Revision 1.3  2010-06-07 11:29:59  johansson
 Added.

 Revision 1.2  2009/03/06 13:30:35  johansson
 Added subsetAndPairs and RemappingFE.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.18  2008/04/02 11:25:21  richard
 Abstract list encoder as parameter. New I/O.

 Revision 1.17  2008/03/11 13:28:30  richard
 Backwards compatible with 1.5.

 Revision 1.16  2007/10/23 12:54:57  richard
 Decompose.

 Revision 1.15  2007/10/12 06:38:36  richard
 Cleaned up test code.

 Revision 1.14  2007/10/08 12:31:34  richard
 Refactored again to remove atoms.

 Revision 1.13.2.1  2007/10/08 12:19:04  richard
 Refactored again to remove atoms.

 Revision 1.13  2007/10/04 12:32:46  richard
 Serial version, enclosed classes serializable.

 Revision 1.12  2007/09/28 14:47:53  richard
 Refactored to atom-based.

 Revision 1.11  2007/09/21 14:54:28  richard
 Serial version on COPY.

 Revision 1.10  2007/08/17 10:24:36  richard
 Temporary solution to allow set-valued features.

 Revision 1.9  2007/05/22 13:12:21  richard
 Removed printout.

 Revision 1.8  2007/04/05 13:34:16  richard
 Removes encoder.

 Revision 1.7  2007/02/20 13:42:08  richard
 Changed comparator.

 Revision 1.6  2006/10/11 12:05:31  richard
 Protected empty constructor.

 Revision 1.5  2006/09/28 08:08:51  richard
 Added setMaxFeatureIndex.

 Revision 1.4  2006/09/27 18:56:08  richard
 Removed debug printout.

 Revision 1.3  2006/09/27 16:39:28  richard
 Debug.

 Revision 1.2  2006/09/26 15:08:41  richard
 Uses a feature encoder.

 Revision 1.1  2006/09/22 08:05:42  richard
 Renamed.

 Revision 1.13  2006/09/22 07:19:47  richard
 Optimized numeric encoding loop.

 Revision 1.12  2006/09/19 13:29:19  richard
 Added getFeatureList and getFeatureExtractor.

 Revision 1.11  2006/09/19 08:30:50  richard
 New constructor.

 Revision 1.10  2006/09/18 15:17:17  richard
 Serializable.

 Revision 1.9  2006/09/18 08:05:54  richard
 Renamed encoding methods. Added freeze.

 Revision 1.8  2006/09/17 14:02:54  richard
 Moved FeatureTriple to top level. Cleaned up.

 Revision 1.7  2006/09/14 06:12:57  richard
 Simplified the FeatureTriple.

 Revision 1.6  2006/09/09 16:18:24  richard
 Uses the new SparseVector and GNU Trove.

 Revision 1.5  2006/09/08 18:25:37  richard
 String parameters.

 Revision 1.4  2006/09/08 18:21:16  richard
 Uses string constructor.

 Revision 1.3  2006/09/08 16:04:16  richard
 Feature triples instead of strings.

 Revision 1.2  2006/09/08 11:54:39  richard
 Quadratic kernel experiments.

 Revision 1.1  2006/09/07 17:54:18  richard
 Added files.


 */
package se.lth.cs.nlp.nlputils.ml_long;

import se.lth.cs.nlp.nlputils.core.*;

import java.util.*;
import java.io.*;

import gnu.trove.*;

/**
 * Cost-insensitive multiclass classification with a finite (practical) number
 * of labels.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class MulticlassDefinition<I, V> extends ProblemDefinition<I, V> implements Serializable {    

    private static final long serialVersionUID = 3L;

    protected FeatureExtractor<I> extractor;
    protected FeatureList list;
    protected FeatureDeclaration decl;

    protected TObjectIntHashMap<V> classIDs = new TObjectIntHashMap();
    protected V[] range;
    
    protected NBestItem<V>[] nbestItems;

    protected FeatureListEncoder listEncoder;
    protected SparseVectorClassEncoder svEncoder = new SparseVectorClassEncoder();
    protected SparseVector sv1 = new SparseVector();
    
    public FeatureExtractor getFeatureExtractor() {
        return extractor;
    }

    public FeatureDeclaration getFeatureDecl() {
        return decl;
    }

    protected MulticlassDefinition() {
    }

    public void setMaxFeatureIndex(int maxIndex) {
    	System.out.println("Warning: setMaxFeatureIndex not implemented...");
    }
    
    public MulticlassDefinition(Set<V> range, FeatureDeclaration decl, FeatureExtractor<I> extractor) {
        init(range, decl, extractor, new SimpleFeatureListEncoder());
    }
    
    public MulticlassDefinition(List<Pair<I, V>> trainingSet, FeatureDeclaration decl, FeatureExtractor<I> extractor) {
        HashSet<V> range = new HashSet();
        for(Pair<?, V> ex: trainingSet)
            range.add(ex.right);
        init(range, decl, extractor, new SimpleFeatureListEncoder());
    }
    
    public MulticlassDefinition(List<Pair<I, V>> trainingSet, 
            FeatureDeclaration decl, 
            FeatureExtractor<I> extractor,
            FeatureListEncoder listEncoder) {
        HashSet<V> range = new HashSet();
        for(Pair<?, V> ex: trainingSet)
            range.add(ex.right);
        init(range, decl, extractor, listEncoder);
    }

    /*

    protected FeatureExtractor<I> extractor;
    protected FeatureList list;
    protected FeatureDeclaration decl;

    protected TObjectIntHashMap<V> classIDs = new TObjectIntHashMap();
    protected V[] range;
    
    protected NBestItem<V>[] nbestItems;

    protected FeatureListEncoder listEncoder;
    protected SparseVectorClassEncoder svEncoder = new SparseVectorClassEncoder();
    protected SparseVector sv1 = new SparseVector();

     */
    
    public static <I, V> MulticlassDefinition<I, V> remap(MulticlassDefinition<I, V> s,
    		TIntIntHashMap m) {
    	if(s.getClass() != MulticlassDefinition.class)
    		throw new IllegalArgumentException("can't handle subclasses!");
    	
    	MulticlassDefinition<I, V> out = new MulticlassDefinition();
    	out.extractor = s.extractor;
    	out.list = s.list.clone();
    	out.decl = s.decl;
    	out.classIDs = s.classIDs.clone();
    	out.range = s.range.clone();
    	out.nbestItems = s.nbestItems.clone();
    	    	
    	out.listEncoder = s.listEncoder.clone();
    	out.listEncoder.remap(m); // TODO MÅSTE MODIFIERA I STÄLLET FÖR ATT KOPIERA
    	
    	return out;
    }
    
    public int getNClasses() {
        return range.length;
    }
    
    public Set<V> range() {
        HashSet<V> out = new HashSet();
        for(V v: range)
            out.add(v);
        return out;
    }
    
    protected void init(Set<V> range, FeatureDeclaration decl, 
            FeatureExtractor<I> extractor, FeatureListEncoder listEncoder) {
        this.decl = decl;
        this.extractor = extractor;
        this.list = new FeatureList(decl);
        this.listEncoder = listEncoder;
        
        this.range = (V[]) new Object[range.size()];
        this.nbestItems = new NBestItem[range.size()];
        for(int i = 0; i < range.size(); i++)
            this.nbestItems[i] = new NBestItem();
        
        int i = 0;        
        for(V cl: range) {
            classIDs.put(cl, i + 1);
            this.range[i] = cl;
            i++;
        }        
    }
    
    public double cost(I x, V y, V ybar) {
        return y.equals(ybar)? 0: 1;
    }

    protected boolean frozen = false;
    
    public void freeze() {
        sorted.clear();
        
        // 100512
        if(frozen)
        	System.err.println("Warning: MulticlassDefinition already frozen");
        else {
        	if(listEncoder != null)
        		listEncoder.freeze();
        	if(svEncoder != null)
        		svEncoder.freeze();
        }
        // end 100512
    }

    public int encodeClass(V y) {
        int yid = classIDs.get(y);
        return yid == 0? -1: yid;
    }

    public V decodeClass(int cl) {
        return range[cl-1];
    }

    public void encodeInstance(I x, SparseVector sv) {
        extractor.extractFeatures(x, list);
        listEncoder.encode(list, sv);
        //System.out.println("in encodeInstance: x = " + x);
        //sv.sortIndices();
        //System.out.println("in encodeInstance: sv = " + sv);
    }

    public void encodeFeatures(FeatureList features, SparseVector sv) {
        listEncoder.encode(features, sv);
        //System.out.println("in encodeFeatures: fs = " + features);
        //sv.sortIndices();
        //System.out.println("in encodeFeatures: sv = " + sv);
    }
    
    public void encode(I x, V y, SparseVector sv) {
        extractor.extractFeatures(x, list);
        listEncoder.encode(list, sv);
        int cl = encodeClass(y);
        if(cl == -1)
            throw new IllegalArgumentException("The class " + y + " is unknown.");
        svEncoder.encode(sv, cl);
    }

    /* TODO should also return score. */
    public V argmax(I x, Model model, SparseVector sv, double[] result) {
    	extractor.extractFeatures(x, list);
        listEncoder.encode(list, sv1);
        int bestyid = -1;
        double bestScore = Double.NEGATIVE_INFINITY;
        for(int yid = 0; yid < range.length; yid++) {
        	sv.clear();
        	svEncoder.encode(sv1, yid + 1, sv);
            double score = model.score(sv);
            if(score > bestScore) {
                bestyid = yid;
                bestScore = score;
            }
        }
        sv.clear();
        if(bestyid == -1)
            return null;
        svEncoder.encode(sv1, bestyid + 1, sv);
        result[0] = bestScore;
        return range[bestyid];
    }

    protected static class NBestItem<V> implements Serializable {
        private static final long serialVersionUID = 0L;
        double score;
        int cl;
    }
    
    protected static class NBestComp implements Comparator<NBestItem>, Serializable {
        private static final long serialVersionUID = 0L;
        public int compare(NBestItem t1, NBestItem t2) {
            int c1 = -Double.compare(t1.score, t2.score);
            if(c1 != 0)
                return c1;
            return t1.cl - t2.cl;
        }
    }

    protected static Comparator<NBestItem> ordering = new NBestComp();

    protected TreeSet<NBestItem<V>> sorted = new TreeSet<NBestItem<V>>(ordering);

    /**
     * Returns up to n top-scoring outputs, not necessarily in order.
     * 
     */
    public int findNBest(I x, Model model, int n, V[] ybars, SparseVector[] svs, double[] scores) {
        extractor.extractFeatures(x, list);

        listEncoder.encode(list, sv1);
        
        if(n >= range.length) {
            for(int i = 0; i < range.length; i++) {
                svEncoder.encode(sv1, i + 1, svs[i]);
                scores[i] = model.score(svs[i]);
                ybars[i] = range[i];
            }
            return range.length;
        } else {

            sorted.clear();
            for(int i = 0; i < range.length; i++) {
                svEncoder.encode(sv1, i + 1, svs[0]);
                double score = model.score(svs[0]);
                nbestItems[i].score = score;
                nbestItems[i].cl = i;
                sorted.add(nbestItems[i]);
            }

            int i = 0;
            for(NBestItem<V> p: sorted) {
                scores[i] = p.score;
                ybars[i] = range[p.cl];
                svs[i].clear();
                svEncoder.encode(sv1, p.cl + 1, svs[i]);
                i++;
                if(i == n)
                    return n;
            }
        }
        return 0; // Won't end up here.
    }    
    
    public V maxLoss(I x, V y, Model model, SparseVector sv, double[] result) {
        extractor.extractFeatures(x, list);
        listEncoder.encode(list, sv1);

        int bestyid = -1;
        double maxLoss = Double.NEGATIVE_INFINITY;
        double maxLossScore = 0;
        double maxLossCost = 0;
        for(int yid = 0; yid < range.length; yid++) {
        	sv.clear();
        	svEncoder.encode(sv1, yid + 1, sv);
            double score = model.score(sv);

            //System.out.println(" - score = " + score);
            
            double cost = cost(x, y, range[yid]);
            double loss = score + Math.sqrt(cost);
            if(loss > maxLoss) {
                bestyid = yid;
                maxLoss = loss;
                maxLossScore = score;
                maxLossCost = cost;
            }
        }
        sv.clear();
        if(bestyid == -1)
            return null;
        svEncoder.encode(sv1, bestyid + 1, sv);
        
        /*double xxx = model.score(sv);
        if(Math.abs(xxx - maxLossScore) > 1e-6) {
        	System.out.println("*** ERROR: ");

        	System.out.println("xxx = " + xxx);
        	System.out.println("maxLossScore = " + maxLossScore);

        	System.out.println("bestyid = " + bestyid);
        	
        	System.exit(1);
        }*/
        
        result[0] = maxLossScore;
        result[1] = maxLossCost;
        return range[bestyid];
    }

    private HashMap<I, Decomposition> cachedDecompositions = new HashMap();
    
    private class MulticlassDecomposition implements Decomposition {

        private double cachedLogPartition = Double.NaN;
        private DecompositionPart[] parts;
        private double[] thetas;
        
        MulticlassDecomposition(I x, V y) {
            thetas = new double[range.length];
            parts = new DecompositionPart[range.length];
            for (int i = 0; i < range.length; i++) {
                SparseVector sv = new SparseVector();
                encode(x, range[i], sv);
                DecompositionPart p = new MulticlassDecompositionPart(this,
                        i, range[i].equals(y), sv);
                parts[i] = p;
            }            
        }
        
        public Iterator<DecompositionPart> iterator() {
            return CollectionUtils.iterator(parts);
        }
        
        public int size() {
            return parts.length;
        }
        
        public double getLogPartition() {
            if (Double.isNaN(cachedLogPartition)) {
                double maxTheta = Double.NEGATIVE_INFINITY;
                for (int i = 0; i < thetas.length; i++)
                    if (thetas[i] > maxTheta)
                        maxTheta = thetas[i];
                double sumExpTheta = 0;
                for (int i = 0; i < thetas.length; i++)
                    sumExpTheta += Math.exp(thetas[i] - maxTheta);
                cachedLogPartition = maxTheta + Math.log(sumExpTheta);
                if(Double.isInfinite(cachedLogPartition) || Double.isNaN(cachedLogPartition)) {
                    throw new RuntimeException("cachedLogPartition = " + cachedLogPartition);
                }
            }
            return cachedLogPartition;
        }
    }
    
    private class MulticlassDecompositionPart implements DecompositionPart {

        private MulticlassDecomposition decomp;
        
        private double[] thetas;
        private int index;
        private boolean isCorrect;
        private SparseVector sv;
        
        MulticlassDecompositionPart(MulticlassDecomposition decomp, int index,
                boolean isCorrect, SparseVector sv) {
            this.decomp = decomp;
            this.thetas = decomp.thetas;
            this.index = index;
            this.isCorrect = isCorrect;
            this.sv = sv;
        }
        
        public double getTheta() {
            return thetas[index];
        }
        
        public void setTheta(double value) {
            thetas[index] = value;
            decomp.cachedLogPartition = Double.NaN;
        }
        
        public double getCost() {
            return isCorrect? 0.0: 1.0;
        }
        
        public SparseVector encode() {
            return sv;
        }
        
        public double getMarginalProbability() {
            return Math.exp(thetas[index] - decomp.getLogPartition());
        }

    };

    public Decomposition decompose(I x, V y) {
        Decomposition decomp = cachedDecompositions.get(x);
        if (decomp == null) {
            decomp = new MulticlassDecomposition(x, y);
            cachedDecompositions.put(x, decomp);
        }
        return decomp;        
    }
    
    
    public static final FeatureExtractor<FeatureList> ID = new FeatureExtractor<FeatureList>() {
        private static final long serialVersionUID = 1L;

        public void extractFeatures(FeatureList l, FeatureList l2) {
            l2.values = l.values;
        }
    };

    public static final FeatureExtractor<Object[]> COPY = new FeatureExtractor<Object[]>() {        
        private static final long serialVersionUID = 1L;
        public void extractFeatures(Object[] arr, FeatureList l2) {
            System.arraycopy(arr, 0, l2.values, 0, arr.length);
        }
    };
    
    /* Testing. */

    public static Pair<HashSet<String>, HashSet<Pair<String, String>>> 
        parseSubsetAndPairs(String arg) {

        arg = arg.trim();

        if(!arg.startsWith("{") || !arg.endsWith("}"))
            throw new IllegalArgumentException("Illegal set format");

        arg = arg.substring(1, arg.length() - 1).trim();

        String[] ss = arg.split(",\\s*");

        HashSet<String> fs = new HashSet();
        HashSet<Pair<String, String>> ps = new HashSet();

        for(int i = 0; i < ss.length; i++) {        
            String s = ss[i];
            if(s.startsWith("(")) {
                if(i == ss.length - 1)
                    throw new IllegalArgumentException("Illegal pair format " + s);
                String s2 = ss[++i];
                if(!s2.endsWith(")"))
                    throw new IllegalArgumentException("Illegal pair format " + s2);
                s = s.substring(1);
                s2 = s2.substring(0, s2.length() - 1).trim();
                if(s.equals("") || s2.equals(""))
                    throw new IllegalArgumentException("Illegal pair format (empty)");
                ps.add(new Pair(s, s2));
            } else
                fs.add(s);
        }

        return new Pair(fs, ps);
    }

    private static final String SET_SEP = "" + MLConstants.SET_SEPARATOR;
    
    //TODO ta ut
    public static ArrayList<Pair<FeatureList, String>> readDataSet(String file, 
    		FeatureDeclaration decl) throws IOException {    	
    	BufferedReader br = Util.openFileReader(file);    	
        ArrayList<Pair<FeatureList, String>> l = new ArrayList();
        String line = br.readLine();
        while(line != null) {        	
            if(line.startsWith("decl=")) {
                decl = new FeatureDeclaration(line.substring("decl=".length()));
            } else {
                String[] t1 = line.split("\\" + MLConstants.FEATURE_SEPARATOR, 2);
                String[] t2 = t1[1].split("\\" + MLConstants.FEATURE_SEPARATOR);
                if(decl == null) {
                    decl = new FeatureDeclaration();
                    for(int i = 0; i < t2.length; i++)
                        decl.addFeature("F" + i);
                }
                FeatureList fl = new FeatureList(decl);
                for(int i = 0; i < t2.length; i++) {

                	String[] t3 = t2[i].split(SET_SEP);
                	
                	if(t3.length == 1)
                		fl.setFeature(i, t2[i]);
                	else {
                		HashSet<String> ss = new HashSet();
                		for(String t: t3)
                			ss.add(t);
                		fl.setFeature(i, ss);
                	}
                		
                }
                
                if(false) {
                	System.out.println(fl);
                	System.exit(0);
                }
                
                l.add(new Pair<FeatureList, String>(fl, t1[0]));
            }
            line = br.readLine();
        }        
        
        return l;
    }
    
    private static final class RemappingFE extends FeatureExtractor<FeatureList> {
        private static final long serialVersionUID = 1L;
        private int[] mapping;
        RemappingFE(FeatureDeclaration fullDecl, FeatureDeclaration subset) {
            mapping = new int[fullDecl.size()];
            for(int i = 0; i < mapping.length; i++) {
                String fname = fullDecl.getFeatureName(i);
                mapping[i] = subset.getFeatureIndex(fname);
            }
        }
        
        public void extractFeatures(FeatureList l, FeatureList l2) {
            for(int i = 0; i < l.values.length; i++) {
                int ix = mapping[i];
                if(ix != MLConstants.EXCLUDED_MAGIC_NUMBER)
                    l2.values[ix] = l.values[i];
            }                
        }
    };
    
    public static void useSubset(FeatureDeclaration fulldecl,
            FeatureDeclaration subdecl,
            ArrayList<Pair<FeatureList, String>> exs) {
        RemappingFE rfe = new RemappingFE(fulldecl, subdecl);
        for(Pair<FeatureList, String> ex: exs) {
            FeatureList remapped = new FeatureList(subdecl);
            rfe.extractFeatures(ex.left, remapped);
            ex.left = remapped;
        }
    }
    
    private static ArrayList<Pair<FeatureList, String>> readDataSetWeka(String file, 
            FeatureDeclaration decl) throws IOException {
        Scanner sc = new Scanner(new FileInputStream(file));
        ArrayList<Pair<FeatureList, String>> l = new ArrayList();

        while(sc.hasNextLine() && !sc.nextLine().startsWith("@data"));
        
        while(sc.hasNextLine()) {
            String line = sc.nextLine().trim();
            if(line.length() == 0) // 1.5
                continue;
            
            String[] tokens = line.split("\\,\\s*");

            if(decl == null) {
                decl = new FeatureDeclaration();
                for(int i = 0; i < tokens.length-1; i++)
                    decl.addFeature("F" + i);
            }
            
            FeatureList fl = new FeatureList(decl);
            for(int i = 0; i < tokens.length-1; i++)
                fl.setFeature(i, tokens[i]);
            l.add(new Pair<FeatureList, String>(fl, tokens[tokens.length-1]));
        }
        
        return l;
    }

    protected static final String NULL = "<NONE>";
    
    static <I, V> void printStats(MulticlassDefinition<I, V> rep,
    		ArrayList<Pair<I, V>> ts) {
    	Histogram<V> ystats = new Histogram();
    	//double repDotSum = 0;
    	for(Pair<I, V> p: ts) {
    		ystats.add(p.right);
    		/*SparseVector sv1 = new SparseVector();
    		rep.encode(p.left, p.right, sv1);
    		sv1.prepareDot();    		
    		for(Pair<I, V> p2: ts) {
        		SparseVector sv2 = new SparseVector();
        		rep.encode(p2.left, p2.right, sv2);
        		repDotSum += sv1.dot(sv2);
    		}*/
    	}
    	
    	System.out.println("Y histogram: " + ystats.asSortedList());
    	//System.out.println("dot sum: " + repDotSum);
    }
    
    private static double[] computeAUCs(ArrayList<Pair<FeatureList, String>> ts,
    		ArrayList<String> range,
    		ArrayList<double[]> scores) {
    	int[] counters = new int[range.size()];
    	int[] rankcounters = new int[range.size()];

    	for(int i = 0; i < ts.size(); i++) {
    		Pair<FeatureList, String> p = ts.get(i);
    		int pix = range.indexOf(p.right);
    		counters[pix]++;
    	}

    	int mostfreq = CollectionUtils.maxIndex(counters, counters.length);
    	
    	for(int i = 0; i < ts.size(); i++) {
    		Pair<FeatureList, String> p = ts.get(i);
    		int pix = range.indexOf(p.right);
    		if(pix != mostfreq) {
    			double[] psc = scores.get(i);
    			for(int j = 0; j < ts.size(); j++) {
    				Pair<FeatureList, String> p2 = ts.get(j);
    				int p2ix = range.indexOf(p2.right);
    				double[] p2sc = scores.get(j);

    				if(pix != p2ix && psc[pix] > p2sc[pix])
    					rankcounters[pix]++;        		
    			}
    		}
    	}

    	double[] out = new double[range.size()];
    	for(int i = 0; i < out.length; i++) {
    		int np = counters[i];
    		int nn = ts.size() - np;
    		out[i] = (double) rankcounters[i] / (np*nn);
    	}
    	return out;
    	
    	/*
    	for(int i = 0; i < ts.size(); i++) {
    		Pair<FeatureList, String> p = ts.get(i);
    		if(p.right.equals(posVal)) {
    			npos++;
        		double pv = scores.get(i);
    	    	for(int j = 0; j < ts.size(); j++) {
    	    		Pair<FeatureList, String> n = ts.get(j);    	    		
    	    		if(!n.right.equals(posVal)) {
    	        		double nv = scores.get(j);
    	        		if(pv > nv)
    	        			nposrank++;
    	    		}
    	    	}
    		} else
    			nneg++;
    	}    	
    	return (double) nposrank / (npos * nneg);
    	*/
    }
    
    public static void main(String[] argv) {
        try {

            ArrayList<Pair<FeatureList, String>> l = readDataSet(argv[0], null);
            FeatureDeclaration decl = l.get(0).left.decl;

            //ArrayList<Pair<FeatureList, String>> l2 = new ArrayList();
            //for(int i = (int) (0.1*l.size()); i >= 0; i--)
            //    l2.add(l.remove(l.size()-1));

            //System.out.println(l.size());
            //System.out.println(l2.size());
                       
            ArrayList<Pair<String, String>> pairings = new ArrayList();

            /*            
            pairings.add(new Pair("N_PATH", "N_HEAD_POS"));
            pairings.add(new Pair("P_TARGET_LEMMA", "N_HEAD"));
            //pairings.add(new Pair("N_PATH", "P_HAS_SUBJ_IN_VC"));
            pairings.add(new Pair("N_PATH", "N_POSITION"));            
            //pairings.add(new Pair("N_PATH", "P_OPRD_PARENT_HAS_OBJ"));
            pairings.add(new Pair("N_PATH", "P_TARGET_LEMMA"));
            pairings.add(new Pair("N_RIGHT_POS", "N_HEAD"));
            //pairings.add(new Pair("N_PATH", "P_PARENT_WORD"));
            pairings.add(new Pair("N_PATH", "N_HEAD"));
            pairings.add(new Pair("N_PATH", "N_RIGHT_POS"));
            pairings.add(new Pair("N_PATH", "P_PRED_POS"));
            pairings.add(new Pair("P_REL_TO_PARENT", "P_VOICE"));
            */
            
            /*pairings.add(new Pair("OP_EXPR_POLS", "LEX_WORD_POLS"));
            pairings.add(new Pair("WORD", "NEXT_POS"));
            pairings.add(new Pair("LEX_WORD_POLS", "OP_EXPR_SEM"));
            pairings.add(new Pair("WORD", "PREV_LEMMA"));
            pairings.add(new Pair("OP_EXPR_SEM", "LEX_WORD_SEM"));*/

            //pairings.add(new Pair("WORD", "CLOSEST_OPEXPR_BEFORE"));
            //pairings.add(new Pair("WORD", "CLOSEST_OPEXPR_AFTER"));
            
            //FeatureListEncoder lenc = new SimpleFeatureListEncoder();
            FeatureListEncoder lenc = new PairingFeatureListEncoder(decl, pairings);
            
            MulticlassDefinition<FeatureList, String> m
            = new MulticlassDefinition<FeatureList, String>(l, decl, ID, lenc);

            String methodString = argv[2];
            String argString = argv[3];

            LearningAlgorithm method = new AlgorithmFactory().create(methodString, argString);

            long t0 = System.currentTimeMillis();
            Classifier<FeatureList, String> cl = method.train(m, l);
            System.out.println("Training time: " + (System.currentTimeMillis() - t0));

            Histogram<String> h = new Histogram();
            for(Pair<FeatureList, String> p: l)
            	h.add(p.right);
            String mostFreqClass = h.getMostFrequent();
            h = null;
            l = null;
            
            //sc = new Scanner(new FileInputStream(argv[1]));
            int n = 0, ncorrect = 0;            
            int nguesses = 0, ningold = 0, ncorrectGuesses = 0;
            int ncorrUnl = 0;

            long cltime = 0;

            HashMap<Object, int[]> perClassResults = new HashMap();
            HashMap<Object, int[]> perF2Results = new HashMap();
                       
            ArrayList<Pair<FeatureList, String>> l2 = argv[1].equals("NULL")? null: readDataSet(argv[1], decl);

            boolean computeAUC = true;                        
            ArrayList<double[]> scores = null;
            ArrayList<String> range = null;
            if(computeAUC) {
            	scores = new ArrayList();
            	range = new ArrayList(m.range());
            }
            
            if(l2 != null) {

            	System.out.println("Loaded test set.");
            	
            	for(Pair<FeatureList, String> p: l2) {
                n++;
                if(n % 1000 == 0) {
                    System.out.print(".");
                    System.out.flush();
                }

                Object answer;
                if(!computeAUC) {
                	t0 = System.currentTimeMillis();
                	answer = cl.classify(p.left);
                	cltime += System.currentTimeMillis() - t0;
                } else {
                	double[] clscores = new double[range.size()];
                	t0 = System.currentTimeMillis();   	
                	cl.computeScoresRestricted(p.left, range, clscores);
                	answer = range.get(CollectionUtils.maxIndex(clscores));
                	cltime += System.currentTimeMillis() - t0;
                	scores.add(clscores);
                	//System.out.println(p.right + "\t" + answer + "\t" + clscores[0]);                	
                }

                if(answer == null)
                    throw new RuntimeException("Classifier returned null");
                
                if(answer.equals(p.right))
                    ncorrect++;                

                int[] c1 = null;
                if(true /*!answer.equals(mostFreqClass)*/) {
                    c1 = perClassResults.get(answer);
                    if(c1 == null) {
                        c1 = new int[3];
                        perClassResults.put(answer, c1);
                    }
                    c1[0]++;
                    if(answer.equals(p.right))
                    	c1[2]++;
                }
                if(true /*!p.right.equals(mostFreqClass)*/) {
                    int[] c2 = perClassResults.get(p.right);
                    if(c2 == null) {
                        c2 = new int[3];
                        perClassResults.put(p.right, c2);
                    }
                    c2[1]++;
                }

                /* Temporary for holder test. */
                if(false) {
                	String f2 = (String) p.left.values[2];                
                	if(!answer.equals(NULL)) {
                		c1 = perF2Results.get(f2);
                		if(c1 == null) {
                			c1 = new int[3];
                			perF2Results.put(f2, c1);
                		}
                		c1[0]++;
                		if(answer.equals(p.right))
                			c1[2]++;
                	}
                	if(!p.right.equals(NULL)) {
                		int[] c2 = perF2Results.get(f2);
                		if(c2 == null) {
                			c2 = new int[3];
                			perF2Results.put(f2, c2);
                		}
                		c2[1]++;
                }
                /* end */
                }
                
                /* TODO! */
                if(true) {
                	if(!answer.equals(mostFreqClass) && answer.equals(p.right)) {
                		ncorrectGuesses++;
                		//c1[2]++;
                	} 
                	if(!answer.equals(mostFreqClass) && !p.right.equals(mostFreqClass))
                		ncorrUnl++;
                	if(!answer.equals(mostFreqClass))
                		nguesses++;
                	if(!p.right.equals(mostFreqClass))
                		ningold++;
                }
            }

            double acc = (double) ncorrect / n;
            double p = (double) ncorrectGuesses / nguesses;
            double r = (double) ncorrectGuesses / ningold;
            double f1 = 2*p*r/(p+r);
            double pu = (double) ncorrUnl / nguesses;
            double ru = (double) ncorrUnl / ningold;
            double f1u = 2*pu*ru/(pu+ru);

            System.out.println();           
            System.out.println(ncorrect + "/" + n + " = " + acc);
            System.out.println("Classification time: " + cltime);

            if(computeAUC) {
            	double[] aucs = computeAUCs(l2, range, scores);
            	System.out.println("AUC values: " + Arrays.toString(aucs));
            }
            
            if(true) {
                System.out.println("Precision / recall values assuming null class is " + mostFreqClass);
                System.out.println();
                System.out.println("p = " + ncorrectGuesses + "/" + nguesses + " = " + p);
                System.out.println("r = " + ncorrectGuesses + "/" + ningold + " = " + r);
                System.out.println("f1 = " + f1);

                System.out.println();
                System.out.println("pu = " + ncorrUnl + "/" + nguesses + " = " + pu);
                System.out.println("ru = " + ncorrUnl + "/" + ningold + " = " + ru);
                System.out.println("f1u = " + f1u);
            }


            if(true)
                for(Object c: perClassResults.keySet()) {
                    int[] res = perClassResults.get(c);
                    double pc = (double) res[2] / res[0];
                    double rc = (double) res[2] / res[1];
                    double fc = 2*pc*rc/(pc + rc);
                    pc = Math.round(10000.0*pc)/10000.0;
                    rc = Math.round(10000.0*rc)/10000.0;
                    fc = Math.round(10000.0*fc)/10000.0;
                    System.out.println(c + ": p = " + pc + ", r = " + rc + ", f1 = " + fc);
                }

            if(true)
                for(Object c: perF2Results.keySet()) {
                    int[] res = perF2Results.get(c);
                    double pc = (double) res[2] / res[0];
                    double rc = (double) res[2] / res[1];
                    double fc = 2*pc*rc/(pc + rc);
                    pc = Math.round(10000.0*pc)/10000.0;
                    rc = Math.round(10000.0*rc)/10000.0;
                    fc = Math.round(10000.0*fc)/10000.0;
                    System.out.println(c + ": p = " + pc + ", r = " + rc + ", f1 = " + fc);
                }
            }
            
            if(argv.length > 4) {
            	System.out.println("Writing model to " + argv[4]);
            	ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(argv[4]));
            	oos.writeObject(cl);
            	oos.close();
            }
            
        } catch(Exception e) {
            e.printStackTrace();
        }

    }

}