/*
 ProblemDefinition.java
  
 Created on Aug 30, 2006 by Richard Johansson (richard@cs.lth.se).

 $Log: ProblemDefinition.java,v $
 Revision 1.2  2010-06-07 11:29:59  johansson
 Added.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.9  2008/04/10 08:33:02  richard
 Added getModelDensity.

 Revision 1.8  2008/03/14 08:45:44  richard
 Added nbest.

 Revision 1.7  2007/10/23 12:54:14  richard
 Decompose.

 Revision 1.6  2007/09/28 14:48:44  richard
 Refactored to atom-based.

 Revision 1.5  2007/08/17 10:23:44  richard
 Added setExamples.

 Revision 1.4  2007/08/14 12:40:03  richard
 Model is protected.

 Revision 1.3  2006/09/28 14:09:46  richard
 Serial number.

 Revision 1.2  2006/09/28 08:08:51  richard
 Added setMaxFeatureIndex.

 Revision 1.1  2006/09/22 08:04:23  richard
 Renamed.

 Revision 1.8  2006/09/19 08:31:17  richard
 Returns a scoring classifier.

 Revision 1.7  2006/09/18 15:17:17  richard
 Serializable.

 Revision 1.6  2006/09/18 09:08:55  richard
 Comment.

 Revision 1.5  2006/09/18 08:05:21  richard
 Renamed encoding methods. Added freeze.

 Revision 1.4  2006/09/17 13:59:36  richard
 CreateClassifier, representClass, and representInstance.

 Revision 1.3  2006/09/09 16:17:50  richard
 Uses the new SparseVector.

 Revision 1.2  2006/09/08 11:54:39  richard
 Quadratic kernel experiments.

 Revision 1.1  2006/09/07 17:54:18  richard
 Added files.

   
 */
package se.lth.cs.nlp.nlputils.ml_long;

import java.util.*;
import java.io.*;

import se.lth.cs.nlp.nlputils.core.DoubleIntPair;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public abstract class ProblemDefinition<I, V> implements Serializable {

    private static final long serialVersionUID = 0L;
    
    public double cost(I x, V y, V ybar) {
        throw new UnsupportedOperationException("The method cost is unimplemented: " + getClass().getName());
    }
    
    public V argmax(I x, Model model, SparseVector sv, double[] result) {
        throw new UnsupportedOperationException("The method argmax is unimplemented: " + getClass().getName());
    }
    
    public int findNBest(I x, Model model, int n, V[] ybars, SparseVector[] reps, double[] scores) {
        throw new UnsupportedOperationException("The method findNBest is unimplemented: " + getClass().getName());
    }
    
    /* TODO tv� varianter i SVMstruct. */
    public V maxLoss(I x, V y, Model model, SparseVector sv, double[] result) {
        throw new UnsupportedOperationException("The method maxLoss is unimplemented: " + getClass().getName());
    }

    public void encode(I x, V y, SparseVector sv) {
        throw new UnsupportedOperationException("The method encode is unimplemented: "+ getClass().getName());
    }
    
    public void encodeInstance(I x, SparseVector sv) {
        throw new UnsupportedOperationException("The method encodeInstance is unimplemented: "+ getClass().getName());
    }

    public int encodeClass(V y) {
        throw new UnsupportedOperationException("The method encodeClass is unimplemented: "+ getClass().getName());
    }

    public V decodeClass(int cl) {
        throw new UnsupportedOperationException("The method decodeClass is unimplemented: " + getClass().getName());
    }
    
    public void freeze() {
    }

    public void setMaxFeatureIndex(int maxIndex) {
        throw new UnsupportedOperationException("The method setMaxFeatureIndex is unimplemented: " + getClass().getName());
    }

    public Decomposition decompose(I x, V y) {
        throw new UnsupportedOperationException("The method decompose(x, y) is unimplemented: " + getClass().getName());
    }

    public Object corruptInputUniform(I x, Random corruptionRand,
                                       double corruptionRate) {
        throw new UnsupportedOperationException("The method corruptInputUniform is unimplemented: " + getClass().getName());
    }

    public void decorruptInput(I x, Object backup) {
        throw new UnsupportedOperationException("The method decorruptInput is unimplemented: " + getClass().getName());
    }
    
    private void writeObject(ObjectOutputStream oos) throws IOException {
    	freeze();
    }
    
    /*private LinkedList<DoubleIntPair> maxFeaturesHack;
    public LinkedList<DoubleIntPair> getMaxFeaturesHack() {
    	return maxFeaturesHack;
    }*/
    
    /**
     * Creates a classifier that classifies by calling <code>argmax</code> on 
     * the model <code>m</code> using this specification. Calls <code>freeze</code>.
     * @param m the model.
     * @return the classifier.
     */
    public Classifier<I, V> createClassifier(final Model m) {
        freeze();
        if(m instanceof LinearModel)
            System.out.println("model.size = " + ((LinearModel) m).size());
        if(m instanceof LinearModel2)
            System.out.println("(2) model.size = " + ((LinearModel2) m).size());
        if(m instanceof LinearVectorModel) {
            System.out.println("(v) model.size = " + ((LinearVectorModel) m).size());
            //maxFeaturesHack = ((LinearVectorModel) m).getMostProminentFeatures(200);
        }
    	Classifier<I, V> out = new Classifier<I, V>() {
            private static final long serialVersionUID = 0;
            private Model model = m;
            private ProblemDefinition<I, V> spec = ProblemDefinition.this;
            private SparseVector sv = new SparseVector();
            private double[] r = new double[1];
            public V classify(I x) {
                return spec.argmax(x, model, sv, r);
            }
            public V classifyByFeatures(FeatureList fl) {
                throw new RuntimeException("Classification by feature list not generally supported by this classifier!");
            }
            public void computeScoresByFeatures(FeatureList fl, double[] scores) {
                throw new RuntimeException("Scoring by feature list not generally supported by this classifier!");
            }
            /*public int computeScores(I item, double[] scores) {
                throw new RuntimeException("Must use a restricted list of classes.");
            }*/
            public void computeScoresRestricted(I item, Collection<V> classes, double[] scores) {
                int i = 0;
                for(V ybar: classes) {
                	encode(item, ybar, sv);
                    scores[i++] = model.score(sv);
                }
            }
            public V classifyRestricted(I instance, Collection<V> classes) {
            	//throw new UnsupportedOperationException("Restricted classification not supported ("
                //       + this.getClass().getName() + ")");
            	double maxScore = Double.NEGATIVE_INFINITY;
            	V max = null;
            	
                for(V ybar: classes) {
                	try {
                		encode(instance, ybar, sv);
                		double score = model.score(sv);
                		if(score > maxScore) {
                			maxScore = score;
                			max = ybar;
                		}
                	} catch(IllegalArgumentException e) {
                		// ignore
                	}
                }
                return max;
            }
            
            public int nbest(int n, I instance, V[] values, double[] scores, SparseVector[] reps) {
                return spec.findNBest(instance, model, n, values, reps, scores);
            }
            public int nbest(int n, I instance, V[] values, double[] scores) {
                SparseVector[] reps = new SparseVector[n];
                return spec.findNBest(instance, model, n, values, reps, scores);
            }
            public double getModelDensity() {
                if(model instanceof LinearVectorModel)
                    return ((LinearVectorModel) model).getDensity(); 
                else
                    return Double.NaN;
            }
         };
         return out;
    }
    
}
