/*
 SimpleFeatureListEncoder.java
  
 Created on Oct 5, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: SimpleFeatureListEncoder.java,v $
 Revision 1.2  2010-06-07 11:29:59  johansson
 Added.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.2  2008/04/10 07:08:20  richard
 Serializable.

 Revision 1.1  2008/04/02 11:26:22  richard
 Renamed.

 Revision 1.2  2007/10/08 12:31:34  richard
 Refactored again to remove atoms.

 Revision 1.1.2.1  2007/10/08 12:19:04  richard
 Refactored again to remove atoms.

   
 */
package se.lth.cs.nlp.nlputils.ml_long;

import gnu.trove.TIntIntHashMap;

import java.util.HashSet;
import java.io.Serializable;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class SimpleFeatureListEncoder extends FeatureListEncoder 
    implements Serializable {
    
    private static final long serialVersionUID = 1L;
    
    private SymbolEncoder symbolEncoder;
    private IntPairEncoder pairEncoder = new IntPairEncoder();
    
    public SimpleFeatureListEncoder() {
    	this.symbolEncoder = new SymbolEncoder();
    }
    
    public SimpleFeatureListEncoder(SymbolEncoder symbolEncoder) {
    	this.symbolEncoder = symbolEncoder;
    }

    public void encode(FeatureList list, SparseVector sv) {
        sv.clear();
        for(int i = 0; i < list.values.length; i++)
            if(list.values[i] instanceof String)
                sv.put(encodePair(i, list.values[i]), 1.0);
            else 
                for(String s: (HashSet<String>) list.values[i])
                    sv.put(encodePair(i, s), 1.0);
    }

    private int encodePair(int i, Object s) {
        int se = symbolEncoder.encode((String) s);
        return pairEncoder.encode(i, se);
    }
    
    public void remap(TIntIntHashMap m) {    	
    	pairEncoder.remap(m);
    }

    public void freeze() {
        symbolEncoder.freeze();
        pairEncoder.freeze();
    }
    
    public SimpleFeatureListEncoder clone() {
    	SimpleFeatureListEncoder out = new SimpleFeatureListEncoder();
    	out.symbolEncoder = symbolEncoder.clone();
    	out.pairEncoder = pairEncoder.clone();

    	System.out.println("cloned");
    	
    	return out;
    }
    
}
