/*
 SparseVector.java
  
 Created on Sep 9, 2006 by Richard Johansson (richard@cs.lth.se).

 $Log: SparseVector.java,v $
 Revision 1.5  2011-02-10 11:49:54  johansson
 Simple put.

 Revision 1.4  2010-06-07 11:29:59  johansson
 Added.

 Revision 1.3  2009/04/16 12:04:02  johansson
 Added sortedSubtract.

 Revision 1.2  2009/03/06 13:23:19  johansson
 Added sortedSqrLen.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.20  2007/10/23 12:53:53  richard
 Scaled putAll.

 Revision 1.19  2007/10/19 13:49:56  richard
 Added putAll.

 Revision 1.18  2007/10/04 12:31:26  richard
 New way of updating SIZE.

 Revision 1.17  2007/05/29 08:59:59  richard
 Scale and addTail.

 Revision 1.16  2007/05/10 14:34:08  richard
 Added equals.

 Revision 1.15  2007/04/05 13:35:28  richard
 Constructor from string.

 Revision 1.14  2007/02/26 08:07:44  richard
 Added factor.

 Revision 1.13  2007/02/20 13:42:48  richard
 Added copy.

 Revision 1.12  2007/02/08 15:39:22  richard
 Serializable.

 Revision 1.11  2006/10/27 11:09:53  richard
 Added equals.

 Revision 1.10  2006/10/25 07:07:39  richard
 SparseVector as a sequence.

 Revision 1.9  2006/10/05 07:00:17  richard
 Ignores zero keys.

 Revision 1.8  2006/09/27 18:22:32  richard
 Removed debug printouts.

 Revision 1.7  2006/09/27 16:39:34  richard
 Debug.

 Revision 1.6  2006/09/26 14:47:57  richard
 Added standard vector operations to SparseVector.

 Revision 1.5  2006/09/19 13:33:12  richard
 Bug in sortIndices. Added toString.

 Revision 1.4  2006/09/17 17:52:09  richard
 Bug in sorting.

 Revision 1.3  2006/09/17 17:31:39  richard
 Index is changed when sorting.

 Revision 1.2  2006/09/17 14:01:40  richard
 Added sortIndices.

 Revision 1.1  2006/09/09 16:15:59  richard
 Added the file.

   
 */
package se.lth.cs.nlp.nlputils.ml_long;
import java.io.Serializable;

import gnu.trove.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class SparseVector implements Serializable {

    private static final long serialVersionUID = 0L;
    
    /* TODO. The vector should remember if it is sorted. */
    
    private static int SIZE = 8;

    public long[] keys = new long[SIZE];
    public double[] values = new double[SIZE];
    public int index = 0;

    public double factor = 1.0;
    
    public SparseVector next;
    
    public SparseVector() {
    }
    
    public SparseVector(String s) {
        s = s.trim();
        if(s.length() == 0)
            return;
        String[] ts = s.trim().split("\\s+");        
        for(int i = 0; i < ts.length; i++) {
            int ix = ts[i].indexOf(':');
            int k = Integer.parseInt(ts[i].substring(0, ix));
            double v = Double.parseDouble(ts[i].substring(ix + 1));
            put(k, v);
        }
    }
    
    public void clear() {
        index = 0;
    }

    public void put(long key, double value) {
        if(key == 0)
            return;
        if(index == keys.length) {
            while(keys.length >= SIZE)
                SIZE += 1;
            //if(keys.length > SIZE)
            //    throw new RuntimeException("Here: keys.length = " + keys.length + ", SIZE = " + SIZE + ", this = " + this);
            long[] newKeys = new long[SIZE];
            double[] newValues = new double[SIZE];
            System.arraycopy(keys, 0, newKeys, 0, index);
            System.arraycopy(values, 0, newValues, 0, index);
            keys = newKeys;
            values = newValues;
        }
        keys[index] = key;
        values[index] = value;
        index++;
    }

	public void put(long key) {
		put(key, 1.0);
	}    
    
    public void putAll(SparseVector other) {
        for(int i = 0; i < other.index; i++)
        	if(other.values[i] < -TOLERANCE || other.values[i] > TOLERANCE)
        		put(other.keys[i], other.values[i]);
    }
    
    public void putAll(SparseVector other, double scale) {
        for(int i = 0; i < other.index; i++)
        	if(other.values[i] < -TOLERANCE || other.values[i] > TOLERANCE)
        		put(other.keys[i], scale * other.values[i]);
    }
    
    public void sortIndices() {
    	for(int i = 0; i < index; i++) {
    		for(int j = i+1; j < index; j++)
    			if(keys[j] < keys[i]) {
    				long tmpk = keys[i];
    				double tmpv = values[i];
    				keys[i] = keys[j];
    				values[i] = values[j];
    				keys[j] = tmpk;
    				values[j] = tmpv;
    			}
    	}
        
        for(int i = 0; i < index; i++) {
            while(i < index-1 && keys[i] == keys[i+1]) {
                values[i] += values[i+1];
    			for(int j = i+1; j < index-1; j++) {
    				keys[j] = keys[j+1];
    				values[j] = values[j+1];
    			}
    			index--;
            }
    	}
        
        if(next != null)
            next.sortIndices();
    }

    public String toString() {        
        StringBuffer sb = new StringBuffer();
        sb.append("(" + factor + ") ");
        for(int i = 0; i < index; i++) {
            sb.append(keys[i] + ":" + values[i]);
            if(i < index - 1)
                sb.append(" ");
        }
        return sb.toString() + (next == null? "": " ### " + next.toString());
    }
    
    public boolean equals(Object o) {
        /* Egentligen borde det kunna vara en godtycklig ordning, vi fuskar. */
        
        SparseVector sv = (SparseVector) o;
        if(index != sv.index)
            return false;
        for(int i = 0; i < index; i++)
            if(keys[i] != sv.keys[i] || values[i] != sv.values[i])
                return false;
        if(next == null)
            return true;
        return next.equals(sv.next);
    }
    
    /* TODO jag att de b�r vara ickestatiska. */
    
    private /*static*/ TLongDoubleHashMap dotMap;
    public /*static*/ void prepareDot(/*SparseVector y1*/) {
        dotMap = new TLongDoubleHashMap();
        for(int i = 0; i < /*y1.*/index; i++) {
            addToMap(dotMap, /*y1.*/keys[i], /*y1.*/values[i]);
        }
    }

    private static void addToMap(TLongDoubleHashMap m, long k, double v) {
        if(!m.adjustValue(k, v))
            m.put(k, v);
    }

    public /*static*/ double dot(SparseVector y2) {
        double s12 = 0;
        for(int i = 0; i < y2.index; i++) {
            double d2 = dotMap.get(y2.keys[i]);
            s12 += y2.values[i] * d2;
        }
        return s12;
    }
    
    public static double dot(SparseVector y1, SparseVector y2) {
        double s12 = 0;
        TLongDoubleHashMap m = new TLongDoubleHashMap();
        for(int i = 0; i < y1.index; i++)
            addToMap(m, y1.keys[i], y1.values[i]);
        for(int i = 0; i < y2.index; i++)
            s12 += y2.values[i] * m.get(y2.keys[i]);
        return s12;
    }
    
    public static double sqrDiff(SparseVector y1, SparseVector y2) {
        double s1 = 0, s2 = 0, s12 = 0;
        TLongDoubleHashMap m = new TLongDoubleHashMap();
        for(int i = 0; i < y1.index; i++) {
            s1 += y1.values[i] * y1.values[i];
            addToMap(m, y1.keys[i], y1.values[i]);
        }
        for(int i = 0; i < y2.index; i++) {
            s2 += y2.values[i] * y2.values[i];
            s12 += y2.values[i] * m.get(y2.keys[i]);
        }
        return s1 + s2 - 2*s12;
    }

    public static double sqrLength(SparseVector y) {
        /*double sum = 0;
        for(int i = 0; i < y.index; i++)
                sum += y.values[i]*y.values[i];
        return sum;*/
        return dot(y, y);
    }

    public static double sortedSqrLength(SparseVector y) {
        double out = 0;
        for(int i = 0; i < y.index; i++) {
            double v = y.values[i];
            out += v*v;
        }
        return out;
    }

    private static final double TOLERANCE = 1e-15;
    
    public static void sortedSubtract(SparseVector result, 
    		SparseVector v1, SparseVector v2) {
    	int i1 = 0, i2 = 0;
    	while(i1 < v1.index || i2 < v2.index) {
    		long k1 = i1 < v1.index? v1.keys[i1]: Long.MAX_VALUE;
    		long k2 = i2 < v2.index? v2.keys[i2]: Long.MAX_VALUE;
    		
    		if(k1 == k2) {
    			double v = v1.values[i1] - v2.values[i2];
    			if(v > TOLERANCE || v < -TOLERANCE)
    				result.put(k1, v);
    			i1++;
    			i2++;
    		} else if(k1 < k2) {
    			result.put(k1, v1.values[i1]);
    			i1++;
    		} else {
    			result.put(k2, -v2.values[i2]);
    			i2++;
    		}
    	}
    }
    
    public void print() {
        System.out.println(/*"index = " + index + ", data = " + */this);
    }
    
    public void trim() {
    	if(index == keys.length)
    		return;
    	long[] is = new long[index];
    	double[] vs = new double[index];
    	System.arraycopy(keys, 0, is, 0, index);
    	System.arraycopy(values, 0, vs, 0, index);
    	keys = is;
    	values = vs;
    }
    
    public SparseVector copy() {
        SparseVector v = this;
        SparseVector p = null;
        SparseVector out = null;
        SparseVector f = null;
        while(v != null) {
            p = out;
            out = new SparseVector();
            out.index = v.index;
            out.factor = v.factor;
            out.keys = new long[v.keys.length];
            System.arraycopy(v.keys, 0, out.keys, 0, v.keys.length);
            out.values = new double[v.values.length];
            System.arraycopy(v.values, 0, out.values, 0, v.values.length);
            if(p != null)
                p.next = out;
            else
                f = out;
            v = v.next;
        }
        return f;
    }
    
    public boolean equals(SparseVector sv) {
        if(sv == null)
            return false;
        if(index != sv.index)
            return false;
        for(int i = 0; i < index; i++)
            if(keys[i] != sv.keys[i]
               || values[i] != sv.values[i])
                return false;
        if(next != null)
            return next.equals(sv.next);
        else if(sv.next != null)
            return false;
        return true;
    }
    
    public void scale(double alpha) {
        factor *= alpha;
        if(next != null)
            next.scale(alpha);
    }
    
    public void addTail(SparseVector sv) {
        SparseVector prev = this;
        while(prev.next != null)
            prev = prev.next;
        prev.next = sv;
    }

    public void addToArray(double[] arr) {   
        for(int i = 0; i < index; i++)
            arr[(int) keys[i]] = values[i];
    }

    

}