/*
 SparseVectorClassEncoder.java
  
 Created on Oct 5, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: SparseVectorClassEncoder.java,v $
 Revision 1.2  2010-06-07 11:29:59  johansson
 Added.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.3  2008/04/22 11:01:24  richard
 Does not clear second sv.

 Revision 1.2  2007/10/08 12:31:34  richard
 Refactored again to remove atoms.

 Revision 1.1.2.1  2007/10/08 12:19:04  richard
 Refactored again to remove atoms.

   
 */
package se.lth.cs.nlp.nlputils.ml_long;

import java.io.FileOutputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;

import se.lth.cs.nlp.nlputils.core.Pair;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class SparseVectorClassEncoder implements Serializable {

    private static final long serialVersionUID = 0L;

    private static final long NONCOMP_OFFS = 48L;

    private IntPairEncoder pairEncoder;
    
    public SparseVectorClassEncoder(boolean compress) {
	if(compress)
	    pairEncoder = new IntPairEncoder();
    }

    public SparseVectorClassEncoder() {
	this(true);
    }

    public void encode(SparseVector sv, int cl) {
        //throw new RuntimeException("Unimplemented!");
	if(pairEncoder == null) {
	    for(int i = 0; i < sv.index; i++) {
		long k = sv.keys[i] & ((1L << NONCOMP_OFFS)-1L);
		k |= (long) cl << NONCOMP_OFFS;
		sv.keys[i] = k;
	    }
	} else
	    for(int i = 0; i < sv.index; i++)
		sv.keys[i] = pairEncoder.encode(cl, (int) sv.keys[i]);
    }
    
    public void encode(SparseVector from, int cl, SparseVector to) {
    	if(from == null) {
    		System.out.println("from = null");
    		System.exit(1);
    	}
    	if(to == null) {
    		System.out.println("to = null");
    		System.exit(1);
    	}    		
    	
    	//throw new RuntimeException("Unimplemented!");
	if(pairEncoder == null) {
	    for(int i = 0; i < from.index; i++) {
		long k = from.keys[i] & ((1L << NONCOMP_OFFS)-1L);
		k |= (long) cl << NONCOMP_OFFS;
		to.put(k, from.values[i]);
	    }
	} else
	    for(int i = 0; i < from.index; i++)
		to.put(pairEncoder.encode(cl, (int) from.keys[i]), from.values[i]);
    }
    
    public void freeze() {
	if(pairEncoder != null)
	    pairEncoder.freeze();
    }
    
    public int size() {
	if(pairEncoder != null)
	    return pairEncoder.size();
	return -1;
    }
    
}

