/*
 SymbolEncoder.java
  
 Created on Oct 5, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: SymbolEncoder.java,v $
 Revision 1.3  2010-06-07 11:29:59  johansson
 Added.

 Revision 1.2  2009/04/16 12:04:43  johansson
 Added toString.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.5  2008/03/11 08:34:26  richard
 Added inverse.

 Revision 1.4  2007/10/19 13:50:13  richard
 Added decode.

 Revision 1.3  2007/10/08 12:31:34  richard
 Refactored again to remove atoms.

 Revision 1.2.2.1  2007/10/08 12:19:04  richard
 Refactored again to remove atoms.

   
 */
package se.lth.cs.nlp.nlputils.ml_long;

import se.lth.cs.nlp.nlputils.core.*;

import gnu.trove.*;
import java.io.Serializable;
import java.util.ArrayList;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class SymbolEncoder implements Serializable {

    private static final long serialVersionUID = 0L;
    
    private TObjectIntHashMap<String> encoded = new TObjectIntHashMap<String>();
    //private ClosedObjectIntMap<String> encoded = new ClosedObjectIntMap(50000);

    private boolean frozen = false;
    private IntRef ir;
    
    public void freeze() {
        frozen = true;
    }

    public SymbolEncoder(IntRef ir) {
        this.ir = ir;
    }
    
    public SymbolEncoder() {
        this(new IntRef());
    }
    
    public SymbolEncoder clone() {
    	IntRef ir = this.ir.clone();
    	SymbolEncoder out = new SymbolEncoder(ir);
    	out.frozen = frozen;
    	out.encoded = encoded; //encoded.clone();

    	System.out.println("symbolencoder.clone, size = " + out.size());
    	
    	return out;
    }
    
    public int encode(String symbol) {
        // TODO upgrade trove, use adjustOrPut
        int out = encoded.get(symbol);
        if(out == 0) {
            if(frozen)
                return 0;
            ir.increase();
            out = ir.get();
            encoded.put(symbol, out);
        }
        return out;
    }

    /**
     * Linear-time.
     * @param encoded
     * @return
     */
    public String decode(int e) {
        if(e == 0)
            return null;
        for(TObjectIntIterator<String> it = encoded.iterator(); it.hasNext();) {
            it.advance();
            if(it.value() == e)
                return it.key();
        }
        return null;
    }
    
    public ArrayList<String> inverse() {
        ArrayList<String> out = new ArrayList(ir.get() + 1);
        for(int i = 0; i < ir.get() + 1; i++)
            out.add(null);
        for(TObjectIntIterator<String> it = encoded.iterator(); it.hasNext();) {
            it.advance();
            out.set(it.value(), it.key());
        }
        return out;
    }
    
    public String toString() {
        StringBuilder sb = new StringBuilder("{ ");
        for(TObjectIntIterator<String> it = encoded.iterator(); it.hasNext();) {
            it.advance();
            sb.append("\"" + it.key() + "\"->" + it.value()); 
            if(it.hasNext())
                sb.append(", ");
        }
        sb.append(" }");
        return sb.toString();
    }
    
    public int size() {
        return encoded.size();
    }
    
}
