/*
 BatchMEC.java
 
 Created on May 10, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: BatchMEC.java,v $
 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.10  2007/09/28 15:08:59  richard
 Refactored to atom-based.

 Revision 1.9  2007/06/14 12:14:29  richard
 Update.

 Revision 1.8  2007/06/01 13:13:48  richard
 Temporary.

 Revision 1.7  2007/06/01 11:01:45  richard
 Optimized slightly.

 Revision 1.6  2007/05/31 15:10:30  richard
 Added new cost functions.

 Revision 1.5  2007/05/30 09:05:59  richard
 Printouts.

 Revision 1.4  2007/05/29 10:06:54  richard
 Uses maxLoss instead of argmax.

 Revision 1.3  2007/05/29 08:59:14  richard
 Prototype implementation.

 Revision 1.2  2007/05/22 13:12:45  richard
 K, alpha

 Revision 1.1  2007/05/16 14:45:06  richard
 Added.

 
 */
package se.lth.cs.nlp.nlputils.ml_long.algorithms;

import gnu.trove.*;

import java.util.*;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.ml_long.*;
import se.lth.cs.nlp.nlputils.math.*;
import se.lth.cs.nlp.nlputils.optimization.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class BatchMEC<I, V> extends
        LearningAlgorithm<ProblemDefinition<I, V>, I, V> {

    static final int N_INCREMENT = 100;

    private double gamma = 1.0;
    
    public BatchMEC(String[] argv, int offset) {
        
        BLAS.setNative(true);
        
        for(int i = offset; i < argv.length; i++) {
            if(argv[i].equals("gamma"))
                gamma = Double.parseDouble(argv[++i]);
            else
                throw new IllegalArgumentException("Unknown parameter: " + argv[i]);
        }
    }
    
    /* Example vectors. */
    //private SparseVector[] svs;
    private ArrayList<SparseVector> svs;

    /* Example structure. First is the correct vector.
     * examples[i][j] is the index in svs.
     */
    //private int[][] examples;
    private ArrayList<TIntArrayList> examples;

    /* Costs per vector per example. First zero is never used. */
    //private double[][] costs;
    private ArrayList<TDoubleArrayList> costs;
    
    /* index in alphas -> index in svs */
    private TIntArrayList aindices;

    /* index in alphas -> positive index in svs */
    private TIntArrayList apindices;
    
    private TDoubleArrayList maxCosts;
    
    private DVector alphas;
    
    /* Dot product differences.
     * row: index of sv
     * col: index of variable
     */
    private DMatrix K;
    
    private DVector exps;
    private TDoubleArrayList denoms, nums;
   
    /* Example -> index in examples. */
    private TObjectIntHashMap exIndex;

    private SparseVector rep = new SparseVector();
    private double[] result = new double[2];
    
    public Classifier<I, V> train(ProblemDefinition<I, V> def, List<Pair<I, V>> trainingSet) {

        exIndex = new TObjectIntHashMap();
        examples = new ArrayList();
        aindices = new TIntArrayList();
        apindices = new TIntArrayList();
        maxCosts = new TDoubleArrayList();
        svs = new ArrayList();
        costs = new ArrayList();
        
        exps = new DVector(0);
        denoms = new TDoubleArrayList();
        nums = new TDoubleArrayList();
        
        K = new DMatrix(0, 0);
        alphas = new DVector(0);
        
        Model model = new LinearModel();
        boolean foundNew;
            
        do {
            foundNew = false;
            int countNew = 0;

            int count = 0, countCorrect = 0;            
            
            for(Pair<I, V> p: trainingSet) {
                
                //Pair<V, SparseVector> phat = def.argmax(p.left, model);
                V yhat = def.maxLoss(p.left, p.right, model, rep, result);

                /* TODO */
                Pair<V, SparseVector> phat = new Pair(yhat, rep.copy());
                
                double c = result[1];

                count++;
                
                if (c > 0) {

                    int index = exIndex.get(p) - 1;
                    if(index == -1)
                        index = addExample(def, p, model);
                    
                    if(!hasNegative(index, phat)) {
                        System.out.print(".");
                        System.out.flush();

                        addNegative(index, phat, c, model);
                        
                        countNew++;
                        foundNew = true;

                        if (countNew == N_INCREMENT) {
                            model = null;
                            model = createNewModel();
                            countNew = 0;
                        }
                    } else
                        System.out.print("-");
                } else {
                    countCorrect++;
                    System.out.print("+");
                }
                System.out.flush();
            }
            if (countNew > 0) {
                model = null;
                model = createNewModel();
            }
            System.out.println(" [" + K.m + "x" + K.n + "] (" + countCorrect + "/" + count + ")");            
            
        } while (foundNew);

        return def.createClassifier(model);
    }

    private int addExample(ProblemDefinition<I, V> def, Pair<I, V> p, Model model) {
        TIntArrayList ex = new TIntArrayList(2);
        ex.add(svs.size());

        SparseVector sv = new SparseVector();
        
        def.encode(p.left, p.right, sv);        

        svs.add(sv);
        
        int index = examples.size();
        examples.add(ex);

        TDoubleArrayList cs = new TDoubleArrayList(2);
        cs.add(0);
        costs.add(cs);

        exIndex.put(p, index + 1);

        addRow(model, sv);

        denoms.add(0);
        nums.add(0);
        maxCosts.add(0);
        
        return index;
    }
    
    private void addRow(Model model, SparseVector sv) {        
        K.addRows(1);        
        model.prepareDot(sv);
        
        int prevPos = -1;
        double prevPDot = 0;
        for(int j = 0; j < alphas.length; j++) {            
            double kij;
            int ap = apindices.get(j);
            if(ap != prevPos) {
                prevPos = ap;
                prevPDot = model.dot(svs.get(ap));
            }
            kij = prevPDot;
            kij -= model.dot(svs.get(aindices.get(j)));            
            K.set(K.m - 1, j, kij);            
        }
        
        exps.addElements(1);
    }
    
    private void addColumn(Model model, SparseVector pos, SparseVector neg) {
        K.addCols(1);
        model.prepareDot(pos);
        for(int i = 0; i < K.m; i++)
            K.set(i, K.n - 1, model.dot(svs.get(i)));
        model.prepareDot(neg);
        for(int i = 0; i < K.m; i++)
            K.set(i, K.n - 1, K.get(i, K.n - 1) - model.dot(svs.get(i)));       
    }
    
    private boolean hasNegative(int index, Pair<V, SparseVector> phat) {
        TIntArrayList ex = examples.get(index);
        for(int i = 0; i < ex.size(); i++) {
            SparseVector sv = svs.get(ex.get(i));
            if(sv.equals(phat.right))
                return true;
        }
        return false;
    }
    
    private void addNegative(int eindex, Pair<V, SparseVector> phat, double cost, Model model) {
        TIntArrayList ex = examples.get(eindex);
        int sindex = svs.size();
        ex.add(sindex);
        aindices.add(sindex);
        apindices.add(ex.get(0));
        costs.get(eindex).add(cost);
        svs.add(phat.right);
        
        addRow(model, phat.right);
        addColumn(model, svs.get(ex.get(0)), phat.right);
        
        alphas.addElements(1);
        alphas.set(alphas.length - 1, 0);
        if(cost > maxCosts.get(eindex))
            maxCosts.set(eindex, cost);
    }

    //private UnconstrainedOptimizer optimizer = new LBFGS(200);
    private UnconstrainedOptimizer optimizer = new LBFGSGlue(200, 1e-7);
    //private UnconstrainedOptimizer optimizer = new GradientDescent();
    //private UnconstrainedOptimizer optimizer = new GradientDescent2();
    //private UnconstrainedOptimizer optimizer = new PolakRibiere();
    
    private Model createNewModel() {

        Model out = new LinearModel();

        //CostSum cf = new CostSum();
        //LogCostP1Sum cf = new LogCostP1Sum();
        LogPost cf = new LogPost();
        
        //Arrays.fill(alphas.data, 0);
        
        //alphas.set(0, 7.20);
        //alphas.set(1, 4.9);
        //alphas.set(2, 4.0);
        
        //System.out.println("Optimizing: K = \n" + K);
        //System.out.println("svs = " + svs);
        //System.out.println("examples = " + examples);
        //System.out.println("costs = " + costs);
        
        double obj = optimizer.minimize(cf, alphas.data, alphas.length);

        //cf.compute(alphas.data);

        //System.out.println("expected cost = " + (obj - cf.prior));
        //System.out.println("prior = " + cf.prior);
        //System.out.println("alphas = " + alphas);
        
        //System.exit(0);
        
        for (int i = 0; i < alphas.length; i++) {
            out.update(alphas.get(i), svs.get(apindices.get(i)));
            out.update(-alphas.get(i), svs.get(aindices.get(i)));
        }

        return out;
    }

    private class CostSum extends DiffFunction {

        private double prior;
        
        public double compute(double[] x) {
            double costSum = 0;

            //exps.scale(0);
            alphas.data = x;
            
            /*System.out.println(exps.length);
            System.out.println(K);
            System.out.println(alphas.length);*/
            
            exps.addAX(K, alphas, 1.0, 0.0);
            
            for(int i = 0; i < examples.size(); i++) {
                TIntArrayList ex = examples.get(i);
                double expSum = 0;
                double expSumC = 0;
                for(int j = 0; j < ex.size(); j++) {
                    double exp = Math.exp(exps.get(ex.get(j)));
                    //double exp = Math.expm1(exps.get(ex.get(j))) + 1;
                    exps.set(ex.get(j), exp);
                    expSum += exp;
                    if(j > 0)
                        expSumC += exp * costs.get(i).get(j);
                }
                nums.set(i, expSumC);
                denoms.set(i, expSum);
                costSum += expSumC / expSum;
            }
            
            double psum = 0;
            for(int k = 0; k < alphas.length; k++) {
                double a = alphas.get(k);
                psum += gamma * a * a;
            }
            prior = Math.expm1(psum) + 1;

            return costSum + prior;
        }

        public void computeGradient(double[] x, double[] g) {            

            double gsum = 0;
            
            for(int l = 0; l < alphas.length; l++) {
                double gtotal = 0.0;                
                for(int i = 0; i < examples.size(); i++) {
                    TIntArrayList ex = examples.get(i);
                    double gexpSum = 0;
                    double gexpSumC = 0;
                    for(int j = 0; j < ex.size(); j++) {
                        double exp = exps.get(ex.get(j));
                        double kijl = K.get(ex.get(j), l);
                        gexpSum += exp * kijl;
                        if(j > 0)
                            gexpSumC += exp * kijl * costs.get(i).get(j);
                    }
                    double den = denoms.get(i);
                    gtotal += (gexpSumC - nums.get(i)*gexpSum / den) / den;
                }
                g[l] = gtotal + 2*gamma*alphas.get(l)*prior;
                gsum += g[l]*g[l];
            }
            
            System.out.println("Gradient norm: " + gsum);
            
        }

        
    }

    
    private class LogCostP1Sum extends DiffFunction {

        private double prior;
        
        public double compute(double[] x) {
            double costSum = 0;

            //exps.scale(0);
            alphas.data = x;
            
            /*System.out.println(exps.length);
            System.out.println(K);
            System.out.println(alphas.length);*/
            
            exps.addAX(K, alphas, 1.0, 0.0);
            
            for(int i = 0; i < examples.size(); i++) {
                TIntArrayList ex = examples.get(i);
                double expSum = 0;
                double expSumC = 0;
                for(int j = 0; j < ex.size(); j++) {
                    double exp = Math.exp(exps.get(ex.get(j)));
                    //double exp = Math.expm1(exps.get(ex.get(j))) + 1;
                    exps.set(ex.get(j), exp);
                    expSum += exp;
                    if(j > 0)
                        expSumC += exp * costs.get(i).get(j);
                }
                nums.set(i, expSumC);
                denoms.set(i, expSum);
                double mc = maxCosts.get(i);
                costSum -= Math.log(mc*expSum - expSumC) - Math.log(expSum);
            }
            
            double psum = 0;
            for(int k = 0; k < alphas.length; k++) {
                double a = alphas.get(k);
                psum += gamma * a * a;
            }
            prior = psum;
            
            return costSum + prior;
        }

        public void computeGradient(double[] x, double[] g) {            

            double gsum = 0;
            
            for(int l = 0; l < alphas.length; l++) {
                double gtotal = 0.0;                
                for(int i = 0; i < examples.size(); i++) {
                    TIntArrayList ex = examples.get(i);
                    double gexpSum = 0;
                    double gexpSumC = 0;
                    for(int j = 0; j < ex.size(); j++) {
                        double exp = exps.get(ex.get(j));
                        double kijl = K.get(ex.get(j), l);
                        gexpSum += exp * kijl;
                        if(j > 0)
                            gexpSumC += exp * kijl * costs.get(i).get(j);
                    }
                    double mc = maxCosts.get(i);
                    
                    gtotal -= (mc*gexpSum - gexpSumC) / (mc*denoms.get(i) - nums.get(i));
                    gtotal += gexpSum / denoms.get(i);
                    
                }
                g[l] = gtotal + 2*gamma*alphas.get(l);
                gsum += g[l]*g[l];
            }
            
            System.out.println("Gradient norm: " + gsum);
            
        }

        
    }
    
    private class LogPost extends DiffFunction {

        private double prior;
        
        public double compute(double[] x) {

            double total = 0;
            
            //exps.scale(0);
            alphas.data = x;
            
            /*System.out.println(exps.length);
            System.out.println(K);
            System.out.println(alphas.length);*/
            
            //System.out.println("alphas = " + alphas);
            
            exps.addAX(K, alphas, 1.0, 0.0);
            
            BLAS.apply_exp(exps.data, svs.size());
            
            for(int i = 0; i < examples.size(); i++) {
                TIntArrayList ex = examples.get(i);

                double term1 = Math.log(exps.get(ex.get(0)));
                double term2 = 0;
                for(int j = 0; j < ex.size(); j++) {
                    //double exp = Math.exp(exps.get(ex.get(j)));
                    //double exp = Math.expm1(exps.get(ex.get(j))) + 1;
                    double exp = exps.get(ex.get(j));
                    //exps.set(ex.get(j), exp);
                    term2 += exp;
                }
                denoms.set(i, term2);
                
                total -= term1 - Math.log(term2);
            }
            
            double psum = 0;
            for(int k = 0; k < alphas.length; k++) {
                double a = alphas.get(k);
                psum += gamma * a * a;
            }
            prior = psum;
            
            //System.out.println("lp = " + (total + prior));
            
            //System.out.println("returns " + (total + prior));
            
            if(Double.isNaN(total + prior))
                return Double.POSITIVE_INFINITY;
            
            return total + prior;
        }

        public void computeGradient(double[] x, double[] g) {            

            double gsum = 0;
            
            for(int l = 0; l < alphas.length; l++) {
                double gtotal = 0.0;                
                for(int i = 0; i < examples.size(); i++) {
                    TIntArrayList ex = examples.get(i);

                    double term1 = K.get(ex.get(0), l);
                    double term2 = term1*exps.get(ex.get(0));
                    
                    for(int j = 1; j < ex.size(); j++) {
                        double exp = exps.get(ex.get(j));
                        double kijl = K.get(ex.get(j), l);
                        term2 += exp * kijl;
                    }

                    gtotal -= term1;
                    gtotal += term2/denoms.get(i);
                    
                    //System.out.println("gtotal = " + gtotal);
                    
                }
                g[l] = gtotal + 2*gamma*alphas.get(l);

                //System.out.println();
                
                gsum += g[l]*g[l];
            }
            
            /*if(gsum > 1e5) {
                for(int l = 0; l < alphas.length; l++)
                    g[l] = g[l] / gsum * 1e5;
            }*/
            
            System.out.println("Gradient norm: " + gsum);
            
        }

        
    }
    

    
}
