/*
 Dump.java
  
 Created on Apr 1, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: Dump.java,v $
 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.1  2008/04/02 11:21:17  richard
 Added.

   
 */

package se.lth.cs.nlp.nlputils.ml_long.algorithms;

import java.util.*;
import java.io.*;

import se.lth.cs.nlp.nlputils.ml_long.*;
import se.lth.cs.nlp.nlputils.core.*;

public class Dump<I, V> extends LearningAlgorithm<MulticlassDefinition<I, V>, I, V> {

    private String filename;

    public Dump(String[] argv, int offset) {
	if(argv[offset].equals("file")) {
	    filename = argv[offset + 1];
	}
	if(filename == null)
	    throw new IllegalArgumentException("Must specify a file");
    }

    public Classifier<I, V> train(MulticlassDefinition<I, V> spec, List<Pair<I, V>> trainingSet) {

	try {
	    Random rand = new Random(0);
            Collections.shuffle(trainingSet, rand);
            
	    PrintWriter pw = new PrintWriter(filename);

	    FeatureDeclaration decl = spec.getFeatureDecl();
	    FeatureExtractor fe = spec.getFeatureExtractor();
	    FeatureList fl = new FeatureList(decl);

            pw.println("decl=" + decl);
            
	    for(Pair<I, V> p: trainingSet) {
		fe.extractFeatures(p.left, fl);
		pw.println(p.right + (MLConstants.FEATURE_SEPARATOR + fl.toTrainingString()));
	    }

	    pw.close();

	    return null;

	} catch(Exception e) {
	    e.printStackTrace();
	    throw new RuntimeException(e.getMessage());
	}

    }

}
