package se.lth.cs.nlp.nlputils.ml_long.algorithms;

import java.util.*;

import se.lth.cs.nlp.nlputils.core.Pair;
import se.lth.cs.nlp.nlputils.ml_long.*;

public class IterSVMClusterer {

    private int k;
    private int niter1, niter2;
    
    public IterSVMClusterer(int k, int niter1, int niter2) {
        this.k = k;
        this.niter1 = niter1;
        this.niter2 = niter2;        
    }
    
    public int[] run(ArrayList<SparseVector> trainingSet) {
        
        int[] classifications = new KMeans(k, niter1).run(trainingSet).right;
        
        String[] algArgs = "quiet true LibLinearInterface -s 2 -quiet".split(" ");
        
        ArrayList<Pair<SparseVector, Integer>> ts2 = new ArrayList();
        for(int i = 0; i < trainingSet.size(); i++)
            ts2.add(new Pair(trainingSet.get(i), classifications[i]));
        
        for(int i = 0; i < niter2; i++) {
            System.out.println("*** Iteration " + (i+1) + " ***");
            LearningAlgorithm alg = new ClasswiseBinarizer(algArgs, 0);
            EncodedMulticlassDefinition def = new EncodedMulticlassDefinition();
            Classifier cl = alg.train(def, ts2);

            int nchanges = 0;
            for(int j = 0; j < classifications.length; j++) {
                int guess = (Integer) cl.classify(trainingSet.get(j));
                if(guess != classifications[j])
                    nchanges++;
                classifications[j] = guess;
                ts2.get(j).right = guess;
            }
            System.out.println("*** Number of changes: " + nchanges + " ***");
            if(nchanges == 0)
                break;
        }
        
        return classifications;
    }
    
}
