/*
 OnlineMulticlassEG.java
  
 Created on Oct 8, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: OnlineMulticlassEG.java,v $
 Revision 1.2  2010-06-07 11:30:23  johansson
 Major update.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.1  2007/10/26 13:26:46  richard
 Renamed.

 Revision 1.2  2007/10/08 12:31:34  richard
 Refactored again to remove atoms.

 Revision 1.1.2.1  2007/10/08 12:20:24  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.ml_long.algorithms;

import java.util.*;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.ml_long.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class OnlineMulticlassEG<I, V> extends LearningAlgorithm<MulticlassDefinition <I, V>, I, V> {

    private int nRounds = 1;
    private double C = 1;
    private double eta = 1;
    private boolean maxMargin = false;
    
    public OnlineMulticlassEG(String[] argv, int offset) {
        for(int i = offset; i < argv.length; i++) {
            if(argv[i].equals("nRounds"))
                nRounds = Integer.parseInt(argv[++i]);
            else if(argv[i].equals("C"))
                C = Double.parseDouble(argv[++i]);
            else if(argv[i].equals("eta"))
                eta = Double.parseDouble(argv[++i]);
            else if(argv[i].equals("maxMargin"))
                maxMargin = Boolean.parseBoolean(argv[++i]);
            else
                throw new RuntimeException("Unknown argument: " + argv[i]);
        }
    }

    public Classifier<I, V> train(MulticlassDefinition<I, V> spec, List<Pair<I, V>> trainingSet) {            

        int nc = spec.getNClasses();

        double[][] thetas = new double[trainingSet.size()][nc];
        double[][] probs = new double[trainingSet.size()][nc];
        SparseVector[][] svs = new SparseVector[trainingSet.size()][nc];        
        double[][] costs = maxMargin? new double[trainingSet.size()][nc]: null;
        
        Model model = new LinearVectorModel(1000000);
        //Model model = new QuadraticModel();

        Set<V> range = spec.range();
        
        int i = 0;
        for(Pair<I, V> ex: trainingSet) {
            int j = 0;
            for(V v: range) {
                svs[i][j] = new SparseVector();
                spec.encode(ex.left, v, svs[i][j]);
                if(v.equals(ex.right))
                    model.update(1.0, svs[i][j]);
                probs[i][j] = 1.0 / nc;
                model.update(-probs[i][j], svs[i][j]);
                if(maxMargin)
                    costs[i][j] = spec.cost(ex.left, ex.right, v);                
                j++;
            }
            i++;
        }

        Random rand = new Random(0);
        
        for(int round = 0; round < nRounds; round++) {
            for(i = 0; i < thetas.length; i++) {
                double maxtheta = Double.NEGATIVE_INFINITY;
                for(int j = 0; j < thetas[i].length; j++) {
                    double d = model.score(svs[i][j]);                    
                    
                    if(maxMargin)
                        thetas[i][j] -= eta * (-costs[i][j] - d/C);
                    else
                        thetas[i][j] -= eta * (thetas[i][j] - d/C);

                    if(thetas[i][j] > maxtheta)
                        maxtheta = thetas[i][j];
                }

                double sumExpTheta = 0;
                for(int j = 0; j < thetas[i].length; j++)
                    sumExpTheta += Math.exp(thetas[i][j] - maxtheta);

                double logSumExpTheta = maxtheta + Math.log(sumExpTheta);
                
                if(Double.isInfinite(logSumExpTheta) || Double.isNaN(logSumExpTheta)) {
                    throw new RuntimeException("logSumExpTheta = " + logSumExpTheta);
                }
                
                for(int j = 0; j < thetas[i].length; j++) {
                    double pprev = probs[i][j];
                    probs[i][j] = Math.exp(thetas[i][j] - logSumExpTheta);
                    model.update(pprev - probs[i][j], svs[i][j]);
                }
            }
            
            for(i = 0; i < thetas.length; i++) {
                int si = rand.nextInt(thetas.length);
                CollectionUtils.swap(thetas, i, si);
                CollectionUtils.swap(probs, i, si);
                CollectionUtils.swap(svs, i, si);
                if(maxMargin)
                    CollectionUtils.swap(costs, i, si);
            }
            
            System.out.print(".");
            System.out.flush();
        }
        System.out.println();
        
        return spec.createClassifier(model);
    }
    
}
