/*
 Sampler.java
  
 Created on Aug 17, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: Sampler.java,v $
 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.3  2007/11/01 10:06:55  richard
 Prunes destructively.

 Revision 1.2  2007/09/28 14:42:29  richard
 Refactored to atom-based.

 Revision 1.1  2007/08/17 10:25:03  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.ml_long.algorithms;

import se.lth.cs.nlp.nlputils.ml_long.*;
import se.lth.cs.nlp.nlputils.core.*;

import java.util.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class Sampler<I, V> extends LearningAlgorithm<ProblemDefinition<I, V>, I, V> {

    private double p;
    private LearningAlgorithm<ProblemDefinition<I, V>, I, V> alg;
    
    public Sampler(String[] argv, int offset) {
        p = Double.parseDouble(argv[offset]);
        if(p < 0 || p > 1)
            throw new IllegalArgumentException("p must be in [0, 1]");
        AlgorithmFactory<ProblemDefinition<I, V>, I, V> af = new AlgorithmFactory();        
        alg = af.create(argv[offset + 1], argv, offset + 2);        
    }
    
    public Classifier<I, V> train(ProblemDefinition<I, V> spec, List<Pair<I, V>> trainingSet) {
        Random rand = new Random(0);
        System.out.println("Sampler: pruning training set, p = " + p + ", size = " + trainingSet.size());
        ArrayList<Pair<I, V>> tmp = new ArrayList();
        for(Iterator<Pair<I, V>> i = trainingSet.iterator(); i.hasNext(); ) {
            Pair<I, V> ex = i.next();
            if(rand.nextDouble() <= p)
                tmp.add(ex);
        }
        trainingSet.clear();
        trainingSet.addAll(tmp);
        tmp = null;
        System.gc();
        System.out.println("Sampler: done, size = " + trainingSet.size());
        return alg.train(spec, trainingSet);
    }

}
