/*
 BinaryLinearClassifier.java
 
 Created on Mar 27, 2006 by Richard Johansson (richard@cs.lth.se).
 
 $Log: BinaryLinearClassifier.java,v $
 Revision 1.2  2010-06-07 11:30:38  johansson
 Major update.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.2  2008/04/02 11:18:54  richard
 Implemented.

 Revision 1.1  2007/09/28 15:00:51  richard
 Moved.

 Revision 1.6  2006/09/22 08:06:39  richard
 Renamed specification to definition.

 Revision 1.5  2006/09/18 09:08:36  richard
 Uses MulticlassDefinition instead of NumericEncoding.

 Revision 1.4  2006/09/07 17:51:19  richard
 New revision.

 Revision 1.3  2006/04/19 09:30:15  richard
 Type parameter.
 
 Revision 1.2  2006/04/11 08:41:03  richard
 Added documentation.
 
 Revision 1.1  2006/04/05 11:38:05  richard
 Added the file.
 
 
 */
package se.lth.cs.nlp.nlputils.ml_long.classifiers;

import se.lth.cs.nlp.nlputils.ml_long.MulticlassDefinition;

import se.lth.cs.nlp.nlputils.core.*;
import gnu.trove.*;

/**
 * A binary linear classifier.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class BinaryLinearClassifier<I,V> extends BinaryNumericClassifier<I,V> {
    
    private static final long serialVersionUID = 0L;
    
    private double bias;
    private double[] w;

    public BinaryLinearClassifier(MulticlassDefinition<I,V> spec, V[] classes,
            double bias, double[] w) {
        this.spec = spec;
        this.bias = bias;
        this.w = w;        
        this.classes = classes;
//        for(int i = 0; i < w.length; i++)
//            if(!Ax.equal(w[i], 0))
//                System.out.println(i + ":" + w[i]);
    }
    
    public double apply() {
        double out = bias;
        for(int i = 0; i < sv.index; i++)
            if(sv.keys[i] < w.length)
                out += sv.values[i] * w[(int) sv.keys[i]];        
        return out;
    }
    
    public Pair<BinaryLinearClassifier<I, V>, TIntIntHashMap> compress(double tol) {
    	int length = 1;
    	for(int i = 1; i < w.length; i++)
    		if(w[i] > tol || w[i] < -tol)
    			length++;
    	double[] wout = new double[length];
    	wout[0] = w[0];
    	TIntIntHashMap mout = new TIntIntHashMap();
    	mout.put(0, 0);
    	int index = 1;
    	for(int i = 1; i < w.length; i++)
    		if(w[i] > tol || w[i] < -tol) {
    			mout.put(i, index);
    			wout[index] = w[i];
    			index++;
    		}
    	
    	System.out.println("Compressed linear classifier: old size = " 
    			+ w.length + ", new size = " + wout.length);
    	
    	// TODO jag tror att vi måste MODIFIERA spec i stället för att kopiera
    	MulticlassDefinition newSpec = MulticlassDefinition.remap(this.spec, mout);
    	
    	BinaryLinearClassifier<I, V> out 
    		= new BinaryLinearClassifier(newSpec, this.classes, this.bias, wout);    	
    	return new Pair(out, mout);
    }

    public double[] copyWeights() {
        //return w.clone();
        double[] out = new double[w.length];
        System.arraycopy(w, 0, out, 0, w.length);
        return out;
    }

    public double getBias() {
        return bias;
    }
    
    public double getModelDensity() {
        int count = 0;
        for(int i = 0; i < w.length; i++)
            if(!Util.equal(w[i], 0, 1e-3))
                count += 1;
        return (double) count / w.length;
    }
    
}
