/*
 BinaryQuadraticClassifier.java
 
 Created on Mar 27, 2006 by Richard Johansson (richard@cs.lth.se).
 
 $Log: BinaryQuadraticClassifier.java,v $
 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.1  2007/09/28 14:42:11  richard
 Moved to new package.

 Revision 1.7  2006/10/11 12:04:55  richard
 Index bug.

 Revision 1.6  2006/09/22 08:06:39  richard
 Renamed specification to definition.

 Revision 1.5  2006/09/18 09:08:36  richard
 Uses MulticlassDefinition instead of NumericEncoding.

 Revision 1.4  2006/09/07 17:51:19  richard
 New revision.

 Revision 1.3  2006/04/19 09:30:15  richard
 Type parameter.
 
 Revision 1.2  2006/04/11 08:41:22  richard
 Added documentation.
 
 Revision 1.1  2006/04/05 11:39:01  richard
 Added the file.
 
 
 */
package se.lth.cs.nlp.nlputils.ml_long.classifiers;

import se.lth.cs.nlp.nlputils.ml_long.*;

/**
 * A binary quadratic classifier.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class BinaryQuadraticClassifier<I,V> extends BinaryNumericClassifier<I,V> {
    
    private static final long serialVersionUID = 0L;
    
    private double[] lC;
    private int[][] qInd;
    private double[][] qC;
    
    private double rho;
    
    public BinaryQuadraticClassifier(MulticlassDefinition<I,V> spec, V[] labels, double[] lC, 
            int[][] qInd, double[][] qC, double rho) {
        this.spec = spec;
        this.classes = labels;
        this.lC = lC;
        this.qInd = qInd;
        this.qC = qC;
        this.rho = rho;
    }
    
    protected double apply() {
        
        double out = -rho;

        sv.sortIndices();
        
        for(int i = 0; i < sv.index; i++) {
            int index1 = (int) sv.keys[i];//indices[i];

            if(sv.values[i] != 1.0)
                throw new RuntimeException("Sorry, feature values must still be 1.0.");
            
            if(index1 >= lC.length)
                break;
            
            out += lC[index1];
            
            if(qInd[index1] == null)
                continue;
            int[] qInd_i = qInd[index1];
            double[] qC_i = qC[index1];
            
            int i1 = i+1, i2 = 0;
            while(i1 < sv.index /*indices.length*/ && i2 < qInd_i.length) {
                int ind1 = (int) sv.keys[i1]; //indices[i1];
                int ind2 = qInd_i[i2];
                
                if(ind1 < ind2)
                    i1++;
                else if(ind1 > ind2)
                    i2++;
                else {
                    out += qC_i[i2];
                    i1++;
                    i2++;
                }
            }
        }
        
        return out;
    }
    
}
