/*
 PairwiseClassifier.java
 
 Created on Apr 19, 2006 by Richard Johansson (richard@cs.lth.se).
 
 $Log: PairwiseClassifier.java,v $
 Revision 1.2  2010-06-07 11:30:38  johansson
 Major update.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.2  2007/10/08 12:31:34  richard
 Refactored again to remove atoms.

 Revision 1.1.2.1  2007/10/08 12:20:14  richard
 Refactored again to remove atoms.

 Revision 1.1  2007/09/28 14:42:11  richard
 Moved to new package.

 Revision 1.5  2006/09/22 08:07:10  richard
 Renamed specification to definition.

 Revision 1.4  2006/09/18 09:08:36  richard
 Uses MulticlassDefinition instead of NumericEncoding.

 Revision 1.3  2006/09/07 17:51:19  richard
 New revision.

 Revision 1.2  2006/04/19 09:29:53  richard
 Type parameter. Added documentation.
 
 Revision 1.1  2006/04/19 09:21:47  richard
 Added the file.
 
 
 */
package se.lth.cs.nlp.nlputils.ml_long.classifiers;

import se.lth.cs.nlp.nlputils.ml_long.*;

/**
 * Multiclass classifier by applying a binary classifier for each pair of
 * classes.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class PairwiseClassifier<I,V> extends Classifier<I,V> {
    
    private static final long serialVersionUID = 0L;
    
    private BinaryNumericClassifier[] subClassifiers;
    private V[] outValues;
    private SparseVector rep = new SparseVector();
    
    private MulticlassDefinition<I, V> def;
    
    public PairwiseClassifier(MulticlassDefinition<I, V> def, BinaryNumericClassifier[] subClassifiers, V[] outValues) {
    	this.def = def;
    	this.subClassifiers = subClassifiers;
    	this.outValues = outValues;
    	for(BinaryNumericClassifier cl: subClassifiers)
            cl.setSparseVector(rep);
    }
    
    public V classify(I instance) {
        def.encodeInstance(instance, rep);
        return apply();
    }
    
    public V classifyByFeatures(FeatureList features) {
        def.encodeFeatures(features, rep);
        return apply();
    }
    
    public V apply() {        
        double[] values = new double[outValues.length];
        int[] wins = new int[outValues.length];
        
        int index = 0;
        for(int i = 0; i < outValues.length; i++)
            for(int j = i + 1; j < outValues.length; j++) {
                double value = subClassifiers[index].apply(rep);				
                
                values[i] += value;
                values[j] -= value;
                
                if(value > 0)
                    wins[i]++;
                else
                    wins[j]++;

                index++;
            }
        
        int bestWins = -1;
        double bestValue = Double.NEGATIVE_INFINITY;
        int bestIndex = -1;
        
        for(int i = 0; i < outValues.length; i++) {
            if(wins[i] > bestWins) {
                bestWins = wins[i];
                bestValue = values[i];
                bestIndex = i;
            } else if(wins[i] == bestWins && values[i] > bestValue) {
                bestValue = values[i];
                bestIndex = i;
            }
        }
        
        return outValues[bestIndex];
    }
    
}