/*
 SparseBinaryLinearClassifier.java
 
 Created on Apr 16, 2008 by Richard Johansson (richard@cs.lth.se).
 
 $Log: SparseBinaryLinearClassifier.java,v $
 Revision 1.2  2010-06-07 11:30:38  johansson
 Major update.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.1  2008/04/21 08:43:29  richard
 Added.

 
 */
package se.lth.cs.nlp.nlputils.ml_long.classifiers;

import se.lth.cs.nlp.nlputils.ml_long.MulticlassDefinition;
import gnu.trove.*;

/**
 * A binary linear classifier.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class SparseBinaryLinearClassifier<I,V> extends BinaryNumericClassifier<I,V> {
    
    private static final long serialVersionUID = 0L;
    
    private static final double TOLERANCE = 1e-9;
    
    private double bias;
    private TLongDoubleHashMap w;    
    
    public SparseBinaryLinearClassifier(MulticlassDefinition<I,V> spec, V[] classes,
            double bias, double[] w) {
        this.spec = spec;
        this.bias = bias;
        this.w = new TLongDoubleHashMap();
        for(int i = 0; i < w.length; i++)
            if(Math.abs(w[i]) > TOLERANCE)
                this.w.put(i, w[i]);

        if(true)
        	System.out.println("Size of sparse classifier: " + this.w.size());

        this.classes = classes;
    }
    
    public double apply() {
        double out = bias;
        for(int i = 0; i < sv.index; i++)
            out += sv.values[i] * w.get(sv.keys[i]);
        return out;
    }
    
}
