/*
 UnaryClassifier.java
  
 Created on Apr 1, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: UnaryClassifier.java,v $
 Revision 1.2  2010-06-07 11:30:38  johansson
 Major update.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.2  2008/04/21 08:43:55  richard
 Uses log probs.

 Revision 1.1  2008/04/02 11:18:36  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.ml_long.classifiers;

import java.util.Collection;

import se.lth.cs.nlp.nlputils.ml_long.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class UnaryClassifier<I, V> extends BinaryNumericClassifier<I, V> {

    private static final long serialVersionUID = 0L;
    
    protected V value;
    
    public UnaryClassifier(V v) {
        this.value = v;
    }
    
    public V classify(I instance) {
        return value;
    }
    
    public V classifyRestricted(I instance, Collection<V> classes) {
        if(classes.contains(value))
            return value;
        else
            return classes.iterator().next();
    }
    
    public V classifyByFeatures(FeatureList features) {
        return value;
    }
    
    public V classifyByFeaturesRestricted(FeatureList features, Collection<V> classes) {
        if(classes.contains(value))
            return value;
        else
            return classes.iterator().next();
    }
    
    public void computeScoresRestricted(I instance, Collection<V> classes, double[] scores) {
        if(classes.size() == 0)
            return;
        if(!classes.contains(value)) {
            double val = Math.log(1.0 / classes.size());
            for(int i = classes.size() - 1; i >= 0; i--)
                scores[i] = val;
        } else {
            int i = 0;
            for(V v: classes) {
                if(v.equals(value))
                    scores[i] = 0.0;
                else
                    scores[i] = Double.NEGATIVE_INFINITY;
                i++;
            }
        }
    }
    
    public void computeScoresByFeaturesRestricted(FeatureList features, Collection<V> classes, double[] scores) {
        if(classes.size() == 0)
            return;
        if(!classes.contains(value)) {
            double val = Math.log(1.0 / classes.size());
            for(int i = classes.size() - 1; i >= 0; i--)
                scores[i] = val;
        } else {
            int i = 0;
            for(V v: classes) {
                if(v.equals(value))
                    scores[i] = 0.0;
                else
                    scores[i] = Double.NEGATIVE_INFINITY;
                i++;
            }
        }
    }

    public double apply(I instance) {
    	return 0;
    }
    
	protected double apply() {
		return 0;
	}

	public double applyFeatures(FeatureList features) {
		return 0;
	}
	
}
