/**
 * 
 */
package se.lth.cs.nlp.nlputils.ml_long.featuresearch;

import java.io.IOException;
import java.util.*;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.ml_long.*;

public class FMeasureEvaluator extends Function<FeatureSet, Double> {

	private String zero;

    ////// 
    
    private String trainFileName;
    private int nCVFolds = -1;
    private String methodName;
    private String methodArgs;  

    private List<Pair<FeatureList, String>> trainingSet;
    FeatureDeclaration decl;
    
    public FMeasureEvaluator(String[] argv) {
        trainFileName = argv[0];
        zero = argv[1];
        nCVFolds = Integer.parseInt(argv[2]);
        //String restrictionFeature = argv[2];
        methodName = argv[3];
        methodArgs = argv[4]; // CollectionUtils.subArray(argv, 4, argv.length - 4);
        
        System.out.println("trainFile = " + trainFileName);
        System.out.println("zero = " + zero);
        System.out.println("nCVFolds = " + nCVFolds);
        //System.out.println("restrictionFeature = " + restrictionFeature);
        System.out.println("methodName = " + methodName);
        System.out.println("methodArgs = " + methodArgs);

        try {
            trainingSet = MulticlassDefinition.readDataSet(argv[0], null);
            decl = trainingSet.get(0).left.decl;
        } catch(IOException e) {
            throw new RuntimeException(e);
        }   

    }

    private ArrayList<Pair<SparseVector, Integer>> encodeSet(
            FeatureDeclaration subDecl,
            FeatureListEncoder lenc, SymbolEncoder clEnc) {
        ArrayList<Pair<SparseVector, Integer>> tsEnc = new ArrayList();
        int[] m = decl.getSubsetIndices(subDecl);
        int n = subDecl.size();
        
        for(Pair<FeatureList, String> p: trainingSet) {
            Object[] vs = new Object[n];
            FeatureList flMain = p.left;
            for(int i = 0; i < flMain.values.length; i++) {
                if(m[i] != MLConstants.EXCLUDED_MAGIC_NUMBER)
                    vs[m[i]] = flMain.values[i];
            }
            FeatureList flSub = new FeatureList(subDecl, vs);
            SparseVector sv = new SparseVector();
            lenc.encode(flSub, sv);
            sv.sortIndices();
            sv.trim();
            int ce = clEnc.encode(p.right);             
            tsEnc.add(new Pair(sv, ce));
        }
        return tsEnc;
    }

    private int[] one(ArrayList<Pair<SparseVector, Integer>> trs,
            ArrayList<Pair<SparseVector, Integer>> tes, int zeroEnc) {
        LearningAlgorithm method = new AlgorithmFactory().create(methodName, methodArgs);
        
        EncodedMulticlassDefinition def = new EncodedMulticlassDefinition();

        Classifier cl = method.train(def, trs);
        int ncorrect = 0, ningold = 0, nguess = 0;
        for(Pair<SparseVector, Integer> p: tes) {
            Integer guess = (Integer) cl.classify(p.left);

            if(!guess.equals(zeroEnc)) {
                nguess++;
                if(guess.equals(p.right))
                    ncorrect++;
            }
            
            if(!p.right.equals(zeroEnc))
                ningold++;

        }
        return new int[] { ncorrect, ningold, nguess };
    }
    
    private double cv(ArrayList<Pair<SparseVector, Integer>> tsEnc, int zeroEnc) {
        ArrayList<Pair<SparseVector, Integer>> trs, tes;
        int ncorrect = 0, ningold = 0, nguess = 0;
        for(int i = 0; i < nCVFolds; i++) {
            trs = new ArrayList();
            tes = new ArrayList();
            CollectionUtils.cvSplit(tsEnc, trs, tes, nCVFolds, i);
            int[] stats = one(trs, tes, zeroEnc);
            ncorrect += stats[0];
            ningold += stats[1];
            nguess += stats[2];
        }

        double pr = nguess == 0? 1.0: (double) ncorrect / nguess;
        double re = ningold == 0? 1.0: (double) ncorrect / ningold;
        
        double f1 = (pr+re ==0)? 0.0: 2*pr*re/(pr+re);
        return f1;
    }
    
    public Double apply(FeatureSet features) {
        
        FeatureDeclaration subdecl = decl.subset(features.fs);      
        FeatureListEncoder fe = new PairingFeatureListEncoder(subdecl, features.ps);

        SymbolEncoder clEnc = new SymbolEncoder();
        
        ArrayList<Pair<SparseVector, Integer>> tsEnc 
        = encodeSet(subdecl, fe, clEnc);

        int zeroEnc = clEnc.encode(zero);
        
        return cv(tsEnc, zeroEnc);
    }

    
}