
package se.lth.cs.nlp.nlputils.ml_long.featuresearch;

import java.util.*;

import java.io.*;
import java.lang.reflect.*;

import se.lth.cs.nlp.nlputils.core.*;

public class Main {

	private static Pair<ArrayList<String>, ArrayList<Pair<String, String>>> 
		parseFeatureFile(String fileName) throws IOException {
		BufferedReader br = new BufferedReader(new FileReader(fileName));

		ArrayList<String> fs = new ArrayList();
		ArrayList<Pair<String, String>> ps = new ArrayList();
		
		String line = br.readLine();
		while(line != null) {
			line = line.replaceAll("#.*", "");
			line = line.trim();
			if(!line.equals("")) {
				String[] ts = line.split("\\s+");
				if(ts.length == 1)
					fs.add(line);
				else if(ts.length == 2)
					ps.add(new Pair(ts[0], ts[1]));
				else
					throw new IllegalArgumentException("only single or paired features supported");				
			}
			line = br.readLine();
		}		
		br.close();		
		return new Pair(fs, ps);
	}
	
	public static void search(String[] argv) throws IOException {
		String allFeatureFile = argv[1];
		String startFeatureFile = argv[2];
		int sbPort = Integer.parseInt(argv[3]);
		String evalCl = argv[4];		
		if(argv.length < 5) {
			System.err.println("not enough arguments");
			System.exit(1);
		}		
		Pair<ArrayList<String>, ArrayList<Pair<String, String>>> allFeatures
		= parseFeatureFile(allFeatureFile);
		Pair<ArrayList<String>, ArrayList<Pair<String, String>>> startFeatures;
		if(startFeatureFile.equals("NULL") || startFeatureFile.equals("NONE"))
			startFeatures = new Pair(new ArrayList(), new ArrayList());
		else
			startFeatures = parseFeatureFile(startFeatureFile);		
		FeatureSearch fs = new GreedyForwardFeatureSearch(sbPort);		
		String[] argvRest = CollectionUtils.subArray(argv, 5, argv.length - 5);
		
		// TODO do I really need to create this evaluator?
		Function<FeatureSet, Double> eval = createEvaluator(evalCl, argvRest);
		fs.setStartSet(new FeatureSet(startFeatures.left, startFeatures.right));
		FeatureSet result = fs.search(allFeatures.left, eval);
		System.out.println(result);
	}
	
	private static Function<FeatureSet, Double> createEvaluator(String className,
			String[] argv) {
		try {
            Class cl = Class.forName(className);
            Constructor con = cl.getConstructor(new Class[] { String[].class });
            Function out = (Function) con.newInstance(new Object[] {argv });
            System.out.println("Created: " + out.getClass().getName());
            return out;
        } catch(Exception e) {
            throw new RuntimeException(e);
        }
	}	
	
	public static void main(String[] argv) {
		try {
			if(argv[0].equals("-search")) {
				search(argv);
			} else if(argv[0].equals("-worker")) {
				throw new UnsupportedOperationException("unimplemented");
			}
		} catch(IOException e) {
			e.printStackTrace();
		}
	}

	
}
