/*
 LibSVMInterface.java
  
 Created on Sep 17, 2006 by Richard Johansson (richard@cs.lth.se).

 $Log: LibSVMInterface.java,v $
 Revision 1.2  2010-06-07 11:28:37  johansson
 Major update.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.16  2008/03/19 15:50:35  richard
 Flushes.

 Revision 1.15  2007/09/28 14:53:31  richard
 Refactored to atom-based.

 Revision 1.14  2007/09/05 14:15:59  richard
 Dump option.

 Revision 1.13  2007/08/17 10:25:43  richard
 Uses MulticlassQuadratic.

 Revision 1.12  2007/04/05 13:32:01  richard
 Offset.

 Revision 1.11  2006/10/11 07:50:47  richard
 Waits for process output threads to finish.

 Revision 1.10  2006/10/10 14:14:07  richard
 Reads characters instead of lines from stdin.

 Revision 1.9  2006/09/22 08:10:21  richard
 Renamed specification to definition.

 Revision 1.8  2006/09/19 13:34:04  richard
 Sets the feature extractor.

 Revision 1.7  2006/09/19 08:28:42  richard
 Freezes specification before training.

 Revision 1.6  2006/09/18 09:09:42  richard
 Loads expanded binary quadratic models.

 Revision 1.5  2006/09/18 08:07:00  richard
 Implemented.

 Revision 1.4  2006/09/17 17:51:07  richard
 Added.


 */
package se.lth.cs.nlp.nlputils.ml_long.glue;

import java.util.*;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.ml_long.*;
import se.lth.cs.nlp.nlputils.ml_long.classifiers.BinaryQuadraticClassifier;

import java.io.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class LibSVMInterface<I, V> extends LearningAlgorithm<MulticlassDefinition<I, V>, I, V> {

    /* TODO expanderad kvadratisk f�ruts�tter booleska. */
    
	private static final String TRAINING_COMMAND = "svm-train";// -c 0.5 -t 0 -m 500";

    private String argString = "";
    private boolean quiet = false;
    private boolean expand = false;
    private boolean dump = false;
    
    public LibSVMInterface(String[] argv, int offset) {
        for(int i = offset; i < argv.length; i++) {
            if(argv[i].equals("-quiet"))
                quiet = true;
            else if(argv[i].equals("-expand"))
                expand = true;
            else if(argv[i].equals("-dump"))
                dump = true;
            else
                argString += " " + argv[i];
        }
    }

    private Process process;
    
    public Classifier<I, V> train(MulticlassDefinition<I, V> spec, List<Pair<I, V>> trainingSet) {
        try {
            String prefix = System.getProperty("java.io.tmpdir") 
            + System.getProperty("file.separator");
            String fileName = prefix + "LibSVMInt" + Util.uniqueSymbol();

            PrintWriter pw = new PrintWriter(new FileWriter(fileName));

            SparseVector rep = new SparseVector();
            
            int count = 0;
            for(Pair<I, V> ex: trainingSet) {
                count++;
                
                spec.encodeInstance(ex.left, rep);
                rep.sortIndices();
                
                int cl = spec.encodeClass(ex.right);

                //System.out.println("rep = " + rep + ", cl = " + cl);                
                
                pw.print(cl);

                for(int i = 0; i < rep.index; i++)
                    pw.print(" " + rep.keys[i] + ":" + rep.values[i]);

                pw.println();
            }

            pw.close();
            
            if(dump)
                return null;

            // 100512 spec.freeze();

            String cmdString = TRAINING_COMMAND + argString + " " + fileName + " " + fileName + ".model";

            System.out.println("LibSVMInterface: training, " + count + " instances.");

            if(!quiet) {
                //System.out.println(cmdString);
            }

            process = Runtime.getRuntime().exec(cmdString);

            Thread t1 = new Thread() {
                public void run() {
                    try {
                        BufferedReader in = new BufferedReader(new InputStreamReader(process.getInputStream()));
                        int c;
                        while(true) {
                            c = in.read();
                            if(c == -1)
                                break;
                            if(!quiet) {
                                System.out.print((char) c);
                                System.out.flush();
                            }
                        }
                    } catch(IOException e) {
                        e.printStackTrace();
                    }
                }
            };
            t1.start();
            
            Thread t2 = new Thread() {
                public void run() {
                    Scanner sc = new Scanner(process.getErrorStream());
                    while(sc.hasNextLine())
                        System.out.println(sc.nextLine());
                }
            };
            t2.start();

            int status = process.waitFor();

            if(status != 0)
                throw new IOException("LibSVM was unsuccessful.");

            t1.join();
            t2.join();
            
            Classifier<I, V> cl = null;

            if(expand)
                cl = loadExpanded(fileName + ".model", spec);
            else
                cl = new LibSVMClassifier<I, V>(fileName + ".model", spec);

            if(cl == null)
                throw new RuntimeException("returned null");
            
            if(true) {
                if(!new File(fileName).delete())  
                    System.err.println("Could not delete temporary file.");

                if(!new File(fileName + ".model").delete())
                    System.err.println("Could not delete temporary file.");
            }
 
            System.out.println("LibSVMInterface: training finished.");
            
            return cl;

        } catch(Exception e) {
            e.printStackTrace();
            throw new RuntimeException(e);
        }
    }

    private Classifier<I, V> loadExpanded(String fileName, MulticlassDefinition<I, V> spec) throws Exception {

        Scanner sc = new Scanner(new File(fileName));

        double gamma = 1;
        double coef0 = 0;
        
        V[] labels = null;

        double[] rhos = null;

        int nClasses = -1;
        int degree = -1;
        
        String line = sc.nextLine();
        while(!line.equals("SV")) {
            
            if(line.startsWith("svm_type"))
                Util.check(line.equals("svm_type c_svc"), "Illegal SVM type.");
            else if(line.startsWith("kernel_type"))
                Util.check(line.equals("kernel_type polynomial"), "Illegal kernel type.");
            else if(line.startsWith("degree"))
                degree = Integer.parseInt(line.substring(7));
                //Aux.check(line.equals("degree 2"), "Illegal degree.");
            else if(line.startsWith("nr_class"))
                nClasses = Integer.parseInt(line.substring(9));
            else if(line.startsWith("coef0"))
                //Aux.check(line.equals("coef0 0"), "Illegal kernel coefficient.");
                coef0 = Double.parseDouble(line.substring(5));
            else if(line.startsWith("label")) {
                String[] tokens = line.split(" ");
                //labels = new Object[tokens.length - 1];
                labels = (V[]) java.lang.reflect.Array.newInstance(spec.decodeClass(Integer.parseInt(tokens[1])).getClass(), tokens.length - 1);
                for(int i = 0; i < labels.length; i++)
                    labels[i] = spec.decodeClass(Integer.parseInt(tokens[i+1]));
            } else if(line.startsWith("gamma")) {
                gamma = Double.parseDouble(line.substring(5));
            } else if(line.startsWith("rho")) {
                String[] tokens = line.split(" ");
                rhos = new double[tokens.length - 1];
                for(int i = 0; i < rhos.length; i++)
                    rhos[i] = Double.parseDouble(tokens[i+1]);
            } else if(line.startsWith("probA")) {
                //out.probA = Double.parseDouble(line.substring(5));
            } else if(line.startsWith("probB")) {
                //out.probB = Double.parseDouble(line.substring(5));
            }
            
            line = sc.nextLine();
        }

        if(degree != 1 && degree != 2)
            throw new IOException("Illegal degree.");
        
        if(degree == 1 && nClasses == 2)
            return loadBinaryLinear(sc, rhos[0], labels);
        if(degree == 2 && nClasses == 2)
            return loadBinaryQuadratic(sc, gamma, coef0, rhos[0], labels, spec);
        if(degree == 1)
            return loadLinear(sc, nClasses);
        else 
            //return //loadQuadratic(sc, gamma, coef0, nClasses);
            return MulticlassQuadratic.loadFromLibSVM(fileName, spec);
    }

    private Classifier<I, V> loadBinaryLinear(Scanner sc, double rho, V[] labels) {
        throw new UnsupportedOperationException("Unimplemented!");
    }

    /* TODO expanderad kvadratisk f�ruts�tter booleska. */
    private Classifier<I, V> loadBinaryQuadratic(Scanner sc, double gamma, double coef0, double rho, V[] labels,
            MulticlassDefinition<I, V> spec) {

        double[] lC = null;
        double[][] qC = null;
        int[][] qInd = null;
        
        HashMap<Integer, Double> lTable = new HashMap<Integer, Double>();
        HashMap<Integer, HashMap<Integer, Double>> qTable
        = new HashMap<Integer, HashMap<Integer, Double>>();
        //HashMap<Integer, TreeMap<Integer, Double>> qTable
        //= new HashMap<Integer, TreeMap<Integer, Double>>();
        //HashMap<Integer, IntDoubleMap> qTable
        //= new HashMap<Integer, IntDoubleMap>();
        
        int maxIndex = 0;
        
        int countLines = 0;
        
//      long t0 = System.currentTimeMillis();
//      
//      long tt1_tot = 0, tt2_tot = 0;
        
        while(sc.hasNextLine()) {
            String line = sc.nextLine();
            
            int shIndex = line.indexOf('#');
            if(shIndex != -1)
                line = line.substring(0, shIndex);
            
            //line = line.trim();
            
            countLines++;
            
            String[] tokens = line.split("\\s+");
            
            double alpha = Double.parseDouble(tokens[0]);
            
            rho -= alpha * coef0 * coef0;
            
            ArrayList<Integer> indices = new ArrayList<Integer>();
            
            for(int i = 1; i < tokens.length; i++) {
                int cIndex = tokens[i].indexOf(':');
                Util.check(cIndex != -1, "Illegal format");
                
                String s1 = tokens[i].substring(0, cIndex);
                String s2 = tokens[i].substring(cIndex + 1);
                
                //Aux.check(s2.matches("0*[01](\\.0*)?"), "Illegal feature value!");
                double val = Double.parseDouble(s2);
                int v2;
                if(Math.abs(val) < 1e-6)
                    v2 = 0;
                else if(Math.abs(val - 1) < 1e-6)
                    v2 = 1;
                else
                    throw new RuntimeException("Illegal feature value");
                if(v2 == 1)
                    indices.add(Integer.parseInt(s1));
            }
            
            double l = 2 * alpha * gamma * coef0
            +  alpha * gamma*gamma; // antar boolean
            
            double q = 2 * alpha * gamma*gamma;
            
            for(int i = 0; i < indices.size(); i++) {
//              long tt0 = System.currentTimeMillis();
                
                int index = indices.get(i);
                
                if(index > maxIndex)
                    maxIndex = index;
                
                addToMap(lTable, index, l);
                
                //TreeMap<Integer, Double> qT = qTable.get(index);
                HashMap<Integer, Double> qT = qTable.get(index);
                //IntDoubleMap qT = qTable.get(index);
                if(qT == null) {
                    //qT = new TreeMap<Integer, Double>();
                    qT = new HashMap<Integer, Double>();
                    //qT = new IntDoubleMap();
                    qTable.put(index, qT);
                }
                
//              long tt1 = System.currentTimeMillis();
                
                for(int j = i+1; j < indices.size(); j++) {
                    addToMap(qT, indices.get(j), q);
                    //qT.add(indices.get(j), q);
                }
                
//              long tt2 = System.currentTimeMillis();
//              
//              tt1_tot += tt1 - tt0;
//              tt2_tot += tt2 - tt1;   
            }
            
            
        }
        
//      long t1 = System.currentTimeMillis();
        
        lC = new double[maxIndex+1];
        qC = new double[maxIndex+1][];
        qInd = new int[maxIndex+1][];
        
        for(Map.Entry<Integer, Double> e: lTable.entrySet())
            lC[e.getKey()] = e.getValue();
        
        //for(Map.Entry<Integer, TreeMap<Integer, Double>> e: qTable.entrySet()){   
        for(Map.Entry<Integer, HashMap<Integer, Double>> e: qTable.entrySet()){ 
            //for(Map.Entry<Integer, IntDoubleMap> e: qTable.entrySet()) {
            int index = e.getKey();
            
            HashMap<Integer, Double> qT = e.getValue();
            //TreeMap<Integer, Double> qT = e.getValue();
            //IntDoubleMap qT = e.getValue();
            
            qC[index] = new double[qT.size()];
            qInd[index] = new int[qT.size()];
            int n = 0;
            for(Map.Entry<Integer, Double> e2: qT.entrySet()) {
                qInd[index][n] = e2.getKey();
                qC[index][n] = e2.getValue();
                n++;
            }
            
//          Pair<int[], double[]> p = qT.getContents();
//          qInd[index] = p.left;
//          qC[index] = p.right;
//          int n = qC[index].length;
            
//          for(int p = 1; p < n; p++) {
//          
//          int tmp = qInd[index][p];
//          double dtmp = qC[index][p];
//          
//          int j = p;
//          for(; j > 0 && tmp < qInd[index][j-1]; j--) {
//          qInd[index][j] = qInd[index][j-1];
//          qC[index][j] = qC[index][j-1];
//          }
//          
//          qInd[index][j] = tmp;
//          qC[index][j] = dtmp;
//          
//          }
//          
            for(int i = 0; i < n; i++) {
                
                int min = Integer.MAX_VALUE;
                int mIndex = 0;
                for(int j = i + 1; j < n; j++) {
                    if(qInd[index][j] < min) {
                        min = qInd[index][j];
                        mIndex = j;
                    }
                }
                if(min < qInd[index][i]) {
                    double tmp = qC[index][mIndex];
                    qInd[index][mIndex] = qInd[index][i];
                    qC[index][mIndex] = qC[index][i];
                    qInd[index][i] = min;
                    qC[index][i] = tmp;
                }
                
            }           
            
        }
        
        return new BinaryQuadraticClassifier<I, V>(spec, labels, lC, qInd, qC, rho);
    }

    private Classifier<I, V> loadLinear(Scanner sc, int nClasses) {
        throw new UnsupportedOperationException("Unimplemented!");
    }

    private static void addToMap(Map map, Object key, double value) {
        Double dval = (Double) map.get(key);
        double d = dval == null? value: dval.doubleValue() + value;
        map.put(key, new Double(d));
    }
}
