/*
 SVMStructInterface.java

 Created on Sep 27, 2006 by Richard Johansson (richard@cs.lth.se).

 $Log: SVMStructInterface.java,v $
 Revision 1.4  2010-07-05 14:05:48  johansson
 Refactored.

 Revision 1.3  2010-06-08 07:45:23  johansson
 Added debug methods.

 Revision 1.2  2010-06-07 11:26:56  johansson
 Upgraded to new version of SVMStruct.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.8  2007/09/28 14:53:31  richard
 Refactored to atom-based.

 Revision 1.7  2007/04/05 13:31:39  richard
 Offset.

 Revision 1.6  2007/02/26 08:08:26  richard
 Uses KernelQModel.

 Revision 1.5  2006/10/05 07:02:40  richard
 Can use kernels.

 Revision 1.4  2006/09/28 08:08:34  richard
 Uses setMaxFeatureIndex.

 Revision 1.3  2006/09/27 18:23:13  richard
 Fixed the 1-offset bug.

 Revision 1.2  2006/09/27 16:39:54  richard
 Debug.

 Revision 1.1  2006/09/27 10:48:27  richard
 Added the file.


 */
package se.lth.cs.nlp.nlputils.ml_long.glue;

import gnu.trove.*;
import java.util.*;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.ml_long.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class SVMStructInterface<I, V> extends LearningAlgorithm<ProblemDefinition<I, V>, I, V> {

	/* TODO this class and its JNI are not thread safe. */
	
    static {
        System.loadLibrary("svm_struct_interface");
    }

    private native int callLearnMain(ProblemDefinition<I, V> problem, String[] argv);

    private int sizePsi = -1;

    /* Kernel parameters. */
    private int kernelType = 0;
    private int polyDegree;
    private double gamma;
    private double s;
    private double c;

    private static final int LINEAR_MODEL = 0;
    private static final int LINEAR_VECTOR_MODEL = 1;
    private static final int NEW_LINEAR_MODEL = 2;
    
    private int modelType = LINEAR_MODEL;    
    
    public SVMStructInterface(String[] argv, int offset) {
        ArrayList<String> a = new ArrayList<String>();
    	for(int i = offset; i < argv.length; i++) {
            if(argv[i].equals("-quiet")) {
                /* Just an alias for SVM^struct's -v 0. */
                a.add("-v");
                a.add("0");
            }
            else if(argv[i].equals("-linear"))
            	modelType = LINEAR_MODEL;
            else if(argv[i].equals("-vector"))
            	modelType = LINEAR_VECTOR_MODEL;
            else if(argv[i].equals("-newLinear"))
            	modelType = NEW_LINEAR_MODEL;
            else if(argv[i].equals("-sizePsi"))
                sizePsi = Integer.parseInt(argv[++i]);
            else if(argv[i].equals("-t")) {
                kernelType = Integer.parseInt(argv[++i]);
                a.add("-t");
                a.add(argv[i]);
            } else if(argv[i].equals("-d")) {
                polyDegree = Integer.parseInt(argv[++i]);
                a.add("-d");
                a.add(argv[i]);
            } else if(argv[i].equals("-s")) {
                s = Double.parseDouble(argv[++i]);
                a.add("-s");
                a.add(argv[i]);
            } else if(argv[i].equals("-r")) {
                c = Double.parseDouble(argv[++i]);
                a.add("-r");
                a.add(argv[i]);
            } else if(argv[i].equals("-g")) {
                gamma = Double.parseDouble(argv[++i]);
                a.add("-g");
                a.add(argv[i]);
            } else
                a.add(argv[i]);
        }
    	if(sizePsi < 1)
            throw new IllegalArgumentException("sizePsi must be given a positive value");

        /* SVM^struct wants a file name and a model name. */
        a.add("dummy_file_name");
        a.add("dummy_model_name");
        
        this.argv = a.toArray(new String[0]);
    }

    private String[] argv;
    private Object[] examples;
    private Model model;
    
    Object[] getExamples() {
        return examples;
    }

    SparseVector createSV(int n) {
        SparseVector sv = new SparseVector();
        sv.keys = new long[n];
        sv.values = new double[n];
        sv.index = n;
        return sv;
    }
    
    V maxLoss(I x, V y, ProblemDefinition<I, V> prob, Model m) {
    	SparseVector sv = new SparseVector();
    	double[] res = new double[2];
    	V out = prob.maxLoss(x, y, m, sv, res);
    	return out;
    }

    void encode(ProblemDefinition<I, V> prob, I x, V y, SparseVector sv) {
    	SparseVector svTmp = new SparseVector();    	
    	prob.encode(x, y, svTmp);
    	encodeSV(svTmp, sv);
    }
    
    void setWeight(Model model, int index, double weight) {
    	if(index > encInv.size())
    		return;
     	long decodedIndex = encInv.get(index - 1);
    	model.set(decodedIndex, weight);
    }

    private void encodeSV(SparseVector from, SparseVector to) {
    	to.clear();
    	for(int i = 0; i < from.index; i++) {
    		int encodedKey = enc.get(from.keys[i]);
    		if(encodedKey == 0) {
    			encodedKey = encInv.size() + 1;
    			encInv.add(from.keys[i]);
    			enc.put(from.keys[i], encodedKey);
    			
    			if(encInv.size() % 100000 == 0) {
    				System.out.println();
    				System.out.println("Number of features: " + encInv.size());
    			}
    			
    		}
    		to.put(encodedKey, from.values[i]);
    	}    	
    	to.sortIndices();
    }
    
    private TLongIntHashMap enc;
    private TLongArrayList  encInv;
    
    public Classifier<I, V> train(ProblemDefinition<I, V> spec, List<Pair<I, V>> trainingSet) {

    	enc = new TLongIntHashMap();
    	encInv = new TLongArrayList();
    	
        examples = trainingSet.toArray();

        /* �r detta meningsfullt om vi anv�nder k�rna? */
        spec.setMaxFeatureIndex(sizePsi);

        System.out.println(s);
        System.out.println(c);
        
        if(kernelType == 0) {
        	switch(modelType) {
        	case LINEAR_MODEL:
        		model = new LinearModel();
        		break;
        	case LINEAR_VECTOR_MODEL:
        		model = new LinearVectorModel(sizePsi + 1);
        		break;
        	case NEW_LINEAR_MODEL:
        		model = new LinearModel2(sizePsi + 1);
        		break;
        	}
        } else if(kernelType == 1 && polyDegree == 2 && s == 1 && c == 0) {
            model = null; //new KernelQModel();
        } else
            model = new KernelModel(kernelType, polyDegree, gamma, s, c);
        
        System.out.println("Model type: " + model.getClass().getName());
        
        int status = callLearnMain(spec, argv);
        if(status != 0) {
            System.err.println("SVM^struct failed.");
            return null;
        }
        // 100512 spec.freeze();

        return spec.createClassifier(model);
    }
    
    public void debugPrint(Object o) {
    }
    
    
    private Class c1, c2;
    public void debugTest(Object o1, Object o2) {
    	if(c1 == null) {
    		c1 = o1.getClass();
    		c2 = o2.getClass();
    	} else {
    		if(c1 != o1.getClass())
    			throw new RuntimeException("c1 error");
    		if(c2 != o2.getClass())
    			throw new RuntimeException("c2 error");
    	}
    }
    
}
