package se.lth.cs.nlp.nlputils.optimization;

import java.util.*;
import se.lth.cs.nlp.nlputils.core.*;

public class BestCombinations {

	private static class Item<T> implements Comparable<Item<T>> {
		int[] positions;
		int[] lengths;
		
		ArrayList<DoubleObjPair<T>> data;
		double score;
		
		Item(ArrayList<ArrayList<DoubleObjPair<T>>> l) {
			positions = new int[l.size()];
			lengths = new int[l.size()];
			for(int i = 0; i < lengths.length; i++)
				lengths[i] = l.get(i).size();

			data = new ArrayList();
			for(int i = 0; i < positions.length; i++) {
				DoubleObjPair<T> p = l.get(i).get(0);
				data.add(p);
				score += p.left;
			}			
		}
		
		Item(Item item) {
			positions = new int[item.positions.length];
			System.arraycopy(item.positions, 0, positions, 0, positions.length);
			lengths = item.lengths;
		}
		
		Item advance(int index, ArrayList<ArrayList<DoubleObjPair<T>>> l) {
			if(positions[index] + 1 >= lengths[index])
				return null;
			Item out = new Item(this);
			out.positions[index]++;

			out.data = new ArrayList();
			for(int i = 0; i < positions.length; i++) {
				DoubleObjPair<T> p = l.get(i).get(out.positions[i]);
				out.data.add(p);
				out.score += p.left;
			}
			
			return out;
		}

		public int compareTo(Item<T> o) {
			return Double.compare(o.score, score);
		}
	}
	
	/* Assumes sorted input */
	public static <T> ArrayList<ArrayList<DoubleObjPair<T>>> findBestCombinations(ArrayList<ArrayList<DoubleObjPair<T>>> l,
			int n) {
		
		ArrayList<ArrayList<DoubleObjPair<T>>> out = new ArrayList();
		if(n == 0)
			return out;
		
		PriorityQueue<Item> q = new PriorityQueue();
		q.add(new Item(l));

		double prevScore = Double.POSITIVE_INFINITY;

		while(!q.isEmpty()) {
			Item item = q.poll();
			if(prevScore == item.score)
				continue;
			prevScore = item.score;
			out.add(item.data);
			if(out.size() == n)
				break;
			else
				for(int i = 0; i < l.size(); i++) {					
					Item next = item.advance(i, l);
					if(next == null)
						continue;
					q.offer(next);
				}
		}
		
		return out;
	}

	public static void main(String[] argv) {

		ArrayList<ArrayList<DoubleObjPair<String>>> l = new ArrayList();

		ArrayList<DoubleObjPair<String>> a = new ArrayList();
		a.add(new DoubleObjPair(1.0, "a1"));
		a.add(new DoubleObjPair(0.5, "a2"));
		//a.add(new DoubleObjPair(0.1, "a3"));
		l.add(a);
		
		ArrayList<DoubleObjPair<String>> b = new ArrayList();
		b.add(new DoubleObjPair(0.95, "b1"));
		b.add(new DoubleObjPair(0.4, "b2"));
		//b.add(new DoubleObjPair(0.2, "b3"));
		l.add(b);

		System.out.println(findBestCombinations(l, 3));
		
	}
	
}
