package se.lth.cs.nlp.nlputils.pstree;

import java.io.*;
import java.util.*;

import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;

import se.lth.cs.nlp.nlputils.core.*;

public class SimplifiedTigerGraphParser {

	/* This is usable for Talbanken only, since Tiger defines attribute names in the XML.
	 */
	
	private final static class ParserCallback extends DefaultHandler 
    implements ContentHandler {
        
		private HashMap<String, Node> nodeMap;
		private NonterminalNode currentNT;
		private String rootId;
		PhraseStructureTree result;
		private HashMap<Node, String> nodeMapInv;
		
		ParserCallback(HashMap<String, Node> nodeMap) {
			this.nodeMap = nodeMap;
		}
		
        public void startElement(String namespace, String localname,  
                String type, Attributes attributes) {
        	if(type.equals("t")) {
        		String w = attributes.getValue("word");
        		String p = attributes.getValue("pos");
        		String id = attributes.getValue("id");
        		TokenNode tn = new TokenNode(w, p);
        		tn.setPosition(nodeMap.size());
        		nodeMap.put(id, tn);
        	} else if(type.equals("nt")) {
        		if(currentNT != null)
        			throw new IllegalStateException("!!");
        		String l = attributes.getValue("cat");
        		String id = attributes.getValue("id");
        		currentNT = new NonterminalNode(l);        		
        		nodeMap.put(id, currentNT);
        	} else if(type.equals("edge")) {
        		if(currentNT == null)
        			throw new IllegalStateException("!!");
        		String id = attributes.getValue("idref");
        		String l = attributes.getValue("label");
        		Node child = nodeMap.get(id);
        		if(child == null)
        			throw new IllegalStateException("!!!");
        		child.setFunction(l);
        		currentNT.addChild(child);
        	} if(type.equals("graph")) {
        		rootId = attributes.getValue("root");        		
        	}
        }

        public void endElement(String namespace, String localname,  
                String type) {
        	if(type.equals("nt")) {
        		currentNT = null;
        	} else if(type.equals("graph")) {
        		Node root = nodeMap.get(rootId);
        		result = new PhraseStructureTree(root);
        		nodeMapInv = new HashMap();
        		for(String id: nodeMap.keySet())
        			nodeMapInv.put(nodeMap.get(id), id);        		
        	}
        }
                
	}
	
	   private static org.xml.sax.XMLReader makeXMLReader() throws Exception { 
	        javax.xml.parsers.SAXParserFactory saxParserFactory   =  
	            javax.xml.parsers.SAXParserFactory.newInstance(); 

	        //saxParserFactory.setValidating(false);
	        
	        final javax.xml.parsers.SAXParser        saxParser 
	        = saxParserFactory.newSAXParser(); 

	        final org.xml.sax.XMLReader              parser    
	        = saxParser.getXMLReader(); 
	        
	        try {
	            //parser.setFeature("http://xml.org/sax/features/validation", 
	            //        false);
	            //parser.setFeature("http://apache.org/xml/features/nonvalidating/load-dtd-grammar", 
	            //        false);
	            parser.setFeature("http://apache.org/xml/features/nonvalidating/load-external-dtd",
	                    false);
	        } 
	        catch (Exception e) {
	            System.err.println("could not set parser feature");
	        }

	        return parser; 
	    }
	    
	
	public static Pair<PhraseStructureTree, HashMap<Node, String>> parse(String graphString) {
		try {
			HashMap<String, Node> nodeMap = new HashMap();
			org.xml.sax.XMLReader reader = makeXMLReader();
			ParserCallback cb = new ParserCallback(nodeMap);
			reader.setContentHandler(cb);
			reader.parse(new InputSource(new StringReader(graphString)));
			return new Pair(cb.result, cb.nodeMapInv);
		} catch(Exception e) {
			throw new RuntimeException(e);
		}
	}
	
	public static void main(String[] argv) {
		
		try {
			String graphText = Util.readFile(argv[0]);
			Pair<PhraseStructureTree, HashMap<Node, String>> p = parse(graphText);
			System.out.println(p.left.tabbedOutput());
		} catch(Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
	}
	
}
